/*
 * Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <algorithm>
#include <iomanip>
#include <infiniband/ibdiag/ibdiag.h>
#include <infiniband/ibutils/output_stream_primitives.h>


/****************************************************/
/******************private methods*******************/
/****************************************************/
void IBDiag::DumpNodeInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (csv_out.DumpStart(SECTION_NODES))
        IBDIAG_RETURN_VOID;

    stringstream sstream;
    sstream << "NodeDesc," << "NumPorts," << "NodeType," << "ClassVersion," << "BaseVersion," << "SystemImageGUID,"
            << "NodeGUID," << "PortGUID," << "DeviceID," << "PartitionCap," << "revision,"
            << "VendorID," << "LocalPortNum"
            << endl;
    csv_out.WriteBuf(sstream.str());

    // go over all nodes:
    struct SMP_NodeInfo * p_curr_node_info = NULL;
    char buffer[1024];
    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;
        p_curr_node_info = this->fabric_extended_info.getSMPNodeInfo(i);
        if (!p_curr_node_info)          //node info not exist
            continue;
        sstream.str("");
        snprintf(buffer, sizeof(buffer), "\"" STR_FMT "\"" ","   \
                 U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","  \
                 U64H_FMT "," U64H_FMT "," U64H_FMT ","  \
                 U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT,
                 //p_curr_node->getName().c_str(),
                 p_curr_node->description.c_str(),
                 p_curr_node_info->NumPorts,
                 p_curr_node_info->NodeType,
                 p_curr_node_info->ClassVersion,
                 p_curr_node_info->BaseVersion,
                 p_curr_node_info->SystemImageGUID,
                 p_curr_node_info->NodeGUID,
                 p_curr_node_info->PortGUID,
                 p_curr_node_info->DeviceID,
                 p_curr_node_info->PartitionCap,
                 p_curr_node_info->revision,
                 p_curr_node_info->VendorID,
                 p_curr_node_info->LocalPortNum);
        sstream << buffer << endl;
        csv_out.WriteBuf(sstream.str());
    }
    csv_out.DumpEnd(SECTION_NODES);

    IBDIAG_RETURN_VOID;
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::DumpPortInfoToCSV(CSVOut &csv_out, bool show_ports_data_extra)
{
    IBDIAG_ENTER;

    stringstream sstream;
    sstream << "# In the next section the values of the fields LinkSpeedEn/LinkSpeedActv/LinkSpeedSup "
            << "are not real raw data from the wire but as follow: " << endl
            << "# They are 32 bit fields." << endl
            << "# Byte1 is regular speeds," << endl
            << "# Byte2 is extended speeds," << endl
            << "# Byte3 is mlnx ext speeds," << endl
            << "# Byte4 is extended speeds 2."
            << endl << endl;
    csv_out.WriteBuf(sstream.str());

    // cannot be skipped, because can generate errors
    (void) csv_out.DumpStart(SECTION_PORTS);

    sstream.str("");
    sstream << "NodeGuid," << "PortGuid," << "PortNum,"   //these first values are used to match port to node
            << "MKey," << "GIDPrfx," << "MSMLID," << "LID," << "CapMsk,"
            << "M_KeyLeasePeriod," << "DiagCode," << "LinkWidthActv," << "LinkWidthSup," << "LinkWidthEn,"
            << "LocalPortNum," << "LinkSpeedEn," << "LinkSpeedActv," << "LMC,"
            << "MKeyProtBits," << "LinkDownDefState," << "PortPhyState," << "PortState," << "LinkSpeedSup,"
            << "VLArbHighCap," << "VLHighLimit," << "InitType," << "VLCap," << "MSMSL,"
            << "NMTU," << "FilterRawOutb," << "FilterRawInb," << "PartEnfOutb," << "PartEnfInb,"
            << "OpVLs," << "HoQLife," << "VLStallCnt," << "MTUCap," << "InitTypeReply,"
            << "VLArbLowCap," << "PKeyViolations," << "MKeyViolations," << "SubnTmo,"
            << "MulticastPKeyTrapSuppressionEnabled,"
            << "ClientReregister," << "GUIDCap," << "QKeyViolations," << "MaxCreditHint," << "OverrunErrs,"
            << "LocalPhyError," << "RespTimeValue," << "LinkRoundTripLatency," << "OOOSLMask";
    if (show_ports_data_extra)
        sstream << ",CapMsk2,FECActv,RetransActv";
    sstream << endl;
    csv_out.WriteBuf(sstream.str());

    struct SMP_PortInfo *p_curr_port_info = NULL;
    struct SMP_MlnxExtPortInfo *p_curr_mlnx_ext_port_info = NULL;
    char buffer[1024];
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        //port not exist or is not in the scope
        if (!p_curr_port || !p_curr_port->getInSubFabric())
            continue;

        if (!p_curr_port->getPortInfoMadWasSent())
            continue;

        p_curr_port_info = this->fabric_extended_info.getSMPPortInfo(i);
        if (!p_curr_port_info)      //port info not exist
            continue;

        u_int32_t cap_mask                  = p_curr_port_info->CapMsk;
        u_int16_t cap_mask2                 = p_curr_port_info->CapMsk2;
        u_int32_t correct_link_speed_en;
        u_int32_t correct_link_speed_actv;
        u_int32_t correct_link_speed_sup;

        //if it switch than take cap mask from port 0
        if ((p_curr_port->p_node->type == IB_SW_NODE) &&
                p_curr_port->num != 0) {

            IBPort *p_zero_port = p_curr_port->p_node->getPort(0);
            if (!p_zero_port) {
                this->SetLastError("DB error - can not found manage port for switch=%s\n",
                        p_curr_port->p_node->getName().c_str());
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }

            struct SMP_PortInfo *p_zero_port_info =
                    this->fabric_extended_info.getSMPPortInfo(p_zero_port->createIndex);
            if (!p_zero_port_info) {
                stringstream ss;
                ss << "DB error - can not found port info for port=" << p_zero_port->getName();

                this->errors.push_back(
                    new FabricErrPortInfoFail(p_curr_port->p_node, p_curr_port->num, ss.str().c_str()));

                continue;
            }

            cap_mask = p_zero_port_info->CapMsk;
            cap_mask2 = p_zero_port_info->CapMsk2;
        }

        //if the port supports mlnx extended speeds
        p_curr_mlnx_ext_port_info = this->fabric_extended_info.getSMPMlnxExtPortInfo(i);

        //regular speeds + extended + mellanox
        GetTotalSpeeds(p_curr_port_info,
                       p_curr_mlnx_ext_port_info,
                       cap_mask,
                       cap_mask2,
                       &correct_link_speed_sup,
                       &correct_link_speed_en,
                       &correct_link_speed_actv);

        stringstream OOOSLMask_Output("N/A");

        if (p_curr_mlnx_ext_port_info) {

            if (this->capability_module.IsSupportedSMPCapability(p_curr_port->p_node,
                                                                 EnSMPCapIsGlobalOOOSupported)) {
                OOOSLMask_Output.str("");
                OOOSLMask_Output << "0x" << std::hex << p_curr_mlnx_ext_port_info->OOOSLMask;
            }
        }

        sstream.str("");
        snprintf(buffer, sizeof(buffer),
                 U64H_FMT "," U64H_FMT "," U32D_FMT ","    \
                 U64H_FMT "," U64H_FMT ","    \
                 U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","    \
                 U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","    \
                 U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","    \
                 U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","    \
                 U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","    \
                 U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","    \
                 U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","    \
                 U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","    \
                 U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ",%s" ,
                 p_curr_port->p_node->guid_get(),
                 p_curr_port->guid_get(),
                 p_curr_port->num,
                 p_curr_port_info->MKey,
                 p_curr_port_info->GIDPrfx,
                 p_curr_port_info->MSMLID,
                 p_curr_port_info->LID,
                 p_curr_port_info->CapMsk,
                 p_curr_port_info->M_KeyLeasePeriod,
                 p_curr_port_info->DiagCode,
                 p_curr_port_info->LinkWidthActv,
                 p_curr_port_info->LinkWidthSup,
                 p_curr_port_info->LinkWidthEn,
                 p_curr_port_info->LocalPortNum,
                 correct_link_speed_en,
                 correct_link_speed_actv,
                 p_curr_port_info->LMC,
                 p_curr_port_info->MKeyProtBits,
                 p_curr_port_info->LinkDownDefState,
                 p_curr_port_info->PortPhyState,
                 p_curr_port_info->PortState,
                 correct_link_speed_sup,
                 p_curr_port_info->VLArbHighCap,
                 p_curr_port_info->VLHighLimit,
                 p_curr_port_info->InitType,
                 p_curr_port_info->VLCap,
                 p_curr_port_info->MSMSL,
                 p_curr_port_info->NMTU,
                 p_curr_port_info->FilterRawOutb,
                 p_curr_port_info->FilterRawInb,
                 p_curr_port_info->PartEnfOutb,
                 p_curr_port_info->PartEnfInb,
                 p_curr_port_info->OpVLs,
                 p_curr_port_info->HoQLife,
                 p_curr_port_info->VLStallCnt,
                 p_curr_port_info->MTUCap,
                 p_curr_port_info->InitTypeReply,
                 p_curr_port_info->VLArbLowCap,
                 p_curr_port_info->PKeyViolations,
                 p_curr_port_info->MKeyViolations,
                 p_curr_port_info->SubnTmo,
                 p_curr_port_info->MulticastPKeyTrapSuppressionEnabled,
                 p_curr_port_info->ClientReregister,
                 p_curr_port_info->GUIDCap,
                 p_curr_port_info->QKeyViolations,
                 p_curr_port_info->MaxCreditHint,
                 p_curr_port_info->OverrunErrs,
                 p_curr_port_info->LocalPhyError,
                 p_curr_port_info->RespTimeValue,
                 p_curr_port_info->LinkRoundTripLatency,
                 OOOSLMask_Output.str().c_str());

        sstream << buffer;
        if (show_ports_data_extra) {
            //PortInfoCapMsk2, FEC, Retransmission
            if (IS_SUPPORT_PORT_INFO_CAP_MASK2(p_curr_port_info->CapMsk))
                sstream << "," << p_curr_port_info->CapMsk2;
            else
                sstream << ",N/A";
            IBFECMode fec = p_curr_port->get_fec_mode();
            if (fec == IB_FEC_NA)
                sstream << "," <<  IB_FEC_NA_STR << ",";
            else
                sstream << "," << fec << ",";

            if (!p_curr_mlnx_ext_port_info)
                sstream << "N/A";
            else
                sstream << (unsigned int)p_curr_mlnx_ext_port_info->RetransMode;
        }
        sstream << endl;
        csv_out.WriteBuf(sstream.str());
    }
    csv_out.DumpEnd(SECTION_PORTS);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void IBDiag::DumpExtendedNodeInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (csv_out.DumpStart(SECTION_EXTENDED_NODE_INFO))
        IBDIAG_RETURN_VOID;

    stringstream sstream;
    sstream << "NodeGuid,sl2vl_cap,sl2vl_act,num_pcie,num_oob,"
               "AnycastLIDTop,AnycastLidCap,node_type_extended,asic_max_planes"
            << endl;

    csv_out.WriteBuf(sstream.str());

    struct ib_extended_node_info *p_meni;

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {

        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        p_meni = this->fabric_extended_info.getSMPExtNodeInfo(p_curr_node->createIndex);
        if (!p_meni)
            continue;

        sstream.str("");
        sstream << PTR(p_curr_node->guid_get())  << ","
                << PTR(p_meni->sl2vl_cap)        << ","
                << PTR(p_meni->sl2vl_act)        << ","
                << +p_meni->num_pcie             << ","
                << +p_meni->num_oob              << ","
                << +p_meni->AnycastLIDTop        << ","
                << +p_meni->AnycastLidCap        << ","
                << +p_meni->node_type_extended   << ","
                << +p_meni->asic_max_planes      << endl;

        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_EXTENDED_NODE_INFO);

    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpExtendedPortInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (csv_out.DumpStart(SECTION_EXTENDED_PORT_INFO))
        IBDIAG_RETURN_VOID;

    stringstream sstream;
    sstream << "NodeGuid,PortGuid,PortNum,"
            << "StateChangeEnable,PortRecovered,RouterLIDEn,SHArPANEn,AME,LinkSpeedSupported,UnhealthyReason,"
            << "BwUtilization,BwUtilEn,LinkSpeedEnabled,MinBwUtilization,LinkSpeedActive,"
            << "ActiveRSFECParity,ActiveRSFECData,CapabilityMask,SMPDisable,NMXAdminState,"
            << "FECModeActive,RetransMode,FDR10FECModeSupported,FDR10FECModeEnabled,"
            << "FDRFECModeSupported,FDRFECModeEnabled,EDR20FECModeSupported,EDR20FECModeEnabled,"
            << "EDRFECModeSupported,EDRFECModeEnabled,FDR10RetranSupported,FDR10RetranEnabled,"
            << "FDRRetranSupported,FDRRetranEnabled,EDR20RetranSupported,EDR20RetranEnabled,"
            << "EDRRetranSupported,EDRRetranEnabled,IsSpecialPort,SpecialPortType,"
            << "SpecialPortCapabilityMask,IsFNMPort,"
            << "HDRFECModeSupported,HDRFECModeEnabled,OOOSLMask,"
            << "AdaptiveTimeoutSLMask,NDRFECModeSupported,NDRFECModeEnabled"
            << endl;

    csv_out.WriteBuf(sstream.str());

    IBPort *p_curr_port;
    struct SMP_MlnxExtPortInfo *p_mepi;
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
         p_curr_port = this->fabric_extended_info.getPortPtr(i);
         if (!p_curr_port || !p_curr_port->getInSubFabric())
             continue;

         p_mepi = fabric_extended_info.getSMPMlnxExtPortInfo(i);
         if (!p_mepi)
             continue;

         sstream.str("");
         sstream << "0x" << std::hex << setfill('0') << setw(16) << p_curr_port->p_node->guid_get()
                 << ",0x" << setw(16) << p_curr_port->guid_get()
                 << "," << std::dec << +p_curr_port->num
                 << ",0x" << std::hex << setw(2) << +p_mepi->StateChangeEnable
                 << ",0x" << setw(2) << +p_mepi->port_recovered
                 << ",0x" << setw(2) << +p_mepi->RouterLIDEn
                 << ",0x" << setw(2) << +p_mepi->SHArPANEn
                 << ",0x" << setw(2) << +p_mepi->AME
                 << ",0x" << setw(2) << +p_mepi->LinkSpeedSupported
                 << ","   << std::dec << +p_mepi->unhealthy_reason
                 << ","   << std::dec << +p_mepi->bw_utilization
                 << ","   << std::dec << +p_mepi->bw_util_en
                 << ",0x" << std::hex << setw(2) << +p_mepi->LinkSpeedEnabled
                 << ","   << std::dec << +p_mepi->min_bw_utilization
                 << ",0x" << std::hex << setw(2) << +p_mepi->LinkSpeedActive
                 << ",0x" << setw(4) << p_mepi->ActiveRSFECParity
                 << ",0x" << setw(4) << p_mepi->ActiveRSFECData
                 << ",0x" << setw(4) << p_mepi->CapabilityMask
                 << ",0x" << setw(2) << +p_mepi->SMPDisable
                 << ",0x" << setw(2) << +p_mepi->NMXAdminState
                 << ",0x" << setw(2) << +p_mepi->FECModeActive
                 << ",0x" << setw(2) << +p_mepi->RetransMode
                 << ",0x" << setw(4) << p_mepi->FDR10FECModeSupported
                 << ",0x" << setw(4) << p_mepi->FDR10FECModeEnabled
                 << ",0x" << setw(4) << p_mepi->FDRFECModeSupported
                 << ",0x" << setw(4) << p_mepi->FDRFECModeEnabled
                 << ",0x" << setw(4) << p_mepi->EDR20FECModeSupported
                 << ",0x" << setw(4) << p_mepi->EDR20FECModeEnabled
                 << ",0x" << setw(4) << p_mepi->EDRFECModeSupported
                 << ",0x" << setw(4) << p_mepi->EDRFECModeEnabled
                 << ",0x" << setw(2) << +p_mepi->FDR10RetranSupported
                 << ",0x" << setw(2) << +p_mepi->FDR10RetranEnabled
                 << ",0x" << setw(2) << +p_mepi->FDRRetranSupported
                 << ",0x" << setw(2) << +p_mepi->FDRRetranEnabled
                 << ",0x" << setw(2) << +p_mepi->EDR20RetranSupported
                 << ",0x" << setw(2) << +p_mepi->EDR20RetranEnabled
                 << ",0x" << setw(2) << +p_mepi->EDRRetranSupported
                 << ",0x" << setw(2) << +p_mepi->EDRRetranEnabled
                 << "," << std::dec << +p_mepi->IsSpecialPort
                 << ",";
                 if (p_mepi->IsSpecialPort)
                     sstream << +p_mepi->SpecialPortType;
                 else
                     sstream << "N/A";
         sstream << ",0x" << std::hex << setw(2) << +p_mepi->SpecialPortCapabilityMask
                 << ","   << std::dec << +p_mepi->IsFNMPort
                 << ",0x" << std::hex << setw(4) << p_mepi->HDRFECModeSupported
                 << ",0x" << setw(4) << p_mepi->HDRFECModeEnabled
                 << ",0x" << setw(4) << p_mepi->OOOSLMask
                 << ",0x" << setw(4) << p_mepi->AdaptiveTimeoutSLMask
                 << ",0x" << setw(4) << p_mepi->NDRFECModeSupported
                 << ",0x" << setw(4) << p_mepi->NDRFECModeEnabled
                 << endl;

         csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_EXTENDED_PORT_INFO);
    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpPortInfoExtendedToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (csv_out.DumpStart(SECTION_PORT_INFO_EXTENDED))
        IBDIAG_RETURN_VOID;

    stringstream sstream;
    sstream << "NodeGuid,PortGuid,PortNum,FECModeActive,FDRFECModeSupported,FDRFECModeEnabled,"
            << "EDRFECModeSupported,EDRFECModeEnabled,HDRFECModeSupported,HDRFECModeEnabled,"
            << "NDRFECModeSupported,NDRFECModeEnabled,CapabilityMask" << endl;

    csv_out.WriteBuf(sstream.str());

    IBPort *p_curr_port;
    struct SMP_PortInfoExtended *p_port_info_extended;
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
         p_curr_port = this->fabric_extended_info.getPortPtr(i);
         if (!p_curr_port || !p_curr_port->getInSubFabric())
             continue;

         p_port_info_extended = fabric_extended_info.getSMPPortInfoExtended(i);
         if (!p_port_info_extended)
             continue;

         sstream.str("");
         sstream << "0x" << std::hex << setfill('0') << setw(16) << p_curr_port->p_node->guid_get()
                 << ",0x" << setw(16) << p_curr_port->guid_get()
                 << "," << std::dec << +p_curr_port->num
                 << ",0x" << std::hex << setw(4) << p_port_info_extended->FECModeActive
                 << ",0x" << setw(4) << p_port_info_extended->FDRFECModeSupported
                 << ",0x" << setw(4) << p_port_info_extended->FDRFECModeEnabled
                 << ",0x" << setw(4) << p_port_info_extended->EDRFECModeSupported
                 << ",0x" << setw(4) << p_port_info_extended->EDRFECModeEnabled
                 << ",0x" << setw(4) << p_port_info_extended->HDRFECModeSupported
                 << ",0x" << setw(4) << p_port_info_extended->HDRFECModeEnabled
                 << ",0x" << setw(4) << p_port_info_extended->NDRFECModeSupported
                 << ",0x" << setw(4) << p_port_info_extended->NDRFECModeEnabled
                 << ",0x" << setw(8) << p_port_info_extended->CapMsk
                 << endl;

         csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_PORT_INFO_EXTENDED);
    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpFECModeToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (csv_out.DumpStart(SECTION_FEC_MODE))
        IBDIAG_RETURN_VOID;

    stringstream sstream;
    sstream << "NodeGuid,PortGuid,PortNum,FECActv,FDRFECSup,FDRFECEn,EDRFECSup,EDRFECEn,"
            "HDRFECSup,HDRFECEn,NDRFECSup,NDRFECEn,FDR10FECSup,FDR10FECEn,EDR20FECSup,EDR20FECEn" << endl;
    csv_out.WriteBuf(sstream.str());

    char buffer[1024];
    IBPort *p_curr_port = NULL;
    struct SMP_PortInfoExtended *p_port_info_ext = NULL;
    struct SMP_MlnxExtPortInfo *p_mepi = NULL;
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port || !p_curr_port->getInSubFabric())
            continue;

        p_port_info_ext = fabric_extended_info.getSMPPortInfoExtended(i);
        p_mepi = fabric_extended_info.getSMPMlnxExtPortInfo(i);

        bool advancedFECModes = this->capability_module.IsSupportedSMPCapability(
                p_curr_port->p_node, EnSMPCapIsAdvancedRetransAndFECModesSupported);

        if ((!p_mepi || !advancedFECModes) && !p_port_info_ext)    // no FEC info
            continue;

        sstream.str("");
        snprintf(buffer, sizeof(buffer),
                 U64H_FMT "," U64H_FMT "," U32D_FMT "," U32D_FMT ",",
                 p_curr_port->p_node->guid_get(),
                 p_curr_port->guid_get(),
                 p_curr_port->num,
                 p_curr_port->get_fec_mode());

        sstream << buffer;

        if (p_mepi && advancedFECModes) {
            snprintf(buffer, sizeof(buffer),
                    U16H_FMT "," U16H_FMT "," U16H_FMT "," U16H_FMT "," U16H_FMT "," U16H_FMT ","
                    U16H_FMT "," U16H_FMT "," U16H_FMT "," U16H_FMT "," U16H_FMT "," U16H_FMT ,
                    p_mepi->FDRFECModeSupported,
                    p_mepi->FDRFECModeEnabled,
                    p_mepi->EDRFECModeSupported,
                    p_mepi->EDRFECModeEnabled,
                    p_mepi->HDRFECModeSupported,
                    p_mepi->HDRFECModeEnabled,
                    p_mepi->NDRFECModeSupported,
                    p_mepi->NDRFECModeEnabled,
                    p_mepi->FDR10FECModeSupported,
                    p_mepi->FDR10FECModeEnabled,
                    p_mepi->EDR20FECModeSupported,
                    p_mepi->EDR20FECModeEnabled);
        } else {    // no mepi, use port_info_ext.
            snprintf(buffer, sizeof(buffer),
                    U16H_FMT "," U16H_FMT "," U16H_FMT "," U16H_FMT "," U16H_FMT ","
                    U16H_FMT "," U16H_FMT "," U16H_FMT ",N/A,N/A,N/A,N/A",
                    p_port_info_ext->FDRFECModeSupported,
                    p_port_info_ext->FDRFECModeEnabled,
                    p_port_info_ext->EDRFECModeSupported,
                    p_port_info_ext->EDRFECModeEnabled,
                    p_port_info_ext->HDRFECModeSupported,
                    p_port_info_ext->HDRFECModeEnabled,
                    p_port_info_ext->NDRFECModeSupported,
                    p_port_info_ext->NDRFECModeEnabled);
        }

        sstream << buffer << endl;
        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_FEC_MODE);
    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpSwitchInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if  (csv_out.DumpStart(SECTION_SWITCHES))
        IBDIAG_RETURN_VOID;

    stringstream sstream;
    sstream << "NodeGUID," << "LinearFDBCap," << "RandomFDBCap," << "MCastFDBCap,"
            << "LinearFDBTop," << "DefPort," << "DefMCastPriPort," << "DefMCastNotPriPort,"
            << "LifeTimeValue," << "PortStateChange," << "OptimizedSLVLMapping," << "LidsPerPort,"
            << "PartEnfCap," << "InbEnfCap," << "OutbEnfCap," << "FilterRawInbCap,"
            << "FilterRawOutbCap," << "ENP0," << "MCastFDBTop"
            << endl;
    csv_out.WriteBuf(sstream.str());

    char buffer[1024];
    IBNode *p_curr_node = NULL;
    SMP_SwitchInfo *p_curr_switch_info = NULL;

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node  || !p_curr_node->getInSubFabric())
            continue;

        p_curr_switch_info = this->fabric_extended_info.getSMPSwitchInfo(i);
        if (!p_curr_switch_info)        // switch info not exist
            continue;

        sstream.str("");

        snprintf(buffer, sizeof(buffer),
                 U64H_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," \
                 U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," \
                 U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," \
                 U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT,
                 p_curr_node->guid_get(),
                 p_curr_switch_info->LinearFDBCap,
                 p_curr_switch_info->RandomFDBCap,
                 p_curr_switch_info->MCastFDBCap,
                 p_curr_switch_info->LinearFDBTop,
                 p_curr_switch_info->DefPort,
                 p_curr_switch_info->DefMCastPriPort,
                 p_curr_switch_info->DefMCastNotPriPort,
                 p_curr_switch_info->LifeTimeValue,
                 p_curr_switch_info->PortStateChange,
                 p_curr_switch_info->OptimizedSLVLMapping,
                 p_curr_switch_info->LidsPerPort,
                 p_curr_switch_info->PartEnfCap,
                 p_curr_switch_info->InbEnfCap,
                 p_curr_switch_info->OutbEnfCap,
                 p_curr_switch_info->FilterRawInbCap,
                 p_curr_switch_info->FilterRawOutbCap,
                 p_curr_switch_info->ENP0,
                 p_curr_switch_info->MCastFDBTop);
        sstream << buffer << endl;
        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_SWITCHES);
    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpPortHierarchyInfoStream(ostream & stream, const char * header_preffix)
{
    IBDIAG_ENTER;

    stream << header_preffix
            << "NodeGUID,PortGUID,TemplateGUID,PortNum,Label,ExtendedLabel,Bus,Device,Function,Type,SlotType,SlotValue,"
            << "NumOfPlanes,Plane,APort,IsCageManager,NumberOnBaseBoard,ASIC,Cage,Port,Split,IBPort,PortType,AsicName,"
            << "DeviceNumOnCpuNode,CpuNodeNumber,BoardType,ChassisSlotIndex,TrayIndex,TopologyID"
            << endl;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_curr_node = nI->second;

        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName");
            IBDIAG_RETURN_VOID;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        for (u_int32_t i = 1; i <= p_curr_node->numPorts; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);

            if (!p_curr_port)
                continue;

            if (!p_curr_port->p_port_hierarchy_info)
                continue;

            stream << PTR(p_curr_node->guid_get()) << ","
               << PTR(p_curr_port->guid_get()) << ","
               << PTR(p_curr_port->p_port_hierarchy_info->m_template_guid) << ","
               << DEC(p_curr_port->num) << ","
               << QUOTED(p_curr_port->p_port_hierarchy_info->label()) << ","
               << QUOTED(p_curr_port->p_port_hierarchy_info->extendedLabel());

            #define PRINT_CSV( x )                                          \
                stream << ',';                                              \
                if (p_curr_port->p_port_hierarchy_info->x == -1)            \
                    stream << "N/A";                                        \
                else                                                        \
                    stream << DEC(p_curr_port->p_port_hierarchy_info->x)

            PRINT_CSV(bus());
            PRINT_CSV(device());
            PRINT_CSV(function());
            PRINT_CSV(m_type);
            PRINT_CSV(m_slot_type);
            PRINT_CSV(m_slot_value);
            PRINT_CSV(m_num_of_planes);
            PRINT_CSV(m_plane);
            PRINT_CSV(m_aport);
            PRINT_CSV(m_is_cage_manager);
            PRINT_CSV(m_number_on_base_board);
            PRINT_CSV(m_asic);
            PRINT_CSV(m_cage);
            PRINT_CSV(m_port);
            PRINT_CSV(m_split);
            PRINT_CSV(m_ibport);
            PRINT_CSV(m_port_type);
            PRINT_CSV(m_asic_name);
            PRINT_CSV(m_device_num_on_cpu_node);
            PRINT_CSV(m_cpu_node_number);
            PRINT_CSV(m_board_type);
            PRINT_CSV(m_chassis_slot_index);
            PRINT_CSV(m_tray_index);
            PRINT_CSV(m_topology_id);

            #undef PRINT_CSV

            stream << endl;
        }
    }
    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpPortHierarchyInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (csv_out.DumpStart(SECTION_PORT_HIERARCHY_INFO))
        IBDIAG_RETURN_VOID;

    stringstream sstream;
    DumpPortHierarchyInfoStream(sstream, "");
    csv_out.WriteBuf(sstream.str());

    csv_out.DumpEnd(SECTION_PORT_HIERARCHY_INFO);
    IBDIAG_RETURN_VOID;
}

int IBDiag::WritePortHierarchyInfoFile(const std::string & file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("Port Hierarchy Info",
                            file_name, sout,
                            false, //no append
                            "#"); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    this->DumpPortHierarchyInfoStream(sout, "# ");

    this->CloseFile(sout);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void IBDiag::DumpPhysicalHierarchyInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (csv_out.DumpStart(SECTION_PHYSICAL_HIERARCHY_INFO))
        IBDIAG_RETURN_VOID;

    stringstream sstream;
    sstream << "NodeGUID,CampusSerialNum,RoomSerialNum,RackSerialNum,SystemType,SystemTopUNum,"
            << "BoardType,BoardSlotNum,DeviceSerialNum,DeviceNumOnCPUNode,CPUNodeNumber,"
            << "ChassisSlotIndex,TrayIndex,TopologyID"
            << endl;
    csv_out.WriteBuf(sstream.str());

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {
        IBNode *p_curr_node = *nI;

        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN_VOID;
        }

        if (!p_curr_node->getInSubFabric() || !p_curr_node->p_physical_hierarchy_info)
            continue;

        stringstream ss;
        ss << PTR(p_curr_node->guid_get());

        #define PRINT_CSV( x )                                          \
            ss << ',';                                                  \
            if (p_curr_node->p_physical_hierarchy_info->x == -1)        \
                ss << "N/A";                                            \
            else                                                        \
                ss << DEC(p_curr_node->p_physical_hierarchy_info->x)

        PRINT_CSV(m_campus_serial_num);
        PRINT_CSV(m_room_serial_num);
        PRINT_CSV(m_rack_serial_num);
        PRINT_CSV(m_system_type);
        PRINT_CSV(m_system_topu_num);
        PRINT_CSV(m_board_type);
        PRINT_CSV(m_board_slot_num);
        PRINT_CSV(m_device_serial_num);
        PRINT_CSV(m_device_num_on_cpu_node);
        PRINT_CSV(m_cpu_node_number);
        PRINT_CSV(m_chassis_slot_index);
        PRINT_CSV(m_tray_index);
        PRINT_CSV(m_topology_id);

        #undef PRINT_CSV

        ss << endl;

        csv_out.WriteBuf(ss.str());
    }

    csv_out.DumpEnd(SECTION_PHYSICAL_HIERARCHY_INFO);
    IBDIAG_RETURN_VOID;
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::DumpLinksToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    //first mark in each port some field we will use latter
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        // go over all ports of this node:
        for (int i = 1; i <= p_curr_node->numPorts; ++i) {    //start from port 1, no need to dump port 0
                IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
                if (!p_curr_port)       //port not exist
                    continue;
                p_curr_port->counter1 = 0;
        }
    }

    if (csv_out.DumpStart(SECTION_LINKS))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGuid1," << "PortNum1,"
            << "NodeGuid2," << "PortNum2"
            << endl;
    csv_out.WriteBuf(sstream.str());

    // go over all nodes:
    char buffer[1024];
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            csv_out.DumpEnd(SECTION_LINKS);
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        // go over all ports of this node:
        for (u_int32_t i = 1; i <= p_curr_node->numPorts; ++i) { //start from port 1, no need to dump port 0
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port || !p_curr_port->getInSubFabric())
                continue;

            IBPort *p_remote_port = p_curr_port->p_remotePort;
            if (!p_remote_port || !p_remote_port->getInSubFabric())
                continue;

            IBNode *p_remote_node = p_remote_port->p_node;
            if (!p_remote_node) {
                this->SetLastError("DB error - found port with no node %s",
                                   p_remote_port->getName().c_str());
                csv_out.DumpEnd(SECTION_LINKS);
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }

            //don't write same connection more than one time
            if ((p_curr_port->counter1 == 1) || (p_remote_port->counter1 == 1))
                continue;
            p_curr_port->counter1 = 1;
            p_remote_port->counter1 = 1;

            sstream.str("");
            snprintf(buffer, sizeof(buffer),
                     U64H_FMT "," U32D_FMT ","   \
                     U64H_FMT "," U32D_FMT,
                     p_curr_node->guid_get(), p_curr_port->num,
                     p_remote_node->guid_get(), p_remote_port->num);
            sstream << buffer << endl;
            csv_out.WriteBuf(sstream.str());
        }
    }

    csv_out.DumpEnd(SECTION_LINKS);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void IBDiag::DumpARInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (csv_out.DumpStart(SECTION_AR_INFO))
        IBDIAG_RETURN_VOID;

    stringstream sstream;
    sstream << "NodeGUID,e,is_arn_sup,is_frn_sup,is_fr_sup,fr_enabled,rn_xmit_enabled,is_ar_trials_supported,"
            << "sub_grps_active,group_table_copy_sup,direction_num_sup,is4_mode,"
            << "glb_groups,by_sl_cap,by_sl_en,by_transp_cap,dyn_cap_calc_sup,group_cap,"
            << "group_top,group_table_cap,string_width_cap,ar_version_cap,rn_version_cap,"
            << "sub_grps_supported,enable_by_sl_mask,by_transport_disable,ageing_time_value,"
            << "is_hbf_supported,by_sl_hbf_en,enable_by_sl_mask_hbf,is_whbf_supported,whbf_en,whbf_granularity,"
            << "is_symmetric_hash_supported,is_dceth_hash_supported,is_bth_dqp_hash_supported,"
            << "is_pfrn_supported,pfrn_enabled"
            << endl;
    csv_out.WriteBuf(sstream.str());

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {
        IBNode *p_curr_node = *nI;

        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN_VOID;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        adaptive_routing_info* ar_info = this->GetIBDMExtendedInfoPtr()->getARInfo(p_curr_node->createIndex);
        if (!ar_info)        // AR info not exist
            continue;

        sstream.str("");

        sstream << PTR(p_curr_node->guid_get())       << ","
                << +ar_info->e                        << ","
                << +ar_info->is_arn_sup               << ","
                << +ar_info->is_frn_sup               << ","
                << +ar_info->is_fr_sup                << ","
                << +ar_info->fr_enabled               << ","
                << +ar_info->rn_xmit_enabled          << ","
                << +ar_info->is_ar_trials_supported   << ","
                << +ar_info->sub_grps_active          << ","
                << +ar_info->group_table_copy_sup     << ","
                << +ar_info->direction_num_sup        << ","
                << +ar_info->is4_mode                 << ","
                << +ar_info->glb_groups               << ","
                << +ar_info->by_sl_cap                << ","
                << +ar_info->by_sl_en                 << ","
                << +ar_info->by_transp_cap            << ","
                << +ar_info->dyn_cap_calc_sup         << ","
                << +ar_info->group_cap                << ","
                << +ar_info->group_top                << ","
                << +ar_info->group_table_cap          << ","
                << +ar_info->string_width_cap         << ","
                << +ar_info->ar_version_cap           << ","
                << +ar_info->rn_version_cap           << ","
                << +ar_info->sub_grps_supported       << ","
                << +ar_info->enable_by_sl_mask        << ","
                << +ar_info->by_transport_disable     << ","
                << +ar_info->ageing_time_value        << ","
                //"is_hbf_supported,by_sl_hbf_en,enable_by_sl_mask_hbf,is_whbf_supported,whbf_en,whbf_granularity"
                << +ar_info->is_hbf_supported          << ","
                << +ar_info->by_sl_hbf_en              << ","
                << +ar_info->enable_by_sl_mask_hbf     << ","
                << +ar_info->is_whbf_supported         << ","
                << +ar_info->whbf_en                   << ","
                << +ar_info->whbf_granularity          << ","
                //"is_symmetric_hash_supported,is_dceth_hash_supported,is_bthdqp_hash_supported"
                << +ar_info->is_symmetric_hash_supported << ","
                << +ar_info->is_dceth_hash_supported     << ","
                << +ar_info->is_bth_dqp_hash_supported   << ","
                << +ar_info->is_pfrn_supported           << ","
                << +ar_info->pfrn_enabled                << endl;

        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_AR_INFO);
    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpPortDRToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    direct_route_t return_dr = {};
    stringstream sstream;
    set<const IBNode*> visited;


    if (csv_out.DumpStart(SECTION_PORT_DRS))
        IBDIAG_RETURN_VOID;

    sstream << "NodeGuid," << "PortGuid," << "Lid,"   //these first values are used to match port to node
            << "DirectRoute," << "ReturnDirectRoute"
            << endl;
    csv_out.WriteBuf(sstream.str());

    for (IBPort * p_curr_port: this->discovered_fabric.PortByLid) {
        //port not exist or is not in the scope
        if (!p_curr_port || !p_curr_port->getInSubFabric() || !p_curr_port->is_data_worthy(DATA_WORTHY_SMP_MAD))
            continue;

        list_p_direct_route& routes = this->GetDirectRoutesByNodeGuid(p_curr_port->guid_get());
        if (routes.empty())
            continue;

        direct_route_t * chosen = nullptr;

        for (direct_route_t *curr_dr : routes) {
            if (curr_dr) {
                if (!chosen || chosen->length > curr_dr->length)
                    chosen = curr_dr;
                else if (curr_dr->length == chosen->length)
                    for (int i = 0; i < chosen->length; ++i)
                        if (chosen->path.BYTE[i] > curr_dr->path.BYTE[i]) {
                            chosen = curr_dr;
                            break;
                        }
            }
        }

        if (chosen) {
            sstream.str("");
            CLEAR_STRUCT(return_dr);

            if (!GetReverseDirectRoute(&return_dr, chosen, true)) {
                string curr_dr_str = Ibis::ConvertDirPathToStr(chosen);
                string ret_dr_str = Ibis::ConvertDirPathToStr(&return_dr);
                // remove [] and replace all ',' with '.'
                std::replace(curr_dr_str.begin(), curr_dr_str.end(), ',', '.');
                std::replace(ret_dr_str.begin(), ret_dr_str.end(), ',', '.');
                curr_dr_str = curr_dr_str.substr(2, curr_dr_str.length() - 3);
                ret_dr_str = ret_dr_str.substr(2, ret_dr_str.length() - 3);

                sstream << PTR(p_curr_port->guid_get())             << ","
                        << PTR(p_curr_port->p_node->guid_get())     << ","
                        << DEC(p_curr_port->base_lid)               << ","
                        << curr_dr_str                              << ","
                        << ret_dr_str
                        << endl;

                csv_out.WriteBuf(sstream.str());
            }
        }
    }

    csv_out.DumpEnd(SECTION_PORT_DRS);

    IBDIAG_RETURN_VOID;
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::DumpNetwork(ofstream &sout)
{
    IBDIAG_ENTER;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
         nI != this->discovered_fabric.NodeByName.end(); ++nI) {

        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName " \
                               "map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (DumpNetworkNodeHeader(sout, p_curr_node))
            continue;

        // go over all ports of this node:
        // start from port 1, no need to dump port 0
        for (int i = 1; i <= p_curr_node->numPorts; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port)
                continue;

            DumpNetworkPort(sout, p_curr_port);
        }
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::DumpNetworkAggregated(ofstream &sout)
{
    IBDIAG_ENTER;
    this->discovered_fabric.unvisitAllAPorts();

    for (auto& system_name_pair : this->discovered_fabric.SystemByName) {
        IBSystem *p_curr_system = system_name_pair.second;
        if (!p_curr_system)
            continue;

        if (DumpNetworkSystemHeader(sout, p_curr_system))
            continue;
        for (auto& node_name_pair : p_curr_system->NodeByName) {
            IBNode *p_curr_node = node_name_pair.second;
            if (!p_curr_node) {
                this->SetLastError("DB error - found null node in NodeByName " \
                                "map for key = %s", node_name_pair.first.c_str());
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }
            // go over all ports of this node:
            // if an APort was already visited skip it
            for (int i = 1; i <= p_curr_node->numPorts; ++i) {
                IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
                if (!p_curr_port)
                    continue;

                APort *p_curr_aport = p_curr_port->p_aport;
                if (p_curr_aport) {
                    if (p_curr_aport->visited)
                        continue;
                    for (size_t plane = 1; plane < p_curr_aport->ports.size(); ++plane) {
                        if (!p_curr_aport->ports[plane])
                            continue;
                        DumpNetworkPortBySys(sout, p_curr_aport->ports[plane]);
                    }
                    p_curr_aport->visited = true;
                } else {
                    DumpNetworkPortBySys(sout, p_curr_port);
                }
            }
        }
    }


    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::DumpNetworkNodeHeader(ostream &sout, IBNode *p_node)
{
    IBDIAG_ENTER;
    char port_line_header[1024] = {'\0'};

    snprintf(port_line_header, 1024,
             "  %-10s : %-3s : %-4s : %-10s : MTU : %-7s : " \
             "%-7s : %-19s : %-6s : %-18s : %-10s : %-4s : %s",
             "#", "IB#", "Sta", "PhysSta", "LWA", "LSA", "FEC mode", "Retran",
             "Neighbor Guid", "N#", "NLID", "Neighbor Description");

    // go over all nodes:
    struct SMP_NodeInfo * p_curr_node_info = NULL;
    if (!p_node || !p_node->getInSubFabric())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);

    p_curr_node_info =
        this->fabric_extended_info.getSMPNodeInfo(p_node->createIndex);
    if (!p_curr_node_info)          //node info not exist
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    if (p_node->type != IB_SW_NODE)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);

    sout << '\"' << p_node->description << "\", "
            << (Ibis::IsVenMellanox(p_curr_node_info->VendorID)? "" : "Non ") << "Mellanox"
            << ", 0x" << hex << p_node->guid_get() << dec
            << ", LID ";

    IBPort *p_port = p_node->getPort((phys_port_t)1);

    if (p_port)
        sout << p_port->base_lid;
    else
        sout << " <ERROR, Port 1 is NULL>";

    sout << endl << port_line_header << endl;

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::DumpNetworkSystemHeader(ostream &sout, IBSystem *p_system)
{
    IBDIAG_ENTER;
    char port_line_header[1024] = {'\0'};
    uint64_t sys_guid = 0;
    IBNodeType type = IB_UNKNOWN_NODE_TYPE;
    if (!p_system)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    for (const auto& pair : p_system->NodeByName) {
        IBNode *p_curr_node = pair.second;
        if (!p_curr_node)
            continue;

        if (p_curr_node->system_guid_get())
            sys_guid = p_curr_node->system_guid_get();
        // TODO: change the way type is chosen
        if (p_curr_node->type != IB_UNKNOWN_NODE_TYPE)
            type = p_curr_node->type;

        if (type != IB_UNKNOWN_NODE_TYPE && sys_guid)
            break;
    }

    sout << nodetype2char(type) << " SystemGUID: " << PTR(sys_guid) << " Description: "
         << p_system->name << endl;

    snprintf(port_line_header, 1024,
             " %-10s : %-4s : %-3s : %-4s : %-10s : MTU : %-7s : " \
             "%-7s : %-19s : %-6s : %-18s : %-10s : %-4s : %s",
             "Label", "#", "IB#", "Sta", "PhysSta", "LWA", "LSA", "FEC mode", "Retran",
             "Neighbor Guid", "N#", "NLID", "Neighbor Description");

    sout << port_line_header << endl;

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::DumpNetworkPort(ostream &sout, IBPort *p_port)
{
    IBDIAG_ENTER;
    char port_line[1024] = {'\0'};
    //port not exist or not in the scope
    if (!p_port || !p_port->getInSubFabric())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);

    if (!p_port->isValid())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);

    struct SMP_PortInfo * p_curr_port_info =
        this->fabric_extended_info.getSMPPortInfo(p_port->createIndex);
    if (!p_curr_port_info)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
                    
    struct SMP_MlnxExtPortInfo *p_mlnx_ext_port_info =
        this->fabric_extended_info.getSMPMlnxExtPortInfo(p_port->createIndex);

    stringstream mtu_str;
    mtu_str << +p_curr_port_info->NMTU;

    stringstream remote_guid;
    stringstream remote_num;
    stringstream remote_lid;
    stringstream remote_desc;

    if (p_port->p_remotePort) {
        remote_guid << "0x" << hex << p_port->p_remotePort->guid_get();
        remote_num << p_port->p_remotePort->numAsString();
        remote_lid << p_port->p_remotePort->base_lid;
        remote_desc << '\"'
                << p_port->p_remotePort->p_node->description.c_str() << '\"';
    }

    snprintf(port_line, 1024,
                "  %-10s : %-3u : %-4s : %-10s : %-3s : %-7s : " \
                "%-7s : %-19s : %-6s : %-18s : %-10s : %-4s : %s\n",
                p_port->numAsString().c_str(),
                p_port->num,
                portstate2char(p_port->get_internal_state()),
                portphysstate2char(
                    (IBPortPhysState)p_curr_port_info->PortPhyState),
                p_port->get_internal_state() == IB_PORT_STATE_DOWN ?
                        NOT_AVAILABLE : mtu_str.str().c_str(),
                p_port->get_internal_state() == IB_PORT_STATE_DOWN ?
                        NOT_AVAILABLE : width2char(p_port->get_internal_width()),
                p_port->get_internal_state() == IB_PORT_STATE_DOWN ?
                        NOT_AVAILABLE : speed2char(p_port->get_internal_speed()),
                fec2char(p_port->get_fec_mode()),
                !p_mlnx_ext_port_info? RETRANS_MODE_NA : retransmission2char(
                        (EnRetransmissionMode)p_mlnx_ext_port_info->RetransMode),
                remote_guid.str().c_str(),
                remote_num.str().c_str(),
                remote_lid.str().c_str(),
                remote_desc.str().c_str()
                );
    sout << port_line;

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::DumpNetworkPortBySys(ostream &sout, IBPort *p_port)
{
    IBDIAG_ENTER;
    char port_line[1024] = {'\0'};
    //port not exist or not in the scope
    if (!p_port || !p_port->getInSubFabric())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);

    if (!p_port->isValid())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);

    struct SMP_PortInfo * p_curr_port_info =
        this->fabric_extended_info.getSMPPortInfo(p_port->createIndex);
    if (!p_curr_port_info)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
                    
    struct SMP_MlnxExtPortInfo *p_mlnx_ext_port_info =
        this->fabric_extended_info.getSMPMlnxExtPortInfo(p_port->createIndex);

    stringstream mtu_str;
    string aggregated_label = p_port->getLabel();
    int plane_number = -1;
    mtu_str << +p_curr_port_info->NMTU;

    stringstream remote_guid;
    stringstream remote_num;
    stringstream remote_lid;
    stringstream remote_desc;

    if (p_port->p_aport) {
        aggregated_label = p_port->p_aport->getAggregatedLabel();
        plane_number = p_port->p_port_hierarchy_info->m_plane;
    }

    if (p_port->p_remotePort) {
        remote_guid << "0x" << hex << p_port->p_remotePort->guid_get();
        remote_num << p_port->p_remotePort->numAsString();
        remote_lid << p_port->p_remotePort->base_lid;
        remote_desc << '\"'
                << p_port->p_remotePort->p_node->description.c_str() << '\"';
    }

    snprintf(port_line, 1024,
                " %-10s :  %-4d : %-3u : %-4s : %-10s : %-3s : %-7s : " \
                "%-7s : %-19s : %-6s : %-18s : %-10s : %-4s : %s\n",
                aggregated_label.c_str(),
                plane_number,
                p_port->num,
                portstate2char(p_port->get_internal_state()),
                portphysstate2char(
                    (IBPortPhysState)p_curr_port_info->PortPhyState),
                p_port->get_internal_state() == IB_PORT_STATE_DOWN ?
                        NOT_AVAILABLE : mtu_str.str().c_str(),
                p_port->get_internal_state() == IB_PORT_STATE_DOWN ?
                        NOT_AVAILABLE : width2char(p_port->get_internal_width()),
                p_port->get_internal_state() == IB_PORT_STATE_DOWN ?
                        NOT_AVAILABLE : speed2char(p_port->get_internal_speed()),
                fec2char(p_port->get_fec_mode()),
                !p_mlnx_ext_port_info? RETRANS_MODE_NA : retransmission2char(
                        (EnRetransmissionMode)p_mlnx_ext_port_info->RetransMode),
                remote_guid.str().c_str(),
                remote_num.str().c_str(),
                remote_lid.str().c_str(),
                remote_desc.str().c_str()
                );
    sout << port_line;

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

/****************************************************/
/*******************public methods*******************/
/****************************************************/
//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR / ERR_CODE_DB_ERR
int IBDiag::DumpInternalToCSV(CSVOut &csv_out, bool show_ports_data_extra)
{
    IBDIAG_ENTER;
    //if (!this->IsDiscoveryDone())
    //    IBDIAG_RETURN(ERR_CODE_DISCOVERY_NOT_SUCCESS);

    this->DumpNodeInfoToCSV(csv_out);
    int rc = this->DumpPortInfoToCSV(csv_out, show_ports_data_extra); //here we can fail
    if (rc)
        IBDIAG_RETURN(rc);

    this->DumpExtendedNodeInfoToCSV(csv_out);
    this->DumpExtendedPortInfoToCSV(csv_out);
    this->DumpPortInfoExtendedToCSV(csv_out);
    this->DumpFECModeToCSV(csv_out);
    this->DumpSwitchInfoToCSV(csv_out);
    this->DumpPortHierarchyInfoToCSV(csv_out);
    this->DumpPhysicalHierarchyInfoToCSV(csv_out);
    this->DumpARInfoToCSV(csv_out);
    this->DumpPortDRToCSV(csv_out);

    rc = this->DumpLinksToCSV(csv_out); //here we can fail because we are get the links information
                                        //only in this stage and after that we dump it

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IBDM_ERR
int IBDiag::WriteLSTFile(const string &file_path, bool write_with_lmc)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("LST", file_path, sout,
                            false, //no append
                            "#"); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    rc = this->discovered_fabric.dumpLSTFile(sout, write_with_lmc);

    this->CloseFile(sout);

    if (rc) {
        this->SetLastError("Failed to dump lst file");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::WriteNetDumpFile(const string &file_path)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("Network dump",
                            file_path, sout,
                            false, //no append
                            "#"); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    IBFabric::GetSwitchLabelPortNumExplanation(sout);
    this->DumpNetwork(sout);

    this->CloseFile(sout);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::WriteNetDumpAggregatedFile(const string &file_path)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (!this->GetDiscoverFabricPtr()->IsHaveAPorts())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    ofstream sout;
    int rc = this->OpenFile("Network dump aggregated",
                            file_path, sout,
                            false, //no append
                            "#"); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    IBFabric::GetSwitchLabelPortNumExplanation(sout);
    this->DumpNetworkAggregated(sout);

    this->CloseFile(sout);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::DumpHBFConfigToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (csv_out.DumpStart(SECTION_HBF_CONFIG))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,hash_type,seed_type,seed,fields_enable" << endl;
    csv_out.WriteBuf(sstream.str());

    for (set_pnode::const_iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {
        const IBNode *p_curr_node = *nI;

        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        const hbf_config *p_hbf = this->GetIBDMExtendedInfoPtr()->getHBFConfig(p_curr_node->createIndex);
        if (!p_hbf)
            continue;

        sstream.str("");

        sstream << PTR(p_curr_node->guid_get())             << ","
                << (p_hbf->hash_type ? "XOR" : "CRC")       << ","
                << (p_hbf->seed_type ? "Random" : "Config") << ","
                << PTR(p_hbf->seed)                         << ","
                << PTR(p_hbf->fields_enable)                << endl;

        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_HBF_CONFIG);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::DumpHBFCountersToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_HBF_PORT_COUNTRERS))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,PortGUID,PortNumber,rx_pkt_forwarding_static,"
            << "rx_pkt_forwarding_hbf, rx_pkt_forwarding_ar, rx_pkt_hbf_fallback_local,"
            << "rx_pkt_hbf_fallback_remote,rx_pkt_forwarding_hbf_sg0,rx_pkt_forwarding_hbf_sg1,"
            << "rx_pkt_forwarding_hbf_sg2,rx_pkt_forwarding_ar_sg0,rx_pkt_forwarding_ar_sg1,"
            << "rx_pkt_forwarding_ar_sg2"
            << endl;
    csv_out.WriteBuf(sstream.str());

    for (set_pnode::const_iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        const IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric() || !p_curr_node->isHBFSupported())
            continue;

        for (u_int8_t port = 1;  port <= p_curr_node->numPorts ; ++port) {
             const IBPort *p_port = p_curr_node->getPort(port);

             // the following counters shouldn't be printed for special ports, unconnected SW ports
             if (!p_port ||
                 p_port->isSpecialPort() ||
                 !p_port->p_remotePort ||
                 !p_port->p_remotePort->p_node)
                continue;

             const port_routing_decision_counters* p_routing_decision_counters =
                    this->fabric_extended_info.getRoutingDecisionCounters(p_port->createIndex);

             if (!p_routing_decision_counters)
                 continue;

             sstream.str("");
             sstream << PTR(p_curr_node->guid_get())   << ','
                    << PTR(p_port->guid_get())         << ','
                    << +p_port->num                    << ','
                    <<  p_routing_decision_counters->rx_pkt_forwarding_static    << ','
                    <<  p_routing_decision_counters->rx_pkt_forwarding_hbf       << ','
                    <<  p_routing_decision_counters->rx_pkt_forwarding_ar        << ','
                    <<  p_routing_decision_counters->rx_pkt_hbf_fallback_local   << ','
                    <<  p_routing_decision_counters->rx_pkt_hbf_fallback_remote  << ','
                    <<  p_routing_decision_counters->rx_pkt_forwarding_hbf_sg0   << ','
                    <<  p_routing_decision_counters->rx_pkt_forwarding_hbf_sg1   << ','
                    <<  p_routing_decision_counters->rx_pkt_forwarding_hbf_sg2   << ','
                    <<  p_routing_decision_counters->rx_pkt_forwarding_ar_sg0    << ','
                    <<  p_routing_decision_counters->rx_pkt_forwarding_ar_sg1    << ','
                    <<  p_routing_decision_counters->rx_pkt_forwarding_ar_sg2    << endl;

             csv_out.WriteBuf(sstream.str());
        }
    }

    csv_out.DumpEnd(SECTION_HBF_PORT_COUNTRERS);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}
