/*
 * Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
#include "infiniband/ibutils/output_stream_primitives.h"
#include "infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h"
#include <infiniband/ibdiag/ibdiag_dfp.h>
using namespace std;

int DFPTopology::DumpToStream(ostream& stream) const
{
    IBDIAGNET_ENTER;

    if (this->p_nonComputeIsland)
       stream << "DFP island: " << this->p_nonComputeIsland->GetId()
              << " has less roots as the rest of islands"
              << " and will be considered as non-compute one"
              << endl;

     stream << endl;
    //firstly we dump islands internal data
    for (size_t i = 0; i < this->islands.size(); i++) {
         const DFPIsland *p_island = islands[i];
         if (!p_island) {
             ERR_PRINT("Cannot dump a DFP island: NULL pointer\n");
             IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
         }

         int rc = p_island->DumpToStream(stream);
         if (rc)
             IBDIAGNET_RETURN(rc);
    }

    //secondly we dump iland's connectivity table
    for (size_t i = 0; i < this->islands.size(); i++) {
         const DFPIsland *p_island = islands[i];
         if (!p_island) {
             ERR_PRINT("Cannot provide connectivity details of DFP island: NULL pointer\n");
             IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
         }

         int rc = p_island->ConnectivityDetailsToStream(stream);
         if (rc)
             IBDIAGNET_RETURN(rc);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

DFPTopology::~DFPTopology()
{
    IBDIAGNET_ENTER;

    for (size_t i = 0; i < this->islands.size(); i++) {
         DFPIsland *p_island = islands[i];
         delete p_island;
    }
    islands.clear();

    IBDIAGNET_RETURN_VOID;
}

int DFPTopology::Build(map_guid_pnode &dfp_roots)
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;
    set_pnode visited_switches;
    int index = 0;
    for (map_guid_pnode::iterator it = dfp_roots.begin(); it != dfp_roots.end(); it++) {
         IBNode *p_root = it->second;
         if (visited_switches.find(p_root) != visited_switches.end())
             continue;

         DFPIsland *p_island = new DFPIsland(*this, index++);
         this->islands.push_back(p_island);
         queue_p_node queue_switches;
         queue_switches.push(p_root);
         while (!queue_switches.empty()) {
                IBNode *p_switch = queue_switches.front();
                queue_switches.pop();
                std::pair < set_pnode::iterator, bool > ret = visited_switches.insert(p_switch);
                // If switch is already visited
                if (!ret.second)
                    continue;

                bool is_root = (dfp_roots.find(p_switch->guid_get()) != dfp_roots.end())
                                     ? true : false;
                if (is_root)
                    p_island->AddRoot(p_switch);
                else
                    p_island->AddLeaf(p_switch);

                for (u_int32_t i = 1 ; i <=  p_switch->numPorts ; i++) {
                     IBPort *p_curr_port = p_switch->getPort((phys_port_t)i);
                     if (!p_curr_port)                       //port not exist
                         continue;
                     //Skip ports that are down or have undefined state
                     //or are not in the scope of the fabric.
                     if (!p_curr_port->is_data_worthy(DATA_WORTHY_GMP_MAD))
                         continue;
                     if (!p_curr_port->p_remotePort || !p_curr_port->p_remotePort->p_node)
                         continue;
                     IBNode *p_remote_switch = p_curr_port->p_remotePort->p_node;
                     if (p_remote_switch->type != IB_SW_NODE)
                         continue;

                     this->edges.Add(p_switch, p_remote_switch);

                     if (is_root) {
                         if (dfp_roots.find(p_remote_switch->guid_get()) == dfp_roots.end())
                             queue_switches.push(p_remote_switch);
                     } else {
                         queue_switches.push(p_remote_switch);
                     }
                }
         }
    }

    IBDIAGNET_RETURN(rc);
}

bool DFPTopology::IsConnected(const IBNode *n1, const IBNode *n2) const
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(this->edges.Contains(n1, n2));
}

int DFPTopology::Validate(u_int32_t &warnings, u_int32_t &errors)
{
    IBDIAGNET_ENTER;

    if (islands.empty()) {
        ERR_PRINT("No island was found in the DFP topology\n");
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    INFO_PRINT("%lu DFP islands were detected\n", this->islands.size());
    this->FindNonComputeIsland(errors);

    int rc = IBDIAG_SUCCESS_CODE;
    LOG_PRINT("\n");
    for (size_t i = 0; i < this->islands.size(); i++) {
         DFPIsland *p_island = this->islands[i];
         if (!p_island) {
             ERR_PRINT("Cannot validate NULL pointer DFP island\n");
             IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
         }

         rc = p_island->Validate(warnings, errors);
         if (rc)
             IBDIAGNET_RETURN(rc);
    }

    bool isConnected = true;
    rc = this->CheckTopologyConnectivity(warnings, errors, isConnected);
    if (rc)
        IBDIAGNET_RETURN(rc);

    bool isSymmetric = false;
    if (isConnected) {
        rc = this->CheckTopologySymmetric(warnings, errors, isSymmetric);
        if (rc)
            IBDIAGNET_RETURN(rc);
    }
    else
        ERR_PRINT("DFP Non symmetrical switch connectivity discovered\n");

    //The "resilient check" excludes the non-compute island from the topology.
    //So, if a compute island is not connected to a non-compute one
    //we could still have positive "resilient checking" result.
    rc = this->ResilientReport();
    if (rc)
        IBDIAGNET_RETURN(rc);

    rc = this->BandwidthReport(errors);
    if (rc)
        IBDIAGNET_RETURN(rc);

    //The "check medium" excludes the non-compute island from the topology.
    //So, if a compute island is not connected to a non-compute one
    //we could still have positive "medium checking" result.
    //But a non-connected network cannot be medium and it needs recabeling
    if (!isConnected)
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);

    bool isMedium = true;
    bool couldBeMedium = true;
    //The calculation of "medium" is based on the fact that any island
    //is connected to all others. So, we exclude a non-compute island from the test and
    //simply compare switch's free ports with number of compute islands it is not connected yet.
    //It is getting more complicated if a non-compute island is not connected to a compute one.
    //Firstly -- it will need free ports to connect to the compute island.
    //Secondly -- the copmute islands also needs a port to connect to the non-compute one.
    rc = this->CheckMediumTopology(warnings, errors, isMedium, couldBeMedium);
    if (rc)
        IBDIAGNET_RETURN(rc);

    if (isMedium)
        INFO_PRINT("DFP medium topology was detected\n");
    else {
        if (!isSymmetric) { //not symmetric
            if (couldBeMedium)
                INFO_PRINT("Topology size enables medium DFP connectivity\n");
            else
                INFO_PRINT("DFP large topology was detected\n");
        } else { //symmetrical
            if (couldBeMedium) {  //symmetric and could be medium
                warnings++;
                WARN_PRINT("DFP large topology was detected; may be implemented as medium\n");
            } else
                INFO_PRINT("DFP large topology was detected\n");
        }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int DFPTopology::FillIslandsSizeMap(islands_size_map &islandsSizeMap, u_int32_t &errors) const
{
    IBDIAGNET_ENTER;

    for (size_t i = 0; i < this->islands.size(); i++) {
         DFPIsland *p_island = this->islands[i];
         if (!p_island) {
             errors++;
             ERR_PRINT("Cannot count islands roots: one of DFP islands is NULL\n");

             IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
         }

         islandsSizeMap[p_island->GetRoots().size()].push_back(p_island);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DFPTopology::IslandsToStream(ostream& stream, const dfp_islands_vec &vec) const
{
    IBDIAGNET_ENTER;

    dfp_islands_vec::const_iterator itr_last =  --vec.end();
    for (dfp_islands_vec::const_iterator vecItr = vec.begin(); vecItr != itr_last; vecItr++)
         stream << (*vecItr)->GetId() << ',';

    stream << (*itr_last)->GetId();

    IBDIAGNET_RETURN_VOID;
}

int DFPTopology::FindNonComputeIsland(u_int32_t &errors)
{
    IBDIAGNET_ENTER;

    if (this->islands.size() <= 1)
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);

    islands_size_map sizePerIslandsMap;
    int rc = this->FillIslandsSizeMap(sizePerIslandsMap, errors);
    if (rc)
       IBDIAGNET_RETURN(rc);

    if (sizePerIslandsMap.size() == 1)
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);

    dfp_islands_vec &minSizeVec = sizePerIslandsMap.begin()->second;
    if (minSizeVec.size() > 1) {
        ERR_PRINT("Some of the islands have less roots than others\n");
        errors++;

        for (std::map<size_t, dfp_islands_vec>::const_iterator itr = sizePerIslandsMap.begin();
             itr != sizePerIslandsMap.end(); itr++) {

            const dfp_islands_vec &vec = itr->second;
            stringstream stream;
            this->IslandsToStream(stream, itr->second);

            LOG_ERR_PRINT("\t%s: (%s) %s %llu roots per island\n",
                        (vec.size() > 1 ? "islands" : "island"), stream.str().c_str(),
                        (vec.size() > 1 ? "have" : "has"), itr->first);
        }
    }
    else
        this->p_nonComputeIsland = *minSizeVec.begin();

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int DFPTopology::CheckMediumTopology(u_int32_t &warnings, u_int32_t &errors,
                        bool &isMedium, bool &couldBeMedium)
{
    IBDIAGNET_ENTER;

    LOG_PRINT("\n");

    isMedium = true;
    couldBeMedium = true;

    for (size_t i = 0; i < this->islands.size(); i++) {
        const DFPIsland *p_island = this->islands[i];
        if (!p_island) {
           ERR_PRINT("Cannot check medium topology: one of DFP islands is NULL\n");
           IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Checkimg Medium Topology for DFP island id=%d\n",
                p_island->GetId());

        if (p_island == this->p_nonComputeIsland)
           continue;

        bool isIslandMedium = true;
        bool islandToMedium = true;
        int rc = p_island->CheckMedium(this->p_nonComputeIsland,
                                this->islands.size(), isIslandMedium, islandToMedium);
        if (rc) {
            errors++;
            isMedium = false;
            couldBeMedium = false;
            IBDIAGNET_RETURN(rc);
        }

        isMedium = isMedium && isIslandMedium;
        couldBeMedium = couldBeMedium && islandToMedium;
    }

    if (isMedium)
        couldBeMedium = false;

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int DFPTopology::CheckTopologyConnectivity(u_int32_t &warnings,
                                u_int32_t &errors, bool &isConnected)
{
    IBDIAGNET_ENTER;

    LOG_PRINT("\n");

    PairsContainer<const DFPIsland*> checkedPairs;

    for (size_t i = 0; i < this->islands.size(); i++) {
        DFPIsland *p_i = this->islands[i];
        if (!p_i) {
           ERR_PRINT("Cannot check topology connectivity: one of DFP islands is NULL\n");
           IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        for (size_t j = 0; j < this->islands.size(); j++) {
             DFPIsland *p_j = this->islands[j];
             if (!p_j) {
               ERR_PRINT("Cannot check topology connectivity: one of DFP islands is NULL\n");
               IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
             }

             if (p_i == p_j || checkedPairs.Contains(p_i, p_j))
                continue;

             checkedPairs.Add(p_i, p_j);

             bool isIslandConnected = false;
             int rc = p_i->FillConnectivityData(*p_j, isIslandConnected);
             if (rc)
                 IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);

             if (!isIslandConnected) {
                 errors++;
                 isConnected = false;
                 ERR_PRINT("DFP island-%d and island-%d are not directly connected\n",
                        p_i->GetId(), p_j->GetId());
             }

             rc = p_j->FillConnectivityData(*p_i);
             if (rc)
                 IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
        }

        //if a root switch is not connected to any island related to it
        //NodeData.resilientToAll will be "true". reset it to "false" in this case.
        p_i->UpdateResilient();
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int DFPTopology::CheckTopologySymmetric(u_int32_t &warnings, u_int32_t &errors,
                                bool &isSymmetric) const
{
    IBDIAGNET_ENTER;

    global_link_id_map linksToIdMap;
    for (size_t i = 0; i < this->islands.size(); i++) {
         const DFPIsland *p_island = islands[i];
         if (!p_island) {
             ERR_PRINT("Cannot check DFP symmetric -- NULL pointer DFP island\n");
             IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
         }

         int globalLinks = p_island->CountGlobalLinks(this->p_nonComputeIsland, warnings);

         if (p_island != this->p_nonComputeIsland)
             linksToIdMap[globalLinks].push_back(p_island->GetId());
    }

    if (linksToIdMap.size() == 1) {
        isSymmetric = true;
        INFO_PRINT("DFP Symmetrical switch connectivity discovered, global links per island: %d\n",
                linksToIdMap.begin()->first);

        if (this->p_nonComputeIsland)
            LOG_INFO_PRINT("One island (island-%d) has less roots as the rest of islands\n",
                        this->p_nonComputeIsland->GetId());
    } else if (linksToIdMap.size() > 1) {
        errors++;
        isSymmetric = false;
        ERR_PRINT("DFP Non symmetrical switch connectivity discovered\n");
        this->ExternalLinksReport(linksToIdMap);

        //the non compute islands is included in symmetry calculations.
        //we would like to report on islands roots.
        if (this->p_nonComputeIsland) {
            int rc = this->IslandRootsReport(errors);
            if (rc)
                IBDIAGNET_RETURN(rc);
        }
    } else {
        errors++;
        isSymmetric = false;
        ERR_PRINT("Failed to check DFP symmetrical connectivity\n");
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int DFPTopology::IslandRootsReport(u_int32_t &errors) const
{
    IBDIAGNET_ENTER;

    std::map<size_t, dfp_islands_vec> sizePerIslandsMap;
    int rc = this->FillIslandsSizeMap(sizePerIslandsMap, errors);
    if (rc)
        IBDIAGNET_RETURN(rc);

    if (sizePerIslandsMap.empty()) {
        errors++;
        ERR_PRINT("Cannot report on islands roots\n");

        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    LOG_INFO_PRINT("Different number of roots per island found:\n");
    for (std::map<size_t, dfp_islands_vec>::const_iterator itr = sizePerIslandsMap.begin();
         itr != sizePerIslandsMap.end(); itr++) {

        const dfp_islands_vec &vec = itr->second;
        stringstream stream;
        this->IslandsToStream(stream, vec);

        LOG_INFO_PRINT("\t%s: (%s) %s %llu roots per island\n",
                    (vec.size() > 1 ? "islands" : "island"), stream.str().c_str(),
                    (vec.size() > 1 ? "have" : "has"), itr->first);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DFPTopology::ExternalLinksReport(const global_link_id_map& linksToIdMap) const
{
    IBDIAGNET_ENTER;

    LOG_ERR_PRINT("Different number of global links per island found:\n");
    for (global_link_id_map::const_iterator itr = linksToIdMap.begin();
         itr != linksToIdMap.end(); itr++) {
        const id_list& ids = itr->second;

        stringstream stream;
        id_list::const_iterator itr_last =  --ids.end();
        for (id_list::const_iterator idItr = ids.begin(); idItr != itr_last; idItr++)
             stream << *idItr << ',';

        stream << *itr_last;

        LOG_ERR_PRINT("\t%s: (%s) %s %d global links per island\n",
                    (ids.size() > 1 ? "islands" : "island"), stream.str().c_str(),
                    (ids.size() > 1 ? "have" : "has"), itr->first);
    }

    IBDIAGNET_RETURN_VOID;
}

int DFPTopology::ResilientReport() const
{
    IBDIAGNET_ENTER;

    bool topoResilient = true;
    bool topoPartialResilient = false;
    for (size_t i = 0; i < this->islands.size(); i++) {
         const DFPIsland *p_island = islands[i];
         if (!p_island) {
             ERR_PRINT("Cannot report on resilient connection -- NULL pointer DFP island\n");
             IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
         }

         if (p_island == this->p_nonComputeIsland)
             continue;

         bool islResilient = true;
         bool islPartialResilient = false;
         int rc = p_island->CheckResilient(this->p_nonComputeIsland,
                                        islResilient, islPartialResilient);
         if (rc)
             IBDIAGNET_RETURN(rc);

         topoResilient = topoResilient && islResilient;
         topoPartialResilient = topoPartialResilient || islPartialResilient;
    }

    if (topoResilient)
        INFO_PRINT("DFP resilient discovered\n");
    else if (topoPartialResilient)
             INFO_PRINT("Partially resilient DFP discovered\n");

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int DFPTopology::BandwidthReport(u_int32_t &errors) const
{
    IBDIAGNET_ENTER;

    LOG_PRINT("\n");

    bandwidth_id_map bandwidthToIdMap;
    for (size_t i = 0; i < this->islands.size(); i++) {
         const DFPIsland *p_island = islands[i];
         if (!p_island) {
             ERR_PRINT("Cannot calculate bandwidth -- NULL pointer DFP island\n");
             IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
         }

         bandwidthToIdMap[p_island->GetBandwidth()].push_back(p_island->GetId());
    }

    #define BANDWIDTH_FORMAT "%.0f"

    if (bandwidthToIdMap.size() == 1) {
        double bandwidth = bandwidthToIdMap.begin()->first;
        INFO_PRINT("All DFP islands have the same bandwidth: "
                BANDWIDTH_FORMAT " Gbps\n",
                bandwidth);
        INFO_PRINT("Theoretical DFP network bisection bandwidth: "
                BANDWIDTH_FORMAT " Gbps\n",
                this->CalculateNetworkBandwidth(bandwidth));
    } else if (bandwidthToIdMap.size() > 1) {
        double minBandwidth = bandwidthToIdMap.begin()->first;
        INFO_PRINT("Minimal island's bandwidth: " BANDWIDTH_FORMAT " Gbps (island-%d)\n",
                minBandwidth, bandwidthToIdMap.begin()->second.front());
        INFO_PRINT("Minimal DFP network theoretical bisection bandwidth: "
                BANDWIDTH_FORMAT " Gbps\n", this->CalculateNetworkBandwidth(minBandwidth));

        double maxBandwidth = (bandwidthToIdMap.rbegin())->first;
        INFO_PRINT("Maximal island's bandwidth: " BANDWIDTH_FORMAT " Gbps (island-%d)\n",
                maxBandwidth, bandwidthToIdMap.rbegin()->second.front());
        INFO_PRINT("Maximal DFP network theoretical bisection bandwidth: "
                BANDWIDTH_FORMAT " Gbps\n", this->CalculateNetworkBandwidth(maxBandwidth));
    } else {
        ERR_PRINT("Failed to calculated DFP network's bandwidth\n");
        errors++;
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

double DFPTopology::CalculateNetworkBandwidth(double bandwidth) const
{
    IBDIAGNET_ENTER;

    int size  = static_cast<int>(this->islands.size());

    IBDIAGNET_RETURN(bandwidth*(size + (size & 0x1))/(2*size - 1));
}

void DFPIsland::AddRoot(IBNode *p_root)
{
    IBDIAGNET_ENTER;

    this->nodes[p_root->guid_get()] = p_root;
    this->roots[p_root->guid_get()] = p_root;

    IBDIAGNET_RETURN_VOID;
}

void DFPIsland::AddLeaf(IBNode *p_leaf)
{
    IBDIAGNET_ENTER;

    this->nodes[p_leaf->guid_get()] = p_leaf;
    this->leaves[p_leaf->guid_get()] = p_leaf;

    IBDIAGNET_RETURN_VOID;
}

int DFPIsland::Validate(u_int32_t &warnings, u_int32_t &errors) const
{
    IBDIAGNET_ENTER;

    if (this->roots.empty()) {
        ERR_PRINT("The DFP island-%d has not roots\n", this->id);
        errors++;
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    int rc = this->CheckNotConnectedNodes(0, this->roots, warnings, errors);
    if (rc)
       IBDIAGNET_RETURN(rc);

    rc = this->CheckNotConnectedNodes(1, this->leaves, warnings, errors);
    if (rc)
       IBDIAGNET_RETURN(rc);

    //this should be modified for 3 level fat-tree
    rc = CheckFullyCoonnetedLeavesAndRoots(warnings, errors);
    if (rc)
        IBDIAGNET_RETURN(rc);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int DFPIsland::CheckNotConnectedNodes(int rank, const map_guid_pnode &map,
                    u_int32_t &warnings, u_int32_t &errors) const
{
    IBDIAGNET_ENTER;

    PairsContainer<const IBNode*> checkedPairs;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "DFP island id=%d checking connection between nodes with rank: %d\n",
               this->id, rank);

    for (map_guid_pnode::const_iterator itr_i = map.begin(); itr_i != map.end(); itr_i++) {
         const IBNode *p_node1 =  itr_i->second;
         if (!p_node1) {
             ERR_PRINT("Cannot validate DFP island-%d. "
                       "A node associated with the GUID: " U64H_FMT " is NULL\n",
                       this->id, itr_i->first);
             IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
         }

         for (map_guid_pnode::const_iterator itr_j = map.begin(); itr_j != map.end(); itr_j++) {
              const IBNode *p_node2 = itr_j->second;
              if (!p_node2) {
                  ERR_PRINT("Cannot validate DFP island-%d. "
                            "A node associated with the GUID: " U64H_FMT " is NULL\n",
                            this->id, itr_j->first);
                  IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
              }

              if (p_node1 == p_node2 || checkedPairs.Contains(p_node1, p_node2))
                  continue;

              IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "\tNodes "
                           "GUID=" U64H_FMT " and GUID=" U64H_FMT "\n",
                            p_node1->guid_get(), p_node2->guid_get());

              checkedPairs.Add(p_node1, p_node2);

              if (this->topology.IsConnected(p_node1, p_node2)) {
                  ERR_PRINT("DFP island-%d "
                           "invalid connection between switch ( GUID: " U64H_FMT " rank: %d ) "
                           "and switch ( GUID: " U64H_FMT " rank: %d )\n",
                           this->id, p_node1->guid_get(), rank, p_node2->guid_get(), rank);
                  errors++;
              }
         }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int DFPIsland::CheckFullyCoonnetedLeavesAndRoots(u_int32_t &warnings, u_int32_t &errors) const
{
    IBDIAGNET_ENTER;

    if (this->leaves.empty()) {
        errors++;
        ERR_PRINT("DFP island-%d has no leaves\n", this->id);
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
    }

    for (map_guid_pnode::const_iterator itr_leaf = this->leaves.begin();
        itr_leaf != this->leaves.end(); itr_leaf++) {
         const IBNode *p_leaf = itr_leaf->second;
         if (!p_leaf) {
            ERR_PRINT("Cannot validate DFP island-%d. "
                    "A leaf associated with the GUID: " U64H_FMT " is NULL\n",
                    this->id, itr_leaf->first);
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
         }

        for (map_guid_pnode::const_iterator itr_root = this->roots.begin();
            itr_root != roots.end(); itr_root++) {
             const IBNode *p_root = itr_root->second;
             if (!p_root) {
                 ERR_PRINT("Cannot validate DFP island-%d. "
                        "A root associated with the GUID: " U64H_FMT " is NULL\n",
                         this->id, itr_root->first);
                 IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
             }

             if (!this->topology.IsConnected(p_leaf, p_root)) {
                 errors++;
                 ERR_PRINT("DFP island-%d "
                        "missing connection between switch ( GUID: " U64H_FMT " rank: 0 ) "
                        "and switch ( GUID: " U64H_FMT " rank: 1 )\n",
                        this->id, p_root->guid_get(), p_leaf->guid_get());
             }
        }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int DFPIsland::CountGlobalLinks(const DFPIsland *p_nonComputeIsland, u_int32_t &warnings) const
{
    IBDIAGNET_ENTER;

    int globalLinks = 0;
    std::set<int> linksSet;
    for (connectivity_map::const_iterator itr_node = this->rootsConnectivityMap.begin();
        itr_node != this->rootsConnectivityMap.end(); itr_node++) {
         const remote_islands_map &map = itr_node->second.remoteIslandsMap;

         int switchGlobalLinks = 0;
         for (remote_islands_map::const_iterator itr_ils = map.begin();
             itr_ils != map.end(); itr_ils++) {

             globalLinks += itr_ils->second.globalLinks;
             switchGlobalLinks += itr_ils->second.globalLinks;
         }

        linksSet.insert(switchGlobalLinks);
    }

    if (linksSet.size() > 1 && this != p_nonComputeIsland) {
        stringstream stream;
        set<int>::const_iterator last_itr = --linksSet.end();
        for (set<int>::const_iterator itr = linksSet.begin(); itr != last_itr; itr++)
            stream << *itr << ',';

        stream << *last_itr;

        warnings++;
        LOG_WARN_PRINT("DFP island-%d has different number of global links: (%s) per switches\n",
                    this->id, stream.str().c_str());
    }

    IBDIAGNET_RETURN(globalLinks);
}

void DFPIsland::UpdateResilient()
{
    for (connectivity_map::iterator itr = this->rootsConnectivityMap.begin();
         itr != this->rootsConnectivityMap.end(); itr++) {
        if (itr->second.globalLinksToAll == 0)
            itr->second.resilientToAll = false;
    }
}

int DFPIsland::FillConnectivityData(const DFPIsland &other)
{
    IBDIAGNET_ENTER;

    bool unused = false;
    int rc = FillConnectivityData(other, unused);

    IBDIAGNET_RETURN(rc);

}

int DFPIsland::FillConnectivityData(const DFPIsland &other, bool &isConnected)
{
    IBDIAGNET_ENTER;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "DFP islands id=%d and id=%d are being checked for connectivity\n",
               this->id, other.id);

    isConnected = false;

    for (map_guid_pnode::const_iterator itr_mine = this->roots.begin();
         itr_mine != this->roots.end(); itr_mine++) {
        const IBNode *p_mine = itr_mine->second;
        if (!p_mine) {
            ERR_PRINT("Cannot check DFP island: %d connections"
                    "A root associated with the GUID: " U64H_FMT " is NULL\n",
                    this->id, itr_mine->first);
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        NodeData &nodeData = this->rootsConnectivityMap[p_mine];

        int freePortsCount = 0;
        bool resilient = false;
        set<const IBNode*> connectedNodes;
        for (phys_port_t i = 1 ; i <= p_mine->numPorts; i++) {
            const IBPort *p_port = p_mine->getPort(i);

            if (nodeData.freePorts < 0 && p_port && !p_port->p_remotePort)
                freePortsCount++;

            if (!p_port || !p_port->p_remotePort || !p_port->p_remotePort->p_node)
                continue;

            const IBNode *p_remoteNode = p_port->p_remotePort->p_node;
            if (p_remoteNode->type != IB_SW_NODE ||
                other.nodes.find(p_remoteNode->guid_get()) == other.nodes.end())
                continue;

            isConnected = true;
            connectedNodes.insert(p_remoteNode);

            RemoteIslandConnectivityData &remoteIslandData = nodeData.remoteIslandsMap[other.GetId()];
            remoteIslandData.globalLinks++;
            nodeData.globalLinksToAll++;
            if (connectedNodes.size() >= 2) {
                remoteIslandData.resilient = true;
                resilient = true;
            }

            this->bandwidth += speed2double(p_port->get_internal_speed());
        }

        if (nodeData.freePorts < 0)
            nodeData.freePorts = freePortsCount;

        if (!connectedNodes.empty())
            nodeData.resilientToAll = nodeData.resilientToAll && resilient;
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int DFPIsland::CheckMedium(const DFPIsland *p_nonComputeIsland,
                    size_t islandsCount, bool &isMedium, bool &toMedium) const
{
    IBDIAGNET_ENTER;

    isMedium = true;
    toMedium = true;

    for (connectivity_map::const_iterator itr_node = this->rootsConnectivityMap.begin();
         itr_node != this->rootsConnectivityMap.end(); itr_node++) {
        const NodeData &nodeData = itr_node->second;

        size_t mapSize = nodeData.remoteIslandsMap.size();
        int delta = 0;
        if (!p_nonComputeIsland)
            delta = static_cast<int>(islandsCount - 1 - mapSize);
        else {
            if (nodeData.remoteIslandsMap.find(p_nonComputeIsland->GetId()) !=
                nodeData.remoteIslandsMap.end())
                 delta = static_cast<int>(islandsCount - 1 - mapSize);
            else
                 delta = static_cast<int>(islandsCount - 2 - mapSize);
        }

        if (delta < 0) {
            ERR_PRINT("Wrong connectivity data for the island=%d\n", this->id);
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
        } else if (delta > 0) {
            isMedium = false;
            toMedium = toMedium && (nodeData.freePorts >= delta);
        }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int DFPIsland::CheckResilient(const DFPIsland* p_nonComputeIsland,
                            bool &islResilient, bool &islPartialResilient) const
{
    IBDIAGNET_ENTER;

    islResilient = true;
    islPartialResilient = false;

    for (connectivity_map::const_iterator itr_node = this->rootsConnectivityMap.begin();
         itr_node != this->rootsConnectivityMap.end(); itr_node++) {

        //skip updating if a root is only connected to a non-compute island
        //or is not connected to any island at all
        const remote_islands_map &remoteIslandsMap = itr_node->second.remoteIslandsMap;
        if (remoteIslandsMap.empty() || (remoteIslandsMap.size() == 1 &&
            p_nonComputeIsland && remoteIslandsMap.begin()->first == p_nonComputeIsland->GetId()))
            continue;

        bool rootPartial = true;
        for (remote_islands_map::const_iterator itr_isl = remoteIslandsMap.begin();
             itr_isl != remoteIslandsMap.end(); itr_isl++) {

            if (p_nonComputeIsland && p_nonComputeIsland->GetId() == itr_isl->first)
                continue;

            if (!itr_isl->second.resilient) {
                islResilient = rootPartial = false;
                break;
            }
        }

        islPartialResilient = islPartialResilient || rootPartial;
    }

    //consider the situation: a compute island is connected only to a non compute one.
    //we skip checking of the non compute island and may end with saying
    //that the island is resiliently connected.
    islResilient = islResilient && islPartialResilient;

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int DFPIsland::DumpToStream(ostream& stream) const
{
    IBDIAGNET_ENTER;

    stream << "island: " << id  << endl;
    int rc = this->DumpNodesToStream(stream, 0, this->roots);
    if (rc)
       IBDIAGNET_RETURN(rc);

    rc = this->DumpNodesToStream(stream, 1, this->leaves);
    if (rc)
       IBDIAGNET_RETURN(rc);

    stream << endl << endl;

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int DFPIsland::DumpNodesToStream(ostream& stream,
                        int rank, const map_guid_pnode&  map) const
{
    IBDIAGNET_ENTER;

    stream << " rank: " << rank
           << " (size: " << map.size() << ')'
           << endl;

    for (map_guid_pnode::const_iterator itr = map.begin(); itr != map.end(); itr++) {
        const IBNode* p_node = itr->second;
        const char *ptr = p_node ?  p_node->getName().c_str() : "NULL is associated to the guid";
        stream << "\t" << PTR(itr->first) << ' ' <<  ptr  << endl;
    }

    stream << endl;

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int DFPIsland::ConnectivityDetailsToStream(ostream& stream) const
{
    IBDIAGNET_ENTER;

    int globalLinksToAll = 0;
    bool  resilientToAll = true;

    for (connectivity_map::const_iterator itr = this->rootsConnectivityMap.begin();
         itr != this->rootsConnectivityMap.end(); itr++) {
         globalLinksToAll += itr->second.globalLinksToAll;
         resilientToAll = resilientToAll && itr->second.resilientToAll;
    }

    stream << endl << "island " << this->id
           << ", bandwidth " << this->bandwidth << " [Gb/s]"
           << ", global links " << globalLinksToAll
           << ", resilient connection to all: "
           << (globalLinksToAll ? BOOL_TO_YES_NO_STR(resilientToAll) : BOOL_TO_YES_NO_STR(false))
           << endl;

    for (connectivity_map::const_iterator itr = this->rootsConnectivityMap.begin();
         itr != this->rootsConnectivityMap.end(); itr++) {
         const IBNode *p_node = itr->first;
         if (!p_node) {
            ERR_PRINT("Cannot provide connectivity details of a switch, NULL pointer\n");
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
         }

         const NodeData &nodeData = itr->second;
         stream << "\t" << "switch " << PTR(p_node->guid_get())
                << ", global links: " << nodeData.globalLinksToAll
                << ", resilient connection to all: " << BOOL_TO_YES_NO_STR(nodeData.resilientToAll)
                << ", connected islands: " << nodeData.remoteIslandsMap.size()
                << ", free ports: " << nodeData.freePorts
                << endl;

         for (remote_islands_map::const_iterator itr_isl = nodeData.remoteIslandsMap.begin();
              itr_isl != nodeData.remoteIslandsMap.end(); itr_isl++){
              const RemoteIslandConnectivityData& remoteIslandData= itr_isl->second;
              stream << "\t\t" << "island: " <<  itr_isl->first
                     << ", global links: " << remoteIslandData.globalLinks
                     << ", resilient: " << BOOL_TO_YES_NO_STR(remoteIslandData.resilient)
                     << endl ;
         }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}
