/*--
* SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
* SPDX-License-Identifier: LicenseRef-NvidiaProprietary
*
* NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
* property and proprietary rights in and to this material, related
* documentation and any modifications thereto. Any use, reproduction,
* disclosure or distribution of this material and related documentation
* without an express license agreement from NVIDIA CORPORATION or
* its affiliates is strictly prohibited.
--*/

#include <math.h>

#include "ibis/key_mngr.h"
#include "infiniband/ibdiag/ibdiag_types.h"
#include "infiniband/ibdiag/key_updater.h"
#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibdm/Fabric.h"
#include "infiniband/ibdm/IBDiagnetFileNames.h"
#include "infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h"

KeyUpdater::KeyUpdater(IBDiag* p_ibdiag) : p_ibdiag(p_ibdiag)
{

}

KeyUpdater::~KeyUpdater() {}

static string getDefaultFileName(IBISKeyType type)
{
    switch (type) {
        case IBIS_AM_KEY:  return "";
        case IBIS_VS_KEY:  return GUID2VSKEY_FILE_NAME;
        case IBIS_CC_KEY:  return GUID2CCKEY_FILE_NAME;
        case IBIS_CLASS_C_KEY: return GUID2_M2N_KEY_FILE_NAME;
        case IBIS_PM_KEY:  return GUID2_PM_KEY_FILE_NAME;
        default : return "";
    }
}

static set <IBNodeType> getFilter(IBISKeyType type)
{
    switch (type) {
        case IBIS_CC_KEY:  return {IB_RTR_NODE};
        case IBIS_CLASS_C_KEY: return {IB_RTR_NODE, IB_CA_NODE};
        case IBIS_PM_KEY:  return {IB_RTR_NODE};
        default : return {};
    }
}

void KeyUpdater::SetKeyPerPort(const map_guid_2_key_t &guid_2_key, IBISKeyType key_type,
                                   const string &key_type_str, const set <IBNodeType> &filter)
{
    IBDIAGNET_ENTER;
    
    if (!p_ibdiag)
        IBDIAGNET_RETURN_VOID;
    if (!p_ibdiag->GetDiscoverFabricPtr() || guid_2_key.empty())
        IBDIAGNET_RETURN_VOID;

    set < uint64_t > checked_guids;
    int count = 0;
    const int limit = 5;

    for (map_str_pnode::const_iterator itr = this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.begin();
            itr != this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.end(); itr++) {
        const IBNode *p_node = itr->second;
        if (!p_node)
           continue;

        phys_port_t start = (p_node->type == IB_SW_NODE ? 0 : 1);
        phys_port_t end = (p_node->type == IB_SW_NODE ? 0 : p_node->numPorts);
        for (phys_port_t i = start; i <= end; ++i) {
            const IBPort *p_port = p_node->getPort(i);
            if (!p_port)
                continue;

            if (!p_port->guid_get())
                continue;

            checked_guids.insert(p_port->guid_get());

            map_guid_2_key_t::const_iterator itr_key = guid_2_key.find(p_port->guid_get());
            if (itr_key == guid_2_key.end()) {
                //we have filter here because we are collecting all port guids
                if (filter.find(p_node->type) == filter.end()) {
                    if (count++ < limit)
                        ERR_PRINT("%s - Port GUID: " U64H_FMT " is not associated with any key in fabric\n",
                                key_type_str.c_str(), p_port->guid_get());
                    else
                        LOG_ERR_PRINT("%s - Port GUID: " U64H_FMT " is not associated with any key in fabric\n",
                                    key_type_str.c_str(), p_port->guid_get());
                }

                continue;
            }

            lid_t range = (lid_t)(p_port->base_lid + pow(2, p_port->lmc));
            for (lid_t lid = p_port->base_lid; lid < range; lid++)
                this->p_ibdiag->GetIbisPtr()->GetKeyManager().SetKey(lid, key_type,
                                                                     itr_key->second);
        }
    }

    if (count > limit)
        ERR_PRINT("%s - were found %d Port GUIDs that are not associated with any key in fabric\n",
                        key_type_str.c_str(), count);

    count = 0;
    for (map_guid_2_key_t::const_iterator itr_guids = guid_2_key.begin();
            itr_guids != guid_2_key.end(); itr_guids++) {
        set < uint64_t >::const_iterator itr_found = checked_guids.find(itr_guids->first);
        if (itr_found != checked_guids.end())
            continue;

        if (count++ < limit)
            WARN_PRINT("%s - Port GUID: " U64H_FMT " doesn't exist in fabric\n",
                        key_type_str.c_str(), itr_guids->first);
        else
            LOG_WARN_PRINT("%s - Port GUID: " U64H_FMT " doesn't exist in fabric\n",
                        key_type_str.c_str(), itr_guids->first);
    }

    if (count > limit)
        WARN_PRINT("%s - were found %d Port GUIDs that do not exist in fabric\n",
                        key_type_str.c_str(), count); 

    IBDIAGNET_RETURN_VOID;
}

void KeyUpdater::SetAMKeyPerPort(const map_guid_2_key_t &guid_2_key)
{
    IBDIAGNET_ENTER;

    if (!p_ibdiag)
        IBDIAGNET_RETURN_VOID;

    if (guid_2_key.empty())
        IBDIAGNET_RETURN_VOID;

    set < uint64_t > checked_guids;
    int count = 0;
    const int limit = 5;

    for (map_str_pnode::const_iterator itr = this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.begin();
            itr != this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.end(); itr++) {
        const IBNode *p_node = itr->second;
        if (!p_node)
           continue;

        phys_port_t start = (p_node->type == IB_SW_NODE ? 0 : 1);
        phys_port_t end = (p_node->type == IB_SW_NODE ? 0 : p_node->numPorts);
        for (phys_port_t i = start; i <= end; ++i) {
            IBPort *p_port = p_node->getPort(i);
            if (!p_port)
                continue;

            if (!p_port->guid_get())
                continue;

            checked_guids.insert(p_port->guid_get());

            map_guid_2_key_t::const_iterator itr_key =
                            guid_2_key.find(p_port->guid_get());
            if (itr_key == guid_2_key.end())
                continue;

            if (p_port->p_node && p_port->p_node->isSpecialNode())
                p_port->am_key.SetKey(itr_key->second);
            else
                if (count++ < limit)
                    WARN_PRINT("GUID 2 AM key - Port GUID: " U64H_FMT " doesn't belong to "
                           "Aggregation node\n",  p_port->guid_get());
                else
                    LOG_WARN_PRINT("GUID 2 AM key - Port GUID: " U64H_FMT " doesn't belong to "
                           "Aggregation node\n",  p_port->guid_get());
        }
    }

    if (count > limit)
        WARN_PRINT("GUID 2 AM key - were found %d Port GUIDs that do not belong to "
                "Aggregation node\n", count);

    count = 0;
    for (map_guid_2_key_t::const_iterator itr_guids = guid_2_key.begin();
            itr_guids != guid_2_key.end(); itr_guids++) {
        set < uint64_t >::const_iterator itr_found = checked_guids.find(itr_guids->first);
        if (itr_found != checked_guids.end())
            continue;

        if (count++ < limit)
            WARN_PRINT("GUID 2 AM key - Port GUID: " U64H_FMT " doesn't exist in fabric\n",
                       itr_guids->first);
        else
            LOG_WARN_PRINT("GUID 2 AM key - Port GUID: " U64H_FMT " doesn't exist in fabric\n",
                       itr_guids->first);
    }

    if (count > limit)
        WARN_PRINT("GUID 2 AM key - were found %d Port GUIDs that do not exist in fabric\n",
                count);

    IBDIAGNET_RETURN_VOID;
}

static int getFilePathAndStat(string& file_path, const string key_type_str,  
                              const string &default_file_name,bool should_print_error, 
                              struct stat& file_path_stat)
{
    // Check type of file: directory or regular file
    for (int i = 0; i < 2; i++) {
         if (stat(file_path.c_str(),&file_path_stat) == 0) {
             if ((file_path_stat.st_mode & S_IFDIR) && (i < 1)) {
                  file_path += SEPERATOR + default_file_name;
                  continue;
             } else if (!(file_path_stat.st_mode & S_IFREG)) {
                 if (i > 0) {
                     if (should_print_error)
                         ERR_PRINT("%s path with default file name: %s is not regular file\n",
                                   key_type_str.c_str(), file_path.c_str());
                     else
                         WARN_PRINT("%s path with default file name: %s is not regular file\n",
                                    key_type_str.c_str(), file_path.c_str());
                 } else {
                     if (should_print_error)
                     ERR_PRINT("%s path: %s is not regular file or directory\n",
                               key_type_str.c_str(), file_path.c_str());
                     else
                     WARN_PRINT("%s path: %s is not regular file or directory\n",
                                key_type_str.c_str(), file_path.c_str());

                 }
                 IBDIAGNET_RETURN(IBDIAG_ERR_CODE_IO_ERR);
             }
             break;
         } else {
             int ret_v = 0;
             int errsv = errno;
             char buff[1024];

             if (errsv == ENOENT) {
                 snprintf(buff, sizeof(buff), "%s file '%s' wasn't loaded, reason: %s\n",
                          key_type_str.c_str(), file_path.c_str(), strerror(errsv));

                 ret_v = IBDIAG_ERR_CODE_FILE_NOT_EXIST;
             } else {
                 snprintf(buff, sizeof(buff), "Failed to get %s file stat: %s, reason: %s\n",
                          key_type_str.c_str(), file_path.c_str(), strerror(errsv));

                 ret_v = IBDIAG_ERR_CODE_IO_ERR;
             }

             if (should_print_error)
                 ERR_PRINT("%s", buff);
             else
                 WARN_PRINT("%s", buff);

             IBDIAGNET_RETURN(ret_v);
         }
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int KeyUpdater::ParseGuid2Key(KeyEntry& key_entry,
                             string &key_type_str, bool should_print_error,
                             const string &default_file_name, bool init)
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    struct stat file_path_stat;
    string file_path = key_entry.path;
    map_guid_2_key_t& guid_2_key = key_entry.key_map;

    rc = getFilePathAndStat(file_path, key_type_str, 
                            default_file_name, should_print_error,
                            file_path_stat);
    if (rc)
        IBDIAG_RETURN(rc);

    uint64_t last_modified_nsec = static_cast<uint64_t>(file_path_stat.st_mtim.tv_sec) * 1000000000 
                      + static_cast<uint64_t>(file_path_stat.st_mtim.tv_nsec);
    key_entry.last_modified = last_modified_nsec;
    IBDiag::PrintFileTimestamp(file_path, key_type_str);

    ifstream fs(file_path.c_str());
    if (fs.fail()) {
        ERR_PRINT("Failed to open %s file: %s\n", key_type_str.c_str(),
                  file_path.c_str());
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_FILE_NOT_OPENED);
    }

    // For every line in guid2key file, parse guid and its key.
    // Example :
    //  0x0002c90000000041 0x0000000000000001
    u_int64_t line_num = 0;
    //parse file
    while (fs.good()) {
        char line[1024] = {'\0'};
        fs.getline(line, 1024);
        line_num++;

        char* p_line = line;
        // skip whitespaces
        while(isspace(*p_line))
              ++p_line;

        // skip empty lines and comments
        if (!p_line[0] || p_line[0] == '#')
            continue;

        u_int64_t guid;
        u_int64_t key;
        char tail[100] = {0};
        if (sscanf(p_line, "0x%16" SCNx64 " 0x%16" SCNx64 " %s", &guid, &key, tail) != 2) {
             rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
             WARN_PRINT("Unsupported syntax at line " U64D_FMT ": %s\n", line_num, line);
        } else {
            if (init) {
                if (!guid_2_key.insert(pair< u_int64_t, u_int64_t > (guid, key)).second) {
                        WARN_PRINT("Multiple GUID: " U64H_FMT " in %s file: %s\n",
                                guid, key_type_str.c_str(), file_path.c_str());
                        rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
                }
            }
            else {
                guid_2_key[guid] = key;
            }
        }
    }

    fs.close();

    if (guid_2_key.empty()) {
        // If file is empty
        if (rc)
             WARN_PRINT("GUID 2 Key database of %s file is empty: %s\n",
                        key_type_str.c_str(), file_path.c_str());
        else
             WARN_PRINT("%s file is empty: %s\n",
                        key_type_str.c_str(), file_path.c_str());
        rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
    }

    IBDIAGNET_RETURN(rc);
}

bool KeyUpdater::IsKeyAvailable(IBISKeyType type)
{
    return (key_maps.find(type) != key_maps.end());
}

int KeyUpdater::SetKey(IBISKeyType type, const string& path)
{
    bool init = false;
    auto it = key_maps.find(type);
    if (it == key_maps.end()) {
        key_maps[type] = KeyEntry();
        key_maps[type].path = path;
        key_maps[type].last_modified = 0;
        init = true;
    }
    else {
        it->second.path = path;
    }
    string key_type_str = KeyManager::GetTypeName(type);
    IBDIAG_RETURN(ParseGuid2Key(key_maps[type],
                    key_type_str,
                    type == IBIS_AM_KEY,
                    getDefaultFileName(type), init));
}

void KeyUpdater::UpdateKeysPerPort(const std::set<IBISKeyType>& include_list)
{
    for (std::pair<const IBISKeyType, KeyEntry>& it : key_maps) {
        if (include_list.find(it.first) == include_list.end())
            continue;
        if (it.first == IBIS_AM_KEY) {
            SetAMKeyPerPort(it.second.key_map);
            continue;
        }
        string key_str_name = KeyManager::GetTypeName(it.first);
        SetKeyPerPort(it.second.key_map, 
                      it.first, key_str_name, 
                      getFilter(it.first));
    }
}

int KeyUpdater::UpdateSetKeysIfNeeded()
{
    int rc = 0;
    for (std::pair<const IBISKeyType, KeyEntry>& it : key_maps) {
        struct stat sb;
        string path = it.second.path;
        if (path.empty())
            continue;

        string key_type_str =  KeyManager::GetTypeName(it.first);
        rc = getFilePathAndStat(path, key_type_str, 
                    getDefaultFileName(it.first),
                    it.first == IBIS_AM_KEY,
                    sb);

        if (rc == IBDIAG_ERR_CODE_FILE_NOT_OPENED || rc == IBDIAG_ERR_CODE_IO_ERR)
            IBDIAG_RETURN(rc);

        uint64_t last_modified_nsec = static_cast<uint64_t>(sb.st_mtim.tv_sec) * 1000000000 
                      + static_cast<uint64_t>(sb.st_mtim.tv_nsec);

        if (last_modified_nsec <= it.second.last_modified)
            continue;
        it.second.last_modified = last_modified_nsec;
        
        rc = ParseGuid2Key(it.second, key_type_str,
                           it.first == IBIS_AM_KEY,
                           getDefaultFileName(it.first), false);

        if (rc == IBDIAG_ERR_CODE_FILE_NOT_OPENED || rc == IBDIAG_ERR_CODE_IO_ERR)
            IBDIAG_RETURN(rc);

        if (rc == IBDIAG_ERR_CODE_PARSE_FILE_FAILED) {
            WARN_PRINT("Parse %s file was finished with warnings\n\n", key_type_str.c_str());
        } else if (rc == IBDIAG_SUCCESS_CODE) {
            INFO_PRINT("Parse %s file was finished successfully\n\n", key_type_str.c_str());
        }

        if (it.first == IBIS_AM_KEY)
            SetAMKeyPerPort(it.second.key_map);
        else
            SetKeyPerPort(it.second.key_map, it.first, key_type_str, getFilter(it.first));

    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

