/*
 * Copyright (c) 2004-2020 Mellanox Technologies LTD. All rights reserved
 * Copyright (C) 2021-2024, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibdiag/ibdiag_clbck.h"

extern IBDiagClbck ibDiagClbck;


/****************************************************/
/******************private methods*******************/
/****************************************************/
void DumpSMInfoSection(ofstream &sout, list_p_sm_info_obj& p_sm_list, string header)
{
    IBDIAG_ENTER;

    char buffer[1024];
    set < APort* > aports_set;

    sout << endl << header << endl;

    for (auto const& it : p_sm_list) {

        IBPort* p_port = it->p_port;
        SMP_SMInfo sm_info = it->smp_sm_info;

        if (p_port->p_aport) {
            if (!aports_set.count(p_port->p_aport))
                aports_set.insert(p_port->p_aport);
            else
                continue;
        }

        CLEAR_STRUCT(buffer);
        snprintf(buffer, sizeof(buffer), "    Port=" U32D_FMT " lid=" U16H_FMT " guid=" U64H_FMT \
                 " dev=" U32D_FMT " priority:" U32D_FMT,
                 p_port->num,
                 p_port->base_lid,
                 p_port->guid_get(),
                 p_port->p_node->devId,
                 sm_info.Priority);
        sout << buffer << endl;
    }

    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpSMInfo(ofstream &sout)
{
    IBDIAG_ENTER;

    list_p_sm_info_obj p_sm_not_active_list;
    list_p_sm_info_obj p_sm_discover_list;
    list_p_sm_info_obj p_sm_standby_list;
    list_p_sm_info_obj p_sm_masters_list;
    list_p_sm_info_obj p_sm_unknown_list;

    sout << "ibdiagnet fabric SM report" << endl;

    //collect info
    for (list_p_sm_info_obj::iterator it = this->fabric_extended_info.getSMPSMInfoListRef().begin();
            it != this->fabric_extended_info.getSMPSMInfoListRef().end(); ++it) {
        switch ((*it)->smp_sm_info.SmState) {
        case IBIS_IB_SM_STATE_NOT_ACTIVE: {
            p_sm_not_active_list.push_back(*it);
            break;
        }
        case IBIS_IB_SM_STATE_DISCOVER: {
            p_sm_discover_list.push_back(*it);
            break;
        }
        case IBIS_IB_SM_STATE_STANDBY: {
            p_sm_standby_list.push_back(*it);
            break;
        }
        case IBIS_IB_SM_STATE_MASTER: {
            p_sm_masters_list.push_back(*it);
            break;
        }
        default: {
            p_sm_unknown_list.push_back(*it);
            break;
        }
        }
    }

    if (p_sm_masters_list.size() > 0)
        DumpSMInfoSection(sout, p_sm_masters_list, "  SM - master");

    if (p_sm_standby_list.size() > 0)
        DumpSMInfoSection(sout, p_sm_standby_list, "  SM - standby");

    if (p_sm_discover_list.size() > 0)
        DumpSMInfoSection(sout, p_sm_discover_list, "  SM - discover");

    if (p_sm_not_active_list.size() > 0)
        DumpSMInfoSection(sout, p_sm_not_active_list, "  SM - not active");

    if (p_sm_unknown_list.size() > 0)
        DumpSMInfoSection(sout, p_sm_unknown_list, "  SM - unknown state");

    p_sm_not_active_list.clear();
    p_sm_discover_list.clear();
    p_sm_standby_list.clear();
    p_sm_masters_list.clear();
    p_sm_unknown_list.clear();

    IBDIAG_RETURN_VOID;
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildSMInfo(list_p_fabric_general_err& ports_errors)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    // go over all nodes:
    struct SMP_SMInfo curr_sm_info;
    struct SMP_PortInfo *p_curr_port_info = NULL;

    CLEAR_STRUCT(curr_sm_info);

    ibDiagClbck.Set(this, &fabric_extended_info, &ports_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPSMInfoMadGetClbck>;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        u_int8_t start_port_idx;
        u_int8_t end_port_idx;
        if (p_curr_node->type == IB_SW_NODE) {
            start_port_idx = 0;
            end_port_idx = 0;   //only query manage port
        } else {   //CA
            start_port_idx = 1;
            end_port_idx = p_curr_node->numPorts;
        }
        // go over all ports of this node, if SW than query only manage port
        // use unsigned int index to prevent overwrap
        for (unsigned int port_num = start_port_idx; port_num <= end_port_idx; ++port_num) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)port_num);
            if (!p_curr_port)                                       //port not exist
                continue;
            //Skip ports that are down or have undefined state
            //or are not in the scope of the fabric.
            if ((port_num != 0) && !p_curr_port->is_data_worthy(DATA_WORTHY_SMP_MAD))
                continue;

            //get the port info for this port
            p_curr_port_info = this->fabric_extended_info.getSMPPortInfo(p_curr_port->createIndex);
            if (!p_curr_port_info) {
                stringstream ss;
                ss << "DB error - found connected port=" << p_curr_port->getName() << " without SMPPortInfo";

                ports_errors.push_back(
                        new FabricErrPortInfoFail(p_curr_node, p_curr_port->num, ss.str().c_str()));

                continue;
            }
            if (!(p_curr_port_info->CapMsk & 0x00000002))           //no sm port
                continue;

            //get direct route for this port
            direct_route_t *p_curr_direct_route = this->GetDR(p_curr_port);
            if (!p_curr_direct_route) {
                this->SetLastError("DB error - can't find direct route to node=%s", p_curr_node->getName().c_str());
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }

            clbck_data.m_data1 = p_curr_port;
            this->ibis_obj.SMPSMInfoMadGetByDirect(p_curr_direct_route,
                                                   &curr_sm_info, &clbck_data);

            if(ibDiagClbck.GetState())
                goto exit;

        }
    }
exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!ports_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED
int IBDiag::CheckSMInfo(list_p_fabric_general_err& sm_errors)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    list_p_sm_info_obj p_sm_masters_list;
    list_p_sm_info_obj p_sm_not_masters_list;
    set < APort* >     aports_set;

    //find all master sm
    for (list_p_sm_info_obj::iterator it = this->fabric_extended_info.getSMPSMInfoListRef().begin();
            it != this->fabric_extended_info.getSMPSMInfoListRef().end(); ++it) {

        sm_info_obj_t* cur_sm_obj = *it;
        IBPort* p_port = cur_sm_obj->p_port;
        SMP_SMInfo sm_info = cur_sm_obj->smp_sm_info;

        if ( (sm_info.SmState == IBIS_IB_SM_STATE_NOT_ACTIVE) ||
                (sm_info.SmState == IBIS_IB_SM_STATE_DISCOVER) )
            continue; //nothing to do

        // check if this APort already checked.
        if (p_port->p_aport) {
            if (!aports_set.count(p_port->p_aport))
                aports_set.insert(p_port->p_aport);
            else
                continue;
        }

        // collect master sm
        if (sm_info.SmState == IBIS_IB_SM_STATE_MASTER) {

                p_sm_masters_list.push_back(cur_sm_obj);

        // collect not master sm
        } else if (sm_info.SmState == IBIS_IB_SM_STATE_STANDBY) {

                p_sm_not_masters_list.push_back(cur_sm_obj);

        } else {
            sm_errors.push_back(new FabricErrSMUnknownState((cur_sm_obj)));
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        }
    }

    //no master
    if (p_sm_masters_list.empty()) {
        sm_errors.push_back(new FabricErrSMNotFound(NULL));
        IBDIAG_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    //one sm
    if (p_sm_masters_list.size() == 1) {
        //verify correct sm
        sm_info_obj_t *p_sm_master_obj = p_sm_masters_list.front();
        for (list_p_sm_info_obj::iterator it = p_sm_not_masters_list.begin();
                    it != p_sm_not_masters_list.end(); ++it) {
            //check priority
            if ( ((*it)->smp_sm_info.Priority > p_sm_master_obj->smp_sm_info.Priority) ||
                    (((*it)->smp_sm_info.Priority == p_sm_master_obj->smp_sm_info.Priority) &&
                            ((*it)->smp_sm_info.GUID < p_sm_master_obj->smp_sm_info.GUID)) ) {

                sm_errors.push_back(new FabricErrSMNotCorrect((*it)));
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
                break;  //already in error
            }
        }
    } else {    //more than one sm
        for (list_p_sm_info_obj::iterator it = p_sm_masters_list.begin();
                it != p_sm_masters_list.end(); ++it) {

            sm_errors.push_back(new FabricErrSMManyExists((*it)));
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        }
    }

    p_sm_masters_list.clear();
    p_sm_not_masters_list.clear();

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpSMInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_SM_INFO))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID," << "PortGUID," "PortNumber,"
            << "GUID," << "Sm_Key," << "ActCount,"
            << "SmState," << "Priority"
            << endl;
    csv_out.WriteBuf(sstream.str());

    char buffer[1024];
    for (list_p_sm_info_obj::iterator it = this->fabric_extended_info.getSMPSMInfoListRef().begin();
            it != this->fabric_extended_info.getSMPSMInfoListRef().end(); ++it) {
        sstream.str("");
        snprintf(buffer, sizeof(buffer), U64H_FMT "," U64H_FMT "," U32D_FMT ","   \
                 U64H_FMT "," U64H_FMT "," U32D_FMT ","   \
                 U32D_FMT "," U32D_FMT,
                 (*it)->p_port->p_node->guid_get(),
                 (*it)->p_port->guid_get(),
                 (*it)->p_port->num,
                 (*it)->smp_sm_info.GUID,
                 (*it)->smp_sm_info.Sm_Key,
                 (*it)->smp_sm_info.ActCount,
                 (*it)->smp_sm_info.SmState,
                 (*it)->smp_sm_info.Priority);
        sstream << buffer << endl;
        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_SM_INFO);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR
int IBDiag::WriteSMFile(const string &file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("Subnet Manager",
                            file_name, sout,
                            false, //no append
                            "#"); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    this->DumpSMInfo(sout);

    this->CloseFile(sout);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}
