/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved
 * Copyright (C) 2021-2025, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#ifndef _IBDIAGNET_STAGES_H_
#define _IBDIAGNET_STAGES_H_


#include <stdlib.h>

#include "infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h"
#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibdiag/sharp_mngr.h"
#include "ibdiagapp_types.h"
#include "ibdiagnet_plugins.h"
#include "infiniband/ibdm/cable/CableFW.h"
#include "infiniband/ibdm/cable/PrtlRecord.h"

/******************************************************/
class DiscoveryStage: public Stage {
private:

protected:
    ////////////////////
    //members
    ////////////////////
    string device_name;
    u_int8_t port_num;
    u_int64_t port_guid;
    u_int8_t max_hops;
    bool to_check_dup_guids;
    bool to_check_dup_node_desc;
    bool initiliazed_with_device;
    bool initiliazed_with_guid;
    string psl_dump_file;
    string output_capability_file;
    string output_full_capability_file;
    string input_capability_file;
    bool show_ports_data_db2; //show all data on ports added in db2 format
                              //CapMsk2,FEC,Retransmission
    bool dump_lst_with_lmc; //write the LST file wite the ports lmc.
    string node_name_map_file;
    map<uint16_t, uint64_t> lid2guid_map;
    FilesBasedMKeyManager *p_mkey_manager;

    string load_from_file_path;
    bool load_from_file;

    u_int32_t num_of_switches;
    u_int32_t num_of_channel_adapters;
    u_int32_t num_of_channel_adapters_sdm;
    u_int32_t num_of_routers;
    u_int32_t num_of_aggregation_nodes;
    u_int32_t num_of_GPUs;

    bool check_switch_duplicated_guids;

    ////////////////////
    //methods
    ////////////////////
    void PrintDuplicatedGuidsDetectErrors();
    virtual int PostDiscover() = 0;

    void CheckAPortSymmetry(APort* p_aport);
    /**
     * @brief This function checks the following conditions for each APort:
     *        1. Planes are either all connected, or all disconnected
     *        2. Each remote plane has the same plane number as the local plane conneted to it
     *        3. Local and remote APorts have the same number of planes
     *        Note - this function assumes the 'visited' bit
                     was unset on all APorts before it's first invocation.
     * 
     * @param p_aport - The aport to preform the check upon
     * @param sys_guid - system guid (for error reporting)
     * @param aport_index - aport index (for error reporting)
     * @param visited - to avoid reporting the same error on the remote IBPort,
     *                  this function will push each checked IBPort to a set.
     */
    void CheckRemotePlanes(APort* p_aport);
    int APortChecks();
    int PortsPlaneSymmetryChecks();
    int CageManagerChecks(list_p_fabric_general_err& cage_manager_errs);
    int DumpAPortsReport(bool invalid_aports_only);
    void DumpAPort(ofstream& sout, APort* p_aport);
    void CopyAPortErrorsToList(list_p_fabric_general_err& errs);

    /**
     * @brief Checks that all planes in the APort have the same
     *        attribute (a value returned by an accessor function). if an APort's planes
     *        return different results, or all the results returned are 'unknown_value',
     *        add an error.
     * 
     * @tparam T - the attributes type
     * @param p_aport - aport who's planes will be checked 
     * @param attribute_name - attribute name for error reporting
     * @param accessor - a function that returns an IBPort's attribute value
     * @param unknown_value
     * @param t_to_str - function that converts T type to string, for error reporting.
     */
    template<class T>
    int CheckAttributeSymmetry(APort* p_aport,
                               string attribute_name,
                               const ibport_accessor_t<T>& accessor,
                               T unknown_value,
                               string (*t_to_str)(const T)) {
        if (!p_aport)
            return 1;
        vec_p_fabric_err& aport_errors = this->p_ibdiag->GetErrorsByAPort(p_aport);

        if (!p_aport->isEqualAttribute(accessor)) {
            aport_errors.push_back(new APortUnequalAttribute(
                                        p_aport, attribute_name,
                                        p_aport->getAttributeArrayStr(accessor, t_to_str)));
            p_aport->asymmetry_mask.set(APORT_UNEQUAL_ATTRIBUTE);
            return 1;
        } else if (!p_aport->isKnownAttribute(accessor, unknown_value)) {
            aport_errors.push_back(new APortNoValidAttribute(
                                        p_aport, attribute_name));
            p_aport->asymmetry_mask.set(APORT_NO_VALID_ATTRIBUTE);
            return 1;
        }

        return 0;
    };

    /**
     * @brief Wrapper for previous function for getter use.
     */
    template<class T>
    int CheckAttributeSymmetry(APort* p_aport,
                               string attribute_name,
                               T (IBPort::*getter)() const,
                               T unknown_value,
                               string (*t_to_str)(const T)) {
        ibport_accessor_t<T> accessor = std::mem_fn(getter);
        return this->CheckAttributeSymmetry(p_aport, attribute_name, accessor, unknown_value, t_to_str);
    };

    /**
     * @brief Check that the given FNM Nodes are connected in
     *        a circle that contains all the nodes in the system.
     */
    bool CheckValidFNMConnections(list_p_fabric_general_err& errs, const IBSystem* p_system);

    /**
     * @brief Check that all fnm ports have the same connection speed
     */
    bool CheckFNMSpeeds(list_p_fabric_general_err& errs, const IBSystem* p_system);

public:
    ////////////////////
    //methods
    ////////////////////
    DiscoveryStage(IBDiag *p_ibdiag);
    virtual ~DiscoveryStage() {}

    inline void SetOutputCapabilityFile(const string &file_name) {
        this->output_capability_file = file_name;
    }
    inline void SetOutputFullCapabilityFile(const string &file_name) {
        this->output_full_capability_file = file_name;
    }
    inline void SetInputCapabilityFile(const string &file_name) {
        this->input_capability_file = file_name;
    }
    inline void SetNodeNameMapFile(const string &file_name) {
        this->node_name_map_file = file_name;
    }
    void CreateLid2GuidDB();

    inline bool GetDUPGUIDsDetection() { return this->to_check_dup_guids; }
    inline void TurnOFFDUPGUIDsDetection() { this->to_check_dup_guids = false; }
    inline void TurnOFFDUPNodeDescDetection() { this->to_check_dup_node_desc = false; }
    inline void SetDeviceName(string device_name) {
        this->device_name = device_name;
        initiliazed_with_device = true;
    }
    inline const string & GetDeviceName() const {
        return this->device_name;
    }
    inline void SetPortNum(u_int8_t port_num) {
        this->port_num = port_num;
        initiliazed_with_device = true;
    }
    inline uint8_t GetPortNum() const {
        return this->port_num;
    }
    inline void SetPortGUID(u_int64_t port_guid) {
        this->port_guid = port_guid;
        if (this->port_guid == OPTION_GUID_DEF_VAL)
            initiliazed_with_guid = false;
        else
            initiliazed_with_guid = true;
    }
    inline u_int64_t GetPortGUID() const {
        return this->port_guid;
    }
    inline bool IsInitiliazedWithDevice() const {
        return this->initiliazed_with_device;
    }
    inline bool IsInitiliazedWithGUID() const {
        return this->initiliazed_with_guid;
    }
    inline void SetMaxHops(u_int8_t max_hops) {
        this->max_hops = max_hops;
    }
    inline bool IsNeedDisplayAllPortsGuids() {
        if ((initiliazed_with_guid == true) && this->port_guid == 0) {
            return true;
        }
        return false;
    }
    inline void SetPSLDumpFile(const string &file_name) {
        this->psl_dump_file = file_name;
    }
    inline void SetDumpLSTWithLMC() {
        this->dump_lst_with_lmc = true;
    }

    //determine if to show new db format according to back_compat_db
    inline void SetShowPortsDataFormat(const string &format) {
        double version = atof(format.c_str());

        if (format == OPTION_BACK_COMPAT_DB_DEF_VAL)
            this->show_ports_data_db2 = true;
        else if (version < 2)
            this->show_ports_data_db2 = false;
    }

    inline void SetMKeyManager(FilesBasedMKeyManager *p_mkey_manager) {

        this->p_mkey_manager = p_mkey_manager;
    }

    inline void SetCheckSwitchDuplicatedGuids(bool val) {
        this->check_switch_duplicated_guids = val;
    }

    inline bool GetCheckSwitchDuplicatedGuids() {
        return this->check_switch_duplicated_guids;
    }

        /**
     * @brief this function checks FNM Ports validity for all systems in the fabric:
     *        first it checks that for each APort-containing system, there exactly 2 FNM ports.
     *        For these systems, we check that the fnm ports are connected in a circle.
     * 
     * @return int 
     */
    void FNMPortsChecks(list_p_fabric_general_err& errs);

    int CountNodeTypes();
    virtual bool GetFabricSummarySection(stringstream &section);

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
    virtual int DiscoverFabric() = 0;
    virtual int CleanResources();
};

/******************************************************/
class DiscoveryFabricStage: public DiscoveryStage {
private:
    string scope_port_guid_file;
    string exclude_scope_port_guid_file;

    string unhealthy_ports_dump_file;
    string healthy_ports_policy_file;
    //defines an action for unhealthy switch's ports.
    //by default no action is taken
    bool switch_action;
    //defines an action for unhealthy CA's ports.
    //by default no action is taken
    bool ca_action;

    virtual int PostDiscover();
    int MarkOutUnhealthyPorts(int &unhealthy_ports);
    bool GetDefaultUnhealthyAction(const string &value, bool &action);
public:
    ////////////////////
    //methods
    ////////////////////
    DiscoveryFabricStage(IBDiag *p_ibdiag);
    ~DiscoveryFabricStage() {}
    int Prepare();
    virtual int DiscoverFabric();

    inline void SetScopePortGuidFile(const string &file_name) {
        this->scope_port_guid_file = file_name;
    }

    inline void SetExcludeScopePortGuidFile(const string &file_name) {
        this->exclude_scope_port_guid_file = file_name;
    }

    inline void SetHealthyPortsPolicyFile(const string &file_name) {
        this->healthy_ports_policy_file = file_name;
    }

    inline const string & GetHealthyPortsPolicyFile() const {
        return this->healthy_ports_policy_file;
    }

    inline void SetUnhealthyPortsDumpFile(const string &file_name) {
        this->unhealthy_ports_dump_file = file_name;
    }

    inline const string & GetUnhealthyPortsDumpFile() const {
        return this->unhealthy_ports_dump_file;
    }

    bool SetUnhealthyPortsSwitchAction(const string &value);
    bool SetUnhealthyPortsCAAction(const string &value);

    inline void SetLoadFromFilePath(const string &file_name) {
        this->load_from_file_path = file_name;
        load_from_file = true;
    }

    inline const string & GetLoadFromFilePath() const {
        return this->load_from_file_path;
    }

    inline bool IsLoadFromFile() {
        return load_from_file;
    }
};

/******************************************************/
class LidsStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////

public:
    ////////////////////
    //methods
    ////////////////////
    LidsStage(IBDiag *p_ibdiag);
    ~LidsStage() {}

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class TopologyMatchStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    IBFabric spec_fabric;

    string output_topo_file;
    string input_topo_file;
    string topo_ibnl_dir;
    bool use_smart_names;

public:
    ////////////////////
    //methods
    ////////////////////
    TopologyMatchStage(IBDiag *p_ibdiag);
    ~TopologyMatchStage() {}

    inline IBFabric * GetSpecFabricPtr() {
        return &(this->spec_fabric);
    }

    inline void SetOutputTopoFile(const string &file_name) {
        this->output_topo_file = file_name;
        this->MarkStageAsActive();
    }
    inline void SetInputTopoFile(const string &file_name) {
        this->input_topo_file = file_name;
        this->MarkStageAsActive();
    }
    inline void SetTopoIbnlDir(const string &dir_name) {
        this->topo_ibnl_dir = dir_name;
    }
    inline void ValidateTopoParams() {
        /*
        if ((this->input_topo_file == "") && (this->anchor_sys_name != "")) {
            WARN_PRINT("Local system name is specified, but topology is not given." \
                    "The former is ignored.\n");
            return;
        }
        */
        return;
    }

    inline void SetSmartNames(bool val) {
        this->use_smart_names = val;
    }

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class LinksStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    IBLinksInfo ib_links_info;

public:
    ////////////////////
    //methods
    ////////////////////
    LinksStage(IBDiag *p_ibdiag);
    ~LinksStage() {}

    void printLinksSummary(stringstream &summary_str,
                           string name,
                           uint32_t num_of_links,
                           vec_vec_uint32 &link_width_speed_matrix);

    virtual bool GetFabricSummarySection(stringstream &section);

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class SMStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////

public:
    ////////////////////
    //methods
    ////////////////////
    SMStage(IBDiag *p_ibdiag);
    ~SMStage() {}

    virtual bool GetFabricSummarySection(stringstream &section);

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
    virtual int CleanResources();
};


/******************************************************/
class PMStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    bool to_reset_counters;
    bool to_reset_histogram_data;
    bool to_calc_ber;
    u_int32_t m_check_counters_bitset;
    //bool to_check_ext_speeds_counters;
    //bool to_check_ext_speeds_on_hca;
    bool en_per_lane_cnts;
    bool en_per_slvl_cntrs;
    bool en_per_slvl_sup_warn;
    bool en_port_samples_result;

    u_int64_t ber_threshold;
    u_int32_t pause_time_between_samples;
    double diff_time_between_samples; //elapsed seconds between 2 samples

    map_str_uint64          counters_to_threshold_map;
    vector_p_pm_info_obj    first_sample_pm_info_obj_vector;
    vec_slvl_cntrs          m_slvl_cntrs_vec;

    bool can_send_mads_by_lid;

    u_int32_t total_pm_symbol_errors;


    ////////////////////
    //methods
    ////////////////////
    void Sleep(u_int32_t seconds);
    void PreparePerSlVlCounters();
    void CountTotalPMSymbolError(u_int64_t pm_threshold);
public:
    ////////////////////
    //methods
    ////////////////////
    PMStage(IBDiag *p_ibdiag);
    ~PMStage();

    inline map_str_uint64& GetRefToCountersToThresholdMap() {
        return this->counters_to_threshold_map;
    }
    inline void SetResetCounters(bool set_val) {
        this->to_reset_counters = set_val;
    }
    inline void SetResetHistogramData(bool set_val) {
        this->to_reset_histogram_data = set_val;
    }
    inline void SetBERTest(bool set_val) {
        this->to_calc_ber = set_val;
    }
    inline void SetBERThreshold(u_int64_t ber_threshold) {
        if (ber_threshold == 0) {   /* ZERO is considered as overflow value ==> print all calculations */
            this->ber_threshold = OVERFLOW_VAL_64_BIT;
            return;
        }
        this->ber_threshold = ber_threshold;
    }

    inline void SetPauseTimeBetweenSamples(u_int32_t seconds) {
        this->pause_time_between_samples = seconds;
    }

    inline void SetCheckExtSpeedsCounters(const char * dev_type) {
        if (dev_type && strcmp(dev_type,"none")==0) {
            m_check_counters_bitset &= UNSET_EXT_SPEEDS_COUNTERS;
        } else {
            m_check_counters_bitset |= CHECK_EXT_SPEEDS_COUNTERS_ON_SW;
            if (dev_type && strcmp(dev_type,"all")==0) {
                m_check_counters_bitset |= CHECK_EXT_SPEEDS_COUNTERS_ON_ALL;
            }
        }
    }
    inline void SetPrintLLRCounters() {
        m_check_counters_bitset |= PRINT_LLR_COUNTERS;
    }
    inline void SetPerLaneCounters(bool set_val) {
        en_per_lane_cnts = set_val;
    }

    inline void SetPerSLVLCounters(bool set_val) {
        en_per_slvl_cntrs = set_val;
    }

    inline void SetPerSLVLSupWarn(bool set_val) {
        en_per_slvl_sup_warn = set_val;
    }

    inline void SetPortSamplesResult(bool set_val) {
        en_port_samples_result = set_val;
    }

    virtual bool GetFabricSummarySection(stringstream &section);

    virtual int CleanResources();

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};

/******************************************************/
class NVLinkStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    bool m_reset_counters;

public:
    ////////////////////
    //methods
    ////////////////////
    NVLinkStage(IBDiag *p_ibdiag);
    ~NVLinkStage() {}

    void SetResetCounters(bool value) { m_reset_counters = value; }

    int CleanResources() override;

    int Prepare();
    int RetrieveInfo();
    int RunCheck()      {return 0;}
};



/******************************************************/
class RoutingStage: public Stage {
private:
    void DumpHBFTables();

    ////////////////////
    //members
    ////////////////////
    bool retrieved_switch_info_ok;
    bool retrieved_uc_fdbs_ok;
    bool retrieved_mc_fdbs_ok;
    bool retrieved_plft_data_ok;
    bool retrieved_ar_data_ok;
    bool retrieved_rn_data_ok;
    bool ar_enabled;
    bool is_fat_tree;
    bool use_vlr;
    bool use_vs_route;
    bool check_mcast_credloop;
    bool check_sl;
    bool dump_only;
    bool dump_only_skip_routing_tables;
    bool static_ca2ca;
    bool skip_slvl;
    u_int8_t  default_sl;
    u_int32_t num_ar_switches;
    u_int32_t num_hbf_switches;

    string sa_dump_file;

    bool is_entry_plane_filter_sup;
    bool is_rail_filter_sup;
    bool is_end_port_plane_filter_sup;

public:
    ////////////////////
    //methods
    ////////////////////
    RoutingStage(IBDiag *p_ibdiag);
    ~RoutingStage() {}

    //going to be deleted in the future
    inline void SetFatTree(bool set_val) {
        this->is_fat_tree = set_val;
    }
    inline void UseVLR() {
        this->use_vlr = true;
    }
    void SkipVSRoute() {
        use_vs_route = false;
    }
    void CheckMcastCredLoop() {
        check_mcast_credloop = true;
    }
    void SetCheckSL() {
        check_sl = true;
    }
    void SetDefaultSL(u_int8_t sl) {
        default_sl = sl;
    }
    void SetSADumpFile(const string &file_name) {
        this->sa_dump_file = file_name;
    }
    void DumpOnly(){
        dump_only = true;
    }
    void DumpOnlySkipRoutingTables() {
        dump_only_skip_routing_tables = true;
    }
    void StaticCA2CA() {
        static_ca2ca = true;
    }
    void SkipSLVL() {
        skip_slvl = true;
    }

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class NodesInfoStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    bool can_send_mads_by_lid;

public:
    ////////////////////
    //methods
    ////////////////////
    NodesInfoStage(IBDiag *p_ibdiag);
    ~NodesInfoStage() {}

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class Speed_WidthStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    string expected_link_speed;
    string expected_link_width;

public:
    ////////////////////
    //methods
    ////////////////////
    Speed_WidthStage(IBDiag *p_ibdiag);
    ~Speed_WidthStage() {}

    inline void SetExpectedLinkSpeed(const string &expected_link_speed) {
	if (expected_link_speed == IB_DISABLE_LINK_SPEED_STR)
            this->expected_link_speed = "";
        else
            this->expected_link_speed = expected_link_speed;
    }
    inline void SetExpectedLinkWidth(const string &expected_link_width) {
        if (expected_link_width == IB_DISABLE_LINK_WIDTH_STR)
            this->expected_link_width = "";
        else
            this->expected_link_width = expected_link_width;
    }

    int Prepare();
    int RetrieveInfo() { return 0; }
    int RunCheck();
};


/******************************************************/
class PKeyStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    bool can_send_mads_by_lid;
    map_pkey_p_port_membership pkey_to_p_port_mem;
    bool skip_aport_membership;

public:
    ////////////////////
    //methods
    ////////////////////
    PKeyStage(IBDiag *p_ibdiag);
    ~PKeyStage();

    void SetSkipAPortMembership(bool val) {
        skip_aport_membership = val;
    }

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class AGUIDStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    bool can_send_mads_by_lid;

public:
    ////////////////////
    //methods
    ////////////////////
    AGUIDStage(IBDiag *p_ibdiag);
    ~AGUIDStage();

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class MlnxCntrsStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    bool can_send_mads_by_lid;
    bool to_reset_counters;
public:
    ////////////////////
    //methods
    ////////////////////
    MlnxCntrsStage(IBDiag *p_ibdiag);
    ~MlnxCntrsStage();

    inline void SetResetCounters(bool set_val) {
        this->to_reset_counters = set_val;
    }

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class VirtualizationStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    bool can_send_mads_by_lid;
public:
    ////////////////////
    //methods
    ////////////////////
    VirtualizationStage(IBDiag *p_ibdiag);
    ~VirtualizationStage();

    virtual int CleanResources();

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};

/******************************************************/
class TemperatureSensing: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
public:
    ////////////////////
    //methods
    ////////////////////
    TemperatureSensing(IBDiag *p_ibdiag);
    ~TemperatureSensing();

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


class FLIDsManager;

class RoutersStage : public Stage {
private:
     void CheckFLIDs();

      FLIDsManager *p_FLIDsManager;
public:
    RoutersStage(IBDiag *p_ibdiag);
    ~RoutersStage();

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
    virtual const char* GetLastError();
};

/******************************************************/
class QoSStage : public Stage {
private:
public:
    QoSStage(IBDiag *p_ibdiag);
    ~QoSStage();

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};

/******************************************************/
class SharpStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    bool can_send_mads_by_lid;
    bool dump_counters;
    bool dump_hba_port_counters;
    bool clear_counters;
    bool check_trees;

    u_int8_t version;

    SharpMngr *p_sharp_mngr;

    vec_uint32 sharp_versions_counter;
    u_int32_t num_of_llts;
    u_int32_t num_of_sats;
    u_int32_t num_of_active_jobs;

    map_guid_2_key_t guid_2_am_key;

public:
    ////////////////////
    //methods
    ////////////////////
    SharpStage(IBDiag *p_ibdiag);
    ~SharpStage();

    void SetDumpCounters() { dump_counters = true; }
    void SetDumpHBAPortCounters() { dump_hba_port_counters = true; }
    void SetClearCounters() { clear_counters = true; }
    void SetCheckTrees() { check_trees = true; }

    void SetVersion(u_int8_t ver) { version = ver; }
    inline map_guid_2_key_t& GetAMKeyMap() {
        return this->guid_2_am_key;
    }
    void CountSharpVersionsAndNumOfActiveJobs();
    void CountSharpTreeTypes();
    virtual bool GetFabricSummarySection(stringstream &section);

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class CongestionControlStage : public Stage {
private:
    bool to_get_congestion_counters;
    bool to_clear_congestion_counters;
    string ppcc_algo_path;
    PPCCAlgoDatabase ppcc_algo_database;

    u_int64_t enabled_SWs;
    u_int64_t enabled_CAs;
    u_int64_t enabled_algo_CAs;
    u_int64_t enabled_algo_CA_ports;
    u_int64_t disabled_algo_CA_ports;

    vec_slvl_cntrs          m_slvl_cntrs_vec;

    void PreparePerSlVlCounters();

public:
    CongestionControlStage(IBDiag *p_ibdiag);
    ~CongestionControlStage();

    inline void SetCongestionCountersGet(bool to_get_congestion_counters) {
        this->to_get_congestion_counters = to_get_congestion_counters;
    }

    inline void SetClearCongestionCounters(bool to_clear_congestion_counters) {
        this->to_clear_congestion_counters = to_clear_congestion_counters;
    }

    inline void SetPPCCAlgoPath(const string &ppcc_algo_path) {
        this->ppcc_algo_path = ppcc_algo_path;
    }

    virtual bool GetFabricSummarySection(stringstream &summary_str);

    virtual int CleanResources();

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};

/******************************************************/
class FastRecoveryStage : public Stage {

public:
    FastRecoveryStage(IBDiag *p_ibdiag);
    ~FastRecoveryStage();

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};

class CheckRailOptimizedTopologyStage : public Stage {
private:

    class ExcludedData {
    public:
          ExcludedData() : socketDirect(0), noAddress(0) {}
          size_t socketDirect;
          size_t noAddress;
    };

public:
    CheckRailOptimizedTopologyStage(IBDiag *p_ibdiag);
    ~CheckRailOptimizedTopologyStage();

    int Prepare() override;
    int RetrieveInfo() override;
    int RunCheck() override;

    const char* GetLastError() override { return lastError.c_str(); }
    void SetRailFlow(size_t r) { railFlow = r; }
    bool SetRailRegex(const string &val);
    void SetBDFMode(PCI_BDF_SOURCE s) { source = s;}
    void SetBDFFactor(size_t f) { factor = f; };

private:
    int RetrieveRailsData(list_p_fabric_general_err &errorsList,
                          ExcludedData &excludedData,
                          PCI_LeafSwitchesInfoVec &leafSwitchInfoVec);
    int GetPCIRelatedData(const IBNode &switch_node, PCI_AddressToNodesMap &pciAddresses,
                          ExcludedData &excludedCards, list_p_fabric_general_err &errorsList,
                          PCI_BDF_SOURCE &source);
    PCI_BDF_SOURCE ResolveBDFSource(const IBNode &curr_node) const;
    bool ShouldFilterNode(const string &description) const;
    int PrintValidationReport(const PCI_LeafSwitchesInfoVec &pciToNodesVec,
                              const ExcludedData &excludedData);
    int ReportOnLeafSwitch(const PCI_LeafSwitchInfo &leafSwicthInfo, ostream &ofs);
    int GetSwitchAndNodePorts(IBNode *p_node, IBPort *&p_switch_port, IBPort *& p_node_port);
    void PrintSwitchData(const PCI_LeafSwitchInfo &leafSwitchInfo, ostream &ofs);
    int PrintRailData(const PCI_Address& address, const vector_p_node &nodes,
                     int index, ostream &ofs);
    int ReportOnExcludedData(const ExcludedData &excludedData, ostream &ofs);

    PCI_BDF_SOURCE source;
    size_t factor;
    string lastError;
    size_t railFlow;
    regExp *p_railRegex;
};


class PostReportsStage : public Stage {
public:
    PostReportsStage(IBDiag *p_ibdiag);
    ~PostReportsStage();

    int Prepare();
    int RetrieveInfo();
    int RunCheck();

    virtual bool GetFabricSummarySection(stringstream &section);

    void SetCheckSM(bool val) {
        to_check_sm = val;
    }

    void SetGeneratedClassName(const std::string &name) {
        generated_class_name = name;
    }

    void ParseNodeGUIDs(const std::string &guids);

private:

    bool to_check_sm;
    uint32_t num_of_diff_val;
    string generated_class_name;
    set< u_int64_t > simulator_guids;

    void PrintWarnings(const warnings_list& warnings) const;
};

class CheckDFPTopologyStage : public Stage {
private:
    int maxCAsOnSpine;
    bool is_roots_from_file;
    map_guid_pnode dfp_roots;

    int BuildAndValidate();

public: //todo not active by default
    CheckDFPTopologyStage(IBDiag *p_ibdiag);
    CheckDFPTopologyStage();

    int Prepare();
    int RetrieveInfo();
    int RunCheck();

    int GetMaxCAsOnSpine() const { return maxCAsOnSpine; }
    void SetMaxCAsOnSpine(int val) { maxCAsOnSpine = val; }

    int DiscoverDFPRootsBySMDB();
    int DiscoverDFPRootsByMaxCAsOnSpine();

    bool IsRootFromFile() const { return is_roots_from_file; }
    void SetIsRootFromFile(bool val) { is_roots_from_file = val; }
};


class CheckFatTreeTopologyStage : public Stage {
public:
    CheckFatTreeTopologyStage(IBDiag *p_ibdiag);
    ~CheckFatTreeTopologyStage() {
        if (p_rootsRegExp)
            delete p_rootsRegExp;
    }

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
    const char* GetLastError() { return lastError.c_str(); }
    bool SetRootsRegex(const string &val);
    void SetRootsFromFile(bool val) { rootsFromFile = val; }

    int retries;
    int equalResults;
    regExp *p_rootsRegExp;
    bool rootsFromFile;
    string lastError;

};

class CombinedCableInfo;
class ReportCableInfoStage : public Stage {
private:
    typedef std::map < u_int16_t, u_int16_t > map_cdr_vendor_latency_t;
    typedef std::map < string, u_int16_t > map_pn_vendor_latency_t;

    class LocalPhyLatencyKey {
    public:
        LocalPhyLatencyKey(SerDesVersion inVer = SerDes_NA, IBNodeType inType = IB_UNKNOWN_NODE_TYPE,
                           IBLinkWidth inWidth = IB_UNKNOWN_LINK_WIDTH,
                           IBFECMode inFec = IB_FEC_NO_FEC) : version(inVer), type(inType),
                           width(inWidth), fec_mode(inFec) {}

        SerDesVersion version;
        IBNodeType type;
        IBLinkWidth width;
        IBFECMode fec_mode;

        bool operator<(const LocalPhyLatencyKey &other) const
        {
            if (this == &other)
                return false;

            // Compare first members
            //
            if (version < other.version)
                return true;
            else if (other.version < version)
                return false;

            if (type < other.type)
                return true;
            else if (other.type < type)
                return false;

            if (width < other.width)
                return true;
            else if (other.width < width)
                return false;

            if (fec_mode < other.fec_mode)
                return true;
            else if (other.fec_mode < fec_mode)
                return false;

            // All members are equal, so return false
            return false;
        }
    };
    typedef std::map < LocalPhyLatencyKey, u_int16_t > map_local_phy_latency_t;

public:
    ReportCableInfoStage(IBDiag *p_ibdiag);

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
    bool IsSkipped();
    const char* GetLastError() { return lastError.c_str(); }

    void SetPrtlLatencyFromTable(bool val) { prtl_latency_from_table = val; }
    void SetPNVendorPRTLLatencyFile(const string &val) {
        parse_prtl_mod_dp_latency = true;
        pn_vendor_latency_file = val;
    }
    int DumpPNVendorPRTLLatencyTable(const string &dump_file) const;
    void SetCDRVendorPRTLLatencyFile(const string &val) {
        parse_prtl_mod_dp_latency = true;
        cdr_vendor_latency_file = val;
    }
    int DumpCDRVendorPRTLLatencyTable(const string &dump_file) const;
    void SetPRTLLocalPHYLatencyTable(const string &val) {
        parse_prtl_mod_dp_latency = true;
        local_phy_latency_file = val;
    }
    int DumpPRTLLocalPHYLatencyTable(const string &dump_file) const;


private:
    void InitCDRVendorTable();
    void InitPNVendorTable();
    void InitLocalPhyLatancyTable();
    bool UpdatePrtlLatencyData(PrtlRecord &prtl, const IBPort &port, string &message);
    void SetPRTLLength(const IBPort *p_curr_port, const IBPort *p_remote_port,
                          PairsContainer<const IBPort*> &checkedPairs,
                          list_p_fabric_general_err &errorsList);
    void SetPRTLLength(const IBPort *p_curr_port, const IBPort *p_remote_port,
                       list_p_fabric_general_err &errorsList);
    void SetPRTLLength(const PrtlRecord *p_prtl,
                       const IBPort *p_port, const IBPort *p_other_port,
                       list_p_fabric_general_err &errorsList);
    int ParsePNVendorLatencyFile();
    int ParseCDRVendorLatencyFile();
    int ParsePRTLLocalPHYLatencyFile();

    int GeneralReport(list_p_fabric_general_err &errorsList);
    int FWVersionReport(list_p_fabric_general_err &errorsList);
    int CollectMaxFWVersions(cable_id_to_fw_map &maxFWVersionsMap);
    int CheckFWVersions(const cable_id_to_fw_map &maxFWVersionsMap,
                        list_p_fabric_general_err &errorsList);
    int TemperatureErrorsReport(list_p_fabric_general_err &errorsList);
    int CheckTemperature(int tempRet, const string &message, const IBPort &port,
                         list_p_fabric_general_err &errorsList, bool &isDetected);
    void ReleaseCableInfo();
    void ReportToCSV(const IBPort &port, const CombinedCableInfo &combined);
    void ReportToStream(ostream &stream,
                        const IBPort &port, const CombinedCableInfo &combined) const;
    void InitFileStream(ofstream &fileStream);
    u_int16_t GetPhyLatency(const IBPort &port) const;

    map_local_phy_latency_t local_phy_latency_table;
    map_cdr_vendor_latency_t cdr_vendor_prtl_latency_table;
    string cdr_vendor_latency_file;
    map_pn_vendor_latency_t  pn_vendor_prtl_latency_table;
    string pn_vendor_latency_file;
    string local_phy_latency_file;
    bool parse_prtl_mod_dp_latency;
    bool prtl_latency_from_table;

    string lastError;
};

class ExportDataStage : public Stage {
public:
    ExportDataStage(IBDiag *p_ibdiag, list_p_symbols_plugin *p_loaded_plugins_list);
    ~ExportDataStage();

    int Prepare();
    int RetrieveInfo();
    int RunCheck();

private:
    list_p_symbols_plugin *p_loaded_plugins_list;
};

class ScopeBuilderStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    int ParseHostsString(const char *p_buffer, set_pnode &dest_nodes) const;

    regExp *p_filter;


    string host_file;
public:
    ////////////////////
    //methods
    ////////////////////
    ScopeBuilderStage(IBDiag *p_ibdiag);
    ~ScopeBuilderStage() { delete p_filter; }

    int Prepare();
    int RetrieveInfo();
    int RunCheck();

    void SetNodeDescriptionFilter(string &filter) {
        p_filter = new regExp(filter.c_str());
    }
    void SetHostFile(string &file) { host_file = file; }
    int ReadHostFile(set_pnode &dest_nodes) const;
};

class ParseKeyFilesStage: public Stage {
public:
    using set_key_ports_func_t = std::function<void ()>;
private:
    ////////////////////
    //members
    ////////////////////

    bool is_long_iteration;

public:
    ////////////////////
    //methods
    ////////////////////
    ParseKeyFilesStage(IBDiag *p_ibdiag);
    ~ParseKeyFilesStage() {  }

    int Prepare();
    int RetrieveInfo();
    int RunCheck();


    void SetIsLongIteration(bool is_long_iteration) {
        this->is_long_iteration = is_long_iteration;
    }

};

#endif          /* not defined _IBDIAGNET_STAGES_H_ */
