/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 2021-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#ifndef _ACC_REGISTER_H_
#define _ACC_REGISTER_H_

#include "acc_reg_key.h"
#include "phy_diag_types.h"
#include "phy_export_api.h"

#include "infiniband/ibdm/cable/PhyCableRecord.h"

class PhyDiag;

union AccessRegister {
    struct SMP_AccessRegister smp_reg;
    struct GMP_AccessRegister gmp_reg;
};


typedef union
{
    struct pcam_reg pcam;

    struct mgir_reg mgir;
    struct slrg_reg slrg;
    struct sllm_reg sllm;
    struct slrp_reg slrp;
    struct sltp_reg sltp;
    struct slsir_reg slsir;
    struct ptas_reg ptas;
    struct pmdr_reg pmdr;
    struct pphcr_reg pphcr;
    struct ppll_reg ppll;
    struct mpein_reg mpein;
    struct mpir_reg mpir;
    struct pmcr_reg pmcr;
    struct ptys_reg ptys;
    struct ppsls_reg ppsls;
    struct mpscr_reg mpscr;
    struct mtsr_reg mtsr;
    struct pltc_reg pltc;
    struct paos_reg paos;

    // fans registers
    struct mfcr_reg mfcr;
    struct mfsm_reg mfsm;
    struct mfsl_reg mfsl;
    struct fore_reg fore;
    struct mfnr_reg mfnr;
    // temperature registers
    struct mtcap_reg mtcap;
    struct mtmp_reg mtmp;
    struct mtwe_reg mtwe;
    // current/voltage register
    struct mvcap_reg mvcap;
    struct mvcr_reg mvcr;
    // power supplies register
    struct msps_reg msps;
    // general information register
    struct msgi_reg msgi;
    struct msci_reg msci;

    struct slrip_reg slrip;
    struct slreg_reg slreg;
    struct slcct_reg slcct;
    struct ppamp_reg ppamp;
    struct peucg_reg peucg;
    struct pddr_reg pddr;
    struct pemi_reg pemi;
    struct mpcnt_reg mpcnt;
    struct ppcnt_reg ppcnt;

    struct prtl_reg prtl;
    struct ppbmp_reg ppbmp;

    struct mrcs_reg mrcs;
    struct ppslc_reg ppslc;
    struct ppsld_reg ppsld;
    struct mlpc_reg mlpc;
    struct mppgs_reg mppgs;
    struct mppdr_reg mppdr;
    struct mppgh_reg mppgh;
    struct ppspc_reg ppspc;
    struct ppspi_reg ppspi;
    struct ppspgc_reg ppspgc;
    struct ppslg_reg ppslg;
    struct mppgcr_reg mppgcr;
    struct mppcr_reg mppcr;

    // PDDR Layouts
    pddr_module_info    pddr_module;
    module_latched_flag_info   pddr_latched;
    pddr_link_up_info_page pddr_link_up;
    pddr_phy_info_page  pddr_phyinfo;
    pddr_troubleshooting_page pddr_troubleshooting;
    pddr_operation_info_page  pddr_operation_info;
    pddr_link_down_info_page  pddr_link_down_info;

} reg_list;

typedef struct
{
    uint8_t raw[sizeof(reg_list) + 160];
    union {
        struct {
            uint8_t module_info_ext;
        }; //here you may add structures to keep data related to different
        // registers
        // struct {
        //         uint32_t xxxxxxx;
        //         uint32_t zzzzzzz;
        //}
    };
} reg_raw_data;

struct acc_reg_data {
    //This union should help access the data without pointer casting
    //and this fixes the strict aliasing rule error/warning.
    union {
        reg_raw_data data;
        reg_list regs;
    };
};

typedef map < AccRegKey *, acc_reg_data, bool (*)(AccRegKey *, AccRegKey *)> map_akey_areg;
typedef map < IBNode *, u_int8_t > map_pnode_to_max_group_num_t;
class AccRegHandler;

bool keycomp(AccRegKey *k1, AccRegKey *k2);


//Register class defines common members to all registers.
//Each register needs to have its dedicated class, that inherits Register
//and defines its own PackData and DumpRegisterData.
class Register {

protected:
    PhyDiag   * m_phy_diag;
    uint32_t    m_register_id;
    uint32_t    m_fields_num;
    //That is a bit to set in case register isn't supported
    //and to avoid sending the register more than once to the same node.
    uint64_t    m_not_supported_bit;
    string      m_section_name;
    string      m_header;
    string      m_name;
    PhyPluginSupportedNodesType m_support_nodes;
    bool        m_retrive_disconnected;
    bool        m_dump_enabled;
    AccRegVia_t m_acc_reg_type;
    AccRegVia_t m_acc_type_priority;

    unpack_data_func_t unpack_data_func;

public:
    Register(PhyDiag * phy_diag,
             uint32_t reg_id,
             const unpack_data_func_t p_unpack_func,
             string sn,
             string name,
             uint32_t fn,
             uint64_t nsb,
             string header = "",
             PhyPluginSupportedNodesType support_nodes = SUPPORT_SW_CA,
             bool dump_enabled = true,
             bool retrive_disconnected = false,
             AccRegVia_t acc_reg_type = VIA_SMP,
             AccRegVia_t acc_type_priority = VIA_GMP);

    virtual ~Register(){}

    int PackDataSMP(AccRegKey *p_key, SMP_AccessRegister *acc_reg, const IBNode *p_node);
    int PackDataGMP(AccRegKey *p_key, GMP_AccessRegister *acc_reg, const IBNode *p_node);
    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) = 0;

    virtual int UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer) {
        IBDIAGNET_ENTER;
        unpack_data_func(data_to_unpack, unpacked_buffer);
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
    }

    bool IsEnabledByPCAM(const IBNode* p_node) const;

    inline PhyDiag *        GetPhyDiag()                    { return m_phy_diag;                }

    inline uint64_t         GetNotSupportedBit() const      { return m_not_supported_bit;       }
    inline const string&    GetName() const                 { return m_name;                    }
    inline const string&    GetHeader() const               { return m_header;                  }
    inline const string&    GetSectionName() const          { return m_section_name;            }
    inline uint32_t         GetRegisterID() const           { return m_register_id;             }
    inline uint32_t         GetFieldsNum() const            { return m_fields_num;              }
    inline bool             GetDumpEnabled() const          { return m_dump_enabled;            }
    inline AccRegVia_t      GetRegisterType() const         { return m_acc_reg_type;            }
    inline AccRegVia_t      GetRegisterPriority() const     { return m_acc_type_priority;       }

    inline bool             GetRetriveDisconnected() const  { return m_retrive_disconnected;    }
    inline void             SetRetriveDisconnected(bool retrive_disconnected) {
        this->m_retrive_disconnected = retrive_disconnected;
    }

    inline bool IsRegSupportNodeType(const IBNode* type);

    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header);

    virtual void DumpRegisterData(const struct acc_reg_data &, stringstream &sstream, const AccRegKey &key) const = 0;
    virtual void GetGroupData(struct acc_reg_data& areg, u_int8_t& group, u_int16_t& indices);
    virtual int AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors);
    static int SensorsBitsToList(u_int64_t sensors, u_int8_t max_sensors, list_uint8& sensors_list);
    static int SensorsCountToList(u_int8_t sensor_count, list_uint8& sensors_list);

    virtual int BuildDB(AccRegHandler* p_handler,
                        list_p_fabric_general_err &phy_errors,
                        ProgressBar *progress_bar);

    virtual bool IsValidPakcet(const struct acc_reg_data & reg_data) const { return true; }

    int HandleNodeNotSupportAccReg(PhyDiag *p_phy_diag,
                                   IBNode* p_node,
                                   uint64_t not_supported_bit);

    virtual AccRegVia_t Validation(IBNode *p_node, int &rc); // todo - change to abstract

    virtual void HandleData(IBNode* p_node, AccRegKey* p_key, acc_reg_data& reg_data) const { }

    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key);

    virtual bool CheckRegisterKey(export_data_phy_port_t *p_export_data_phy_port,
                                  export_data_phy_node_t *p_export_data_phy_node,
                                  AccRegKey *p_key);

    int FillLocalPortAndPlaneInd(const AccRegKey* p_key, uint8_t * local_port, uint8_t * plane_ind);
    int FillLocalPortAndPlaneInd(IBPort *p_port, uint8_t * local_port, uint8_t * plane_ind);
    bool isCageManager(IBPort* p_port);
};


class PCAMRegister : public Register {
public:
    uint8_t m_access_reg_group;
    uint8_t m_feature_group;

public:
    PCAMRegister(PhyDiag * phy_diag);
    virtual ~PCAMRegister() {}

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual int UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer);
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const;
    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key);

    virtual AccRegVia_t Validation(IBNode *p_node, int &rc);
};

class MGIRRegister : public Register {
public:
    MGIRRegister(PhyDiag * phy_diag);
    virtual ~MGIRRegister() {}

    static void DumpRegisterHeader_hw_info(stringstream &sstream);
    static void DumpRegisterData_hw_info(stringstream &sstream, const mgir_reg &mgir);

    static void DumpRegisterHeader_fw_info(stringstream &sstream);
    static void DumpRegisterData_fw_info(stringstream &sstream, const mgir_reg &mgir);

    static void DumpRegisterHeader_sw_info(stringstream &sstream);
    static void DumpRegisterData_sw_info(stringstream &sstream, const mgir_reg &mgir);

    static void DumpRegisterHeader_dev_info(stringstream &sstream);
    static void DumpRegisterData_dev_info(stringstream &sstream, const mgir_reg &mgir);

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const;
    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key);
};


class SLRegister : public Register {
private:
    map_akey_areg *p_mpein_map; // for PCI mode only (pnat = 3)
    map_akey_areg *p_mpir_map;  // for PCI mode only (pnat = 3)

public:
    SLRegister(PhyDiag * phy_diag,
               uint32_t reg_id,
               const unpack_data_func_t p_unpack_func,
               string sn,
               string name,
               uint32_t fn,
               uint64_t nsb,
               map_akey_areg *mpein_map,
               map_akey_areg *mpir_map);
    virtual ~SLRegister(){}

    virtual void DumpRegisterData(const struct acc_reg_data &, stringstream &sstream, const AccRegKey &key) const = 0;

    virtual int BuildDB(AccRegHandler* p_handler,
                        list_p_fabric_general_err &phy_errors,
                        ProgressBar *progress_bar);

    virtual bool CheckRegisterKey(export_data_phy_port_t *p_export_data_phy_port,
                                  export_data_phy_node_t *p_export_data_phy_node,
                                  AccRegKey *p_key);
};

class SLTPRegister : public SLRegister {
private:
        u_int8_t    m_pnat;

public:
        SLTPRegister(PhyDiag * phy_diag,
                     u_int8_t pnat = ACC_REG_PNAT_IB_PORT,
                     string sn = ACC_REG_SLTP_INTERNAL_SECTION_NAME,
                     map_akey_areg *mpein_map = NULL,
                     map_akey_areg *mpir_map = NULL);
        virtual ~SLTPRegister(){}


        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

        void Header_Dump_40nm_28nm(stringstream &sstream) const;
        void Header_Dump_16nm(stringstream &sstream) const;
        void Header_Dump_7nm(stringstream &sstream) const;
        void Header_Dump_5nm(stringstream &sstream) const;

        void Dump_40nm_28nm(const struct sltp_reg &sltp, stringstream &sstream) const;
        void Dump_16nm(const struct sltp_reg &sltp, stringstream &sstream) const;
        void Dump_7nm(const struct sltp_reg &sltp, stringstream &sstream) const;
        void Dump_5nm(const struct sltp_reg &sltp, stringstream &sstream) const;

        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
};

class SLRGRegister : public SLRegister {
private:
        u_int8_t    m_pnat;

public:
        SLRGRegister(PhyDiag * phy_diag,
                     u_int8_t pnat = ACC_REG_PNAT_IB_PORT,
                     string sn = ACC_REG_SLRG_INTERNAL_SECTION_NAME,
                     map_akey_areg *mpein_map = NULL,
                     map_akey_areg *mpir_map = NULL);
        virtual ~SLRGRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;


        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

        void Dump_40nm_28nm(const struct slrg_reg &slrg, stringstream &sstream) const;
        void Dump_16nm(const struct slrg_reg &slrg, stringstream &sstream) const;
        void Dump_7nm(const struct slrg_reg &slrg, stringstream &sstream) const;
        void Dump_5nm(const struct slrg_reg &slrg, stringstream &sstream) const;

        void Header_Dump_40nm_28nm(stringstream &sstream) const;
        void Header_Dump_16nm(stringstream &sstream) const;
        void Header_Dump_7nm(stringstream &sstream) const;
        void Header_Dump_5nm(stringstream &sstream) const;

        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
};

class SLLMRegister : public SLRegister {
private:
        u_int8_t    m_pnat;
public:
    SLLMRegister(PhyDiag * phy_diag,
                 u_int8_t pnat = ACC_REG_PNAT_IB_PORT,
                 string sn = ACC_REG_SLLM_INTERNAL_SECTION_NAME,
                 map_akey_areg *mpein_map = NULL,
                 map_akey_areg *mpir_map = NULL);

    virtual ~SLLMRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

    void Dump_16nm(const struct sllm_reg &sllm, stringstream &sstream) const;
    void Dump_7nm(const struct sllm_reg &sllm, stringstream &sstream) const;

    void Header_Dump_16nm(stringstream &sstream);
    void Header_Dump_7nm(stringstream &sstream);

    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key) override;
};

class SLRPRegister : public SLRegister {
private:
        u_int8_t    m_pnat;

public:
        SLRPRegister(PhyDiag * phy_diag,
                     u_int8_t pnat = ACC_REG_PNAT_IB_PORT,
                     string sn = ACC_REG_SLRP_INTERNAL_SECTION_NAME,
                     map_akey_areg *mpein_map = NULL,
                     map_akey_areg *mpir_map = NULL);
        virtual ~SLRPRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

        void Header_Dump_40nm_28nm(stringstream &sstream) const;
        void Header_Dump_16nm(stringstream &sstream) const;
        void Header_Dump_7nm(stringstream &sstream) const;
        void Header_Dump_5nm(stringstream &sstream) const;

        void Dump_40nm_28nm(const struct slrp_reg &slrp, stringstream &sstream) const;
        void Dump_16nm(const struct slrp_reg &slrp, stringstream &sstream) const;
        void Dump_7nm(const struct slrp_reg &slrp, stringstream &sstream) const;
        void Dump_5nm_scc(const struct slrp_reg &slrp, stringstream &sstream) const;

        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
};

class SLSIRRegister : public SLRegister {
private:
    u_int8_t    m_pnat;

public:
    SLSIRRegister(PhyDiag * phy_diag,
                  u_int8_t pnat = ACC_REG_PNAT_IB_PORT,
                  string sn = ACC_REG_SLSIR_INTERNAL_SECTION_NAME,
                  map_akey_areg *mpein_map = NULL,
                  map_akey_areg *mpir_map = NULL);

    virtual ~SLSIRRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

    void Header_Dump_7nm(stringstream &sstream) const;
    void Header_Dump_5nm(stringstream &sstream) const;

    void Dump_7nm(const struct slsir_reg &slsir, stringstream &sstream) const;
    void Dump_5nm(const struct slsir_reg &slsir, stringstream &sstream) const;

    virtual AccRegVia_t Validation(IBNode *p_node, int &rc) override;

    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key) override;
};

class PTASRegister : public Register {
public:

        PTASRegister(PhyDiag * phy_diag);
        virtual ~PTASRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
};

class PMDRRegister : public Register {
private:
        u_int8_t    m_pnat;

public:
        PMDRRegister(PhyDiag * phy_diag);
        virtual ~PMDRRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
};

class PMCRRegister : public Register {
public:
        PMCRRegister(PhyDiag * phy_diag);
        virtual ~PMCRRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
};

class PTYSRegister : public Register {
private:
        u_int8_t    m_pnat;

public:
        PTYSRegister(PhyDiag * phy_diag);
        virtual ~PTYSRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
};

class PPSLSRegister : public Register {
private:
        u_int8_t    m_pnat;

public:
        PPSLSRegister(PhyDiag * phy_diag);
        virtual ~PPSLSRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
};

class MPSCRRegister : public Register {
public:
        MPSCRRegister(PhyDiag * phy_diag);
        virtual ~MPSCRRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
};

class MTSRRegister : public Register {
public:
        MTSRRegister(PhyDiag * phy_diag);
        virtual ~MTSRRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
};

class PLTCRegister : public Register {
private:
        u_int8_t    m_pnat;
public:
        PLTCRegister(PhyDiag * phy_diag);
        virtual ~PLTCRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
};

class PAOSRegister : public Register {
private:
        u_int8_t    m_pnat;
public:
        PAOSRegister(PhyDiag * phy_diag);
        virtual ~PAOSRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
};

class PPHCRRegister : public Register {
private:
        u_int8_t    m_pnat;

public:
        PPHCRRegister(PhyDiag * phy_diag);
        virtual ~PPHCRRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
};

// PEUCG - UPHY
class PEUCGRegister : public Register {
protected:
        u_int8_t           m_pnat;
        UPHY::MadBuilder*  m_cur_mad_builder;

public:
        PEUCGRegister(PhyDiag * phy_diag, string section_name);
        virtual ~PEUCGRegister(){}

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
};

class PEUCG_Ver_Register : public PEUCGRegister {
public:
        PEUCG_Ver_Register(PhyDiag * phy_diag);
        virtual ~PEUCG_Ver_Register(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
        virtual void HandleData(IBNode* p_node, AccRegKey* p_key, acc_reg_data& reg_data) const;
};

class PEUCG_DLN_Register : public PEUCGRegister {

public:
        PEUCG_DLN_Register(PhyDiag * phy_diag);
        virtual ~PEUCG_DLN_Register(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);

        virtual int BuildDB(AccRegHandler* p_handler,
                            list_p_fabric_general_err &phy_errors,
                            ProgressBar *progress_bar);
};

class PEUCG_CLN_Register : public PEUCGRegister {

    map_akey_areg *p_pmdr_map;

public:
    PEUCG_CLN_Register(PhyDiag * phy_diag, map_akey_areg *pmdr_map);
    virtual ~PEUCG_CLN_Register(){}

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);

    virtual int BuildDB(AccRegHandler* p_handler,
                        list_p_fabric_general_err &phy_errors,
                        ProgressBar *progress_bar);
};

class PPLLRegister : public Register {
public:

        PPLLRegister(PhyDiag * phy_diag);
        virtual ~PPLLRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

        void Header_Dump_28nm(stringstream &sstream) const;
        void Header_Dump_16nm(stringstream &sstream) const;
        void Header_Dump_7nm(stringstream &sstream) const;
        void Header_Dump_5nm(stringstream &sstream) const;

        void Dump_28nm(const struct ppll_reg &ppll, stringstream &sstream) const;
        void Dump_16nm(const struct ppll_reg &ppll, stringstream &sstream) const;
        void Dump_7nm(const struct ppll_reg &ppll, stringstream &sstream) const;
        void Dump_5nm(const struct ppll_reg &ppll, stringstream &sstream) const;

        void Dump_pll_28nm(const struct pll_28nm &pll_28, stringstream &sstream) const;
        void Dump_pll_16nm(const struct pll_status_data &pll_16, stringstream &sstream) const;

        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
};

class MPEINRegister : public Register {
public:
        MPEINRegister(PhyDiag * phy_diag);
        virtual ~MPEINRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

        virtual int BuildDB(AccRegHandler* p_handler,
                            list_p_fabric_general_err &phy_errors,
                            ProgressBar *progress_bar) override;

        virtual bool IsValidPakcet(const struct acc_reg_data & reg_data) const override;

        int LoopPCIDepth(const clbck_data_t &clbck_data,
                         int rec_status,
                         void *p_attribute_data);

        int LoopPCINode(const clbck_data_t &clbck_data,
                        int rec_status,
                        void *p_attribute_data);

        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;

        virtual bool CheckRegisterKey(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      AccRegKey *p_key) override;
};

class MPIRRegister : public Register {

    map_akey_areg &mpein_map;

public:
    MPIRRegister(PhyDiag * phy_diag, map_akey_areg &mpein_map);
    virtual ~MPIRRegister(){}

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

    virtual int BuildDB(AccRegHandler* p_handler, // handler containing the source map of mpein
                        list_p_fabric_general_err &phy_errors,
                        ProgressBar *progress_bar) override;
};


// fans registers
class MFCRRegister : public Register {
public:
    MFCRRegister(PhyDiag * phy_diag);
    virtual ~MFCRRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;
    virtual int AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors_list);
};

class MFSMRegister : public Register {
public:
    MFSMRegister(PhyDiag * phy_diag);
    virtual ~MFSMRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;

    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key);
};

class MFSLRegister : public Register {
public:
    MFSLRegister(PhyDiag * phy_diag);
    virtual ~MFSLRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;
};

class FORERegister : public Register {
public:
    FORERegister(PhyDiag * phy_diag);
    virtual ~FORERegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;
};

class MFNRRegister : public Register {
public:
    MFNRRegister(PhyDiag * phy_diag);
    virtual ~MFNRRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;
};

// temperature registers
class MTCAPRegister : public Register {
public:
    MTCAPRegister(PhyDiag * phy_diag);
    virtual ~MTCAPRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;
    virtual int AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors_list);
};

class MTMPRegister : public Register {
public:
    MTMPRegister(PhyDiag * phy_diag);
    virtual ~MTMPRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;

    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key);
};

class MTWERegister : public Register {
public:
    MTWERegister(PhyDiag * phy_diag);
    virtual ~MTWERegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;
};

// current/voltage registers
class MVCAPRegister : public Register {
public:
    MVCAPRegister(PhyDiag * phy_diag);
    virtual ~MVCAPRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const;
    virtual int AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors_list);
};

class MVCRRegister : public Register {
public:
    MVCRRegister(PhyDiag * phy_diag);
    virtual ~MVCRRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;

    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key);
};

// power supplies register
class MSPSRegister : public Register {
public:
    MSPSRegister(PhyDiag * phy_diag);
    virtual ~MSPSRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;
};

// general information register
class MSGIRegister : public Register {
public:
    MSGIRegister(PhyDiag * phy_diag);
    virtual ~MSGIRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;

    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key);
};

class MSCIRegister : public Register {
public:
    MSCIRegister(PhyDiag* phy_diag);
    virtual ~MSCIRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);

    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header)  override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;

    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key);
};

/*
 * Registers for HDR bringup
 * */

class SLRIPRegister : public SLRegister {
private:
    u_int8_t    m_pnat;

public:
     SLRIPRegister(PhyDiag * phy_diag,
        u_int8_t pnat = ACC_REG_PNAT_IB_PORT,
        string sn = ACC_REG_SLRIP_INTERNAL_SECTION_NAME,
        map_akey_areg *mpein_map = NULL,
        map_akey_areg *mpir_map = NULL);

    virtual ~SLRIPRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

    void Header_Dump_16nm(stringstream &sstream) const;
    void Header_Dump_7nm(stringstream &sstream) const;
    void Header_Dump_5nm(stringstream &sstream) const;

    void Dump_16nm(const struct slrip_reg &slrip, stringstream &sstream) const;
    void Dump_7nm(const struct slrip_reg &slrip, stringstream &sstream) const;
    void Dump_5nm(const struct slrip_reg &slrip, stringstream &sstream) const;

    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key) override;
};

class SLREGRegister : public Register {
public:
    SLREGRegister(PhyDiag * phy_diag);
    virtual ~SLREGRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

    void Header_Dump_16nm(stringstream &sstream) const;
    void Header_Dump_7nm(stringstream &sstream) const;
    void Header_Dump_5nm(stringstream &sstream) const;

    void Dump_16nm(const struct slreg_reg &slreg, stringstream &sstream) const;
    void Dump_7nm(const struct slreg_reg &slreg, stringstream &sstream) const;
    void Dump_5nm(const struct slreg_reg &slreg, stringstream &sstream) const;
};

class SLCCTRegister : public Register {
public:
    SLCCTRegister(PhyDiag * phy_diag);
    virtual ~SLCCTRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
};

class PPAMPRegister : public Register {
public:
    PPAMPRegister(PhyDiag * phy_diag);
    virtual ~PPAMPRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void GetGroupData(struct acc_reg_data& areg, u_int8_t& group, u_int16_t& indices) override;
};


class PDDRRegister : public Register {
protected:
     u_int8_t m_port_type;
     u_int8_t m_lp_msb;
     u_int8_t m_pnat;
     //u_int8_t m_local_port; // will be set by key
     u_int8_t m_page_select;
     u_int8_t m_module_info_ext;

public:
    PDDRRegister(PhyDiag * phy_diag,
                 uint32_t page_id,
                 const unpack_data_func_t p_unpack_func,
                 string sn,
                 string name,
                 uint32_t fn,
                 string header,
                 PhyPluginSupportedNodesType support_nodes,
                 bool enabled_dump = true,
                 bool retrieve_disconnected = false)

        : Register(phy_diag,
                    ACCESS_REGISTER_ID_PDDR,
                    p_unpack_func,
                    sn,
                    name,
                    fn,
                    NSB::get(this),
                    header,
                    support_nodes,
                    enabled_dump,
                    retrieve_disconnected,
                    VIA_GMP,
                    VIA_GMP),

         m_port_type(0),
         m_lp_msb(0),
         m_pnat(ACC_REG_PNAT_IB_PORT),
         m_page_select(page_id),
         m_module_info_ext(0)
    {
    }

    virtual ~PDDRRegister(){}

    virtual int UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer);
    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
};

class PDDRModuleInfoRegister : public PDDRRegister {
    private:
        static string transmitter_technology_arr[];
        static string ib_compliance_code_arr[];
    public:
        PDDRModuleInfoRegister(PhyDiag * phy_diag);
        virtual ~PDDRModuleInfoRegister() { }

        static void DumpRawDataHeader(stringstream &sstream);
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        static void DumpLayout(stringstream &sstream, const pddr_module_info &module_info);

        static string ConvertCableTemperatureToStr(const struct pddr_module_info &module_info, u_int16_t temp);
        static string ConvertVoltageToStr(u_int16_t vcc);
        static string ConvertMaxPowerToStr(const struct pddr_module_info &module_info);
        static const string& ConvertTransmitterTechnologyToStr(u_int8_t transmitter_technology);
        static string ConvertCableTechnologyBitsToStr(const struct pddr_module_info &module_info);
        static string ConvertCableIdentifierToStr(const struct pddr_module_info &module_info);
        static string ConvertCableLengthToStr(const struct pddr_module_info &module_info);
        static string ConvertCableLengthSMFiberToStr(const struct pddr_module_info &module_info);
        static string ConvertCableLengthOMXToStr(const struct pddr_module_info &module_info, u_int8_t om_index);
        static bool IsCMISCable(uint8_t identifier);
        static string ConvertIBComplianceCodeToStr(u_int16_t ib_compliance_code);
        static u_int8_t SelectNominalBR(const struct pddr_module_info &module_info);
        static u_int8_t SelectTransmitterTechnology(const struct pddr_module_info &module_info);
        static bool IsModule(const struct pddr_module_info &module_info);
        static bool IsActiveCable(const struct pddr_module_info &module_info);
        static bool IsPassiveCable(const struct pddr_module_info &module_info);
        static string ConvertCDREnableTxRxToStr(const struct pddr_module_info &module_info, bool is_rx,
                                                bool is_csv);
        static string ConvertTxEQRxAMPRxEMPToStr(const struct pddr_module_info &module_info, u_int8_t val,
                                                bool is_csv);
        static string ConvertAttenuationToStr(const struct pddr_module_info &module_info, bool is_csv);
        static string ConvertFWVersionToStr(const struct pddr_module_info &module_info, bool is_csv);
        static string ConvertDateCodeToStr(const struct pddr_module_info &module_info);
        static string ConvertRevisionToStr(const struct pddr_module_info &module_info);
        static double MWToDBM(double mW);
        static void DumpModuleInfoHeader(stringstream &sstream);
        static void DumpModuleInfoData(stringstream &sstream, const struct pddr_module_info *p_module_info,
                                    u_int16_t module_info_ext);
        static void DumpModuleInfoData(ofstream &sout, const struct pddr_module_info *p_module_info,
                                    u_int16_t module_info_ext);
        static PhyCableRecord::ModuleRecord *CreateRecord(const struct pddr_module_info *p_module_info,
                                                        u_int16_t module_info_ext);
        static void CopyTo(const pddr_module_info &from, PhyCableRecord::ModuleRecord &to);

};

class PDDRLinkUpInfoRegister : public PDDRRegister {
public:
    PDDRLinkUpInfoRegister(PhyDiag * phy_diag);
    virtual ~PDDRLinkUpInfoRegister() {}

    static void DumpLayout(stringstream &sstream, const struct pddr_link_up_info_page &linkup_info);
    static void DumpLinkUpInfoRealHeader(stringstream &sstream, const string& header);
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
};

class PDDRLatchedFlagInfoRegister : public PDDRRegister {
public:
    PDDRLatchedFlagInfoRegister(PhyDiag * phy_diag);
    virtual ~PDDRLatchedFlagInfoRegister() { }

    static void DumpRawDataHeader(stringstream &sstream);
    static void DumpLayout(stringstream &sstream, const module_latched_flag_info &latched_flag_info);
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

    static void DumpLatchedFlagInfoHeader(stringstream &sstream);
    static void DumpLatchedFlagInfoData(stringstream &sstream, const struct module_latched_flag_info *p_latched_flag_info);
    static void DumpLatchedFlagInfoData(ofstream &sout, const struct module_latched_flag_info *p_latched_flag_info);
    static PhyCableRecord::LatchedRecord *CreateRecord(const struct module_latched_flag_info *p_latched_flag_info);
    static void CopyTo(const module_latched_flag_info &from , PhyCableRecord::LatchedRecord &to);
};

class PDDRPhyInfoRegister: public PDDRRegister {
public:
    PDDRPhyInfoRegister(PhyDiag * phy_diag);
    virtual ~PDDRPhyInfoRegister() {}

    static void DumpLayout(stringstream &sstream, const struct pddr_phy_info_page &phy_info);
    static void DumpPhyInfoRealHeader(stringstream &sstream, const string& header);
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
};

// PDDR Troubleshooting Info Register
class PDDRTroubleshootingInfoRegister: public PDDRRegister {
public:
    PDDRTroubleshootingInfoRegister(PhyDiag * phy_diag);
    virtual ~PDDRTroubleshootingInfoRegister() {}

    static void DumpLayout(stringstream &sstream, const pddr_troubleshooting_page &troubleshooting_info);
    static void DumpRealHeader(stringstream &sstream, const string& header);
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
};

// PDDR Operation Info Register
class PDDROperationInfoRegister: public PDDRRegister {
public:
    PDDROperationInfoRegister(PhyDiag * phy_diag);
    virtual ~PDDROperationInfoRegister() {}

    static void DumpLayout(stringstream &sstream, const pddr_operation_info_page &operation_info);
    static void DumpRealHeader(stringstream &sstream, const string& header);
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
};

// PDDR Link Down Info Register
class PDDRLinkDownInfoRegister: public PDDRRegister {
public:
    PDDRLinkDownInfoRegister(PhyDiag * phy_diag);
    virtual ~PDDRLinkDownInfoRegister() {}

    static void DumpLayout(stringstream &sstream, const pddr_link_down_info_page &link_down_info);
    static void DumpRealHeader(stringstream &sstream, const string& header);
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
};


class PEMIRegister : public Register {
private:
     u_int8_t m_lp_msb;
     u_int8_t m_pnat;
     //u_int8_t m_local_port; // will be set by key
     u_int8_t m_page_select;
     u_int8_t m_module_info_ext;

public:
     PEMIRegister(PhyDiag * phy_diag,
                 uint32_t page_id,
                 const unpack_data_func_t p_unpack_func,
                 string sn,
                 string name,
                 uint32_t fn,
                 string header = "",
                 PhyPluginSupportedNodesType support_nodes = SUPPORT_SW_CA,
                 bool enabled_dump = true,
                 bool retrieve_disconnected = false)

        : Register(phy_diag,
                    ACCESS_REGISTER_ID_PEMI,
                    p_unpack_func,
                    sn,
                    name,
                    fn,
                    NSB::get(this),
                    header,
                    support_nodes,
                    enabled_dump,
                    retrieve_disconnected,
                    VIA_GMP,
                    VIA_GMP),

         m_lp_msb(0),
         m_pnat(ACC_REG_PNAT_IB_PORT),
         m_page_select(page_id),
         m_module_info_ext(0)
    {
    }

    virtual ~PEMIRegister(){}

    virtual int UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer);
    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
};

class PEMI_Module_Samples_Register : public PEMIRegister {
    public:
        PEMI_Module_Samples_Register(PhyDiag * phy_diag);
        virtual ~PEMI_Module_Samples_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };


class PEMI_SNR_Samples_Register : public PEMIRegister {
    public:
        PEMI_SNR_Samples_Register(PhyDiag * phy_diag);
        virtual ~PEMI_SNR_Samples_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };


class PEMI_Laser_Samples_Register : public PEMIRegister {
    public:
        PEMI_Laser_Samples_Register(PhyDiag * phy_diag);
        virtual ~PEMI_Laser_Samples_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };

class PEMI_PAM4_Samples_Register : public PEMIRegister {
    public:
        PEMI_PAM4_Samples_Register(PhyDiag * phy_diag);
        virtual ~PEMI_PAM4_Samples_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };

class PEMI_PRE_FEC_BER_Samples_Register : public PEMIRegister {
    public:
        PEMI_PRE_FEC_BER_Samples_Register(PhyDiag * phy_diag);
        virtual ~PEMI_PRE_FEC_BER_Samples_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };

class PEMI_FERC_Samples_Register : public PEMIRegister {
    public:
        PEMI_FERC_Samples_Register(PhyDiag * phy_diag);
        virtual ~PEMI_FERC_Samples_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };

class PEMI_Laser_Source_Module_Essential_Samples_Register : public PEMIRegister {
    public:
        PEMI_Laser_Source_Module_Essential_Samples_Register(PhyDiag * phy_diag);
        virtual ~PEMI_Laser_Source_Module_Essential_Samples_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };

class PEMI_Module_Properties_Register : public PEMIRegister {
    public:
        PEMI_Module_Properties_Register(PhyDiag * phy_diag);
        virtual ~PEMI_Module_Properties_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };

class PEMI_SNR_Properties_Register : public PEMIRegister {
    public:
        PEMI_SNR_Properties_Register(PhyDiag * phy_diag);
        virtual ~PEMI_SNR_Properties_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };


class PEMI_Laser_Properties_Register : public PEMIRegister {
    public:
        PEMI_Laser_Properties_Register(PhyDiag * phy_diag);
        virtual ~PEMI_Laser_Properties_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };


class PEMI_PAM4_Properties_Register : public PEMIRegister {
    public:
        PEMI_PAM4_Properties_Register(PhyDiag * phy_diag);
        virtual ~PEMI_PAM4_Properties_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };


class PEMI_PRE_FEC_BER_Properties_Register : public PEMIRegister {
    public:
        PEMI_PRE_FEC_BER_Properties_Register(PhyDiag * phy_diag);
        virtual ~PEMI_PRE_FEC_BER_Properties_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };


class PEMI_FERC_Properties_Register : public PEMIRegister {
    public:
        PEMI_FERC_Properties_Register(PhyDiag * phy_diag);
        virtual ~PEMI_FERC_Properties_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };

class PEMI_Laser_Source_Module_Essential_Properties_Register : public PEMIRegister {
    public:
    PEMI_Laser_Source_Module_Essential_Properties_Register(PhyDiag * phy_diag);
        virtual ~PEMI_Laser_Source_Module_Essential_Properties_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };


class PPCNTRegister : public Register {
private:
     u_int8_t m_swid;
     u_int8_t m_lp_msb;
     u_int8_t m_pnat;
     //u_int8_t m_local_port; // will be set by key
     u_int8_t m_grp;
     u_int8_t m_lp_gl;
     u_int8_t m_prio_tc;
     u_int8_t m_port_type;
     u_int8_t m_plane_ind;

public:
     PPCNTRegister(PhyDiag * phy_diag,
                 uint32_t page_id,
                 const unpack_data_func_t p_unpack_func,
                 string sn,
                 string name,
                 uint32_t fn,
                 string header = "",
                 PhyPluginSupportedNodesType support_nodes = SUPPORT_SW_CA,
                 bool enabled_dump = true,
                 bool retrieve_disconnected = false)

        : Register(phy_diag,
                    ACCESS_REGISTER_ID_PPCNT,
                    p_unpack_func,
                    sn,
                    name,
                    fn,
                    NSB::get(this),
                    header,
                    support_nodes,
                    enabled_dump,
                    retrieve_disconnected,
                    VIA_GMP,
                    VIA_GMP),
         m_swid(0),
         m_lp_msb(0),
         m_pnat(ACC_REG_PNAT_IB_PORT),
         m_grp(page_id),
         m_lp_gl(1),
         m_prio_tc(0),
         m_port_type(0),
         m_plane_ind(0)
    {
    }

    virtual ~PPCNTRegister(){}

    virtual int UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer);
    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
};

class PPCNT_PLR_Counters : public PPCNTRegister {
    public:
        PPCNT_PLR_Counters(PhyDiag * phy_diag);
        virtual ~PPCNT_PLR_Counters() { }

        static void DumpFieldNames(stringstream &sstream);
        static void DumpData(const struct ppcnt_plr_counters& reg, stringstream &sstream, bool has_cap_max_ret_rate);
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class PPCNT_Packet_Type_Counters : public PPCNTRegister {
    public:
        PPCNT_Packet_Type_Counters(PhyDiag * phy_diag);
        virtual ~PPCNT_Packet_Type_Counters() { }

        static void DumpFieldNames(stringstream &sstream);
        static void DumpData(const struct ppcnt_packet_type_counter& reg, stringstream &sstream);
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class PPCNT_InfiniBand_General_Counters : public PPCNTRegister {
    public:
        PPCNT_InfiniBand_General_Counters(PhyDiag * phy_diag);
        virtual ~PPCNT_InfiniBand_General_Counters() { }

        static void DumpFieldNames(stringstream &sstream);
        static void DumpData(const struct ppcnt_infiniband_general_counter& reg, stringstream &sstream);
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class PPCNT_RS_Fec_Histograms_Counters : public PPCNTRegister {
    public:
        PPCNT_RS_Fec_Histograms_Counters(PhyDiag * phy_diag);
        virtual ~PPCNT_RS_Fec_Histograms_Counters() { }

        static void DumpFieldNames(stringstream &sstream);
        static void DumpData(const struct ppcnt_rs_fec_histograms_counters& reg, stringstream &sstream);
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class PPCNT_Phys_Layer_Cntrs : public PPCNTRegister {
    public:
        PPCNT_Phys_Layer_Cntrs(PhyDiag * phy_diag);
        virtual ~PPCNT_Phys_Layer_Cntrs() { }

        static void DumpFieldNames(stringstream &sstream);
        static void DumpData(const struct phys_layer_cntrs& reg, stringstream &sstream);
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class PPCNT_Phys_Layer_Stat_Cntrs : public PPCNTRegister {
    public:
        PPCNT_Phys_Layer_Stat_Cntrs(PhyDiag * phy_diag);
        virtual ~PPCNT_Phys_Layer_Stat_Cntrs() { }

        static void DumpFieldNames(stringstream &sstream);
        static void DumpData(const struct phys_layer_stat_cntrs& reg, stringstream &sstream, 
                             bool is_effective_counter_supported, bool is_raw_ber_per_lane_supported);
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class PPCNT_L1_General_Counters : public PPCNTRegister {
    public:
        PPCNT_L1_General_Counters(PhyDiag * phy_diag);
        virtual ~PPCNT_L1_General_Counters() { }

        static void DumpFieldNames(stringstream &sstream);
        static void DumpData(const struct ppcnt_l1_general_counters& reg, stringstream &sstream);
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class PPCNT_Recovery_Counters : public PPCNTRegister {
    public:
        PPCNT_Recovery_Counters(PhyDiag * phy_diag);
        virtual ~PPCNT_Recovery_Counters() { }

        void DumpFieldNames(stringstream &sstream);
        void DumpData(const struct ppcnt_recovery_counters& reg, stringstream &sstream) const;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class PRTLRegister : public Register {
    public:
        PRTLRegister(PhyDiag * phy_diag);
        virtual ~PRTLRegister() { }

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
        virtual int UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer) override;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key) override;
    };


class MPCNTRegister : public Register {
private:
     map_akey_areg &m_mpein_map;

     u_int8_t m_depth;
     u_int8_t m_pcie_index;
     u_int8_t m_node;
     u_int8_t m_grp;

public:
     MPCNTRegister(PhyDiag * phy_diag,
                     map_akey_areg &mpein_map,
                     uint32_t grp,
                     const unpack_data_func_t p_unpack_func,
                     string sn,
                     string name,
                     uint32_t fn,
                     string header = "",
                     PhyPluginSupportedNodesType support_nodes = SUPPORT_CA,
                     bool enabled_dump = true,
                     bool retrieve_disconnected = false)

        : Register(phy_diag,
                    ACCESS_REGISTER_ID_MPCNT,
                    p_unpack_func,
                    sn,
                    name,
                    fn,
                    NSB::get(this),
                    header,
                    support_nodes,
                    enabled_dump,
                    retrieve_disconnected,
                    VIA_GMP,
                    VIA_GMP),

         m_mpein_map(mpein_map),
         m_depth(0),
         m_pcie_index(0),
         m_node(0),
         m_grp(grp)
    {
    }

    virtual ~MPCNTRegister(){}

    virtual int UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer);
    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);

    virtual int BuildDB(AccRegHandler* p_handler, // handler containing the source map of mpein
                        list_p_fabric_general_err &phy_errors,
                        ProgressBar *progress_bar);
};



class MPCNT_PCIe_Counters_Register : public MPCNTRegister {
    public:
        MPCNT_PCIe_Counters_Register(PhyDiag * phy_diag, map_akey_areg &mpein_map);
        virtual ~MPCNT_PCIe_Counters_Register() { }

        static void DumpFieldNames(stringstream &sstream);
        static void DumpData(const struct pcie_perf_counters & reg, stringstream &sstream);
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    };


class MPCNT_PCIe_Lane_Counters_Register : public MPCNTRegister {
    public:
        MPCNT_PCIe_Lane_Counters_Register(PhyDiag * phy_diag, map_akey_areg &mpein_map);
        virtual ~MPCNT_PCIe_Lane_Counters_Register() { }

        static void DumpFieldNames(stringstream &sstream);
        static void DumpData(const struct pcie_lanes_counters & reg, stringstream &sstream);
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    };

class MPCNT_PCIe_Timers_Register : public MPCNTRegister {
    public:
        MPCNT_PCIe_Timers_Register(PhyDiag * phy_diag, map_akey_areg &mpein_map);
        virtual ~MPCNT_PCIe_Timers_Register() { }

        static void DumpFieldNames(stringstream &sstream);
        static void DumpData(const struct pcie_timers_states & reg, stringstream &sstream);
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    };

class MPCNT_PCIe_Phys_Layer_Register : public MPCNTRegister {
    public:
        MPCNT_PCIe_Phys_Layer_Register(PhyDiag * phy_diag, map_akey_areg &mpein_map);
        virtual ~MPCNT_PCIe_Phys_Layer_Register() { }

        static void DumpData(const struct pcie_physical_layer & reg, stringstream &sstream);
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    };


class PPBMPRegister : public Register {
private:
        u_int8_t    m_pnat;

public:
    PPBMPRegister(PhyDiag * phy_diag);
    virtual ~PPBMPRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
};


class MRCSRegister : public Register {
public:
    MRCSRegister(PhyDiag * phy_diag);
    virtual ~MRCSRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
};

class PPSLCRegister : public Register {
private:
    u_int8_t    m_pnat;
public:
    PPSLCRegister(PhyDiag * phy_diag);
    virtual ~PPSLCRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class PPSLDRegister : public Register {
private:
    u_int8_t m_pnat;
public:
    PPSLDRegister(PhyDiag * phy_diag);
    virtual ~PPSLDRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class MLPCRegister : public Register {
private:
    u_int8_t m_pnat;
public:
    MLPCRegister(PhyDiag * phy_diag);
    virtual ~MLPCRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class MPPGSRegister : public Register {
public:
    MPPGSRegister(PhyDiag * phy_diag);
    virtual ~MPPGSRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key) override;
};
class MPPDRRegister : public Register {
    public:
        MPPDRRegister(PhyDiag * phy_diag);
        virtual ~MPPDRRegister() { }

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;

};
class MPPGHRegister : public Register {
public:
    MPPGHRegister(PhyDiag * phy_diag);
    virtual ~MPPGHRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class PPSPCRegister : public Register {
private:
    u_int8_t m_pnat;
public:
    PPSPCRegister(PhyDiag * phy_diag);
    virtual ~PPSPCRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class PPSPGCRegister : public Register {

public:
    PPSPGCRegister(PhyDiag * phy_diag);
    virtual ~PPSPGCRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class PPSPIRegister : public Register {
private:
    u_int8_t m_pnat;
public:
    PPSPIRegister(PhyDiag * phy_diag);
    virtual ~PPSPIRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key) override;

};


class PPSLGRegister : public Register {
private:
    u_int8_t m_pnat;
    u_int8_t m_page_select;
public:
    PPSLGRegister(PhyDiag * phy_diag,
                 uint32_t page_id,
                 const unpack_data_func_t p_unpack_func,
                 string sn,
                 string name,
                 uint32_t fn,
                 string header = "",
                 PhyPluginSupportedNodesType support_nodes = SUPPORT_SW_CA,
                 bool enabled_dump = true,
                 bool retrieve_disconnected = false)

        : Register(phy_diag,
                   ACCESS_REGISTER_ID_PPSLG,
                   p_unpack_func,
                   sn,
                   name,
                   fn,
                   NSB::get(this),
                   header,
                   support_nodes,
                   enabled_dump,
                   retrieve_disconnected,
                   VIA_GMP,
                   VIA_GMP)
                ,m_pnat(ACC_REG_PNAT_IB_PORT)
                ,m_page_select(page_id)
{

}
    virtual ~PPSLGRegister() { }

    virtual int UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer);
    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);

};


class PPSLG_L1_Capabilities_And_Status : public PPSLGRegister {
    public:
        PPSLG_L1_Capabilities_And_Status(PhyDiag * phy_diag);
        virtual ~PPSLG_L1_Capabilities_And_Status() { }

        void DumpFieldNames(stringstream &sstream);
        void DumpData(const struct ppslg_l1_capabilities_and_status& reg, stringstream &sstream) const;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class PPSLG_L1_Configuration : public PPSLGRegister {
    public:
        PPSLG_L1_Configuration(PhyDiag * phy_diag);
        virtual ~PPSLG_L1_Configuration() { }

        void DumpFieldNames(stringstream &sstream);
        void DumpData(const struct ppslg_l1_configuration& reg, stringstream &sstream) const;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class PPSLG_L1_Debug : public PPSLGRegister {
    public:
        PPSLG_L1_Debug(PhyDiag * phy_diag);
        virtual ~PPSLG_L1_Debug() { }

        void DumpFieldNames(stringstream &sstream);
        void DumpData(const struct ppslg_l1_debug_page& reg, stringstream &sstream) const;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class PPSLG_L1_Statistical_Counters : public PPSLGRegister {
    public:
        PPSLG_L1_Statistical_Counters(PhyDiag * phy_diag);
        virtual ~PPSLG_L1_Statistical_Counters() { }

        void DumpFieldNames(stringstream &sstream);
        void DumpData(const struct ppslg_l1_statistical_counters& reg, stringstream &sstream) const;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class MPPGCRRegister : public Register {
public:
    MPPGCRRegister(PhyDiag * phy_diag);
    virtual ~MPPGCRRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual int UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer) override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key) override;

};

class MPPCRRegister : public Register {
private:
    void DumpProfileConfig(const power_profile_config& profile_conf, stringstream &sstream) const;
public:
    MPPCRRegister(PhyDiag * phy_diag);
    virtual ~MPPCRRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key) override;

};

// AccRegHandler defines a generic way of reading registers' data across the fabric.
// There are 4 types of access to register:
// by port
// by port + lane
// by node
// by pll group
class AccRegHandler {
protected:
    list_p_fabric_general_err   *phy_errors;
    bool                        can_send_mads_by_lid;
    int                         clbck_error_state;
    Register                    *p_reg;
    string                      handler_header;

public:
    map_akey_areg               data_map;

    AccRegHandler(Register *p_register,  string hh);
    virtual ~AccRegHandler();

    virtual int BuildDB(list_p_fabric_general_err &phy_errors) = 0;


    int PCIAddressCollectGetClbck(const clbck_data_t &clbck_data,
                                  int rec_status,
                                  void *p_attribute_data);

    int SMPAccessRegisterHandlerGetClbck(const clbck_data_t &clbck_data,
                                         int rec_status,
                                         void *p_attribute_data);

    int GMPAccessRegisterHandlerGetClbck(const clbck_data_t &clbck_data,
                                         int rec_status,
                                         void *p_attribute_data);

    void DumpCSV(CSVOut &csv_out);

    Register * GetPReg() {
        return p_reg;
    }

    bool GetDumpEnabled() {
        return p_reg->GetDumpEnabled();
    }

    void SetHeader(string header) {
        handler_header = header;
    }

    int SendSMPReg(IBNode *p_curr_node,
                   phys_port_t port_num,
                   SMP_AccessRegister *smp_reg,
                   AccRegKey *p_key,
                   ProgressBar *p_progress_bar,
                   clbck_data_t *call_back_data = NULL);

    int SendGMPReg(IBNode *p_curr_node,
                   uint16_t lid,
                   GMP_AccessRegister *gmp_reg,
                   AccRegKey *p_key,
                   ProgressBar *p_progress_bar,
                   clbck_data_t *call_back_data = NULL);

    int SendAccReg(AccRegVia_t acc_reg_via,
                   IBNode *p_curr_node,
                   phys_port_t port_num,
                   uint16_t lid,
                   AccessRegister &acc_reg,
                   AccRegKey *p_key,
                   ProgressBar *p_progress_bar,
                   clbck_data_t *call_back_data = NULL);

    virtual void ExportData(export_data_phy_port_t *p_export_data_phy_port,
                            export_data_phy_node_t *p_export_data_phy_node);

protected:
    void ExportDataImplNode(export_data_phy_node_t *p_export_data_phy_node);

    template<typename KeyType>
    void ExportDataImplPort(export_data_phy_port_t *p_export_data_phy_port);
};

class AccRegNodeHandler : public AccRegHandler {
public:
    AccRegNodeHandler(Register *p_reg):
        AccRegHandler(p_reg, NODE_HEADER){}
    virtual ~AccRegNodeHandler(){}

    virtual int BuildDB(list_p_fabric_general_err &phy_errors);

    virtual void ExportData(export_data_phy_port_t *p_export_data_phy_port,
                            export_data_phy_node_t *p_export_data_phy_node);
};

class AccRegPortHandler : public AccRegHandler {
public:
    AccRegPortHandler(Register *p_reg):
        AccRegHandler(p_reg, PORT_HEADER){}
    virtual ~AccRegPortHandler(){}

    virtual AccRegKey* CreateKey(uint64_t ng, uint64_t pg, uint8_t pn);

    virtual int BuildDB(list_p_fabric_general_err &phy_errors);

    virtual void ExportData(export_data_phy_port_t *p_export_data_phy_port,
                            export_data_phy_node_t *p_export_data_phy_node);
};

class AccRegPortIndexHandler : public AccRegPortHandler {
private:

    vector_uint64  indexes;
    uint64_t       cur_index;

public:
    AccRegPortIndexHandler(Register *p_reg, vector_uint64& idxs) :
        AccRegPortHandler(p_reg), indexes(idxs), cur_index(0)
    {
        handler_header = PORT_INDEX_HEADER;
    }

    virtual ~AccRegPortIndexHandler(){}

    virtual AccRegKey* CreateKey(uint64_t ng, uint64_t pg, uint8_t pn);
    virtual int BuildDB(list_p_fabric_general_err &phy_errors);
};

class AccRegGroupHandler : public AccRegHandler {
private:
    uint32_t group_max_idx;
public:
    AccRegGroupHandler(Register *p_reg):
        AccRegHandler(p_reg, GROUP_HEADER), group_max_idx(0){}
    virtual ~AccRegGroupHandler(){}

    virtual int BuildDB(list_p_fabric_general_err &phy_errors);

    virtual void ExportData(export_data_phy_port_t *p_export_data_phy_port,
                            export_data_phy_node_t *p_export_data_phy_node);
};

class AccRegNodeSensorHandler : public AccRegHandler {
private:
    AccRegHandler *p_acc_reg_handler; // pointer for the capability register

public:
    AccRegNodeSensorHandler(Register *p_reg, AccRegHandler *p_acc_reg):
        AccRegHandler(p_reg, NODE_SENSOR_HEADER),
        p_acc_reg_handler(p_acc_reg){}
    virtual ~AccRegNodeSensorHandler(){}

    virtual int BuildDB(list_p_fabric_general_err &phy_errors);

    virtual void ExportData(export_data_phy_port_t *p_export_data_phy_port,
                            export_data_phy_node_t *p_export_data_phy_node);
};

class AccRegTypeGroupHandler : public AccRegHandler {
private:

    uint8_t max_type;
    uint16_t max_indices;

public:
    AccRegTypeGroupHandler(Register *p_reg, uint8_t types, uint16_t indices)
        : AccRegHandler(p_reg, TYPE_GROUP_HEADER), max_type(types), max_indices(indices){}
    virtual ~AccRegTypeGroupHandler(){}

    virtual int BuildDB(list_p_fabric_general_err &phy_errors);
};


class AccRegSpecificHandler : public AccRegHandler {
public:
    AccRegSpecificHandler(Register *p_reg, string header = PORT_HEADER)
        : AccRegHandler(p_reg, header){}
    virtual ~AccRegSpecificHandler(){}

    virtual int BuildDB(list_p_fabric_general_err &phy_errors);

    virtual void ExportData(export_data_phy_port_t *p_export_data_phy_port,
                            export_data_phy_node_t *p_export_data_phy_node);
};


class AccRegPortLaneHandler : public AccRegHandler {
    u_int8_t max_idx_in_lane;
public:
    AccRegPortLaneHandler(Register *p_reg, u_int8_t max_idx = 1):
        AccRegHandler(p_reg, PORT_LANE_HEADER), max_idx_in_lane(max_idx){}

    virtual ~AccRegPortLaneHandler() { }

    virtual int BuildDB(list_p_fabric_general_err &phy_errors);

    virtual void ExportData(export_data_phy_port_t *p_export_data_phy_port,
                            export_data_phy_node_t *p_export_data_phy_node);
};



#endif //_ACC_REGISTER_H_

