/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 2021-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#ifndef _PHY_DIAG_H_
#define _PHY_DIAG_H_

#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>
#include "../internal_packets/internal_packets_layouts.h"
#include "acc_reg.h"
#include "diagnostic_data.h"
#include "uphy/uphy_common.h"

#include <bitset>

#define NSB_BITSET(node) ((PHYNodeData*)(node->p_phy_data))->nsb_bitset
#define SET_NSB(node, bit) ((PHYNodeData*)(node->p_phy_data))->nsb_bitset.set(bit)
//typedef vector <struct SMP_AccessRegister *>    vec_p_acceess_reg_t;
typedef vector < struct VS_DiagnosticData * >   vec_p_diagnostic_data_t;
typedef vector < vec_p_diagnostic_data_t >      vec_vec_diagnostic_data_t;
typedef list < string >                         list_string_t;

typedef map < AccRegKey *, struct VS_DiagnosticData *, bool (*)(AccRegKey *, AccRegKey *)>
        map_akey_diagnostic_data_t;
typedef vector < map_akey_diagnostic_data_t *>   vec_map_akey_diagnostic_data_t;

typedef vector <long double *>                  vec_long_double_t;

typedef vector < struct slrg_reg * >            vector_p_slrg_reg_tbl;
typedef vector < vector_p_slrg_reg_tbl >        vector_v_slrg_reg_tbl;
typedef vector < struct slrp_reg * >            vector_p_slrp_reg_tbl;
typedef vector < vector_p_slrp_reg_tbl >        vector_v_slrp_reg_tbl;
typedef vector < struct sltp_reg * >            vector_p_sltp_reg_tbl;
typedef vector < vector_p_sltp_reg_tbl >        vector_v_sltp_reg_tbl;

typedef vector < struct ptas_reg * >            vector_p_ptas_reg_tbl;

typedef vector < class AccRegHandler * > vec_pareg;
typedef vector < class DiagnosticDataInfo * > vec_diagnostic_data_t;

struct BER_thresholds_warning_error {
    double warning;
    double error;
    double normal;

    BER_thresholds_warning_error(): warning(0), error(0), normal(0) {}
    BER_thresholds_warning_error(double warning, double error, double normal)
        : warning(warning), error(error), normal(normal) {}
};

typedef enum {
    MT_UNKNOWN     = 0,
    MT_DACS,
    MT_ACC,
    MT_ACTIVE,
    MT_ACTIVE_DID,
    MT_NUM_OF_MEDIA_TYPES
} MediaType;

struct BER_thresholds_key {
    SerDesVersion technology;
    IBLinkSpeed   speed;
    IBFECMode     fec_mode;
    MediaType     media_type;

    BER_thresholds_key(SerDesVersion inTech, IBLinkSpeed inSpeed, IBFECMode inFec, MediaType inMedia) :
        technology(inTech), speed(inSpeed), fec_mode(inFec), media_type(inMedia){}


    bool operator< (const BER_thresholds_key& other) const {
        if (this == &other)   // Short?cut self?comparison
            return false;

        if (technology < other.technology)
            return true;
        else if (other.technology < technology)
            return false;

        if (speed < other.speed)
            return true;
        else if (other.speed < speed)
            return false;

        if (fec_mode < other.fec_mode)
            return true;
        else if (other.fec_mode < fec_mode)
            return false;

        if (media_type < other.media_type)
            return true;
        else if (other.media_type < media_type)
            return false;

        // All members are equal, so return false
        return false;;
    }
} ;

typedef vector < struct BER_thresholds_warning_error > vec_ber_thresholds_val_t;
typedef map < BER_thresholds_key, vec_ber_thresholds_val_t > map_ber_thresholds_t;

void PhyCountersGetDelegator(const clbck_data_t &clbck_data,
                             int rec_status,
                             void *p_attribute_data);
void PhyCountersResetDelegator(const clbck_data_t &clbck_data,
                               int rec_status,
                               void *p_attribute_data);
void PCICountersGetDelegator(const clbck_data_t &clbck_data,
                             int rec_status,
                             void *p_attribute_data);
void SMPAccessRegisterGetDelegator(const clbck_data_t &clbck_data,
                                   int rec_status,
                                   void *p_attribute_data);
void SMPAccessRegisterPTASGetDelegator(const clbck_data_t &clbck_data,
                                   int rec_status,
                                   void *p_attribute_data);

typedef void (*pack_port_lane_reg_t) (u_int32_t port,
                                      u_int32_t lane,
                                      u_int8_t *buff);

//data for registers with access by port and lane
typedef struct portlane_reg {
    u_int16_t            reg_id;
    u_int64_t            not_support_bit;
    pack_port_lane_reg_t pack;
    string               section;
} portlane_reg_t;


struct ModuleInfoExt {
    const pddr_module_info *p_module_info;
    u_int16_t module_info_ext;

    ModuleInfoExt(): p_module_info(NULL), module_info_ext(0){}
};
/******************************************************/
/******************************************************/
/******************************************************/
class PhyDiag : public Plugin {
public:
    typedef std::map <AccRegKey *,
                      pair<ModuleInfoExt,
                           const struct module_latched_flag_info*>,
                      bool (*)(AccRegKey *, AccRegKey *)> acc_reg_cable_info_map_t;

private:
    ////////////////////
    //members
    ////////////////////
    IBFabric                     *p_discovered_fabric;
    Ibis                         *p_ibis_obj;
    IBDMExtendedInfo             *p_ibdm_extended_info;
    CapabilityModule             *p_capability_module;
    list_p_fabric_general_err     phy_errors;
    list_p_fabric_general_err     phy_effective_ber_errors;
    list_p_fabric_general_err     phy_symbol_ber_errors;
    list_p_fabric_general_err     pci_errors;

    bool                          can_send_mads_by_lid;
    bool                          to_get_phy_info;
    bool                          to_get_ppamp;
    bool                          to_get_slrg;
    bool                          to_get_slreg;
    bool                          to_use_reg;
    bool                          to_dump_cap_reg;
    bool                          to_reset_counters;
    bool                          to_get_pci_info;
    bool                          to_reset_pci_counters;
    bool                          to_show_ber_warnings;
    bool                          to_load_ber_thresholds_table;
    bool                          to_get_cable_disconnected_ports;
    bool                          to_get_power_telemetry_registers;
    bool                          to_use_cage_manager;
    int                           clbck_error_state;
    u_int32_t                     total_ber_errors;
    string                        ber_thresholds_table_file;
    int                           module_info_dd_index;
    int                           module_info_acc_reg_index;

    vector_p_node                 nodes_vector;
    vector_p_port                 ports_vector;

    //port_cidx_and_dd_idx_to_dd_page[port_creation_idx][dd_idx]
    vec_vec_diagnostic_data_t     port_cidx_and_dd_idx_to_dd_page;
    vec_vec_diagnostic_data_t     node_cidx_and_dd_idx_to_dd_page;
    vec_map_akey_diagnostic_data_t dd_idx_and_akey_to_dd_page;

    vec_long_double_t             ber_vec;
    vec_long_double_t             eff_ber_vec;

    vec_pareg                     reg_handlers_vec;
    vec_pareg                     pci_reg_handlers_vec;
    vec_diagnostic_data_t         diagnostic_data_vec;

    map_ber_thresholds_t          ber_thresholds_table;

    set_string_t                  enabled_regs;
    set_string_t                  disabled_regs;
    map_string_2_set_string_t     register_dependencies;

    list_string_t                 uphy_dumps;

    PF_export_data_phy_node       pf_export_data_phy_node;
    PF_export_data_phy_port       pf_export_data_phy_port;

    ////////////////////
    //methods
    ////////////////////
    //Returns: IBDIAGNET_SUCCESS_CODE / IBDIAGNET_ERR_CODE_DB_ERR /
    //IBDIAGNET_ERR_CODE_NO_MEM / IBDIAGNET_ERR_CODE_FABRIC_ERROR
    int BuildPhyCounters(list_p_fabric_general_err &phy_errors,
                           u_int32_t dd_idx);
    //Returns: IBDIAGNET_SUCCESS_CODE / IBDIAGNET_ERR_CODE_DB_ERR /
    //IBDIAGNET_ERR_CODE_NO_MEM / IBDIAGNET_ERR_CODE_FABRIC_ERROR
    int BuildPCICounters(list_p_fabric_general_err &phy_errors,
                           u_int32_t dd_idx,
                           map_akey_areg *data_map);
    /*
     * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
     *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
     */
    inline u_int32_t getNodesVectorSize() {
        return (u_int32_t)nodes_vector.size(); }
    IBNode* getNodePtr(u_int32_t node_index);
    inline u_int32_t getPortsVectorSize() {
        return (u_int32_t)ports_vector.size(); }
    IBPort* getPortPtr(u_int32_t port_index);

    template <typename T>
    void addPtrToVec(vector<T*>& vector_obj, T *p_obj);

    template <typename T, typename V>
    int addDataToVec(vector<T*>& vector_obj,
            T *p_obj,
            vector<V*>& vector_data,
            V& data);

    template <typename T>
    T * getPtrFromVec(std::vector<T*> & vector_obj, u_int32_t idx);

    template <typename T, typename V>
    int addDataToVecInVec(vector<T*>& vector_obj,
                          T *p_obj,
                          vector< vector<V*> >& vec_of_vectors,
                          u_int32_t data_idx,
                          V& data);

    template <typename T>
    T * getPtrFromVecInVec(std::vector< std::vector<T*> > & vec_of_vectors,
                           u_int32_t idx1,
                           u_int32_t idx2);

    int addDataToMapInVec(AccRegKey *p_acc_reg_key,
                          vec_map_akey_diagnostic_data_t& vec_of_maps,
                          u_int32_t data_idx,
                          struct VS_DiagnosticData& data);

    bool getPhysStatIndex(u_int32_t& idx);
    bool getPhysStatForPort(IBPort* p_curr_port, u_int32_t idx, phys_layer_stat_cntrs& phy_statistics);

    int addPhysLayerPortCounters(
    IBPort *p_port,
    struct VS_DiagnosticData &physCounters,
    u_int32_t dd_idx);

    struct VS_DiagnosticData * getPhysLayerPortCounters(u_int32_t port_index,
                                                        u_int32_t dd_idx);

    int addPhysLayerNodeCounters(IBNode *p_node,
                                 struct VS_DiagnosticData &physCounters,
                                 u_int32_t dd_idx);

    struct VS_DiagnosticData* getPhysLayerNodeCounters(u_int32_t node_index,
                                                        u_int32_t dd_idx);

    int addPhysLayerPCICounters(AccRegKey *p_acc_reg_key,
                                struct VS_DiagnosticData &physCounters,
                                u_int32_t dd_idx);

    int addBER(IBPort *p_port, long double ber);
    long double* getBER(u_int32_t port_index);

    int addEffBER(IBPort *p_port, long double ber);
    long double* getEffBER(u_int32_t port_index);

    void DumpCSVPhyCounters(CSVOut &csv_out, u_int32_t section_id);

    bool CollectAccRegCableInfo(acc_reg_cable_info_map_t & info);

    void DumpCSV_AccRegCableInfo(CSVOut &csv_out);

    void DumpCablesForSimulator(CSVOut &csv_out, bool is_module);
    void DumpCablesForSimulator();

    int DumpFile_AccRegCableInfo(const string &file_name);
    void DumpFile_AccRegCableInfo(ofstream &sout);

    void DumpCSVPCICounters(CSVOut &csv_out, u_int32_t dd_type);
    void DumpCSVRawBER(CSVOut &csv_out);
    void DumpCSVEffectiveBER(CSVOut &csv_out);
    void DumpCSVSLRGExternalInfo(CSVOut &csv_out, AccRegHandler *p_areg_handler);
    void CalcRawBER();
    int CalcEffBER(u_int64_t ber_threshold_reciprocal_val,
                   list_p_fabric_general_err& ber_errors);

    int HandleSpecialPorts(IBNode *p_curr_node,
                           IBPort *p_curr_port,
                           u_int32_t i);

    int DumpNetDumpExt();
    int DumpNetDumpExtAgg();
    int DumpNetDumpExtPort(ostream &sout, IBPort *p_curr_port, IBNode *p_curr_node,
        u_int32_t idx, bool is_agg);
    int DumpNetDumpExtPortBySys(ostream &sout, IBPort *p_curr_port,
            IBNode *p_curr_node, u_int32_t dd_idx);
    void ReportPortBERErrors(IBPort *p_curr_port, enum IBBERType ber_entry, double ber_report,
        string ber_description, list_p_fabric_general_err *phy_ber_errors,
        bool found_phys_layer_stat_cntrs);

    int DumpCSVSocketDirect();

    int DumpCSV_UPHY_Versions();
    int DumpCSV_UPHY(string section_name);

    bool isSupportFwBER(IBPort *p_port);

    void InitRegisterDependencies();
    void InitBERTable();
    void Init7NMBERTable();
    void Init16NMBERTable();

    const vec_ber_thresholds_val_t *GetBerThresholdEntry(const IBPort &port);
    MediaType GetMediaType(const IBPort &port);
    bool GetModuleInfo(const IBPort &port, pddr_module_info &container,
                       const pddr_module_info *&p_module_info);
    bool GetModuleInfoByAccRegister(const IBPort &port, const pddr_module_info *&p_module_info);
    bool GetModuleInfoByDiagnosticData(const IBPort &port,
                                       pddr_module_info &container,
                                       const pddr_module_info *&p_module_info);

    int ParseRegistersList(const string &registers_list_str, set_string_t & registers_list);
    void CheckRegisterDependencies();
    int LoadUPHYFile(const string &filename);
    int ParseUPHYFileList(const string &uphy_path);
    int ParseUPHYDumps(const string &uphy_dumps_args);
    bool IsEnabledByFilter(const string &register_name);
    int ParseBERThresholdTable();
    int CreateBERThresholdTable(const string &file_path);

    int ParseAccRegPriorityValue(string value, string &result);
    void CheckCPLDVersionMismatch();
    int CheckFabricPCISpeedDegradation();

    void ExportToIBPort(IBPort &port, const struct ModuleInfoExt *p_module_info,
                    const struct module_latched_flag_info *p_latched_flag_info);

public:
    ////////////////////
    //methods
    ////////////////////
    PhyDiag(IBDiag *p_ibdiag);
    ~PhyDiag();

    /*
     * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
     *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
     */
    int SMPAccRegGetByDirect(IN direct_route_t *p_direct_route,
                             IN u_int8_t port_num,
                             OUT struct SMP_AccessRegister *p_smp_acc_reg,
                             const clbck_data_t *p_clbck_data);

    int GMPAccRegGet(IN u_int16_t lid,
                     IN u_int32_t register_id,
                     OUT struct GMP_AccessRegister *p_gmp_acc_reg,
                     IN const clbck_data_t *p_clbck_data);


    //You can use this function without having PhyDiag instance.
    static inline bool IsSupportDiagnosticDataCountersFailed(IBNode* p_node,
                                                             u_int8_t not_support_bit)
    {
        return NSB_BITSET(p_node)[NOT_SUPPORT_DIAGNOSTIC_DATA] || NSB_BITSET(p_node)[not_support_bit];
    }


    inline void SetPhyInfoGet(bool set_val) {
        this->to_get_phy_info = set_val;
        this->MarkStage(set_val);
    }

    inline void SetPPAMPGet(bool set_val) {
        this->to_get_ppamp = set_val;
    }

    inline void SetSLRGGet(bool set_val) {
        this->to_get_slrg = set_val;
    }

    inline void SetSLREGGet(bool set_val) {
        this->to_get_slreg = set_val;
    }

    inline void SetUseAccReg(bool set_val) {
        this->to_use_reg = set_val;
    }

    inline void SetUseCageManager(bool set_val) {
        this->to_use_cage_manager = set_val;
    }

    inline bool getUseCageManager() {
        return this->to_use_cage_manager;
    }

    inline void SetCapRegShow(bool set_val) {
        this->to_dump_cap_reg = set_val;
    }

    inline void SetPhyInfoClear(bool set_val) {
        this->to_reset_counters = set_val;
        this->MarkStage(set_val);
    }

    inline bool ShouldResetCounters() {
        return this->to_reset_counters;
    }

    inline void SetPowerTelemetryRegisters(bool set_val) {
        this->to_get_power_telemetry_registers = set_val;
    }

    inline bool GetPowerTelemetryRegisters() const {
        return this->to_get_power_telemetry_registers;
    }

    inline void SetPciInfoGet(bool set_val) {
        this->to_get_pci_info = set_val;
        this->MarkStage(set_val);
    }

    inline void SetPciInfoClear(bool set_val) {
        this->to_reset_pci_counters = set_val;
        this->MarkStage(set_val);
    }

    inline void SetShowBERWarnings(bool to_show_ber_warnings) {
        this->to_show_ber_warnings = to_show_ber_warnings;
    }

    inline void SetBERThresholdTableFile(string ber_thresholds_table_file) {
        this->to_load_ber_thresholds_table = true;
        this->ber_thresholds_table_file = ber_thresholds_table_file;
    }

    inline void SetToGetCableDisconnectedPorts(bool to_get_cable_disconnected_ports) {
        this->to_get_cable_disconnected_ports = to_get_cable_disconnected_ports;
    }

    inline bool GetToGetCableDisconnectedPorts() {
        return this->to_get_cable_disconnected_ports;
    }

    bool IsActive();
    void Init() {}
    int Prepare();
    int RetrieveInfo();
    int RunCheck();

    //Returns: 0 - OK / 1 - error / 2 - OK but need to exit
    int HandleOption(string name, string value);

    //PhyDiag callbacks
    void PhyCountersGetClbck(const clbck_data_t &clbck_data,
                             int rec_status,
                             void *p_attribute_data);
    void PCICountersGetClbck(const clbck_data_t &clbck_data,
                             int rec_status,
                             void *p_attribute_data);

    virtual bool GetFabricSummarySection(stringstream &section);
    virtual int CleanResources();

    static AccRegVia_t acc_reg_priority;
    static inline void SetAccRegPriority(AccRegVia_t val) {
        acc_reg_priority = val;
    }

    int InitPhyDataOnNodes();

    IBFabric * GetFabric() { return p_discovered_fabric; };
    Ibis * GetIbis() { return p_ibis_obj; };

    CapabilityModule * GetCapabilityModule() { return p_capability_module; }
    const CapabilityModule * GetCapabilityModule() const { return p_capability_module; }

    list_string_t & GetUphyDumps() { return uphy_dumps; }
    const list_string_t & GetUphyDumps() const { return uphy_dumps; }

    list_p_fabric_general_err & GetPhyErrors() { return  phy_errors; }
    const list_p_fabric_general_err & GetPhyErrors() const { return  phy_errors; }

    list_p_fabric_general_err & GetEffectiveBerErrors() { return  phy_effective_ber_errors; }
    const list_p_fabric_general_err & GetEffectiveBerErrors() const { return  phy_effective_ber_errors; }

    list_p_fabric_general_err & GetSymbolBerErrors() { return phy_symbol_ber_errors; }
    const list_p_fabric_general_err & GetSymbolBerErrors() const { return phy_symbol_ber_errors; }

    IBNode * GetNode(uint64_t node_guid);
    IBPort * GetPort(uint64_t node_guid, phys_port_t port_num);

    PHYNodeData * GetPHYNodeData(uint64_t node_guid);
    PHYPortData * GetPHYPortData(uint64_t node_guid, phys_port_t port_num);

    int InitExportAPI(list_p_fabric_general_err &export_data_errors);
    int ExportData(export_session_handle_t session_handle,
                   list_p_fabric_general_err &export_data_errors);
};

/******************************************************/
/******************************************************/
/******************************************************/
// the class factories
extern "C" {

Plugin * create(IBDiag *p_ibdiag);
void destroy(Plugin *p);

}

#endif //_PHY_DIAG_H_
