/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <math.h>
#include <string.h>
#include <sstream>
#include <bitset>

#include "infiniband/ibdm/cable/CombinedCable.h"
#include "infiniband/ibdm/cable/CableTypes.h"
#include "infiniband/ibdm/cable/CableErrors.h"
#include "infiniband/ibutils/output_stream_primitives.h"
#include "infiniband/ibutils/trim.h"


#ifndef ARRAY_SIZE
    #define ARRAY_SIZE(arr) (sizeof(arr)/sizeof(arr[0]))
#endif

#ifndef BIT_TO_STR
    #define BIT_TO_STR(val, index) (val & (1 << index) ? "1" : "0")
#else
    #error "the BIT_TO_STR has already been defined"
#endif

// notice should be use only numbers
template <class T>
inline static string _to_ptr_string(T val, int width = (int)(sizeof(T) * 2))
{
    stringstream sstream;
    sstream << PTR(val, width);

    return sstream.str();
}

// Will be remove in C++11, notice should be use only numbers
template <class T>
inline static string _to_string(T val)
{
    stringstream sstream;
    sstream << +val;

    return sstream.str();
}

//The supported format is MT<digits><char><char><digits>
//we cannot use %s format since it will take two chars and anything till input's string ends
string _get_vendor_from_sn(const string& sn)
{
    int x = 0;
    int y = 0;
    char a = 0;
    char b = 0;
    int ret = sscanf(sn.c_str(), "MT%d%c%c%d", &x, &a, &b, &y);
    if (ret != 4)
        return "";

    std::stringstream buffer;
    buffer << a << b;
    return buffer.str();
}

string _to_cvs_quoted(bool is_csv, const string &str)
{
    if (is_csv) {
       stringstream sstream;
       sstream << QUOTED(str);
       return sstream.str();
    }

    return str;
}

bool _is_valid_temperature(int8_t temp, u_int8_t cableType)
{
     return (cableType != CABLE_TYPE_COPPER_UNEQUALIZED       &&
             cableType != CABLE_TYPE_COPPER_PASSIVE_EQUALIZED &&
             temp >= CABLE_MIN_SUPPORT_TEMPERATURE            &&
             temp <= CABLE_MAX_SUPPORT_TEMPERATURE);
}

int _check_temperature_by_threshold(bool isHigh, u_int8_t technology,
                                    int8_t temp, int8_t threshold)
{
    if (!_is_valid_temperature(temp, technology) ||
        !_is_valid_temperature(threshold, technology))
        return 0;

    return isHigh ? (temp > threshold) : (temp < threshold);
}

bool CombinedCableInfo::IsOpticCable() const
{
    if (this->p_cable)
        return this->p_cable->IsOpticCable();

    if (this->p_phy)
        return this->p_phy->IsOpticCable();

    return false;
}

void CombinedCableInfo::SetPrtlLength(float length)
{
    stringstream ss;
    ss << length << " m";

    if (this->p_cable)
        this->p_cable->SetPrtlLength(ss.str());
    else if (this->p_phy)
        this->p_phy->SetPrtlLength(ss.str());
}

void CombinedCableInfo::SetPrtlLength(string length)
{
    if (this->p_cable)
        this->p_cable->SetPrtlLength(length);
    else if (this->p_phy)
        this->p_phy->SetPrtlLength(length);
}

cable_id_t CombinedCableInfo::GetIdObj() const
{
    if (!this->p_cable && !this->p_phy)
        return cable_id_t(); //not valid object is returned

    return (this->p_cable ? this->p_cable->GetIdObj() :
                            this->p_phy->GetIdObj());
}

cable_fw_t CombinedCableInfo::GetFWObj() const
{
    if (!this->p_cable && !this->p_phy)
        return cable_fw_t(); //not valid object is returned

    return (this->p_cable ? this->p_cable->GetFWObj() :
                            this->p_phy->GetFWObj());
}

string CombinedCableInfo::GetTemperatureStr() const {
    if (!this->p_cable && !this->p_phy)
        return "N/A";

    return (this->p_cable ? this->p_cable->GetTemperatureStr() :
                            this->p_phy->GetTemperatureStr());
}

string CombinedCableInfo::GetHighTemperatureThresholdStr() const {
    if (!this->p_cable && !this->p_phy)
        return "N/A";

    return (this->p_cable ? this->p_cable->GetHighTemperatureThresholdStr() :
                            this->p_phy->GetHighTemperatureThresholdStr());
}

string CombinedCableInfo::GetLowTemperatureThresholdStr() const {
    if (!this->p_cable && !this->p_phy)
        return "N/A";

    return (this->p_cable ? this->p_cable->GetLowTemperatureThresholdStr() :
                            this->p_phy->GetLowTemperatureThresholdStr());
}

int CombinedCableInfo::GetTemperatureAlarms() const
{
    if (!this->p_cable && !this->p_phy)
        return -1;

    return (this->p_cable ? this->p_cable->GetTemperatureAlarms() :
                            this->p_phy->GetTemperatureAlarms());
}

int CombinedCableInfo::GetTemperatureErrorsByTreshold() const
{
    if (!this->p_cable && !this->p_phy)
        return -1;

    return (this->p_cable ? this->p_cable->GetTemperatureErrorsByTreshold() :
                            this->p_phy->GetTemperatureErrorsByTreshold());
}

string CombinedCableInfo::TemperatureToStr(u_int8_t cableType,
                            int8_t temp, const string& defaultVal)
{
    if (!_is_valid_temperature(temp, cableType))
        return defaultVal;

    stringstream streamStr;
    streamStr << +temp << 'C';

    return streamStr.str();
}

string CombinedCableInfo::VoltageToStr(u_int16_t vcc)
{
    stringstream streamStr;
    streamStr << vcc * 100 << "uV";

    return streamStr.str();
}

string CombinedCableInfo::SupportedSpeedToStr(u_int16_t supportedSpeed, const string& defaultVal)
{
    static const string ib_compliance_code_arr[] = {
                "SDR/",
                "DDR/",
                "QDR/",
                "FDR/",
                "EDR/",
                "HDR/",
                "NDR/",
                "XDR/"
    };

    stringstream  streamStr;
    u_int8_t size = static_cast<u_int8_t>(ARRAY_SIZE(ib_compliance_code_arr));
    for (u_int8_t i = 0; i < size; i++) {
         if (supportedSpeed & (1 << i))
            streamStr << ib_compliance_code_arr[i];
    }

    string str = streamStr.str();
    if (str.empty())
        str = defaultVal;
    else
        str.erase(str.size() - 1, 1);

    return (str);
}

static const string transmitter_technology_arr[] = {
                "850 nm VCSEL"                                             ,
                "1310 nm VCSEL"                                            ,
                "1550 nm VCSEL"                                            ,
                "1310 nm FP"                                               ,
                "1310 nm DFB"                                              ,
                "1550 nm DFB"                                              ,
                "1310 nm EML"                                              ,
                "1550 nm EML"                                              ,
                "others"                                                   ,
                "1490 nm DFB"                                              ,
                "Copper cable- unequalized"                                ,
                "Copper cable- passive equalized"                          ,
                "Copper cable- near and far end limiting active equalizers",
                "Copper cable- far end limiting active equalizers"         ,
                "Copper cable- near end limiting active equalizers"        ,
                "Copper cable- linear active equalizers"
};

bool CombinedCableInfo::IsOpticCable(u_int8_t type)
{
    //based on transmitter_technology_arr[10] = "Copper cable- unequalized"
    return (type < 10);
}

bool CableRecord::IsOpticCable() const
{
    if (this->cable_return_status != CABLEINFO_VS_STATUS_SUCCESS)
        return false;

    return CombinedCableInfo::IsOpticCable(this->cable_type);
}

bool PhyCableRecord::IsOpticCable() const
{
    if (!this->p_module) 
        return false;

    return CombinedCableInfo::IsOpticCable(p_module->SelectTransmitterTechnology());
}

string CombinedCableInfo::CableTypeToStr(u_int8_t type, const string& defaultVal)
{
    string str = defaultVal;

    u_int8_t size = static_cast<u_int8_t>(ARRAY_SIZE(transmitter_technology_arr));
    if (type < size)
        str = transmitter_technology_arr[type];

    return (str);
}

u_int8_t CombinedCableInfo::GetCDRVendor() const
{
    //0 stands for invalid value; OxFF does not exist in any hash.
    return (this->p_phy ? this->p_phy->GetCDRVendor() : 0xFF);
}

string CombinedCableInfo::GetPNVendor() const
{
    if (!this->p_cable && !this->p_phy)
        return ""; //invalid vendor name

    return trim(this->p_cable ? this->p_cable->GetPNVendor() :
                            this->p_phy->GetPNVendor());
}

string ConvertAutonegValueToStr(u_int8_t autoneg_val)
{
    string result;

    switch (autoneg_val) {
    case AUTONEGO_VALUE_IN_PROGRESS: {
        result = "Auto negotiation is in progress";
        break;
    }
    //we don't suppose to get the following errors,
    //we check them before getting here
    case AUTONEGO_VALUE_NOT_QDR: {
        result = "Port is not QDR";
        break;
    }
    case AUTONEGO_VALUE_LINK_DOWN: {
        result = "Link is down";
        break;
    }
    case AUTONEGO_VALUE_SUCCESS: {
        result = "Success";
        break;
    }
    default: {
        result = "UNKNOWN";
        break;
    }
    }

    return (result);
}

string ConvertCableInfoVSStatusToStr(u_int8_t vs_status)
{
    string result;

    switch (vs_status) {
    case CABLEINFO_VS_STATUS_NO_EEPROM: {
        result = "No eeprom in connected cable";
        break;
    }
    case CABLEINFO_VS_STATUS_INVALID_PORT: {
        result = "Invalid port number";
        break;
    }
    case CABLEINFO_VS_STATUS_INVALID_PAGE_NUM: {
        result = "Invalid page number (>15)";
        break;
    }
    case CABLEINFO_VS_STATUS_INVALID_ADDR_OR_SIZE: {
        result = "Address / size  is invalid. ((addr+size)>255) or size>48 or size==0";
        break;
    }
    case CABLEINFO_VS_STATUS_INVALID_SLAVE_ADDR: {
        result = "Slave address is invalid. (slv_addr != 0x50)";
        break;
    }
    //we don't suppose to get the following errors,
    //we check them before getting here
    case CABLEINFO_VS_STATUS_CABLE_NOT_CONNECTED: {
        result = "Cable not connected";
        break;
    }
    case CABLEINFO_VS_STATUS_PORT_NOT_QSFP: {
        result = "Port type is not QSFP (SFP/CX4 or internal)";
        break;
    }
    case CABLEINFO_VS_STATUS_BAD_QSFP_CABLE: {
        result = "Bad QSFP cable (cable ignore ModSel# signal)";
        break;
    }
    case CABLEINFO_VS_STATUS_SUCCESS: {
        result = "Success";
        break;
    }
    default: {
        result = "UNKNOWN";
        break;
    }
    }

    return (result);
};

cable_id_t CableRecord::GetIdObj() const
{
    cable_id_t cableId;

    if (this->pn != "NA" && !this->pn.empty())
        cableId.pn = this->pn;

    cableId.identifier = this->ConvertCableIdentifierToStr();
    if (cableId.identifier == "NA")
        cableId.identifier = ""; //disable cable_id object

    cableId.vendor = _get_vendor_from_sn(this->sn);

    cableId.revision = this->rev;
    if (cableId.revision == "NA")
        cableId.revision = "";

    return cableId;
}

string CableRecord::GetPNVendor() const
{
    return (this->pn != "NA" && !this->pn.empty()) ? this->pn : "";
}

cable_fw_t  CableRecord::GetFWObj() const
{
    cable_fw_t cableFW;

    if (!this->IsMlnxMmf() && !this->IsMlnxPsm())
        return cableFW;

    //TODO::parse it. task 3325602
    return cableFW;
}

string CableRecord::GetTemperatureStr() const {
    return this->ConvertTemperatureToStr(this->temperature, false);
}

string CableRecord::GetHighTemperatureThresholdStr() const
{
    return this->ConvertTemperatureToStr(this->high_temp_alarm_th, false);
}

string CableRecord::GetLowTemperatureThresholdStr() const
{
    return this->ConvertTemperatureToStr(this->low_temp_alarm_th, false);
}

int CableRecord::GetTemperatureErrorsByTreshold() const
{
    int ret = 0;
    int8_t temp = static_cast<int8_t>(this->temperature >> 8);

    if (_check_temperature_by_threshold(true, this->cable_type, temp,
                                        static_cast<int8_t>(this->high_temp_alarm_th >> 8)))
        ret |= CABLE_HIGH_TEMP_MASK;

    if (_check_temperature_by_threshold(false, this->cable_type, temp,
                                        static_cast<int8_t>(this->low_temp_alarm_th >> 8)))
        ret |= CABLE_LOW_TEMP_MASK;

    return ret;
}

int CableRecord::GetTemperatureAlarms() const
{
    int ret = 0;

    ret |= (this->temperature_alarm_and_Warning & 0x80) >> 7;
    ret |= (this->temperature_alarm_and_Warning & 0x40) >> 5;
    ret |= (this->temperature_alarm_and_Warning & 0x20) >> 3;
    ret |= (this->temperature_alarm_and_Warning & 0x10) >> 1;

    return ret;
}

string CableRecord::ConvertLengthToStr() const
{
    if (!this->prtl_length.empty())
        return prtl_length;

    return this->length_str;

}

string CableRecord::ConvertPrtlLengthToStr() const
{
    if (!this->prtl_length.empty())
        return this->prtl_length;

    return "NA";
 }

string CableRecord::ConvertTemperatureToStr(u_int16_t temp, bool is_csv) const {
    const string na_val = (is_csv ? "NA" : "N/A");

    u_int8_t technology = this->cable_type;
    int8_t temperature = static_cast<int8_t>(temp >> 8);

    return CombinedCableInfo::TemperatureToStr(technology, temperature, na_val);
}

string CableRecord::ConvertVoltageToStr(u_int16_t vcc) const {
    return CombinedCableInfo::VoltageToStr(vcc);
}

string CableRecord::ConvertSupportedSpeedToStr() const
{
    return CombinedCableInfo::SupportedSpeedToStr(this->supported_speed, "NA");
}

string CableRecord::ConvertCableIdentifierToStr() const
{
    string result;

    switch (this->identifier) {
    case QSFP_ID_VALUE:
        result = "QSFP";
        break;
    case QSFP_PLUS_ID_VALUE:
        result = "QSFP+";
        break;
    case QSFP28_ID_VALUE:
        result = "QSFP28";
        break;
    case QSFP_DD_ID_VALUE:
        result = "QSFP-DD";
        break;
    case OSFP_ID_VALUE:
        result = "OSFP";
        break;
    case SFP_DD_ID_VALUE:
        result = "SFP-DD";
        break;
    case DSFP_ID_VALUE:
        result = "DSFP";
        break;
    case QSFP_CMIS_ID_VALUE:
        result = "QSFP_CMIS";
        break;
    default:
        result = "NA";
        break;
    }

    return (result);
}

string CableRecord::ConvertCableTypeToStr() const
{
    string str = "NA";
    if (this->cable_return_status != CABLEINFO_VS_STATUS_SUCCESS) {
        str = "NA - " + ConvertCableInfoVSStatusToStr(this->cable_return_status);
        return (str);
    }

    str = CombinedCableInfo::CableTypeToStr(this->cable_type, "NA");
    return (str);
}

bool CableRecord::IsModule() const
{
    if (this->cable_type != CABLE_TYPE_COPPER_UNEQUALIZED &&
        this->connector != 0x23)
            return (true);

    return (false);
}

bool CableRecord::IsActiveCable() const
{
    if (this->cable_type != CABLE_TYPE_COPPER_UNEQUALIZED &&
        this->connector == 0x23)
        return (true);

    return (false);
}

bool CableRecord::IsMlnxMmf() const
{

    if (this->oui == MLNX_OUI_STR &&
        (this->IsModule() || this->IsActiveCable()) &&
        this->mlnx_vendor_byte == 0xE)
        return (true);

    return (false);
}

bool CableRecord::IsMlnxPsm() const
{
    if (this->oui == MLNX_OUI_STR &&
        (this->IsModule() || this->IsActiveCable()) &&
        this->mlnx_vendor_byte == 0x10)
            return (true);

    return (false);
}

bool CableRecord::IsPassiveCable() const
{
    if (this->cable_type == CABLE_TYPE_COPPER_UNEQUALIZED)
            return (true);

    return (false);
}

string CableRecord::ConvertCDREnableTxRxToStr(bool is_csv) const
{
    string str;
    char buffer[24] = {0};

    if (this->IsModule() || this->IsActiveCable()) {
        if (is_csv) {
            snprintf(buffer, sizeof(buffer), "0x%x", this->cdr_control);
            str = buffer;
        } else {
            //Tx
            if ((this->cdr_present & 0x2) != 0) {
                snprintf(buffer, sizeof(buffer), "0x%x ", this->cdr_control >> 4);
                str = buffer;
            } else
                str = "N/A ";
            //Rx
            memset(buffer, 0, sizeof(buffer));
            if ((this->cdr_present & 0x1) != 0) {
                snprintf(buffer, sizeof(buffer), "0x%x", this->cdr_control & 0x0f);
                str += buffer;
            } else
                str += "N/A";
        }
    } else  {
        if (is_csv)
            str = "\"NA\"";
        else
            str = "N/A N/A";
    }

    return (str);
}

string CableRecord::ConvertInputEqToStr(bool is_csv) const
{
    string str;
    char buffer[24] = {0};

    if (IsModule() || IsActiveCable()) {
        if (!is_csv)
            snprintf(buffer, sizeof(buffer), "%u %u %u %u",
                    (this->input_eq >> 12) & 0xf,
                    (this->input_eq >> 8) & 0xf,
                    (this->input_eq >> 4) & 0xf,
                    this->input_eq & 0xf);
        else
            snprintf(buffer, sizeof(buffer), "%x%x%x%x",
                    (this->input_eq >> 12) & 0xf,
                    (this->input_eq >> 8) & 0xf,
                    (this->input_eq >> 4) & 0xf,
                    this->input_eq & 0xf);
        str = buffer;
    } else {
        if (!is_csv)
            str = "N/A";
        else
            str = "\"NA\"";
    }

    return (str);
}

string CableRecord::ConvertOutputAmpToStr(bool is_csv) const
{
    string str;
    char buffer[24] = {0};

    if (IsModule() || IsActiveCable()) {
        if (!is_csv)
            snprintf(buffer, sizeof(buffer), "%u %u %u %u",
                    (this->output_amp >> 12) & 0xf,
                    (this->output_amp >> 8) & 0xf,
                    (this->output_amp >> 4) & 0xf,
                    this->output_amp & 0xf);
        else
            snprintf(buffer, sizeof(buffer), "%x%x%x%x",
                    (this->output_amp >> 12) & 0xf,
                    (this->output_amp >> 8) & 0xf,
                    (this->output_amp >> 4) & 0xf,
                    this->output_amp & 0xf);
        str = buffer;
    } else {
        if (!is_csv)
            str = "N/A";
        else
            str = "\"NA\"";
    }

    return (str);
}

string CableRecord::ConvertOutputEmpToStr(bool is_csv) const
{
    string str;
    char buffer[24] = {0};

    if (IsModule() || IsActiveCable()) {
        if (!is_csv)
            snprintf(buffer, sizeof(buffer), "%u %u %u %u",
                    (this->output_emp >> 12) & 0xf,
                    (this->output_emp >> 8) & 0xf,
                    (this->output_emp >> 4) & 0xf,
                    this->output_emp & 0xf);
        else
            snprintf(buffer, sizeof(buffer), "%x%x%x%x",
                    (this->output_emp >> 12) & 0xf,
                    (this->output_emp >> 8) & 0xf,
                    (this->output_emp >> 4) & 0xf,
                    this->output_emp & 0xf);

        str = buffer;
    } else {
        if (!is_csv)
            str = "N/A";
        else
            str = "\"NA\"";
    }

    return (str);
}

string CableRecord::ConvertFwVersionToStr(bool is_csv) const
{
    string str;

    if (IsMlnxMmf() || IsMlnxPsm()) {
        str = this->fw_version;
    } else  {
        if (!is_csv)
            str = "N/A";
        else
            str = "\"NA\"";
    }

    return (str);
}

string CableRecord::ConvertAttenuationToStr(bool is_csv) const
{
    string str;
    char buffer[24] = {0};

    if (IsPassiveCable()) {
        if (!is_csv)
            snprintf(buffer, sizeof(buffer), "%u %u %u %u",
                    this->attenuation_2_5g,
                    this->attenuation_5g,
                    this->attenuation_7g,
                    this->attenuation_12g);
        else
            snprintf(buffer, sizeof(buffer), "%u,%u,%u,%u",
                    this->attenuation_2_5g,
                    this->attenuation_5g,
                    this->attenuation_7g,
                    this->attenuation_12g);
        str = buffer;
    } else {
        if (!is_csv)
            str = "N/A N/A N/A N/A";
        else
            str = "\"NA\",\"NA\",\"NA\",\"NA\"";
    }

    return (str);
}

string CableRecord::ConvertDateCodeToStr() const
{
    string str;

    if (this->date_code == "N/A")
        str = this->date_code;
    else
        str = this->date_code.substr(0, 2) + "-" + this->date_code.substr(2, 2) + "-" +
              this->date_code.substr(4, 2);

    return (str);
}

string CableRecord::ConvertRXOutputDisableToStr() const
{
    char buffer[64] = {0};
    string str;

    snprintf(buffer, sizeof(buffer), "0x%01x", (this->adaptive_eq_control >> 4) & 0xf);
    str = buffer;

    return (str);
}

string CableRecord::ConvertTXAdaptiveEqualizationEnableToStr() const
{
    char buffer[64] = {0};
    string str;

    snprintf(buffer, sizeof(buffer), "0x%01x", this->adaptive_eq_control & 0xf);
    str = buffer;

    return (str);
}

double CableRecord::mW_to_dBm(double mW) const
{
    if (mW == 0)
        return -999.999;
    return 10 * log10(mW);
}

void CableRecord::ExportData(cable_record_data_t *p_cable_record_data) const
{
    p_cable_record_data->cable_return_status = this->cable_return_status;

    p_cable_record_data->identifier = this->identifier;
    p_cable_record_data->connector = this->connector;
    p_cable_record_data->supported_speed = this->supported_speed;
    p_cable_record_data->cable_type = this->cable_type;
    p_cable_record_data->lengthsmfiber = this->lengthsmfiber;
    p_cable_record_data->lengthom3 = this->lengthom3;
    p_cable_record_data->lengthom2 = this->lengthom2;
    p_cable_record_data->lengthom1 = this->lengthom1;
    p_cable_record_data->lengthcopper = this->lengthcopper;
    p_cable_record_data->temperature = this->temperature;
    p_cable_record_data->nominal_br_100 = this->nominal_br_100;
    p_cable_record_data->nominal_br = this->nominal_br;
    p_cable_record_data->power_class = this->power_class;
    p_cable_record_data->cdr_control = this->cdr_control;
    p_cable_record_data->cdr_present = this->cdr_present;
    p_cable_record_data->mlnx_vendor_byte = this->mlnx_vendor_byte;
    p_cable_record_data->attenuation_2_5g = this->attenuation_2_5g;
    p_cable_record_data->attenuation_5g = this->attenuation_5g;
    p_cable_record_data->attenuation_7g = this->attenuation_7g;
    p_cable_record_data->attenuation_12g = this->attenuation_12g;
    p_cable_record_data->eth_com_codes_10g_40g = this->eth_com_codes_10g_40g;
    p_cable_record_data->eth_com_codes_ext = this->eth_com_codes_ext;
    p_cable_record_data->mlnx_revision = this->mlnx_revision;
    p_cable_record_data->CDR_TX_RX_loss_indicator = this->CDR_TX_RX_loss_indicator;
    p_cable_record_data->adaptive_equalization_fault = this->adaptive_equalization_fault;
    p_cable_record_data->TX_RX_LOL_indicator = this->TX_RX_LOL_indicator;
    p_cable_record_data->temperature_alarm_and_Warning = this->temperature_alarm_and_Warning;
    p_cable_record_data->voltage_alarm_and_warning = this->voltage_alarm_and_warning;
    p_cable_record_data->transmitter_technology = this->transmitter_technology;
    p_cable_record_data->TX_adaptive_equalization_freeze = this->TX_adaptive_equalization_freeze;
    p_cable_record_data->adaptive_eq_control = this->adaptive_eq_control;

    p_cable_record_data->input_eq = this->input_eq;
    p_cable_record_data->output_amp = this->output_amp;
    p_cable_record_data->output_emp = this->output_emp;
    p_cable_record_data->mellanox_cap = this->mellanox_cap;
    p_cable_record_data->diag_supply_voltage = this->diag_supply_voltage;
    p_cable_record_data->RX_power_alarm_and_warning = this->RX_power_alarm_and_warning;
    p_cable_record_data->TX_bias_alarm_and_warning = this->TX_bias_alarm_and_warning;
    p_cable_record_data->TX_power_alarm_and_warning = this->TX_power_alarm_and_warning;
    p_cable_record_data->lot = this->lot;

    p_cable_record_data->high_temp_alarm_th = this->high_temp_alarm_th;
    p_cable_record_data->high_temp_warning_th = this->high_temp_warning_th;
    p_cable_record_data->low_temp_alarm_th = this->low_temp_alarm_th;
    p_cable_record_data->low_temp_warning_th = this->low_temp_warning_th;

    p_cable_record_data->high_vcc_alarm_th = this->high_vcc_alarm_th;
    p_cable_record_data->high_vcc_warning_th = this->high_vcc_warning_th;
    p_cable_record_data->low_vcc_alarm_th = this->low_vcc_alarm_th;
    p_cable_record_data->low_vcc_warning_th = this->low_vcc_warning_th;

    p_cable_record_data->RX1Power = this->RX1Power;
    p_cable_record_data->RX2Power = this->RX2Power;
    p_cable_record_data->RX3Power = this->RX3Power;
    p_cable_record_data->RX4Power = this->RX4Power;
    p_cable_record_data->TX1Bias = this->TX1Bias;
    p_cable_record_data->TX2Bias = this->TX2Bias;
    p_cable_record_data->TX3Bias = this->TX3Bias;
    p_cable_record_data->TX4Bias = this->TX4Bias;
    p_cable_record_data->TX1Power = this->TX1Power;
    p_cable_record_data->TX2Power = this->TX2Power;
    p_cable_record_data->TX3Power = this->TX3Power;
    p_cable_record_data->TX4Power = this->TX4Power;
    p_cable_record_data->RXpower_type = (u_int8_t)this->RXpower_type; // '0'-OMA, '1'-AVP

    p_cable_record_data->qsfp_options = this->qsfp_options;

    strncpy(p_cable_record_data->vendor, this->vendor.c_str(), sizeof(p_cable_record_data->vendor) - 1);
    p_cable_record_data->vendor[sizeof(p_cable_record_data->vendor) - 1] = 0;

    strncpy(p_cable_record_data->oui, this->oui.c_str(), sizeof(p_cable_record_data->oui) - 1);
    p_cable_record_data->oui[sizeof(p_cable_record_data->oui) - 1] = 0;

    strncpy(p_cable_record_data->pn, this->pn.c_str(), sizeof(p_cable_record_data->pn) - 1);
    p_cable_record_data->pn[sizeof(p_cable_record_data->pn) - 1] = 0;

    strncpy(p_cable_record_data->sn, this->sn.c_str(), sizeof(p_cable_record_data->sn) - 1);
    p_cable_record_data->sn[sizeof(p_cable_record_data->sn) - 1] = 0;

    strncpy(p_cable_record_data->rev, this->rev.c_str(), sizeof(p_cable_record_data->rev) - 1);
    p_cable_record_data->rev[sizeof(p_cable_record_data->rev) - 1] = 0;

    strncpy(p_cable_record_data->length_str, this->ConvertLengthToStr().c_str(), sizeof(p_cable_record_data->length_str) - 1);
    p_cable_record_data->length_str[sizeof(p_cable_record_data->length_str) - 1] = 0;

    strncpy(p_cable_record_data->mlnx_sfg_sn, this->mlnx_sfg_sn.c_str(), sizeof(p_cable_record_data->mlnx_sfg_sn) - 1);
    p_cable_record_data->mlnx_sfg_sn[sizeof(p_cable_record_data->mlnx_sfg_sn) - 1] = 0;

    strncpy(p_cable_record_data->fw_version, this->fw_version.c_str(), sizeof(p_cable_record_data->fw_version) - 1);
    p_cable_record_data->fw_version[sizeof(p_cable_record_data->fw_version) - 1] = 0;

    strncpy(p_cable_record_data->date_code, this->date_code.c_str(), sizeof(p_cable_record_data->date_code) - 1);
    p_cable_record_data->date_code[sizeof(p_cable_record_data->date_code) - 1] = 0;
}

#ifndef U8H_FMT
    #define U8H_FMT     "0x%02x"
#else
    #error "The Macro U8H_FMT has already been defined"
#endif

#ifndef U16H_FMT
    #define U16H_FMT    "0x%04x"
#else
    #error "The Macro U16H_FMT has already been defined"
#endif


#define OLD_FORMAT "Identifier: %s (" U8H_FMT ")\nVendor: %s\nOUI: %s\nPN: %s\nSN: %s\nRev: %s\n" \
            "Length: %s\nLength by PRTL: %s\n Length by Reg: %s\n"\
            "Type: %s\nSupportedSpeed: %s\nTemperature: %s\n" \
            "PowerClass: %u\nNominalBitrate: %u Gb/s\nCDREnableTxRx: %s\n" \
            "InputEq: %s\nOutputAmp: %s\nOutputEmp: %s\nFW Version: %s\n" \
            "Attenuation(2.5,5,7,12): %s\nRX power type: %s\n" \
            "RX1 Power: %.3f mW, %.3f dBm\n" \
            "RX2 Power: %.3f mW, %.3f dBm\n" \
            "RX3 Power: %.3f mW, %.3f dBm\n" \
            "RX4 Power: %.3f mW, %.3f dBm\n" \
            "RX5 Power: N/A N/A\n" \
            "RX6 Power: N/A N/A\n" \
            "RX7 Power: N/A N/A\n" \
            "RX8 Power: N/A N/A\n" \
            "TX1 Bias: %.3f mA\n" \
            "TX2 Bias: %.3f mA\n" \
            "TX3 Bias: %.3f mA\n" \
            "TX4 Bias: %.3f mA\n" \
            "TX5 Bias: N/A mA\n" \
            "TX6 Bias: N/A mA\n" \
            "TX7 Bias: N/A mA\n" \
            "TX8 Bias: N/A mA\n" \
            "TX1 Power: %.3f mW, %.3f dBm\n" \
            "TX2 Power: %.3f mW, %.3f dBm\n" \
            "TX3 Power: %.3f mW, %.3f dBm\n" \
            "TX4 Power: %.3f mW, %.3f dBm\n" \
            "TX5 Power: N/A N/A\n" \
            "TX6 Power: N/A N/A\n" \
            "TX7 Power: N/A N/A\n" \
            "TX8 Power: N/A N/A\n" \
            "CDR Latched TX/RX Loss Indicator: " U8H_FMT "\n" \
            "Latched Adaptive Equalization Fault: " U8H_FMT "\n"\
            "Latched  TX / RX LOL Indicator: " U8H_FMT "\n" \
            "Latched Temperature Alarm and Warning: " U8H_FMT "\n" \
            "Latched Voltage Alarm and Warning: " U8H_FMT "\n" \
            "RX Power Alarm and Warning: " U16H_FMT "\n" \
            "TX Bias Alarm and Warning: " U16H_FMT "\n" \
            "TX Power Alarm and Warning: " U16H_FMT "\n" \
            "Supply Voltage Reporting: %s\n" \
            "Transmitter Technology: " U8H_FMT "\n" \
            "Extended Specification Compliance Codes: " U8H_FMT "\n" \
            "AlarmTemperatureHighThresh: %s\n" \
            "AlarmTemperatureLowThresh: %s\n"  \
            "WarnTemperatureHighThresh: %s\n"  \
            "WarnTemperatureLowThresh: %s\n"   \
            "AlarmVoltageHighThresh: %s\n"     \
            "AlarmVoltageLowThresh: %s\n"      \
            "WarnVoltageHighThresh: %s\n"      \
            "WarnVoltageLowThresh: %s\n"       \
            "RXPowerHighThresh: N/A\n"     \
            "RXPowerLowThresh: N/A\n"      \
            "TXPowerHighThresh: N/A\n"     \
            "TXPowerLowThresh: N/A\n"      \
            "TXBiasHighThresh: N/A\n"      \
            "TXBiasLowThresh: N/A\n"       \
            "DateCode: %s\n" \
            "Lot: " U16H_FMT "\n" \
            "TX Adaptive Equalization Freeze: " U8H_FMT "\n" \
            "RX Output Disable: %s\n" \
            "TX Adaptive Equalization Enable: %s\n"

#define PREPEND_COMBINED_FORMAT  "Source: CABLE\n"

#define APPEND_COMBINED_FORMAT  "OutputPreEmp: N/A\n" \
            "OutputPostEmp: N/A\n"

string CableRecord::c_str(bool isCombined) const
{
    const char *p_format = (isCombined ?
                            PREPEND_COMBINED_FORMAT OLD_FORMAT APPEND_COMBINED_FORMAT :
                            OLD_FORMAT);

    char buffer[2048] = {0};
    snprintf(buffer, sizeof(buffer), p_format,
             this->ConvertCableIdentifierToStr().c_str(),
             this->identifier,
             this->vendor.c_str(),
             this->oui.c_str(),
             this->pn.c_str(),
             this->sn.c_str(),
             this->rev.c_str(),
             this->ConvertLengthToStr().c_str(),
             this->ConvertPrtlLengthToStr().c_str(),
             this->length_str.c_str(),
             this->ConvertCableTypeToStr().c_str(),
             this->ConvertSupportedSpeedToStr().c_str(),
             this->ConvertTemperatureToStr(this->temperature, false).c_str(),
             this->power_class,
             this->nominal_br,
             this->ConvertCDREnableTxRxToStr(false).c_str(),
             this->ConvertInputEqToStr(false).c_str(),
             this->ConvertOutputAmpToStr(false).c_str(),
             this->ConvertOutputEmpToStr(false).c_str(),
             this->ConvertFwVersionToStr(false).c_str(),
             this->ConvertAttenuationToStr(false).c_str(),
             this->RXpower_type? "Average Power" : "OMA",
             this->RX1Power, mW_to_dBm(this->RX1Power),
             this->RX2Power, mW_to_dBm(this->RX2Power),
             this->RX3Power, mW_to_dBm(this->RX3Power),
             this->RX4Power, mW_to_dBm(this->RX4Power),
             this->TX1Bias,
             this->TX2Bias,
             this->TX3Bias,
             this->TX4Bias,
             this->TX1Power, mW_to_dBm(this->TX1Power),
             this->TX2Power, mW_to_dBm(this->TX2Power),
             this->TX3Power, mW_to_dBm(this->TX3Power),
             this->TX4Power, mW_to_dBm(this->TX4Power),
             this->CDR_TX_RX_loss_indicator,
             this->adaptive_equalization_fault,
             this->TX_RX_LOL_indicator,
             this->temperature_alarm_and_Warning,
             this->voltage_alarm_and_warning,
             this->RX_power_alarm_and_warning,
             this->TX_bias_alarm_and_warning,
             this->TX_power_alarm_and_warning,
             this->ConvertVoltageToStr(this->diag_supply_voltage).c_str(),
             this->transmitter_technology,
             this->eth_com_codes_ext,
             this->ConvertTemperatureToStr(this->high_temp_alarm_th, false).c_str(),
             this->ConvertTemperatureToStr(this->low_temp_alarm_th, false).c_str(),
             this->ConvertTemperatureToStr(this->high_temp_warning_th, false).c_str(),
             this->ConvertTemperatureToStr(this->low_temp_warning_th, false).c_str(),
             this->ConvertVoltageToStr(this->high_vcc_alarm_th).c_str(),
             this->ConvertVoltageToStr(this->low_vcc_alarm_th).c_str(),
             this->ConvertVoltageToStr(this->high_vcc_warning_th).c_str(),
             this->ConvertVoltageToStr(this->low_vcc_warning_th).c_str(),
             this->ConvertDateCodeToStr().c_str(),
             this->lot,
             this->TX_adaptive_equalization_freeze,
             this->ConvertRXOutputDisableToStr().c_str(),
             this->ConvertTXAdaptiveEqualizationEnableToStr().c_str());

    return (buffer);
}

#undef U8H_FMT
#undef U16H_FMT

CombinedCableInfo::~CombinedCableInfo()
{
   delete p_cable;
   delete p_phy;
}

bool CombinedCableInfo::ExportData(cable_record_data_t *p_cable_record_data)
{
    if (this->p_cable) {
        this->p_cable->ExportData(p_cable_record_data);
        return true;
    }

    return false;
}

string CombinedCableInfo::GetUniqueKey() const {
    if (this->p_cable)
        return "";

    if (this->p_phy)
        return this->p_phy->GetUniqueKey();

    return "";
}

string CombinedCableInfo::GetCSVHeader()
{
    stringstream sstream;
    sstream <<  "NodeGuid,PortGuid,PortNum,Source,Vendor,OUI,PN,SN,Rev,"
          <<  "LengthSMFiber,LengthOM5,LengthOM4,LengthOM3,LengthOM2,LengthOM1,"
          <<  "LengthCopperOrActive,Identifier,IdentifierStr,Connector,Type,SupportedSpeed,"
          <<  "LengthDesc,LengthDescByPRTL,LengthDescByReg,"
          <<  "TypeDesc,SupportedSpeedDesc,Temperature,PowerClass,NominalBitrate,"
          <<  "CDREnableTxRx,CDREnableTx,CDREnableRx,"
          <<  "InputEq,OutputAmp,OutputEmp,OutputPreEmp,OutputPostEmp,FWVersion,"
          <<  "Attenuation2.5G,Attenuation5G,Attenuation7G,Attenuation12G,Attenuation25G,"
          <<  "RXPowerType,RX1Power,RX2Power,RX3Power,RX4Power,"
          <<  "RX5Power,RX6Power,RX7Power,RX8Power,"
          <<  "TX1Bias,TX2Bias,TX3Bias,TX4Bias,"
          <<  "TX5Bias,TX6Bias,TX7Bias,TX8Bias,"
          <<  "TX1Power,TX2Power,TX3Power,TX4Power,"
          <<  "TX5Power,TX6Power,TX7Power,TX8Power,"
          <<  "RX1LatchedLossIndicator,RX2LatchedLossIndicator,"
          <<  "RX3LatchedLossIndicator,RX4LatchedLossIndicator,"
          <<  "TX1LatchedLossIndicator,TX2LatchedLossIndicator,"
          <<  "TX3LatchedLossIndicator,TX4LatchedLossIndicator,"
          <<  "TX1AdaptiveEqualizationFaultIndicator,TX2AdaptiveEqualizationFaultIndicator,"
          <<  "TX3AdaptiveEqualizationFaultIndicator,TX4AdaptiveEqualizationFaultIndicator,"
          <<  "RX1CDRLOL,RX2CDRLOL,RX3CDRLOL,RX4CDRLOL,"
          <<  "TX1CDRLOL,TX2CDRLOL,TX3CDRLOL,TX4CDRLOL,"
          <<  "HighTemperatureAlarm,LowTemperatureAlarm,"
          <<  "HighTemperatureWarning,LowTemperatureWarning,"
          <<  "InitializationFlagComplete,"
          <<  "HighSupplyVoltageAlarm,LowSupplyVoltageAlarm,"
          <<  "HighSupplyVoltageWarning,LowSupplyVoltageWarning,"
          <<  "HighRX1PowerAlarm,LowRX1PowerAlarm,HighRX1PowerWarning,LowRX1PowerWarning,"
          <<  "HighRX2PowerAlarm,LowRX2PowerAlarm,HighRX2PowerWarning,LowRX2PowerWarning,"
          <<  "HighRX3PowerAlarm,LowRX3PowerAlarm,HighRX3PowerWarning,LowRX3PowerWarning,"
          <<  "HighRX4PowerAlarm,LowRX4PowerAlarm,HighRX4PowerWarning,LowRX4PowerWarning,"
          <<  "HighTX1BiasAlarm,LowTX1BiasAlarm,HighTX1BiasWarning,LowTX1BiasWarning,"
          <<  "HighTX2BiasAlarm,LowTX2BiasAlarm,HighTX2BiasWarning,LowTX2BiasWarning,"
          <<  "HighTX3BiasAlarm,LowTX3BiasAlarm,HighTX3BiasWarning,LowTX3BiasWarning,"
          <<  "HighTX4BiasAlarm,LowTX4BiasAlarm,HighTX4BiasWarning,LowTX4BiasWarning,"
          <<  "HighTX1PowerAlarm,LowTX1PowerAlarm,HighTX1PowerWarning,LowTX1PowerWarning,"
          <<  "HighTX2PowerAlarm,LowTX2PowerAlarm,HighTX2PowerWarning,LowTX2PowerWarning,"
          <<  "HighTX3PowerAlarm,LowTX3PowerAlarm,HighTX3PowerWarning,LowTX3PowerWarning,"
          <<  "HighTX4PowerAlarm,LowTX4PowerAlarm,HighTX4PowerWarning,LowTX4PowerWarning,"
          <<  "SupplyVoltageReporting,TransmitterTechnology,"
          <<  "ActiveWavelengthControl,CooledTransmitterDevice,"
          <<  "ActivePinDetector,TunableTransmitter,"
          <<  "ExtendedSpecificationComplianceCodes,"
          <<  "AlarmTemperatureHighThresh,AlarmTemperatureLowThresh,"
          <<  "WarnTemperatureHighThresh,WarnTemperatureLowThresh,"
          <<  "AlarmVoltageHighThresh,AlarmVoltageLowThresh,"
          <<  "WarnVoltageHighThresh,WarnVoltageLowThresh,"
          <<  "RXPowerHighThresh,RXPowerLowThresh,"
          <<  "TXPowerHighThresh,TXPowerLowThresh,"
          <<  "TXBiasHighThresh,TXBiasLowThresh,"
          <<  "DateCode,Lot,"
          <<  "TX1AdaptiveEqualizationFreeze,TX2AdaptiveEqualizationFreeze,"
          <<  "TX3AdaptiveEqualizationFreeze,TX4AdaptiveEqualizationFreeze,"
          <<  "RXOutputDisable,TXAdaptiveEqualizationEnable,"
          <<  "MaxPower,CdrVendor,MaxFiberLength"
          << endl;

    return sstream.str();
}

void CombinedCableInfo::ToCSVStream(ostream &stream) const
{
    if (this->p_cable)
        this->p_cable->ToCSVStream(stream);
    else if (this->p_phy)
        this->p_phy->ToCSVStream(stream);
    else stream << endl;
}

void CableRecord::ToCSVStream(ostream &stream) const
{
    stream  << QUOTED("CABLE")           << ','
            << QUOTED(this->vendor)      << ','
            << this->oui                 << ','
            << QUOTED(this->pn)          << ','
            << QUOTED(this->sn)          << ','
            << QUOTED(this->rev)         << ','
            << +this->lengthsmfiber      << ','
            << QUOTED("NA")              << ','
            << QUOTED("NA")              << ','
            << +this->lengthom3          << ','
            << +this->lengthom2          << ','
            << +this->lengthom1          << ','
            << +this->lengthcopper       << ','
            << PTR(this->identifier)     << ','
            << QUOTED(this->ConvertCableIdentifierToStr()) << ','
            << PTR(this->connector)      << ','
            << +this->cable_type         << ','
            << +this->supported_speed    << ','
            << QUOTED(this->ConvertLengthToStr())         << ','
            << QUOTED(this->ConvertPrtlLengthToStr())     << ','
            << QUOTED(this->length_str)                   << ','
            << QUOTED(this->ConvertCableTypeToStr())      << ','
            << QUOTED(this->ConvertSupportedSpeedToStr()) << ','
            << QUOTED(this->ConvertTemperatureToStr(this->temperature, true)) << ','
            << +this->power_class                         << ','
            << (int)(this->nominal_br * 1000)             << ','
        //"CDREnableTxRx,CDREnableTx,CDREnableRx,
            << this->ConvertCDREnableTxRxToStr(true)      << ','
            << QUOTED("NA")                               << ','
            << QUOTED("NA")                               << ','
        //"InputEq,OutputAmp,OutputEmp,OutputPreEmp,OutputPostEmp,FWVersion,"
            << this->ConvertInputEqToStr(true)            << ','
            << this->ConvertOutputAmpToStr(true)          << ','
            << this->ConvertOutputEmpToStr(true)          << ','
            << QUOTED("NA")                               << ','
            << QUOTED("NA")                               << ','
            << this->ConvertFwVersionToStr(true)          << ','
            << this->ConvertAttenuationToStr(true)        << ','
            << QUOTED("NA")              << ','
            << this->RXpower_type        << ','
            << FLOAT(this->RX1Power, 3)  << ','
            << FLOAT(this->RX2Power, 3)  << ','
            << FLOAT(this->RX3Power, 3)  << ','
            << FLOAT(this->RX4Power, 3)  << ','
            //RX5Power to RX8Power are not available
            << QUOTED("NA")              << ','
            << QUOTED("NA")              << ','
            << QUOTED("NA")              << ','
            << QUOTED("NA")              << ','
            << FLOAT(this->TX1Bias, 3)   << ','
            << FLOAT(this->TX2Bias, 3)   << ','
            << FLOAT(this->TX3Bias, 3)   << ','
            << FLOAT(this->TX4Bias, 3)   << ','
            //TX5Bias to TX8Bias are not available
            << QUOTED("NA")              << ','
            << QUOTED("NA")              << ','
            << QUOTED("NA")              << ','
            << QUOTED("NA")              << ','
            << FLOAT(this->TX1Power, 3)  << ','
            << FLOAT(this->TX2Power, 3)  << ','
            << FLOAT(this->TX3Power, 3)  << ','
            << FLOAT(this->TX4Power, 3)  << ','
            //TX5Power to TX8Power are not available
            << QUOTED("NA")              << ','
            << QUOTED("NA")              << ','
            << QUOTED("NA")              << ','
            << QUOTED("NA")              << ','
            << BIT_TO_STR(this->CDR_TX_RX_loss_indicator, 0)         << ','
            << BIT_TO_STR(this->CDR_TX_RX_loss_indicator, 1)         << ','
            << BIT_TO_STR(this->CDR_TX_RX_loss_indicator, 2)         << ','
            << BIT_TO_STR(this->CDR_TX_RX_loss_indicator, 3)         << ','
            << BIT_TO_STR(this->CDR_TX_RX_loss_indicator, 4)         << ','
            << BIT_TO_STR(this->CDR_TX_RX_loss_indicator, 5)         << ','
            << BIT_TO_STR(this->CDR_TX_RX_loss_indicator, 6)         << ','
            << BIT_TO_STR(this->CDR_TX_RX_loss_indicator, 7)         << ','
            << BIT_TO_STR(this->adaptive_equalization_fault, 4)      << ','
            << BIT_TO_STR(this->adaptive_equalization_fault, 5)      << ','
            << BIT_TO_STR(this->adaptive_equalization_fault, 6)      << ','
            << BIT_TO_STR(this->adaptive_equalization_fault, 7)      << ','
            << BIT_TO_STR(this->TX_RX_LOL_indicator, 0)              << ','
            << BIT_TO_STR(this->TX_RX_LOL_indicator, 1)              << ','
            << BIT_TO_STR(this->TX_RX_LOL_indicator, 2)              << ','
            << BIT_TO_STR(this->TX_RX_LOL_indicator, 3)              << ','
            << BIT_TO_STR(this->TX_RX_LOL_indicator, 4)              << ','
            << BIT_TO_STR(this->TX_RX_LOL_indicator, 5)              << ','
            << BIT_TO_STR(this->TX_RX_LOL_indicator, 6)              << ','
            << BIT_TO_STR(this->TX_RX_LOL_indicator, 7)              << ','
            << BIT_TO_STR(this->temperature_alarm_and_Warning, 7)    << ','
            << BIT_TO_STR(this->temperature_alarm_and_Warning, 6)    << ','
            << BIT_TO_STR(this->temperature_alarm_and_Warning, 5)    << ','
            << BIT_TO_STR(this->temperature_alarm_and_Warning, 4)    << ','
            << BIT_TO_STR(this->temperature_alarm_and_Warning, 0)    << ','
            << BIT_TO_STR(this->voltage_alarm_and_warning, 7)        << ','
            << BIT_TO_STR(this->voltage_alarm_and_warning, 6)        << ','
            << BIT_TO_STR(this->voltage_alarm_and_warning, 5)        << ','
            << BIT_TO_STR(this->voltage_alarm_and_warning, 4)        << ','
            << BIT_TO_STR(this->RX_power_alarm_and_warning, 15)      << ','
            << BIT_TO_STR(this->RX_power_alarm_and_warning, 14)      << ','
            << BIT_TO_STR(this->RX_power_alarm_and_warning, 13)      << ','
            << BIT_TO_STR(this->RX_power_alarm_and_warning, 12)      << ','
            << BIT_TO_STR(this->RX_power_alarm_and_warning, 11)      << ','
            << BIT_TO_STR(this->RX_power_alarm_and_warning, 10)      << ','
            << BIT_TO_STR(this->RX_power_alarm_and_warning, 9)       << ','
            << BIT_TO_STR(this->RX_power_alarm_and_warning, 8)       << ','
            << BIT_TO_STR(this->RX_power_alarm_and_warning, 7)       << ','
            << BIT_TO_STR(this->RX_power_alarm_and_warning, 6)       << ','
            << BIT_TO_STR(this->RX_power_alarm_and_warning, 5)       << ','
            << BIT_TO_STR(this->RX_power_alarm_and_warning, 4)       << ','
            << BIT_TO_STR(this->RX_power_alarm_and_warning, 3)       << ','
            << BIT_TO_STR(this->RX_power_alarm_and_warning, 2)       << ','
            << BIT_TO_STR(this->RX_power_alarm_and_warning, 1)       << ','
            << BIT_TO_STR(this->RX_power_alarm_and_warning, 0)       << ','
            << BIT_TO_STR(this->TX_bias_alarm_and_warning, 15)       << ','
            << BIT_TO_STR(this->TX_bias_alarm_and_warning, 14)       << ','
            << BIT_TO_STR(this->TX_bias_alarm_and_warning, 13)       << ','
            << BIT_TO_STR(this->TX_bias_alarm_and_warning, 12)       << ','
            << BIT_TO_STR(this->TX_bias_alarm_and_warning, 11)       << ','
            << BIT_TO_STR(this->TX_bias_alarm_and_warning, 10)       << ','
            << BIT_TO_STR(this->TX_bias_alarm_and_warning, 9)        << ','
            << BIT_TO_STR(this->TX_bias_alarm_and_warning, 8)        << ','
            << BIT_TO_STR(this->TX_bias_alarm_and_warning, 7)        << ','
            << BIT_TO_STR(this->TX_bias_alarm_and_warning, 6)        << ','
            << BIT_TO_STR(this->TX_bias_alarm_and_warning, 5)        << ','
            << BIT_TO_STR(this->TX_bias_alarm_and_warning, 4)        << ','
            << BIT_TO_STR(this->TX_bias_alarm_and_warning, 3)        << ','
            << BIT_TO_STR(this->TX_bias_alarm_and_warning, 2)        << ','
            << BIT_TO_STR(this->TX_bias_alarm_and_warning, 1)        << ','
            << BIT_TO_STR(this->TX_bias_alarm_and_warning, 0)        << ','
            << BIT_TO_STR(this->TX_power_alarm_and_warning, 15)      << ','
            << BIT_TO_STR(this->TX_power_alarm_and_warning, 14)      << ','
            << BIT_TO_STR(this->TX_power_alarm_and_warning, 13)      << ','
            << BIT_TO_STR(this->TX_power_alarm_and_warning, 12)      << ','
            << BIT_TO_STR(this->TX_power_alarm_and_warning, 11)      << ','
            << BIT_TO_STR(this->TX_power_alarm_and_warning, 10)      << ','
            << BIT_TO_STR(this->TX_power_alarm_and_warning, 9)       << ','
            << BIT_TO_STR(this->TX_power_alarm_and_warning, 8)       << ','
            << BIT_TO_STR(this->TX_power_alarm_and_warning, 7)       << ','
            << BIT_TO_STR(this->TX_power_alarm_and_warning, 6)       << ','
            << BIT_TO_STR(this->TX_power_alarm_and_warning, 5)       << ','
            << BIT_TO_STR(this->TX_power_alarm_and_warning, 4)       << ','
            << BIT_TO_STR(this->TX_power_alarm_and_warning, 3)       << ','
            << BIT_TO_STR(this->TX_power_alarm_and_warning, 2)       << ','
            << BIT_TO_STR(this->TX_power_alarm_and_warning, 1)       << ','
            << BIT_TO_STR(this->TX_power_alarm_and_warning, 0)       << ','
        //"SupplyVoltageReporting,TransmitterTechnology,"
            << QUOTED(this->ConvertVoltageToStr(this->diag_supply_voltage)) << ','
            //we have to use two digits to fit CIMS format
            << _to_ptr_string(this->transmitter_technology >> 4, 2)  << ','
        //"ActiveWavelengthControl,CooledTransmitterDevice,"
        //"ActivePinDetector,TunableTransmitter,"
            << BIT_TO_STR(this->transmitter_technology, 3)           << ','
            << BIT_TO_STR(this->transmitter_technology, 2)           << ','
            << BIT_TO_STR(this->transmitter_technology, 1)           << ','
            << BIT_TO_STR(this->transmitter_technology, 0)           << ','
        //ExtendedSpecificationComplianceCodes,
            << +this->eth_com_codes_ext                              << ','
        // AlarmTemperatureHighThresh,AlarmTemperatureLowThresh,
        // WarnTemperatureHighThresh,WarnTemperatureLowThresh,
            << QUOTED(this->ConvertTemperatureToStr(this->high_temp_alarm_th, true))   << ','
            << QUOTED(this->ConvertTemperatureToStr(this->low_temp_alarm_th, true))    << ','
            << QUOTED(this->ConvertTemperatureToStr(this->high_temp_warning_th, true)) << ','
            << QUOTED(this->ConvertTemperatureToStr(this->low_temp_warning_th, true))  << ','
        // AlarmVoltageHighThresh,AlarmVoltageLowThresh,
        // WarnVoltageHighThresh,WarnVoltageLowThresh,
            << QUOTED(this->ConvertVoltageToStr(this->high_vcc_alarm_th))   << ','
            << QUOTED(this->ConvertVoltageToStr(this->low_vcc_alarm_th))    << ','
            << QUOTED(this->ConvertVoltageToStr(this->high_vcc_warning_th)) << ','
            << QUOTED(this->ConvertVoltageToStr(this->low_vcc_warning_th))  << ','
        // RXPowerHighThresh,RXPowerLowThresh,
        // TXPowerHighThresh,TXPowerLowThresh,
        // TXBiasHighThresh,TXBiasLowThresh,
            << QUOTED("NA")                                          << ','
            << QUOTED("NA")                                          << ','
            << QUOTED("NA")                                          << ','
            << QUOTED("NA")                                          << ','
            << QUOTED("NA")                                          << ','
            << QUOTED("NA")                                          << ','
        //DateCode,Lot,
            << this->ConvertDateCodeToStr()                          << ','
            << _to_ptr_string(this->lot)                             << ','
        //TX1AdaptiveEqualizationFreeze,TX2AdaptiveEqualizationFreeze,
        //TX3AdaptiveEqualizationFreeze,TX4AdaptiveEqualizationFreeze,
            << BIT_TO_STR(this->TX_adaptive_equalization_freeze, 3)   << ','
            << BIT_TO_STR(this->TX_adaptive_equalization_freeze, 2)   << ','
            << BIT_TO_STR(this->TX_adaptive_equalization_freeze, 1)   << ','
            << BIT_TO_STR(this->TX_adaptive_equalization_freeze, 0)   << ','
        //RXOutputDisable,TXAdaptiveEqualizationEnable
            << this->ConvertRXOutputDisableToStr()                    << ','
            << this->ConvertTXAdaptiveEqualizationEnableToStr()       << ','
        //MaxPower
            << QUOTED("NA")                                    << ','
        //CdrVendor
            << QUOTED("NA")                                    << ','
        //MaxFiberLength
            << QUOTED("NA")                                    << endl;
}

void PhyCableRecord::ToCSVStream(ostream &stream) const
{
    static const string na_val = "\"NA\"";
    stream
    //"NodeGuid,PortGuid,PortNum,Source,Vendor,OUI,PN,SN,Rev,"
        << QUOTED(this->m_source)             << ','
        << QUOTED(this->VendorToStr())        << ','
        << this->VendorOUIToStr()             << ','
        << QUOTED(this->VendorPnToStr())      << ','
        << QUOTED(this->VendorSnToStr())      << ','
        << QUOTED(this->RevisionToStr())      << ','
    //"LengthSMFiber,LengthOM5,LengthOM4,LengthOM3,LengthOM2,LengthOM1,"
        << (this->p_module ? _to_string(this->p_module->smf_length) : na_val) << ','
        << (this->p_module ? _to_string(this->p_module->length_om5) : na_val) << ','
        << (this->p_module ? _to_string(this->p_module->length_om4) : na_val) << ','
        << (this->p_module ? _to_string(this->p_module->length_om3) : na_val) << ','
        << (this->p_module ? _to_string(this->p_module->length_om2) : na_val) << ','
        << (this->p_module ? _to_string(this->p_module->length_om1) : na_val) << ','
    //"LengthCopperOrActive,Identifier,IdentifierStr,Connector,Type,SupportedSpeed
        << (this->p_module ?
                _to_string(this->p_module->cable_length)                  : na_val) << ','
        << (this->p_module ?
                _to_ptr_string(this->p_module->cable_identifier)          : na_val) << ','
        << QUOTED(this->CableIdentifierToStr(true))                                 << ','
        << (this->p_module ?
                _to_ptr_string(this->p_module->connector_type)            : na_val) << ','
        << (this->p_module ?
                _to_string(this->p_module->SelectTransmitterTechnology()) : na_val) << ','
        << (this->p_module ?
                _to_string(this->p_module->ib_compliance_code) : na_val)            << ','
        //,LengthDesc,LengthDescByPRTL,LengthDescByReg"
        << this->CableLengthToStr(true)            << ','
        << this->CableLengthByPrtlToStr(true)      << ','
        << this->CableLengthByRegToStr(true)       << ','
    //"TypeDesc,SupportedSpeedDesc,Temperature, PowerClass,NominalBitrate,"
        << this->TypeToStr(true)             << ','
        << this->SupportedSpeedToStr(true)   << ','
        << (this->p_module ? this->p_module->ConvertTemperatureToStr(
                this->p_module->temperature, true) : na_val) << ','
        << this->CablePowerClassToStr(true)  << ','
        << this->BitrateToStr(true)          << ','
    //"CDREnableTxRx,CDREnableTx,CDREnableRx,"
        << na_val                            << ','
        << (this->p_module ?
                p_module->ConvertCDREnableTxRxToStr(false, na_val, true) : na_val) << ','
        << (this->p_module ?
                p_module->ConvertCDREnableTxRxToStr(true, na_val, true) : na_val) << ','
    //InputEq,OutputAmp,OutputEmp,OutputPreEmp,OutputPostEmp,FWVersion,"
        << this->InputEqToStr(true)          << ','
        << this->OutputAmpToStr(true)        << ','
        << na_val                            << ','
        << this->OuputPreEmpToStr(true)      << ','
        << this->OuputPostEmpToStr(true)     << ','
        << this->FWVersionToStr(true)        << ','
    //"Attenuation2.5G,Attenuation5G,Attenuation7G,Attenuation12G,Attenuation25G,"
        << na_val                            << ','
        //returns 3 strings
        << this->AttenuationToStr(true)      << ','
    //"RXPowerType,RX1Power,RX2Power,RX3Power,RX4Power,"
        << (this->p_module ? _to_string(p_module->rx_power_type) : na_val)  << ','
        << this->RX1PowerToStr(true)         << ','
        << this->RX2PowerToStr(true)         << ','
        << this->RX3PowerToStr(true)         << ','
        << this->RX4PowerToStr(true)         << ','
    //"RX5Power,RX6Power,RX7Power,RX8Power,"
        << this->RX5PowerToStr(true)         << ','
        << this->RX6PowerToStr(true)         << ','
        << this->RX7PowerToStr(true)         << ','
        << this->RX8PowerToStr(true)         << ','
    //"TX1Bias,TX2Bias,TX3Bias,TX4Bias,"
        << this->TXBias1ToStr(true)          << ','
        << this->TXBias2ToStr(true)          << ','
        << this->TXBias3ToStr(true)          << ','
        << this->TXBias4ToStr(true)          << ','
    //"TX5Bias,TX6Bias,TX7Bias,TX8Bias,"
        << this->TXBias5ToStr(true)          << ','
        << this->TXBias6ToStr(true)          << ','
        << this->TXBias7ToStr(true)          << ','
        << this->TXBias8ToStr(true)          << ','
    //"TX1Power,TX2Power,TX3Power,TX4Power"
        << this->TX1PowerToStr(true)         << ','
        << this->TX2PowerToStr(true)         << ','
        << this->TX3PowerToStr(true)         << ','
        << this->TX4PowerToStr(true)         << ','
    //"TX5Power,TX6Power,TX7Power,TX8Power"
        << this->TX5PowerToStr(true)         << ','
        << this->TX6PowerToStr(true)         << ','
        << this->TX7PowerToStr(true)         << ','
        << this->TX8PowerToStr(true)         << ','
    //"RX1LatchedLossIndicator,RX2LatchedLossIndicator,
    //RX3LatchedLossIndicator,RX4LatchedLossIndicator"
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_los, 0) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_los, 1) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_los, 2) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_los, 3) : na_val)  << ','
    //TX1LatchedLossIndicator,TX2LatchedLossIndicator,
    //TX3LatchedLossIndicator,TX4LatchedLossIndicator,
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_los, 0) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_los, 1) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_los, 2) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_los, 3) : na_val)  << ','
    //TX1AdaptiveEqualizationFaultIndicator,TX2AdaptiveEqualizationFaultIndicator,
    //TX3AdaptiveEqualizationFaultIndicator,TX4AdaptiveEqualizationFaultIndicator,"
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_ad_eq_fault, 0) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_ad_eq_fault, 1) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_ad_eq_fault, 2) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_ad_eq_fault, 3) : na_val)  << ','
    //RX1CDRLOL,RX2CDRLOL,RX3CDRLOL,RX4CDRLOL,
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_cdr_lol, 0) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_cdr_lol, 1) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_cdr_lol, 2) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_cdr_lol, 3) : na_val)  << ','
    //TX1CDRLOL,TX2CDRLOL,TX3CDRLOL,TX4CDRLOL,
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_cdr_lol, 0) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_cdr_lol, 1) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_cdr_lol, 2) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_cdr_lol, 3) : na_val)  << ','
    //"HighTemperatureAlarm,LowTemperatureAlarm,HighTemperatureWarning,LowTemperatureWarning,"
        << (this->p_latched ? BIT_TO_STR(p_latched->temp_flags, 0) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->temp_flags, 1) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->temp_flags, 2) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->temp_flags, 3) : na_val)  << ','
    //InitializationFlagComplete,
        << na_val  << ','
    //HighSupplyVoltageAlarm,LowSupplyVoltageAlarm,HighSupplyVoltageWarning,LowSupplyVoltageWarning
        << (this->p_latched ? BIT_TO_STR(p_latched->vcc_flags, 0) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->vcc_flags, 1) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->vcc_flags, 2) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->vcc_flags, 3) : na_val)  << ','
    //HighRX1PowerAlarm,LowRX1PowerAlarm,HighRX1PowerWarning,LowRX1PowerWarning,
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_power_hi_al, 0) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_power_lo_al, 0) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_power_hi_war, 0) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_power_lo_war, 0) : na_val)  << ','
    //HighRX2PowerAlarm,LowRX2PowerAlarm,HighRX2PowerWarning,LowRX2PowerWarning,
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_power_hi_al, 1) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_power_lo_al, 1) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_power_hi_war, 1) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_power_lo_war, 1) : na_val)  << ','
    //HighRX3PowerAlarm,LowRX3PowerAlarm,HighRX3PowerWarning,LowRX3PowerWarning,
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_power_hi_al, 2) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_power_lo_al, 2) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_power_hi_war, 2) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_power_lo_war, 2) : na_val)  << ','
    //HighRX4PowerAlarm,LowRX4PowerAlarm,HighRX4PowerWarning,LowRX4PowerWarning,
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_power_hi_al, 3) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_power_lo_al, 3) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_power_hi_war, 3) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->rx_power_lo_war, 3) : na_val)  << ','
    //HighTX1BiasAlarm,LowTX1BiasAlarm,HighTX1BiasWarning,LowTX1BiasWarning,
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_bias_hi_al, 0) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_bias_lo_al, 0) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_bias_hi_war, 0) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_bias_lo_war, 0) : na_val)  << ','
    //HighTX2BiasAlarm,LowTX2BiasAlarm,HighTX2BiasWarning,LowTX2BiasWarning,
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_bias_hi_al, 1) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_bias_lo_al, 1) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_bias_hi_war, 1) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_bias_lo_war, 1) : na_val)  << ','
    //HighTX3BiasAlarm,LowTX3BiasAlarm,HighTX3BiasWarning,LowTX3BiasWarning,
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_bias_hi_al, 2) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_bias_lo_al, 2) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_bias_hi_war, 2) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_bias_lo_war, 2) : na_val)  << ','
    //HighTX4BiasAlarm,LowTX4BiasAlarm,HighTX4BiasWarning,LowTX4BiasWarning,
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_bias_hi_al, 3) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_bias_lo_al, 3) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_bias_hi_war, 3) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_bias_lo_war, 3) : na_val)  << ','
    //HighTX1PowerAlarm,LowTX1PowerAlarm,HighTX1PowerWarning,LowTX1PowerWarning,
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_power_hi_al, 0) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_power_lo_al, 0) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_power_hi_war, 0) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_power_lo_war, 0) : na_val)  << ','
    //HighTX2PowerAlarm,LowTX2PowerAlarm,HighTX2PowerWarning,LowTX2PowerWarning,
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_power_hi_al, 1) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_power_lo_al, 1) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_power_hi_war, 1) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_power_lo_war, 1) : na_val)  << ','
    //HighTX3PowerAlarm,LowTX3PowerAlarm,HighTX3PowerWarning,LowTX3PowerWarning,
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_power_hi_al, 2) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_power_lo_al, 2) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_power_hi_war, 2) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_power_lo_war, 2) : na_val)  << ','
    //HighTX4PowerAlarm,LowTX4PowerAlarm,HighTX4PowerWarning,LowTX4PowerWarning,
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_power_hi_al, 3) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_power_lo_al, 3) : na_val)   << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_power_hi_war, 3) : na_val)  << ','
        << (this->p_latched ? BIT_TO_STR(p_latched->tx_power_lo_war, 3) : na_val)  << ','
    //SupplyVoltageReporting,TransmitterTechnology,
        << (this->p_module ? this->p_module->ConvertVoltageToStr(
                this->p_module->voltage, true) : na_val) << ','
        << this->TechnologyToStr(true)    << ','
    //ActiveWavelengthControl,CooledTransmitterDevice,ActivePinDetector,TunableTransmitter,
        << (this->p_module ? BIT_TO_STR(p_module->cable_technology, 0) : na_val) << ','
        << (this->p_module ? BIT_TO_STR(p_module->cable_technology, 1) : na_val) << ','
        << (this->p_module ? BIT_TO_STR(p_module->cable_technology, 2) : na_val) << ','
        << (this->p_module ? BIT_TO_STR(p_module->cable_technology, 3) : na_val) << ','
    //ExtendedSpecificationComplianceCodes,
        << this->ComplanceCodeToStr(true) << ','
    // AlarmTemperatureHighThresh,AlarmTemperatureLowThresh,
    // WarnTemperatureHighThresh,WarnTemperatureLowThresh,
        << (this->p_module ? this->p_module->ConvertTemperatureToStr(
                this->p_module->temperature_high_th, true) : na_val) << ','
        << (this->p_module ? this->p_module->ConvertTemperatureToStr(
                this->p_module->temperature_low_th, true) : na_val)  << ','
        << na_val                                                    << ','
        << na_val                                                    << ','
    // AlarmVoltageHighThresh,AlarmVoltageLowThresh,
    // WarnVoltageHighThresh,WarnVoltageLowThresh,
        << (this->p_module ? this->p_module->ConvertVoltageToStr(
                this->p_module->voltage_high_th, true) : na_val) << ','
        << (this->p_module ? this->p_module->ConvertVoltageToStr(
                this->p_module->voltage_low_th, true) : na_val)  << ','
        << na_val                                                << ','
        << na_val                                                << ','
    // RXPowerHighThresh,RXPowerLowThresh,
    // TXPowerHighThresh,TXPowerLowThresh,
    // TXBiasHighThresh,TXBiasLowThresh,
        << (this->p_module ? _to_string(this->p_module->rx_power_high_th)    : na_val) << ','
        << (this->p_module ? _to_string(this->p_module->rx_power_low_th)     : na_val) << ','
        << (this->p_module ? _to_string(this->p_module->tx_power_high_th)    : na_val) << ','
        << (this->p_module ? _to_string(this->p_module->tx_power_low_th)     : na_val) << ','
        << (this->p_module ? _to_string(this->p_module->tx_bias_high_th)     : na_val) << ','
        << (this->p_module ? _to_string(this->p_module->tx_bias_low_th)      : na_val) << ','
    // DateCode,
    //NOTICE DateCodeToStr returns N/A instead of "NA"
    //just fit cable_info format
        << this->DateCodeToStr()          << ','
    //Lot
        << this->LotToStr(true)           << ','
    //TX1AdaptiveEqualizationFreeze,TX2AdaptiveEqualizationFreeze,
    //TX3AdaptiveEqualizationFreeze,TX4AdaptiveEqualizationFreeze,
        << na_val    << ','    << na_val  << ','
        << na_val    << ','    << na_val  << ','
    //RXOutputDisable,TXAdaptiveEqualizationEnable
        << na_val    << ','    << na_val  << ','
    //MaxPower
        << this->MaxPowerToStr(true)      << ','
    //CdrVendor, MaxFiberLength
        << (this->p_module ? _to_string(this->p_module->cdr_vendor)       : na_val) << ','
        << (this->p_module ? _to_string(this->p_module->max_fiber_length) : na_val) << endl;
}

void CombinedCableInfo::ToFileStream(ostream &stream) const
{
    if (this->p_cable)
        this->p_cable->ToFileStream(stream);
    else if (this->p_phy)
        this->p_phy->ToFileStream(stream);
}

void CableRecord::ToFileStream(ostream &stream) const
{
    stream << this->c_str(true);
}

void PhyCableRecord::ToFileStream(ostream &stream) const
{
    stream
        << "Source: "        << this->m_source                 << endl
        //module info part
        << "Identifier: "    << this->CableIdentifierToStr()   << endl
        << "Vendor: "        << this->VendorToStr()            << endl
        << "OUI: "           << this->VendorOUIToStr()         << endl
        << "PN: "            << this->VendorPnToStr()          << endl
        << "SN: "            << this->VendorSnToStr()          << endl
        << "Rev: "           << this->RevisionToStr()          << endl
        << "Length: "        << this->CableLengthToStr()       << endl
        << "Length by PRTL: " << this->CableLengthByPrtlToStr() << endl
        << "Length by Reg: " << this->CableLengthByRegToStr()  << endl
        << "Type: "           << this->TypeToStr()             << endl
        << "SupportedSpeed: " << this->SupportedSpeedToStr()   << endl
        << "Temperature: "    << (this->p_module ? this->p_module->ConvertTemperatureToStr(
                this->p_module->temperature, false) : "N/A")   << endl
        << "PowerClass: "     << this->CablePowerClassToStr()  << endl
        << "NominalBitrate: " << this->BitrateToStr()          << endl
        << "CDREnableTxRx: "  << this->CDREnableTxRXToStr()    << endl
        << "InputEq: "        << this->InputEqToStr()          << endl
        << "OutputAmp: "      << this->OutputAmpToStr()        << endl
        << "OutputEmp: "      << "N/A"                         << endl
        << "FW Version: "
                               << this->FWVersionToStr()       << endl
        << "Attenuation(5,7,12,25): "
                               << this->AttenuationToStr()     << endl
        << "RX power type: "   << this->RXPowerTypeToStr()     << endl
        << "RX1 Power: "       << this->RX1PowerToStr()        << endl
        << "RX2 Power: "       << this->RX2PowerToStr()        << endl
        << "RX3 Power: "       << this->RX3PowerToStr()        << endl
        << "RX4 Power: "       << this->RX4PowerToStr()        << endl
        << "RX5 Power: "       << this->RX5PowerToStr()        << endl
        << "RX6 Power: "       << this->RX6PowerToStr()        << endl
        << "RX7 Power: "       << this->RX7PowerToStr()        << endl
        << "RX8 Power: "       << this->RX8PowerToStr()        << endl
        << "TX1 Bias: "        << this->TXBias1ToStr()         << endl
        << "TX2 Bias: "        << this->TXBias2ToStr()         << endl
        << "TX3 Bias: "        << this->TXBias3ToStr()         << endl
        << "TX4 Bias: "        << this->TXBias4ToStr()         << endl
        << "TX5 Bias: "        << this->TXBias5ToStr()         << endl
        << "TX6 Bias: "        << this->TXBias6ToStr()         << endl
        << "TX7 Bias: "        << this->TXBias7ToStr()         << endl
        << "TX8 Bias: "        << this->TXBias8ToStr()         << endl
        << "TX1 Power: "       << this->TX1PowerToStr()        << endl
        << "TX2 Power: "       << this->TX2PowerToStr()        << endl
        << "TX3 Power: "       << this->TX3PowerToStr()        << endl
        << "TX4 Power: "       << this->TX4PowerToStr()        << endl
        << "TX5 Power: "       << this->TX5PowerToStr()        << endl
        << "TX6 Power: "       << this->TX6PowerToStr()        << endl
        << "TX7 Power: "       << this->TX7PowerToStr()        << endl
        << "TX8 Power: "       << this->TX8PowerToStr()        << endl


        //Latched part
        << "CDR Latched TX/RX Loss Indicator: "
            << this->LatchedTxRxIndicatorToStr()               << endl
        << "Latched Adaptive Equalization Fault: "
            << this->LatchedAdaptiveEqualizationFaultToStr()   << endl
        << "Latched  TX / RX LOL Indicator: "
            << this->LatchedTxRxLolIndicatorToStr()            << endl
        << "Latched Temperature Alarm and Warning: "
            << this->LatchedTempAlarmAndWarningToStr()         << endl
        << "Latched Voltage Alarm and Warning: "
            << this->LatchedVoltageAlarmAndWarningToStr()      << endl
        << "RX Power Alarm and Warning: "
            << this->RXPowerAlarmAndWarningToStr()             << endl
        << "TX Bias Alarm and Warning:  "
            << this->TXBiasAlarmAndWarningToStr()              << endl
        << "TX Power Alarm and Warning: "
            << this->TXPowerAlarmAndWarningToStr()             << endl

        //module info part
        << "Supply Voltage Reporting: "
            << (this->p_module ? this->p_module->ConvertVoltageToStr(
                    this->p_module->voltage, false) : "N/A")   << endl
        << "Transmitter Technology: "
                               << this->TechnologyToStr()      << endl
        << "Extended Specification Compliance Codes: "
                               << this->ComplanceCodeToStr()   << endl

        << "AlarmTemperatureHighThresh: "
            << (this->p_module ? this->p_module->ConvertTemperatureToStr(
                    this->p_module->temperature_high_th, false) : "N/A") << endl
        << "AlarmTemperatureLowThresh: "
            << (this->p_module ? this->p_module->ConvertTemperatureToStr(
                    this->p_module->temperature_low_th, false) : "N/A")  << endl
        << "WarnTemperatureHighThresh: " << "N/A"                        << endl
        << "WarnTemperatureLowThresh: "  << "N/A"                        << endl

        << "AlarmVoltageHighThresh: "
            << (this->p_module ? this->p_module->ConvertVoltageToStr(
                    this->p_module->voltage_high_th, false) : "N/A")     << endl
        << "AlarmVoltageLowThresh: "
            << (this->p_module ? this->p_module->ConvertVoltageToStr(
                    this->p_module->voltage_low_th, false) : "N/A")      << endl
        << "WarnVoltageHighThresh: " << "N/A"                            << endl
        << "WarnVoltageLowThresh: "  << "N/A"                            << endl

        << "RXPowerHighThresh: "     << (this->p_module ?
                _to_string(this->p_module->rx_power_high_th)    : "N/A") << endl
        << "RXPowerLowThresh: "      << (this->p_module ?
                _to_string(this->p_module->rx_power_low_th)     : "N/A") << endl
        << "TXPowerHighThresh: "     << (this->p_module ?
                _to_string(this->p_module->tx_power_high_th)    : "N/A") << endl
        << "TXPowerLowThresh: "      << (this->p_module ?
                _to_string(this->p_module->tx_power_low_th)     : "N/A") << endl
        << "TXBiasHighThresh: "      << (this->p_module ?
                _to_string(this->p_module->tx_bias_high_th)     : "N/A") << endl
        << "TXBiasLowThresh: "       << (this->p_module ?
                _to_string(this->p_module->tx_bias_low_th)      : "N/A") << endl

        << "DateCode: "        << this->DateCodeToStr()        << endl
        << "Lot: "             << this->LotToStr()             << endl

        //missing in PHY
        << "TX Adaptive Equalization Freeze: " << "N/A"        << endl
        << "RX Output Disable: "               << "N/A"        << endl
        << "TX Adaptive Equalization Enable: " << "N/A"        << endl

        //the following entries do not exist for cables output
        //and they replace OutputEmp
        << "OutputPreEmp: "     << this->OuputPreEmpToStr()    << endl
        << "OutputPostEmp: "    << this->OuputPostEmpToStr()   << endl;
}

string PhyCableRecord::LengthOmxToStr(u_int8_t index) const
{
    return (this->p_module ?
        this->p_module->ConvertCableLengthOMXToStr(index) : "N/A");
}

string PhyCableRecord::LengthSMFiberToStr() const
{
    return (this->p_module ?
        this->p_module->ConvertCableLengthSMFiberToStr() : "N/A");
}

string PhyCableRecord::OuputPostEmpToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "\"NA\""  : "N/A");

    return (this->p_module ?
        this->p_module->ConvertTxEQRxAMPRxEMPToStr(p_module->cable_rx_post_emphasis,
                                                is_csv, na_val)
        : na_val);
}

string PhyCableRecord::OuputPreEmpToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "\"NA\"" : "N/A");

    return (this->p_module ?
        this->p_module->ConvertTxEQRxAMPRxEMPToStr(p_module->cable_rx_emphasis,
                                                is_csv, na_val)
        : na_val);
}

uint16_t PhyCableRecord::QuadroToInt(uint8_t hi_al, uint8_t lo_al,
                                        uint8_t hi_wrn, uint8_t lo_wrn) const
{
// it returns 16 bit (0xabcd)
// a is lane 1 :
//      Bit     Value
//       7    bit0(hi_al)
//       6    bit0(lo_al)
//       5    bit0(hi_wrn)
//       4    bit0(lo_wrn)
//
// b is lane 2 :
//      Bit     Value
//       3    bit1(hi_al)
//       2    bit1(lo_al)
//       1    bit1(hi_wrn)
//       0    bit1(lo_wrn)
//
// c is lane 3 :
//      Bit     Value
//       7    bit2(hi_al)
//       6    bit2(lo_al)
//       5    bit2(hi_wrn)
//       4    bit2(lo_wrn)
//
// d is lane 4 :
//      Bit     Value
//       3    bit3(hi_al)
//       2    bit3(lo_al)
//       1    bit3(hi_wrn)
//       0    bit3(lo_wrn)
//

    std::bitset<16> arr;
    for (int i = 0; i < 4; i++) {
        arr[arr.size() - (1 + (i * 4))] = (hi_al >> i) & 0x1;
        arr[arr.size() - (2 + (i * 4))] = (lo_al >> i) & 0x1;
        arr[arr.size() - (3 + (i * 4))] = (hi_wrn >> i) & 0x1;
        arr[arr.size() - (4 + (i * 4))] = (lo_wrn >> i) & 0x1;
    }

    return static_cast<uint16_t>(arr.to_ulong());
}

string PhyCableRecord::TXPowerAlarmAndWarningToStr() const
{
    if (!this->p_latched)
        return "N/A";

    return _to_ptr_string(
                this->QuadroToInt(this->p_latched->tx_power_hi_al, this->p_latched->tx_power_lo_al,
                        this->p_latched->tx_power_hi_war, this->p_latched->tx_power_lo_war));
}

string PhyCableRecord::RXPowerAlarmAndWarningToStr() const
{
    if (!this->p_latched)
        return "N/A";

    return _to_ptr_string(
                this->QuadroToInt(this->p_latched->rx_power_hi_al, this->p_latched->rx_power_lo_al,
                        this->p_latched->rx_power_hi_war, this->p_latched->rx_power_lo_war));
}

string PhyCableRecord::TXBiasAlarmAndWarningToStr() const
{
    if (!this->p_latched)
        return "N/A";

    return _to_ptr_string(
                this->QuadroToInt(this->p_latched->tx_bias_hi_al, this->p_latched->tx_bias_lo_al,
                        this->p_latched->tx_bias_hi_war, this->p_latched->tx_bias_lo_war));
}

string PhyCableRecord::LatchedVoltageAlarmAndWarningToStr() const
{
    if (!this->p_latched)
        return "N/A";

    return _to_ptr_string(this->p_latched->vcc_flags);
}

string PhyCableRecord::LatchedTempAlarmAndWarningToStr() const
{
    if (!this->p_latched)
        return "N/A";

    return _to_ptr_string(this->p_latched->temp_flags);
}

string PhyCableRecord::LatchedTxRxLolIndicatorToStr() const
{
    if (!this->p_latched)
        return "N/A";

    return _to_ptr_string((uint8_t)((this->p_latched->tx_cdr_lol << 4) |
                      (this->p_latched->rx_cdr_lol & 0xf)));
}

string PhyCableRecord::LatchedAdaptiveEqualizationFaultToStr() const
{
    if (!this->p_latched)
        return "N/A";

    return _to_ptr_string((uint8_t)(this->p_latched->tx_ad_eq_fault & 0xf));
}

string PhyCableRecord::LatchedTxRxIndicatorToStr() const
{
    if (!this->p_latched)
        return "N/A";

    return _to_ptr_string((uint8_t)((this->p_latched->tx_los << 4) |
                       (this->p_latched->rx_los & 0xf)));
}

string PhyCableRecord::LotToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "\"NA\""  : "N/A");
    return (this->p_module ?
        _to_ptr_string((uint16_t)(this->p_module->date_code & 0xffff)) : na_val);
}

string PhyCableRecord::DateCodeToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "\"NA\""  : "N/A");
    return (this->p_module ? this->p_module->ConvertDateCodeToStr(na_val) : na_val);
}

string PhyCableRecord::ComplanceCodeToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "\"NA\""  : "N/A");
    if (!this->p_module)
        return na_val;

    return  (is_csv ? _to_string(this->p_module->ext_ethernet_compliance_code) :
                      _to_ptr_string(this->p_module->ext_ethernet_compliance_code));
}

string PhyCableRecord::TechnologyToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "\"NA\""  : "N/A");
    if (!this->p_module)
        return na_val;

    //we have to use two digits because of CIMS cables
    return (is_csv ? _to_ptr_string(this->p_module->SelectTransmitterTechnology()) :
                     _to_ptr_string(this->p_module->cable_technology));
}

string PhyCableRecord::TX1PowerToStr(bool is_csv) const
{
    if (!this->p_module)
        return this->EmptyPowerLineToStr(is_csv);

    return this->PowerLineToStr(p_module->tx_power_lane0, p_module->module_info_ext, is_csv);
}

string PhyCableRecord::TX2PowerToStr(bool is_csv) const
{
    if (!this->p_module)
        return this->EmptyPowerLineToStr(is_csv);

    return this->PowerLineToStr(p_module->tx_power_lane1, p_module->module_info_ext, is_csv);
}

string PhyCableRecord::TX3PowerToStr(bool is_csv) const
{
    if (!this->p_module)
        return this->EmptyPowerLineToStr(is_csv);

    return this->PowerLineToStr(p_module->tx_power_lane2, p_module->module_info_ext, is_csv);
}

string PhyCableRecord::TX4PowerToStr(bool is_csv) const
{
    if (!this->p_module)
        return this->EmptyPowerLineToStr(is_csv);

    return this->PowerLineToStr(p_module->tx_power_lane3, p_module->module_info_ext, is_csv);
}

string PhyCableRecord::TX5PowerToStr(bool is_csv) const
{
    if (!this->p_module)
        return this->EmptyPowerLineToStr(is_csv);

    return this->PowerLineToStr(p_module->tx_power_lane4, p_module->module_info_ext, is_csv);
}

string PhyCableRecord::TX6PowerToStr(bool is_csv) const
{
    if (!this->p_module)
        return this->EmptyPowerLineToStr(is_csv);

    return this->PowerLineToStr(p_module->tx_power_lane5, p_module->module_info_ext, is_csv);
}

string PhyCableRecord::TX7PowerToStr(bool is_csv) const
{
    if (!this->p_module)
        return this->EmptyPowerLineToStr(is_csv);

    return this->PowerLineToStr(p_module->tx_power_lane6, p_module->module_info_ext, is_csv);
}

string PhyCableRecord::TX8PowerToStr(bool is_csv) const
{
    if (!this->p_module)
        return this->EmptyPowerLineToStr(is_csv);

    return this->PowerLineToStr(p_module->tx_power_lane7, p_module->module_info_ext, is_csv);
}

string PhyCableRecord::TXBiasToStr(bool is_csv, double bias_lane) const
{
    stringstream streamStr;
    streamStr << FLOAT((bias_lane / 500.0), 3);

    if (!is_csv)
       streamStr << " mA";

    return streamStr.str();
}

string PhyCableRecord::MaxPowerToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "\"NA\""  : "N/A");
    return (this->p_module ? this->p_module->ConvertMaxPowerToStr(na_val) : na_val);
}

string PhyCableRecord::TXBias1ToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "\"NA\""  : "N/A");
    return (this->p_module ? this->TXBiasToStr(is_csv, this->p_module->tx_bias_lane0) : na_val);
}

string PhyCableRecord::TXBias2ToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "\"NA\""  : "N/A");
    return (this->p_module ? this->TXBiasToStr(is_csv, this->p_module->tx_bias_lane1) : na_val);
}

string PhyCableRecord::TXBias3ToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "\"NA\""  : "N/A");
    return (this->p_module ? this->TXBiasToStr(is_csv, this->p_module->tx_bias_lane2) : na_val);
}

string PhyCableRecord::TXBias4ToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "\"NA\""  : "N/A");
    return (this->p_module ? this->TXBiasToStr(is_csv, this->p_module->tx_bias_lane3) : na_val);
}

string PhyCableRecord::TXBias5ToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "\"NA\""  : "N/A");
    return (this->p_module ? this->TXBiasToStr(is_csv, this->p_module->tx_bias_lane4) : na_val);
}

string PhyCableRecord::TXBias6ToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "\"NA\""  : "N/A");
    return (this->p_module ? this->TXBiasToStr(is_csv, this->p_module->tx_bias_lane5) : na_val);
}

string PhyCableRecord::TXBias7ToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "\"NA\""  : "N/A");
    return (this->p_module ? this->TXBiasToStr(is_csv, this->p_module->tx_bias_lane6) : na_val);
}

string PhyCableRecord::TXBias8ToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "\"NA\""  : "N/A");
    return (this->p_module ? this->TXBiasToStr(is_csv, this->p_module->tx_bias_lane7) : na_val);
}

string PhyCableRecord::EmptyPowerLineToStr(bool is_csv) const
{
    //keep as a function in case we would like to return a pair: N/A N/A
    return (is_csv ?  "\"NA\"" : "N/A N/A");
}

//Module info extended configurations.
//resolution for rx_power, rx_power_high_th, rx_power_low_th
//tx_power, tx_power_high_th, tx_power_low_th in module info page
//0: dbm
//1: uW;     65534 * 0.1µW = 6.55mW = 8.2 dBm
string PhyCableRecord::PowerLineToStr(double lane, uint16_t module_info_ext, bool is_csv) const
{
    stringstream streamStr;
    if (is_csv) {
        if (module_info_ext)
            streamStr << FLOAT(this->lane_to_mW(lane), 3);
        else
            streamStr << FLOAT(this->dBm_to_mW(lane), 3);
    }
    else {
        if (module_info_ext)
            streamStr << FLOAT(this->lane_to_mW(lane), 3) << " mW, "
                    << FLOAT(this->mW_to_dBm(this->lane_to_mW(lane)), 3)   << " dBm";
        else
            streamStr << FLOAT(this->dBm_to_mW(lane), 3) << " mW, "
                    << FLOAT(lane, 3)                         << " dBm";
    }

    return streamStr.str();
}

string PhyCableRecord::RX1PowerToStr(bool is_csv) const
{
    if (!this->p_module)
        return this->EmptyPowerLineToStr(is_csv);

    return this->PowerLineToStr(p_module->rx_power_lane0, p_module->module_info_ext, is_csv);
}

string PhyCableRecord::RX2PowerToStr(bool is_csv) const
{
    if (!this->p_module)
        return this->EmptyPowerLineToStr(is_csv);

    return this->PowerLineToStr(p_module->rx_power_lane1, p_module->module_info_ext, is_csv);
}

string PhyCableRecord::RX3PowerToStr(bool is_csv) const
{
    if (!this->p_module)
        return this->EmptyPowerLineToStr(is_csv);

    return this->PowerLineToStr(p_module->rx_power_lane2, p_module->module_info_ext, is_csv);
}

string PhyCableRecord::RX4PowerToStr(bool is_csv) const
{
    if (!this->p_module)
        return this->EmptyPowerLineToStr(is_csv);

    return this->PowerLineToStr(p_module->rx_power_lane3, p_module->module_info_ext, is_csv);
}

string PhyCableRecord::RX5PowerToStr(bool is_csv) const
{
    if (!this->p_module)
        return this->EmptyPowerLineToStr(is_csv);

    return this->PowerLineToStr(p_module->rx_power_lane4, p_module->module_info_ext, is_csv);
}

string PhyCableRecord::RX6PowerToStr(bool is_csv) const
{
    if (!this->p_module)
        return this->EmptyPowerLineToStr(is_csv);

    return this->PowerLineToStr(p_module->rx_power_lane5, p_module->module_info_ext, is_csv);
}

string PhyCableRecord::RX7PowerToStr(bool is_csv) const
{
    if (!this->p_module)
        return this->EmptyPowerLineToStr(is_csv);

    return this->PowerLineToStr(p_module->rx_power_lane6, p_module->module_info_ext, is_csv);
}

string PhyCableRecord::RX8PowerToStr(bool is_csv) const
{
    if (!this->p_module)
        return this->EmptyPowerLineToStr(is_csv);

    return this->PowerLineToStr(p_module->rx_power_lane7, p_module->module_info_ext, is_csv);
}

string PhyCableRecord::FWVersionToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "\"NA\"" : "N/A");
    return (this->p_module ? p_module->ConvertFWVersionToStr(na_val): na_val);
}

string PhyCableRecord::AttenuationToStr(bool is_csv) const
{
    const string na_val =
                (is_csv ? "\"NA\",\"NA\",\"NA\",\"NA\"" : "N/A N/A N/A N/A");
    return (this->p_module ? p_module->ConvertAttenuationToStr(is_csv) : na_val);
}

string PhyCableRecord::RXPowerTypeToStr() const
{
    return (this->p_module ? (p_module->rx_power_type ? "Average Power" : "OMA") : "N/A");
}

string PhyCableRecord::TypeToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "NA" : "N/A");
    const string non_exist_val = (is_csv ? "\"NA\"" : "N/A");
    return (this->p_module ?
        _to_cvs_quoted(is_csv, p_module->ConvertTransmitterTechnologyToStr(na_val)) : non_exist_val);
}

string PhyCableRecord::SupportedSpeedToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "NA" : "N/A");
    const string non_exist_val = (is_csv ? "\"NA\"" : "N/A");
    return (this->p_module ?
        _to_cvs_quoted(is_csv, p_module->ConvertIBComplianceCodeToStr(na_val)) : non_exist_val);
}

string PhyCableRecord::GetHighTemperatureThresholdStr() const
{
    if (!this->p_module)
        return "N/A";

    return this->p_module->ConvertTemperatureToStr(this->p_module->temperature_high_th, false); 
}

string PhyCableRecord::GetLowTemperatureThresholdStr() const
{
    if (!this->p_module)
        return "N/A";

    return this->p_module->ConvertTemperatureToStr(this->p_module->temperature_low_th, false); 
}

int PhyCableRecord::GetTemperatureErrorsByTreshold() const
{
    if (!this->p_module)
        return 0;

    int ret = 0;
    u_int8_t technology = static_cast<u_int8_t>(this->p_module->cable_technology >> 4);
    int8_t temperature = static_cast<int8_t>(this->p_module->temperature >> 8);

    if (_check_temperature_by_threshold(true, technology, temperature,
                                        static_cast<int8_t>(this->p_module->temperature_high_th >> 8)))
        ret |= CABLE_HIGH_TEMP_MASK;

    if (_check_temperature_by_threshold(false, technology, temperature,
                                        static_cast<int8_t>(this->p_module->temperature_low_th >> 8)))
        ret |= CABLE_LOW_TEMP_MASK;

    return ret;
}

string PhyCableRecord::GetTemperatureStr() const
{
    if (!this->p_module)
        return "N/A";

    return this->p_module->ConvertTemperatureToStr(this->p_module->temperature, false);
}


u_int8_t PhyCableRecord::GetCDRVendor () const
{
    return (this->p_module ? this->p_module->cdr_vendor : 0);
}

string PhyCableRecord::GetPNVendor() const
{
    return (this->p_module ? this->p_module->vendor_pn : "");
}

cable_fw_t PhyCableRecord::GetFWObj() const
{
    cable_fw_t fw;

    if (this->p_module)
        this->p_module->GetFWVersion(fw.hi, fw.mid, fw.lo);

    return fw;
}

cable_id_t PhyCableRecord::GetIdObj() const
{
    cable_id_t cableId;

    if (this->p_module) {
        cableId.pn = this->DescToCsvDesc(this->p_module->vendor_pn, true);

        cableId.identifier = this->p_module->ConvertCableIdentifierToStr();
        if (cableId.identifier == "NA")
            cableId.identifier = ""; //disable cable_id

        cableId.vendor = _get_vendor_from_sn(this->DescToCsvDesc(p_module->vendor_sn, true));
        cableId.revision = this->DescToCsvDesc(p_module->ConvertRevisionToStr(), true);
    }

    return cableId;
}

string PhyCableRecord::CDREnableTxRXToStr() const
{
    return  (this->p_module ? (p_module->ConvertCDREnableTxRxToStr(false, "N/A")
                               + ' ' + p_module->ConvertCDREnableTxRxToStr(true, "N/A"))
                            : "N/A N/A");
}

string PhyCableRecord::InputEqToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "\"NA\"" : "N/A");

    return (this->p_module ?
        p_module->ConvertTxEQRxAMPRxEMPToStr(p_module->cable_tx_equalization, is_csv, na_val) : na_val);
}

string PhyCableRecord::OutputAmpToStr(bool is_csv) const
{
    const string na_val = (is_csv ? "\"NA\"" : "N/A");

    return (this->p_module ?
        p_module->ConvertTxEQRxAMPRxEMPToStr(p_module->cable_rx_amp, is_csv, na_val)  : na_val);
}

string PhyCableRecord::VendorToStr() const
{
    return (this->p_module ?
        trim(p_module->vendor_name) : "NA");
}

string PhyCableRecord::VendorPnToStr() const
{
    return (this->p_module ?
        this->DescToCsvDesc(p_module->vendor_pn) : "NA");
}

string PhyCableRecord::VendorSnToStr() const
{
    return (this->p_module ?
        this->DescToCsvDesc(p_module->vendor_sn) : "NA");
}

string PhyCableRecord::RevisionToStr() const
{
    return (this->p_module ?
        this->DescToCsvDesc(p_module->ConvertRevisionToStr()) : "NA");
}

string PhyCableRecord::CableLengthToStr(bool is_csv) const
{
    //if prtl's pointer is not NULL then it was already checked for being valid
    if (!this->prtl_length.empty()) {
        if (is_csv) {
            stringstream ss;
            ss << '\"'
               << this->prtl_length
               << '\"';
            return ss.str();
        } else
            return this->prtl_length;
    }

    //ConvertCableLengthToStr may return default value and it shell not be quoted
    return (this->p_module ? p_module->ConvertCableLengthToStr(is_csv, "NA") : "NA");
}

string PhyCableRecord::CableLengthByPrtlToStr(bool is_csv) const
{
    //if prtl's pointer is not NULL then it was already checked for being valid
    if (!this->prtl_length.empty()) {
        if (is_csv) {
            stringstream ss;
            ss << '\"'
               << this->prtl_length
               << '\"';
            return ss.str();
        } else
            return prtl_length;
    }

    return "NA";
}

string PhyCableRecord::CableLengthByRegToStr(bool is_csv) const
{
    //ConvertCableLengthToStr may return default value and it shell not be quoted
    return (this->p_module ? p_module->ConvertCableLengthToStr(is_csv, "NA") : "NA");
}

string PhyCableRecord::BitrateToStr(bool is_csv) const
{
    const string non_exist_val = (is_csv ? "\"NA\"" : "NA");
    if (!this->p_module)
        return "N/A";

    stringstream streamStr;
    if (is_csv)
        streamStr << (int)(this->p_module->SelectNominalBR() * 1000);
    else
       streamStr << +this->p_module->SelectNominalBR() << " Gb/s";

    return streamStr.str();
}

string PhyCableRecord::CableIdentifierToStr(bool is_csv) const
{
    const string non_exist_val = (is_csv ? "\"NA\"" : "N/A");
    if (!this->p_module)
        return non_exist_val;

    stringstream streamStr;
    streamStr << this->p_module->ConvertCableIdentifierToStr();
    if (!is_csv)
        streamStr << " (" << _to_ptr_string(this->p_module->cable_identifier) << ')';

    return streamStr.str();
}

string PhyCableRecord::CablePowerClassToStr(bool is_csv) const
{
    const string non_exist_val = (is_csv ? "\"NA\"" : "NA");
    if (!this->p_module)
        return non_exist_val;

    return _to_string(this->p_module->cable_power_class);
}

string PhyCableRecord::VendorOUIToStr() const
{
    if (!this->p_module)
        return "NA";

    stringstream streamStr;
    streamStr << PTR(this->p_module->vendor_oui, 0);

    return streamStr.str();
}


string PhyCableRecord::DescToCsvDesc(const string &desc, bool empty) const
{
    string result = trim(desc);

    if (result.empty())
        return (!empty ? "NA" : "");

    // change ',' into '-'
    size_t idx = result.find(',', 0);
    while (idx != string::npos) {
        result[idx] = '-';
        idx = result.find(',', idx + 1);
    }

    return (result);
}

string PhyCableRecord::ModuleRecord::ConvertRevisionToStr() const
{
    stringstream rev_ss;
    size_t size = sizeof(this->vendor_rev);
    for (size_t i = 0; i < size; i++) {
         if (this->vendor_rev[i])
             rev_ss << this->vendor_rev[i];
    }

    return rev_ss.str();
}

bool PhyCableRecord::ModuleRecord::IsCMISCable() const
{
    switch (this->cable_identifier)
    {
    case CABLE_MODULE_INFO_IDENTIFIER_SFP_DD:
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP_DD:
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP_CMIS:
    case CABLE_MODULE_INFO_IDENTIFIER_OSFP:
    case CABLE_MODULE_INFO_IDENTIFIER_DSFP:
          return (true);

    default:
          return (false);
    }
}

string PhyCableRecord::ModuleRecord::ConvertCableIdentifierToStr() const
{
    string result;

    switch (this->cable_identifier) {
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP28:
        result = "QSFP28";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP_PLUS:
        result = "QSFP+";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_SFP28_SFP_plus:
        result = "SFP28/SFP+";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_QSA:
        result = "QSA (QSFP->SFP)";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_Backplane:
        result = "Backplane";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_SFP_DD:
        result = "SFP-DD";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP_DD:
        result = "QSFP-DD";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP_CMIS:
        result = "QSFP_CMIS";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_OSFP:
        result = "OSFP";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_C2C:
        result = "C2C";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_DSFP:
        result = "DSFP";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP_SPLIT_CABLE:
        result = "QSFP_Split_Cable";
        break;
    default:
        result = "NA";
        break;
    }

    return (result);
}

string PhyCableRecord::ModuleRecord::ConvertCableLengthToStr(bool is_csv,
                                                    const string &na_val) const
{
    const  u_int8_t cable_cmis_length_mul_0_1  = 0;
    const  u_int8_t cable_cmis_length_mul_1    = 0x1;
    const  u_int8_t cable_cmis_length_mul_10   = 0x2;
    const  u_int8_t cable_cmis_length_mul_100  = 0x3;

    if (!this->cable_length)
        return (na_val);

    stringstream cable_length_ss;
    u_int64_t cable_length = (u_int64_t)this->cable_length;
    if (this->IsCMISCable()) {
        u_int8_t cable_length_value = this->cable_length & 0x3f;
        if (!cable_length_value)
            return (na_val);

        u_int8_t cable_length_mul =  static_cast<u_int8_t>(this->cable_length >> 6);
        u_int64_t mul = 0;
        if (cable_length_mul == cable_cmis_length_mul_0_1) {
            double cable_length_float = 0.1 * (double)cable_length_value;

            if (is_csv)
                cable_length_ss << '\"' << cable_length_float  << " m" << '\"';
            else
                cable_length_ss << cable_length_float << " m";

            return (cable_length_ss.str());
        } else {
            if (cable_length_mul == cable_cmis_length_mul_1)
                mul = 1;
            else if (cable_length_mul == cable_cmis_length_mul_10)
                mul = 10;
            else if (cable_length_mul == cable_cmis_length_mul_100)
                mul = 100;
        }

        cable_length = mul * (u_int64_t)cable_length_value;
    }

    if (is_csv)
        cable_length_ss << '\"' << cable_length << " m" << '\"';
    else
        cable_length_ss << cable_length << " m";

    return (cable_length_ss.str());
}

string PhyCableRecord::ModuleRecord::ConvertCableLengthSMFiberToStr() const
{
    u_int16_t length_base = this->smf_length & 0xff;
    u_int16_t unit_bit = (this->smf_length >> 8) & 0x3;
    stringstream smf_length_ss;

    if (!length_base)
        return ("N/A");

    switch (unit_bit) {
    case 0:
        smf_length_ss << length_base << " km";
        break;
    case 0x1:
        smf_length_ss << (double)(length_base * 100) / 1000.0 << " km";
        break;
    default:
        smf_length_ss << "N/A";
    }

    return (smf_length_ss.str());
}

string PhyCableRecord::ModuleRecord::ConvertCableLengthOMXToStr(u_int8_t om_index) const
{
    u_int8_t unit;
    u_int8_t length_omx;
    switch (om_index) {
    case 1:
        length_omx = this->length_om1;
        unit = 10;
        break;
    case 2:
        length_omx = this->length_om2;
        if (this->cable_identifier == CABLE_MODULE_INFO_IDENTIFIER_SFP28_SFP_plus)
            unit = 10;
        else
            unit = 1;
        break;
    case 3:
        length_omx = this->length_om3;
        if (this->cable_identifier == CABLE_MODULE_INFO_IDENTIFIER_SFP28_SFP_plus)
           unit = 10;
        else
           unit = 2;
        break;
    case 4:
        length_omx = this->length_om4;
        if (this->cable_identifier == CABLE_MODULE_INFO_IDENTIFIER_SFP28_SFP_plus)
           unit = 10;
        else
           unit = 2;
        break;
    case 5:
        length_omx = this->length_om5;
        unit = 2;
        break;
    default:
        return ("N/A");
    }

    if (!length_omx)
        return ("N/A");

    stringstream length_omx_ss;
    length_omx_ss << (u_int64_t)length_omx * unit << " m";
    return (length_omx_ss.str());
}

u_int8_t PhyCableRecord::ModuleRecord::SelectTransmitterTechnology() const
{
    if (this->IsCMISCable())
        return (this->cable_technology);

    return static_cast<u_int8_t>(this->cable_technology >> 4);
}

string PhyCableRecord::ModuleRecord::ConvertTransmitterTechnologyToStr(const string &na_val) const
{
    u_int8_t technology = this->SelectTransmitterTechnology();
    return CombinedCableInfo::CableTypeToStr(technology, na_val);
}

string PhyCableRecord::ModuleRecord::ConvertIBComplianceCodeToStr(const string &na_val) const
{
    return CombinedCableInfo::SupportedSpeedToStr(this->ib_compliance_code, na_val);
}

string PhyCableRecord::ModuleRecord::ConvertTemperatureToStr(u_int16_t temp,
                                                             bool is_csv) const
{
    const string na_val = (is_csv ? "NA" : "N/A");

    u_int8_t technology = static_cast<u_int8_t>(this->cable_technology >> 4);
    int8_t temperature = static_cast<int8_t>(temp >> 8);

    return _to_cvs_quoted(is_csv, CombinedCableInfo::TemperatureToStr(
        technology, temperature, na_val));
}

string PhyCableRecord::ModuleRecord::ConvertVoltageToStr(u_int16_t vcc, bool is_csv) const
{
    return _to_cvs_quoted(is_csv, CombinedCableInfo::VoltageToStr(vcc));
}

u_int8_t PhyCableRecord::ModuleRecord::SelectNominalBR() const
{
    if (this->nbr100 == 0xff)
        return (this->nbr250 / 4);

    return (this->nbr100 / 10);
}

string PhyCableRecord::ModuleRecord::ConvertCDREnableTxRxToStr(bool is_rx,
                                                        const string &na_val, bool is_csv) const
{
    stringstream cdr_state_ss;

    if (this->IsModule() || this->IsActiveCable()) {
        u_int8_t cdr_cap = 0;
        u_int8_t cdr_state = 0;
        if (is_rx) {
            cdr_cap = this->rx_cdr_cap;
            cdr_state = this->rx_cdr_state;
        } else {
            cdr_cap = this->tx_cdr_cap;
            cdr_state = this->tx_cdr_state;
        }

        if (is_csv || cdr_cap) {
            cdr_state_ss << PTR(cdr_state);
        } else {
            cdr_state_ss << na_val;
        }
    } else {
        cdr_state_ss << na_val;
    }

    return (cdr_state_ss.str());
}

bool PhyCableRecord::ModuleRecord::IsModule() const
{
    if (((this->cable_technology >> 4) != CABLE_TYPE_COPPER_UNEQUALIZED) &&
        (this->connector_type != CABLE_NO_SEPARABLE_CONNECTOR))
        return (true);

    return(false);
}

bool PhyCableRecord::ModuleRecord::IsActiveCable() const
{
    if (((this->cable_technology >> 4) != CABLE_TYPE_COPPER_UNEQUALIZED) &&
        (this->connector_type == CABLE_NO_SEPARABLE_CONNECTOR))
        return (true);

    return (false);
}

string PhyCableRecord::ModuleRecord::ConvertTxEQRxAMPRxEMPToStr(u_int8_t val,
                                            bool is_csv, const string &na_val) const
{
    stringstream val_ss;
    if (this->IsModule() || this->IsActiveCable()) {
        if (!is_csv)
            val_ss << +val << " " << +val << " " << +val << " " << +val;
        else
            val_ss << HEX(val) << HEX(val) << HEX(val) << HEX(val);
    } else
        val_ss << na_val;

    return (val_ss.str());
}

bool PhyCableRecord::ModuleRecord::GetFWVersion(uint8_t &fw_hi,
                                                 uint8_t &fw_mid,
                                                 uint16_t &fw_lo) const
{
    fw_hi = fw_mid = 0;
    fw_lo = 0;

    if (this->fw_version != 0 && (this->IsModule() || this->IsActiveCable())) {
        fw_hi = (uint8_t)(this->fw_version >> 24);
        fw_mid = (uint8_t)((this->fw_version >> 16) & 0x00ff);
        fw_lo = (uint16_t)(this->fw_version & 0x0000ffff);

        return true;
    }

    return false;

}

string PhyCableRecord::ModuleRecord::ConvertFWVersionToStr(const string& na_val) const
{
    stringstream fw_version_ss;

    uint8_t fw_hi = 0;
    uint8_t fw_mid = 0;
    uint16_t fw_lo = 0;

    if (this->GetFWVersion(fw_hi, fw_mid, fw_lo))
        fw_version_ss << +fw_hi << "." << +fw_mid << "." << fw_lo;
    else
        fw_version_ss << na_val;

    return (fw_version_ss.str());
}

string PhyCableRecord::ModuleRecord::ConvertAttenuationToStr(bool is_csv) const
{
    stringstream attenuation_ss;

    string separator;
    string na_val;

    if (is_csv) {
        na_val = "\"NA\"";
        separator = ",";
    } else {
        separator = " ";
        na_val = "NA";
    }

    if (this->IsPassiveCable())
        attenuation_ss << +this->cable_attenuation_5g  << separator
                       << +this->cable_attenuation_7g  << separator
                       << +this->cable_attenuation_12g << separator
                       << +this->cable_attenuation_25g;
    else
        attenuation_ss << na_val << separator
                       << na_val << separator
                       << na_val << separator
                       << na_val;

    return (attenuation_ss.str());
}

bool PhyCableRecord::ModuleRecord::IsPassiveCable() const
{
    if ((this->cable_technology >> 4) == CABLE_TYPE_COPPER_UNEQUALIZED)
        return (true);

    return (false);
}

string PhyCableRecord::ModuleRecord::ConvertDateCodeToStr(const string &na_val) const
{
    u_int64_t date_code = this->date_code >> 16;
    char *p_date_code = (char *)&date_code;
    for (int i = 0; i < 6; i++) {
         if (!isdigit(p_date_code[i]))
             return na_val;
    }

    stringstream date_code_ss;
    date_code_ss << p_date_code[5] << p_date_code[4] << '-'
                 << p_date_code[3] << p_date_code[2] << '-'
                 << p_date_code[1] << p_date_code[0];

    return (date_code_ss.str());
}

double PhyCableRecord::dBm_to_mW(double dBm) const
{
    if (dBm == 0)
       return -999.999;

    return 1 * pow(10, dBm/10);
}

double PhyCableRecord::mW_to_dBm(double mW) const
{
    if (mW <= 0)
       return -999.999;

    return 10 * log10(mW);
}

string PhyCableRecord::GetUniqueKey() const
{
    string pn = this->VendorPnToStr();
    if (pn == "NA")
        return "";

    string identifier;
    if (this->p_module)
       identifier =  _to_ptr_string(this->p_module->cable_identifier);
    else
       return "";

    string revision = this->RevisionToStr();
    if (revision == "NA")
        return "";

    return pn + '_' + identifier + '_' + revision;
}

string PhyCableRecord::ModuleRecord::ConvertMaxPowerToStr(const string &na_val) const
{
    if (!this->max_power)
        return na_val;

    return _to_string ((double)this->max_power / 4.0) + " W";
}

int PhyCableRecord::GetTemperatureAlarms() const
{
    if (!this->p_latched)
        return 0;

    return this->p_latched->temp_flags;
}

PrtlRecord::PrtlRecord():
    lp_msb(0), local_port(0), rtt_support(0), latency_accuracy(0),
    latency_res(0), local_phy_latency(0), local_mod_dp_latency(0),
    round_trip_latency(0) {}

string PrtlRecord::ToString() const
{
    stringstream ss;
    ss << "(round_trip_latency="   << this->round_trip_latency    << ", "
       << "latency_res="           << +this->latency_res          << ", "
       << "local_mod_dp_latency="  << this->local_mod_dp_latency  << ", "
       << "local_phy_latency="     << this->local_phy_latency     << ')';

    return ss.str();
}

float PrtlRecord::CalculateLength(const PrtlRecord &remote) const
{
    //round_trip_latency is combined of the following parts
    //    1. round trip on a cable
    //    2. delay by FW on both sides
    //    3. delay by HW on both sides
    //(P1.round_trip_latency*P1.latency_res - (P1.local_mod_dp_latency + P2.local_mod_dp_latency) -
    // (P1.local_phy_latency + P2.local_phy_latency))/ 2*5 [ns]

    return ((float)(this->round_trip_latency * this->latency_res) - (float)(this->local_mod_dp_latency + remote.local_mod_dp_latency)
                   - (float)(this->local_phy_latency + remote.local_phy_latency))/ 10;
}

float PrtlRecord::CalculateLength(const PrtlRecord &remote, string &message) const
{
    stringstream ss;
    float length = -1;

    if (this->round_trip_latency == 0xFFFFFF) {
        message = "The PRTL round trip latency exceeds its maximal possible value.";
        return -1;
    }

    if (!this->round_trip_latency || (length = this->CalculateLength(remote)) <= 0) {
        ss << "The cable's length is below latency resolution."
           << " PRTL registers details: local=" << this->ToString()
           << " and remote=" << remote.ToString();
        message = ss.str();
        return -1;
    }

    return length;
}
