/*
 * Copyright (c) 2024-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Configure options
 */
// #define HAVE_HOSTLIST_ASSERT 1
// #define HAVE_HOSTLIST_MT     0
/*
 *
 */

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "infiniband/ibdm/hostlist/HostList.h"

#if defined(HAVE_HOSTLIST_ASSERT)
#if defined(NDEBUG)
#error assert() operation is not available
#endif
#define sys_assert(x) assert(x)
#else
#define sys_assert(x) ((void)0)
#endif /* HAVE_ASSERT */

/**
 * @name Synchronization
 * @brief Multi-threading Synchronization operations.
 */
/** @{ */
#if defined(HAVE_HOSTLIST_MT)
#include <pthread.h>
#define lock_t pthread_mutex_t
#if defined(_GNU_SOURCE)
#define sys_lock_init(x)                                           \
    {                                                              \
        pthread_mutexattr_t attr;                                  \
        pthread_mutexattr_init(&attr);                             \
        pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE); \
        pthread_mutex_init(x, &attr);                              \
        pthread_mutexattr_destroy(&attr);                          \
    }
#else
#define sys_lock_init(x) pthread_mutex_init(x, NULL)
#endif
#define sys_lock_destroy(x) pthread_mutex_destroy(x)
#define sys_lock_acquire(x) pthread_mutex_lock(x)
#define sys_lock_release(x) pthread_mutex_unlock(x)
#else
#define lock_t              int
#define sys_lock_init(x)    ((void)0)
#define sys_lock_destroy(x) ((void)0)
#define sys_lock_acquire(x) ((void)0)
#define sys_lock_release(x) ((void)0)
#endif /* HAVE_MT */
/** @} */

/* hostrange type:
 * A single prefix with 'hi' and 'lo' numeric suffix values
 * Single host is recognized as 'width' is set as SINGLE_RANGE
 */
struct hostrange
{
    /* alphanumeric prefix: */
    char* prefix;

    /* beginning (lo) and end (hi) of suffix range */
    uint32_t lo, hi;

    /* width of numeric output format
     * (pad with zeros up to this width) */
    int width;
};

/* The hostlist type:
 * An array of hostrange values describing hosts
 */
struct hostlist
{
    struct
    {
        /* pointer to hostrange array */
        struct hostrange** array;

        /* current number of elements available in array */
        int size;

        /* number of elements to grow */
        int grow;

        /* count of valid elements stored in array */
        int count;
    } ranges;

    /* current number of hosts stored in hostlist */
    int count;

    lock_t lock;
};

static int _hostrange_create(const char* str, struct hostrange** ranges);
static void _hostrange_destroy(struct hostrange* hr);
static inline size_t _hostrange_count(struct hostrange* hr);
static inline int _hostrange_empty(struct hostrange* hr);
static inline int _hostrange_single(struct hostrange* hr);
static char* _hostrange_n2host(struct hostrange* hr, size_t n);
static char* _hostrange_shift(struct hostrange* hr);
static struct hostrange* _hostrange_copy(struct hostrange* hr);
static int _cmp_range(const void* hr1, const void* hr2);
static char* _next_token(char*, char**);
static int _parse_range_list(const char* str, struct hostrange** ranges);
static int _parse_range_single(const char* str, struct hostrange** ranges);
static int _parse_range(const char* str, char* prefix, struct hostrange* hr);

static hostlist_t _hostlist_create_bracketed(const char* str, char* sep, char* r_op);
static void _hostlist_destroy(hostlist_t hl);
static inline int _hostlist_base(void);
static int _hostlist_resize(hostlist_t hl, size_t size);
static void _hostlist_delete_range(hostlist_t hl, int n);
static int _hostlist_push_range(hostlist_t hl, struct hostrange* hr);

#define SINGLE_RANGE (-1)

/* max host range in array of ranges: for sanity check */
#define MAX_RANGE (64 * 1024)

/* max number of ranges that will be processed between brackets: for sanity check */
#define MAX_RANGES (64 * 1024)

/* grow host range in array */
#define GROW_RANGE (16)

hostlist_t hostlist_create(const char* str)
{
    return _hostlist_create_bracketed(str, (char*)"\t, ", (char*)"-");
}

void hostlist_destroy(hostlist_t hl)
{
    if (!hl) {
        return;
    }
    _hostlist_destroy(hl);
}

int hostlist_count(hostlist_t hl)
{
    int count;

    if (!hl) {
        return -1;
    }

    sys_lock_acquire(&hl->lock);
    count = hl->count;
    sys_lock_release(&hl->lock);

    return count;
}

char* hostlist_n2host(hostlist_t hl, size_t n)
{
    char* host = NULL;
    int i, count;

    if (!hl) {
        return NULL;
    }

    sys_lock_acquire(&hl->lock);
    count = 0;
    for (i = 0; i < hl->ranges.count; i++) {
        int num_in_range = (int)_hostrange_count(hl->ranges.array[i]);

        if (n <= (size_t)(num_in_range - 1 + count)) {
            host = _hostrange_n2host(hl->ranges.array[i], n - (size_t)count);
            break;
        } else
            count += num_in_range;
    }
    sys_lock_release(&hl->lock);

    return host;
}

char* hostlist_shift(hostlist_t hl)
{
    char* host = NULL;

    if (!hl) {
        return NULL;
    }

    sys_lock_acquire(&hl->lock);
    if (hl->count > 0) {
        struct hostrange* hr = hl->ranges.array[0];

        host = _hostrange_shift(hr);
        hl->count--;

        if (_hostrange_empty(hr)) {
            _hostlist_delete_range(hl, 0);
        }
    }
    sys_lock_release(&hl->lock);

    return host;
}

void hostlist_uniq(hostlist_t hl)
{
    int i = 1;

    if (!hl || hl->ranges.count <= 1) {
        return;
    }

    sys_lock_acquire(&hl->lock);
    qsort(hl->ranges.array, hl->ranges.count, sizeof(struct hostrange*), &_cmp_range);

    while (i < hl->ranges.count) {
        struct hostrange* h1 = hl->ranges.array[i - 1];
        struct hostrange* h2 = hl->ranges.array[i];
        int rc = 0;
        int duph = -1;

        rc = strcmp(h1->prefix, h2->prefix);
        if (0 == rc) {
            rc = h1->width - h2->width;
            if (0 == rc) {
                if (_hostrange_single(h1)) {
                    /* both hosts are identical:
                     *  - remove h2
                     *  - decrement total host count
                     */
                    duph = 1;
                } else if (h1->hi == h2->lo - 1) {
                    /* h2 included in h1:
                     *  - remove h2
                     *  - total host count should be the same
                     */
                    h1->hi = h2->hi;
                    duph = 0;
                } else if (h1->hi >= h2->lo) {
                    /* h1 and h2 are overlapped:
                     *  - remove h2
                     *  - decrement total host count
                     */
                    if (h1->hi < h2->hi) {
                        duph = h1->hi - h2->lo + 1;
                        h1->hi = h2->hi;
                    } else {
                        duph = (int)_hostrange_count(h2);
                    }
                }
            }
        }

        /* change number of host ranges or total host count */
        if (duph >= 0) {
            _hostlist_delete_range(hl, i);
            hl->count -= duph;
        } else {
            i++;
        }
    }
    sys_lock_release(&hl->lock);
}

int hostlist_proc(const char* buf, void* arg, int uniq, int (*callback)(const char* hostname, void* arg))
{
    int rc = 0;
    int retval = 0;
    hostlist_t hl;
    char* host = NULL;

    if (!buf || !callback) {
        return -EINVAL;
    }
    hl = hostlist_create(buf);
    if (NULL == hl) {
        return -EINVAL;
    }

    if (uniq) {
        hostlist_uniq(hl);
    }

    while (NULL != (host = hostlist_shift(hl))) {
        rc = callback(host, arg);
        free(host);
        if (rc != 0) {
            retval = (rc > 0 ? -rc : rc);
            break;
        }
        retval++;
    }

    hostlist_destroy(hl);

    return retval;
}

/*
 * Create an array of range objects using input string
 * Note: it is used to push this array to hostlist object later
 */
static int _hostrange_create(const char* str, struct hostrange** ranges)
{
    struct hostrange* hr = NULL;
    int nr = 0;

    sys_assert(str);
    sys_assert(ranges);

    if (strrchr(str, '[')) {
        nr = _parse_range_list(str, &hr);
    } else {
        nr = _parse_range_single(str, &hr);
        sys_assert(nr < 2);
    }

    if (nr > 0) {
        *ranges = hr;
    }

    return nr;
}

/*
 * Destroy hostrange object 'hr'
 */
static void _hostrange_destroy(struct hostrange* hr)
{
    sys_assert(hr);

    if (hr->prefix) {
        free(hr->prefix);
        hr->prefix = NULL;
    }
    free(hr);
}

/*
 * Return number of hosts described by hostrange object 'hr'
 */
static inline size_t _hostrange_count(struct hostrange* hr)
{
    sys_assert(hr);
    return _hostrange_empty(hr) ? 0 : (hr->hi - hr->lo + 1);
}

/*
 * Check if 'hr' is valid
 */
static inline int _hostrange_empty(struct hostrange* hr)
{
    sys_assert(hr);
    return ((hr->hi < hr->lo) || (hr->hi == ((uint32_t)-1)));
}

/*
 * Check if 'hr' is specific hostrange object
 */
static inline int _hostrange_single(struct hostrange* hr)
{
    sys_assert(hr);
    return (SINGLE_RANGE == hr->width);
}

/*
 * Return hostname using 'n' position in the hostrange 'hr'
 */
static char* _hostrange_n2host(struct hostrange* hr, size_t n)
{
    char* host = NULL;

    sys_assert(hr);

    if (_hostrange_single(hr)) {
        sys_assert(n == 0);

        host = strdup(hr->prefix);
    } else {
        size_t size = 0;

        sys_assert(n <= (hr->hi - hr->lo));

        size = strlen(hr->prefix) + hr->width + 16;
        host = (char*)calloc(1, size * sizeof(char));
        if (host) {
            snprintf(host, size, "%s%0*lu", hr->prefix, hr->width, hr->lo + n);
        }
    }

    return host;
}

/*
 * Shift host two single position in hostrange object
 * Return hostname
 */
static char* _hostrange_shift(struct hostrange* hr)
{
    char* host = NULL;

    sys_assert(hr);

    host = _hostrange_n2host(hr, 0);

    if (host) {
        hr->lo++;
    }

    return host;
}

/*
 * Allocate memory for new hostrange object and
 * set that same values as passed 'hr'
 */
static struct hostrange* _hostrange_copy(struct hostrange* hr)
{
    struct hostrange* p_new = NULL;

    if (hr) {
        p_new = (struct hostrange*)calloc(1, sizeof(*hr));
        if (!p_new) {
            return NULL;
        }

        memcpy(p_new, hr, sizeof(*hr));
        p_new->prefix = strdup(hr->prefix);

        if (NULL == p_new->prefix) {
            free(p_new);
            return NULL;
        }
    }

    return p_new;
}

/*
 * Comparing function for two hostrange objects
 * Note: this function is used in qsort()
 */
static int _cmp_range(const void* hr1, const void* hr2)
{
    struct hostrange* h1 = *(struct hostrange**)hr1;
    struct hostrange* h2 = *(struct hostrange**)hr2;
    int rc = 0;

    sys_assert(hr1);
    sys_assert(hr2);

    rc = strcmp(h1->prefix, h2->prefix);
    if (0 == rc) {
        rc = h1->width - h2->width;
        if (0 == rc) {
            rc = h1->lo - h2->lo;
        }
    }

    return rc;
}

/*
 * Returns a pointer to the next token.
 */
static char* _next_token(char* sep, char** str)
{
    char *tok, *parse, *open_bracket, *close_bracket;

    while ((**str != '\0') && (strchr(sep, **str) != NULL)) {
        (*str)++;
    }

    if (**str == '\0') {
        return NULL;
    }

    tok = *str;
    parse = tok;
    while (1) {
        while ((**str != '\0') && (strchr(sep, **str) == NULL)) {
            (*str)++;
        }

    again:
        open_bracket = strchr(parse, '[');
        if ((open_bracket == NULL) || (open_bracket > *str)) {
            break;
        }
        close_bracket = strchr(parse, ']');
        if ((close_bracket == NULL) || (close_bracket < open_bracket)) {
            break;
        }
        if (close_bracket < *str) {
            parse = close_bracket + 1;
            goto again;
        } else {
            *str = close_bracket;
        }
    }

    while ((**str != '\0') && (strchr(sep, **str) != NULL)) {
        *(*str)++ = '\0';
    }

    return tok;
}

/*
 * Parse token in bracketed notation
 * Return ranges array 'ranges' and number of filled ranges
 */
static int _parse_range_list(const char* str, struct hostrange** ranges)
{
    struct hostrange* hr = NULL;
    int nr = 0;
    char* p;

    sys_assert(str);
    sys_assert(ranges);
    p = strrchr((char*)str, '[');
    if (p == NULL) {
        return -EINVAL;
    } else {
        int size = 0;
        char *q, *prefix, *suffix;

        *p++ = '\0';
        prefix = (char*)str;
        suffix = p;
        q = strchr(p, ']');
        if (q) {
            if ((q[1] != ',') && (q[1] != '\0')) {
                return -EINVAL;
            }
            *q = '\0';
            while (suffix) {
                /* get current range */
                p = strchr(suffix, ',');
                if (p) {
                    /* skip sequence of ',' */
                    if (p == suffix) {
                        *p++ = '\0';
                        suffix = p;
                        continue;
                    }
                    *p++ = '\0';
                }
                /* increase ranges buffer */
                if (nr >= size) {
                    size += 16;
                    if (size > MAX_RANGES) {
                        goto error;
                    }//was realloc(hr, size * sizeof(*hr));
                    hr = (hostrange*)realloc(hr, size * sizeof(*hr));
                    if (!hr) {
                        goto error;
                    }
                }
                /* process current range */
                if (0 > _parse_range(suffix, prefix, &hr[nr])) {
                    goto error;
                }
                nr++;
                suffix = p;
            }
        } else {
            return -EINVAL;
        }
    }

    *ranges = hr;

    return nr;

error:
    if (hr) {
        while (nr--) {
            free(hr[nr].prefix);
        }
        free(hr);
    }
    return -EINVAL;
}

/*
 * Parse single token passed as 'node1'
 */
static int _parse_range_single(const char* str, struct hostrange** ranges)
{
    struct hostrange* hr = NULL;
    int nr = 1;
    char* prefix;
    int idx = 0;

    sys_assert(str);
    sys_assert(ranges);

    idx = (int)strlen(str) - 1;

    hr = (struct hostrange*)calloc(1, sizeof(*hr));
    if (!hr) {
        return -ENOMEM;
    }
    prefix = strdup(str);
    if (NULL == prefix) {
        free(hr);
        return -ENOMEM;
    }

    /* find digital part */
    while ((idx >= 0) && isdigit((int)str[idx])) {
        idx--;
    }

    if (idx == ((int)strlen(str) - 1)) {
        hr->prefix = prefix;
        hr->lo = 0L;
        hr->hi = 0L;
        hr->width = SINGLE_RANGE;
    } else {
        prefix[idx + 1] = '\0';
        nr = _parse_range((str + idx + 1), prefix, hr);
        free(prefix);
        if (nr < 1) {
            free(hr);
            return -EINVAL;
        }
    }

    *ranges = hr;

    return nr;
}

/*
 * Parse single range using sting 'str'
 * 'prefix' is passed as a parameter
 * Return number of hosts in this range and filled 'hr'
 */
static int _parse_range(const char* str, char* prefix, struct hostrange* hr)
{
    char *p, *q;

    sys_assert(str);
    sys_assert(prefix);
    sys_assert(hr);

    memset(hr, 0, sizeof(*hr));

    /* expect a digit inside */
    if (!isdigit((int)str[0])) {
        return -EINVAL;
    }

    hr->prefix = strdup(prefix);
    if (!hr->prefix) {
        return -ENOMEM;
    }
    p = strchr((char*)str, '-');
    if (p) {
        *p++ = '\0';
    }

    hr->width = (int)strlen(str);

    hr->hi = hr->lo = (uint32_t)strtoul(str, &q, _hostlist_base());
    if (q == str) {
        goto error;
    }

    if (p && *p) {
        hr->hi = (uint32_t)strtoul(p, &q, _hostlist_base());
    }

    if (q == p || *q != '\0') {
        goto error;
    }

    if (hr->lo > hr->hi) {
        goto error;
    }

    if (hr->hi - hr->lo + 1 > MAX_RANGE) {
        goto error;
    }

    return (hr->hi - hr->lo + 1);

error:
    free(hr->prefix);
    hr->prefix = NULL;
    return -EINVAL;
}

/* Create hostlist object using input string
 */
static hostlist_t _hostlist_create_bracketed(const char* str, char* sep, char* r_op)
{
    hostlist_t hl = NULL;
    struct hostrange* hr = NULL;
    char *tok, *cur, *orig;
    int rc = 0;
    int idx = 0;

    hl = (hostlist_t)calloc(1, sizeof(*hl));
    if (!hl) {
        return NULL;
    }

    hl->ranges.count = 0;
    hl->ranges.grow = GROW_RANGE;
    hl->ranges.size = 0;
    if (0 != _hostlist_resize(hl, hl->ranges.grow)) {
        free(hl);
        return NULL;
    }

    hl->count = 0;
    sys_lock_init(&hl->lock);

    if (str == NULL)
        return hl;
    orig = cur = strdup(str);
    if (!orig) {
        goto error;
    }

    while ((tok = _next_token(sep, &cur)) != NULL) {
        int nr = 0;
        rc = nr = _hostrange_create(tok, &hr);
        if (rc > 0) {
            for (idx = 0; idx < nr; idx++) {
                if (rc > 0) {
                    rc = _hostlist_push_range(hl, &hr[idx]);
                }
                free(hr[idx].prefix);
            }
            free(hr);
        }
        if (rc <= 0) {
            goto error;
        }
    }

    free(orig);
    return hl;

error:
    free(orig);
    hostlist_destroy(hl);
    return NULL;
}

/* Completely destroy hostlist 'hl'
 */
static void _hostlist_destroy(hostlist_t hl)
{
    int i;

    sys_assert(hl);

    for (i = 0; i < hl->ranges.count; i++) {
        _hostrange_destroy(hl->ranges.array[i]);
    }
    free(hl->ranges.array);

    sys_lock_destroy(&hl->lock);
    free(hl);
}

/* Numeric base
 */
static inline int _hostlist_base(void)
{
    return 10;
}

/* Resize the internal array used to store the list of hostrange objects.
 * Size passed as 'size' should be greater then current size
 */
static int _hostlist_resize(hostlist_t hl, size_t size)
{
    int i;
    size_t oldsize;

    sys_assert(hl);

    if (hl->ranges.count > (int)size) {
        return -EINVAL;
    }

    oldsize = hl->ranges.size;
    hl->ranges.size = (int)size;
    //was realloc((void*)hl->ranges.array, (hl->ranges.size * sizeof(struct hostrange*)));
	hl->ranges.array = (hostrange**)realloc((void*)hl->ranges.array, (hl->ranges.size * sizeof(struct hostrange*)));
    if (!hl->ranges.array)
        return -ENOMEM;

    for (i = (int)oldsize; i < hl->ranges.size; i++) {
        hl->ranges.array[i] = NULL;
    }

    return 0;
}

/* Remove hostrange object in 'n' position from hostlist 'hl'
 */
static void _hostlist_delete_range(hostlist_t hl, int n)
{
    int i;
    struct hostrange* old;

    sys_assert(hl);
    sys_assert(n >= 0);

    old = hl->ranges.array[n];
    for (i = n; i < hl->ranges.count - 1; i++) {
        hl->ranges.array[i] = hl->ranges.array[i + 1];
    }
    hl->ranges.count--;
    hl->ranges.array[hl->ranges.count] = NULL;

    _hostrange_destroy(old);
}

/* Push a hostrange object 'hr' onto hostlist 'hl'
 * Returns the number of hosts successfully pushed onto hl
 * or negative failure value
 */
static int _hostlist_push_range(hostlist_t hl, struct hostrange* hr)
{
    struct hostrange* tail;
    int retval;

    sys_assert(hl);
    sys_assert(hr);

    tail = (hl->ranges.count > 0) ? hl->ranges.array[hl->ranges.count - 1] : hl->ranges.array[0];

    if (hl->ranges.size == hl->ranges.count && (0 != _hostlist_resize(hl, hl->ranges.size + hl->ranges.grow))) {
        goto error;
    }

    if (hl->ranges.count > 0 && !_hostrange_single(tail) && !_hostrange_single(hr) && tail->width == hr->width &&
        tail->hi == (hr->lo - 1) && strcmp(tail->prefix, hr->prefix) == 0)
    {
        tail->hi = hr->hi;
    } else {
        struct hostrange* p_new = _hostrange_copy(hr);
        if (!p_new) {
            goto error;
        }
        hl->ranges.array[hl->ranges.count] = p_new;
        hl->ranges.count++;
    }

    retval = (int)_hostrange_count(hr);
    hl->count += retval;

    return retval;

error:
    return -EPERM;
}
