/*
 * Copyright (c) 2020-2020 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 2021-2025, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include "ibis_mads_stat.h"
#include "ibis_mad_names.h"


IbisMadsStat::~IbisMadsStat() {
    clear();
}

IbisMadsStat::histogram_base::histogram_base(const mads_record_t * record)
    : m_max_value(1), m_min_value(0x0FFFFFFFFFFFFFFF), m_total(0),
      m_max_time(0), m_min_time(0x0FFFFFFFFFFFFFFF),
      m_end_time(0), m_is_valid(false)
{
    if( !record )
        return;

    if(record->histogram.empty())
        return;

    for(time_histogram_t::const_iterator x = record->histogram.begin(); x != record->histogram.end(); ++x)
    {
        if(m_min_value > x->second) m_min_value = x->second;
        if(m_max_value < x->second) m_max_value = x->second;

        if(m_min_time > x->first) m_min_time = x->first;
        if(m_max_time < x->first) m_max_time = x->first;

        m_total += x->second;
    }
    m_end_time = record->end_time.tv_sec;

    if(!m_end_time)
    {
        struct timespec timestamp;
        clock_gettime(CLOCK_REALTIME, &timestamp);
        m_end_time = timestamp.tv_sec;
    }

    m_is_valid = true;
}


void IbisMadsStat::add(uint8_t * raw)
{
    if(!mads_record)
        return;

    key_t key(*(umad_hdr_t*)raw);

    if(mads_record->table.empty())
    {
        mads_cache.first = mads_record->table.insert(std::make_pair(key,1)).first;
        ++mads_cache.first->second;
    }
    else
        if(mads_cache.first->first == key)
            ++mads_cache.first->second;
        else
            if(mads_cache.second->first == key)
                ++mads_cache.second->second;
            else
                if(mads_cache.thrid->first == key)
                    ++mads_cache.thrid->second;
                else
                {
                    mads_cache.thrid    = mads_cache.second;
                    mads_cache.second   = mads_cache.first;
                    mads_cache.first    = mads_record->table.insert(std::make_pair(key,0)).first;
                    ++mads_cache.first->second;
                }

    if(is_histogram_enabled)
    {
        struct timespec timestamp;
        clock_gettime(CLOCK_REALTIME, &timestamp);

        if(mads_record->histogram.empty() || mads_record->histogram_entry->first != timestamp.tv_sec)
            mads_record->histogram_entry =
                mads_record->histogram.insert(
                    mads_record->histogram.end(),
                    std::make_pair(timestamp.tv_sec, 0));


        ++mads_record->histogram_entry->second;
    }
}



void IbisMadsStat::start(const std::string & name)
{
    if(mads_record)
        stop();

    mads_record = new mads_record_t(name);

    if(mads_record)
    {
        mads_db.push_back(mads_record);
        mads_cache.first =
            mads_cache.second =
            mads_cache.thrid =
            mads_record->table.begin();
    }
}



void IbisMadsStat::stop()
{
    if(mads_record)
    {
        clock_gettime(CLOCK_REALTIME, &mads_record->end_time);
        mads_record = NULL;
    }
}

void IbisMadsStat::clear()
{
    stop();

    for(mads_db_t::iterator x = mads_db.begin(); x != mads_db.end(); ++x)
        delete *x;

    mads_db.clear();
}



void IbisMadsStat::resume()
{
    if(mads_db.empty() || mads_record != NULL)
        return;

    mads_record                     = mads_db.back();
    mads_record->end_time.tv_sec    = 0;
    mads_record->end_time.tv_nsec   = 0;
}



void IbisMadsStat::aggregate(mads_record_t & record)
{
    if(!mads_db.empty())
    {
        record.start_time = (*mads_db.begin())->start_time;

        for(mads_db_t::const_iterator x = mads_db.begin(); x != mads_db.end(); ++x) {
            for(mads_map_t::const_iterator z = (*x)->table.begin(); z != (*x)->table.end(); ++z)
                record.table[z->first] += z->second;

            if(!(*x)->histogram.empty())
            {
                time_histogram_t::iterator itr = (*x)->histogram.begin();

                if(!record.histogram.empty())
                    if(record.histogram.back().first == itr->first) {
                        record.histogram[record.histogram.size()-1].second += itr->second;
                        ++itr;
                    }

                record.histogram.insert(record.histogram.end(), itr, (*x)->histogram.end());
            }

            record.end_time = (*x)->end_time;
        }
    }
}



std::ostream & IbisMadsStat::output_all_mads_tables(std::ostream & stream, bool skip_empty)
{
    for(mads_db_t::const_iterator x = mads_db.begin(); x != mads_db.end(); ++x) {
        if(skip_empty && (*x)->table.empty())
            continue;

        output_mads_table(stream, *x);
    }

    return stream;
}



std::ostream & IbisMadsStat::output_all_time_histograms(std::ostream & stream, bool skip_empty)
{
    for(mads_db_t::const_iterator x = mads_db.begin(); x != mads_db.end(); ++x) {
        if(skip_empty && (*x)->table.empty())
            continue;

        output_time_histogram(stream, *x);
    }

    return stream;
}



std::ostream & IbisMadsStat::output_all_records(std::ostream & stream, bool skip_empty)
{
    for(mads_db_t::const_iterator x = mads_db.begin(); x != mads_db.end(); ++x) {
        if(skip_empty && (*x)->table.empty())
            continue;

        output_record(stream, *x);
    }

    return stream;
}

std::ostream & IbisMadsStat::output_all_records_csv(std::ostream & stream)
{
    uint64_t total = 0;
    
    std::ios_base::fmtflags flags = stream.flags();
    stream << "Stage,MAD,mgmt_class,attr_id,method,count\n";

    for(mads_db_t::const_iterator x = mads_db.begin(); x != mads_db.end(); ++x) {
        if((*x)->table.empty()) {
            stream << (*x)->name << ",N/A,N/A,N/A,N/A,N/A\n";
            continue;
        }

        uint64_t sub_total = 0;

        for(mads_map_t::const_iterator z = (*x)->table.begin(); z != (*x)->table.end(); ++z) {
            const char* mad_name = IbisMadNames::getMadName(z->first.fields.mgmt_class, z->first.fields.attr_id);
            const char* mad_name_str = mad_name == NULL ? "Unknown MAD Name" : mad_name;
            stream << (*x)->name << ','
                   << mad_name_str << ','
                   << "0x" << std::hex << std::setfill('0') << std::setw(2) << +z->first.fields.mgmt_class << ','
                   << "0x" << std::hex << std::setfill('0') << std::setw(4) << +z->first.fields.attr_id << ','
                   << "0x" << std::hex << std::setfill('0') << std::setw(2) << +z->first.fields.method << ','
                   << std::dec << +z->second << '\n';

            sub_total += z->second;
        }

        stream << (*x)->name << " (Total),N/A,N/A,N/A,N/A," << std::dec << sub_total << '\n';

        total += sub_total;

    }

    stream << "TOTAL,N/A,N/A,N/A,N/A," << total << '\n';

    stream.flags(flags);
    return stream;
}

std::ostream & IbisMadsStat::output_all_perf_records_csv(std::ostream & stream)
{
    stream << "Name,duration_abs,duration_rel,min_bw,max_bw,abs_bw,rel_bw,adapt_bw\n";

    for(mads_db_t::const_iterator x = mads_db.begin(); x != mads_db.end(); ++x)
    {
        histogram_base base((*x));

        stream << (*x)->name;

        if( !base.m_is_valid ) {
            stream << ",N/A,N/A,N/A,N/A,N/A,N/A,N/A\n";
            continue;
        }

        stream << ',' << base.m_end_time - (*x)->start_time.tv_sec + 1
               << ',' << base.m_max_time - base.m_min_time + 1
               << ',' << base.m_min_value
               << ',' << base.m_max_value
               << ',' << (double)base.m_total / (double)(base.m_end_time - (*x)->start_time.tv_sec + 1)
               << ',' << (double)base.m_total / (double)(base.m_max_time - base.m_min_time + 1)
               << ',' << (double)base.m_total / (double)(*x)->histogram.size()
               << '\n';
    }

    return stream;
}

std::ostream & IbisMadsStat::output_mads_table_summary(std::ostream & stream)
{
    mads_record_t record("Summary");
    aggregate(record);
    output_mads_table(stream, &record);
    return stream;
}


std::ostream & IbisMadsStat::output_mads_table(std::ostream & stream, const mads_record_t * record)
{
    if(!record) {
        if(!mads_record)
            return stream;
        else
            record = mads_record;
    }

    if(record->table.empty())
        return stream << output_prefix << "IBIS STATISTICS ['" << record->name << "']: NO MADs DATA FOUND" << std::endl;

    uint64_t total = 0;
    u_int8_t max_mad_name_len = IbisMadNames::getMaxMadNameLen();

    std::ios_base::fmtflags flags = stream.flags();

    stream  << IBISStat_LINE

            << output_prefix
            << " IBIS MADs statistic record '" << record->name << "' count ["
            << record->table.size() << "]" << std::endl

            << IBISStat_LINE

            << output_prefix
            << std::left << std::setw(max_mad_name_len) << "name" << std::right
            << std::setw(IBISStat_COLUMN_WIDTH) << "mgmt_class"
            << std::setw(IBISStat_COLUMN_WIDTH) << "attr_id"
            << std::setw(IBISStat_COLUMN_WIDTH) << "method"
            << std::setw(IBISStat_COLUMN_WIDTH) << "count" << std::endl;

    for(mads_map_t::const_iterator x = record->table.begin(); x != record->table.end(); ++x)
    {
        const char* mad_name = IbisMadNames::getMadName(x->first.fields.mgmt_class, x->first.fields.attr_id);
        const char* mad_name_str = mad_name == NULL ? "Unknown MAD Name" : mad_name;
        stream << output_prefix << std::hex
               << std::setfill(' ') <<  std::left << std::setw(max_mad_name_len)
               << mad_name_str << std::right
               
               << std::setfill(' ') << std::setw(IBISStat_COLUMN_WIDTH - 2)
               << "0x" << std::setfill('0') << std::setw(2) << +x->first.fields.mgmt_class

               << std::setfill(' ') << std::setw(IBISStat_COLUMN_WIDTH - 4)
               << "0x" << std::setfill('0') << std::setw(4) << +x->first.fields.attr_id

               << std::setfill(' ') << std::setw(IBISStat_COLUMN_WIDTH - 2)
               << "0x" << std::setfill('0') << std::setw(2) << +x->first.fields.method

               << std::setfill(' ') << std::setw(IBISStat_COLUMN_WIDTH) << std::dec << x->second << std::endl;

        total += x->second;
    }

    stream << IBISStat_LINE
            << output_prefix << std::setw(IBISStat_COLUMN_WIDTH * 4) << total << std::endl;

    stream.flags(flags);
    return stream;
}

std::ostream & IbisMadsStat::output_time_histogram_summary(std::ostream & stream)
{
    mads_record_t record("Summary");
    aggregate(record);
    output_time_histogram(stream, &record);
    return stream;
}

std::ostream & IbisMadsStat::output_time_histogram(std::ostream & stream, const mads_record_t * record)
{
    if(!record) {
        if(!mads_record)
            return stream;
        else
            record = mads_record;
    }

    if(record->histogram.empty())
        return stream << output_prefix << "IBIS STATISTICS ['" << record->name
               << "']: NO HISTOGRAM DATA FOUND" << std::endl;

    std::ios_base::fmtflags flags = stream.flags();

    stream << IBISStat_LINE

           << output_prefix
           << " IBIS histogram '" << record->name << "' count ["
           << record->histogram.size() << "]" << std::endl

           << IBISStat_LINE;

    if(record->histogram.empty())
    {
        stream << output_prefix << "EMPTY" << std::endl;
        stream.flags(flags);
        return stream;
    }

    histogram_base base(record);

    if (is_histogram_full_output)
    {
        double      k = ((double)IBISStat_HISTOGRAM_SIZE) / ((double)base.m_max_value);
        __time_t    z = -1;

        stream    << output_prefix << std::endl;

        for(time_histogram_t::const_iterator x = record->histogram.begin(); x != record->histogram.end(); ++x)
        {
            if(z != -1)
            {
                __time_t t = x->first - z - 1;
                if(t)
                    stream  << output_prefix << " "
                            << std::setfill('=') << std::setw((int)t) << ""
                            << std::setfill(' ') << "  <" << t << ">" << std::endl;
            }

            stream  << output_prefix << " "
                    << std::setfill('*') << std::setw((int)(k * (double)x->second) + 1) << ""
                    << std::setfill(' ') << "  [" << x->second << "]" << std::endl;

            z = x->first;
        }

        stream << output_prefix << std::endl;
    }

    stream << IBISStat_LINE
           << output_prefix << " Durations         : " << std::endl
           << output_prefix << "    Absolute       : " << base.m_end_time - record->start_time.tv_sec + 1 << std::endl
           << output_prefix << "    Relative       : " << base.m_max_time - base.m_min_time + 1 << std::endl
           << output_prefix << std::endl

           << output_prefix << " MADs per second   : " << std::endl
           << output_prefix << "    Min            : " << base.m_min_value << std::endl
           << output_prefix << "    Max            : " << base.m_max_value << std::endl
           << output_prefix << std::endl

           << output_prefix << "    Absolute BW    : " << (double)base.m_total / (double)(base.m_end_time - record->start_time.tv_sec + 1) << std::endl
           << output_prefix << "    Relative BW    : " << (double)base.m_total / (double)(base.m_max_time - base.m_min_time + 1) << std::endl
           << output_prefix << "    Adaptive BW    : " << (double)base.m_total / (double)record->histogram.size() << std::endl
           << IBISStat_LINE;

    stream.flags(flags);
    return stream;
}


std::ostream & IbisMadsStat::output_summary(std::ostream & stream)
{
    mads_record_t record("Summary");
    aggregate(record);
    output_record(stream, &record);
    return stream;
}

std::ostream & IbisMadsStat::output_record(std::ostream & stream, const mads_record_t * record)
{
    output_mads_table(stream, record);
    output_time_histogram(stream, record);
    return stream;
}
