/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved
 * Copyright (C) 2021-2025, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <cstdint>
#include <stdio.h>
#include <math.h>
#include <signal.h>
#include <map>
#include <list>
#include <iostream>
#include <errno.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "ibis/key_mngr.h"
#include "infiniband/ibdiag/ibdiag_log.h"
#include "infiniband/ibdiag/ibdiag_types.h"
#include "infiniband/ibdiag/key_updater.h"
#include "infiniband/ibdm/TopoMatch.h"
#include "infiniband/ibdm/SysDef.h"
#include "infiniband/ibdm/SysLogger.h"
#include <infiniband/ibdm/CredLoops.h>
#include "infiniband/ibdm/hostlist/HostList.h"
#include <infiniband/ibutils/release_containers.h>
#include "infiniband/ibutils/output_stream_primitives.h"
#include "infiniband/ibdiag/ibdiag_dfp.h"
#include "infiniband/ibdiag/ibdiag_fat_tree.h"
#include "infiniband/ibdiag/ibdiag_flid.h"
#include "infiniband/ibdm/cable/CombinedCable.h"
#include "infiniband/ibdiag/ibdiagnet/ibdiagnet_export_api.h"
#include "infiniband/ibdiag/ibdiagnet/ibdiagnet_version.h"

#include "ibdiagnet_stages.h"
#include "ibdiagnet_types.h"


#define MIN_FNM_PORTS_PER_NODE 2

/****************************************************/
/* globals */
static bool g_vs_cap_gmp_run = false;

/****************************************************/
int check_if_local_link_is_down(IN IBDiag *p_ibdiag,
        OUT bool& result)
{
    IBDIAGNET_ENTER;
    u_int8_t local_port_state;
    if (p_ibdiag->GetLocalPortState(local_port_state)) {
        ERR_PRINT("%s\n", p_ibdiag->GetLastError());
        IBDIAGNET_RETURN(1);
    }

    switch (local_port_state) {
    case IB_PORT_STATE_DOWN: {
        result = true;
        break;
    }
    default: {
        result = false;
        break;
    }
    }
    IBDIAGNET_RETURN(0);
}


/******************************************************/
/////
//DiscoveryFabric private methods:
/////

// an elaborated parser could be added here if needed
bool DiscoveryFabricStage::GetDefaultUnhealthyAction(const string &value, bool &out)
{
    IBDIAGNET_ENTER;

    out = false;

    if (value.empty())
        IBDIAGNET_RETURN(true);

    if (value == "no_discover" || value == "isolate") {
        out = true;
        IBDIAGNET_RETURN(true);
    }

    IBDIAGNET_RETURN(false);
}

bool DiscoveryFabricStage::SetUnhealthyPortsSwitchAction(const string &value)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(this->GetDefaultUnhealthyAction(value, this->switch_action));
}

bool DiscoveryFabricStage::SetUnhealthyPortsCAAction(const string &value)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(this->GetDefaultUnhealthyAction(value, this->ca_action));
}

int DiscoveryFabricStage::MarkOutUnhealthyPorts(int &unhealthy_ports)
{
    IBDIAGNET_ENTER;

    PRINT("\n");
    IBDiag::PrintFileTimestamp(this->healthy_ports_policy_file, "Healthy Ports Policy");

    map_guid_to_ports exclude_ports;
    string parser_output;
    int rc = this->p_ibdiag->ReadUnhealthyPortsPolicy(parser_output, exclude_ports,
                                                    this->healthy_ports_policy_file,
                                                    this->switch_action, this->ca_action);
    LOG_PRINT("%s\n", parser_output.c_str());
    this->ParseIBDMOutput(parser_output);

    if (rc)
        IBDIAGNET_RETURN(rc);

    parser_output = "";

    PRINT("\n");
    IBDiag::PrintFileTimestamp(this->unhealthy_ports_dump_file, "Unhealthy Ports Dump File");

    rc = this->p_ibdiag->MarkOutUnhealthyPorts(parser_output, unhealthy_ports,
                                            exclude_ports, this->unhealthy_ports_dump_file);
    LOG_PRINT("%s\n", parser_output.c_str());
    this->ParseIBDMOutput(parser_output);

    if (!rc)
        IBDIAGNET_RETURN(rc);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int DiscoveryFabricStage::PostDiscover()
{
    IBDIAGNET_ENTER;
    string parser_output;
    int rc;

    string scope_file = this->scope_port_guid_file;
    u_int32_t errors = this->num_errors;
    u_int32_t warnings = this->num_warnings;

    int scope_lines = 0;
    int exclude_scope_lines = 0;

    //work on guids that are in a scope
    if (!scope_file.empty() &&
        access(scope_file.c_str(), F_OK) != -1) {
        PRINT("\n");
        IBDiag::PrintFileTimestamp(scope_file, "Scope");
        rc = p_ibdiag->ParseScopePortGuidsFile(
                            scope_file.c_str(),
                            parser_output,
                            true,
                            scope_lines);

        LOG_PRINT("%s\n", parser_output.c_str());

        //if critical errors found, just exit
        if (rc) {
            ERR_PRINT("Failed to parse Scope file (%s).\n",
                      scope_file.c_str());
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }
        //parse number of errors, warnings
        this->ParseIBDMOutput(parser_output);
        if (this->num_errors > errors) {
            ERR_PRINT("Parsing the scope file finished with errors.\n");
        } else if (this->num_warnings > warnings) {
            WARN_PRINT("Parsing the scope file finished with warnings.\n");
        }
    }
    string exclude_scope_file = this->exclude_scope_port_guid_file;
    //init for exclude scope
    parser_output.clear();
    errors = this->num_errors;
    warnings = this->num_warnings;

    //work on guids that are excluded from scope
    if (!exclude_scope_file.empty() &&
        access(exclude_scope_file.c_str(), F_OK) != -1) {
        PRINT("\n");
        IBDiag::PrintFileTimestamp(exclude_scope_file, "Exclude Scope");
        rc = p_ibdiag->ParseScopePortGuidsFile(
                            exclude_scope_file.c_str(),
                            parser_output,
                            false, //not to include, but to exclude
                            exclude_scope_lines);
        LOG_PRINT("%s\n", parser_output.c_str());

        //if critical errors found, just exit
        if (rc) {
            ERR_PRINT("Failed to parse Exclude Scope file (%s).\n",
                      exclude_scope_file.c_str());
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }
        //parse number of errors, warnings
        this->ParseIBDMOutput(parser_output);
        if (this->num_errors > errors) {
            ERR_PRINT("Parsing the exclude scope file finished"
                      " with errors.\n");
        } else if (this->num_warnings > warnings) {
            WARN_PRINT("Parsing the exclude scope file finished"
                       " with warnings.\n");
        }
    }

    int unhealthy_ports = 0;
    if (!this->unhealthy_ports_dump_file.empty() &&
            !this->healthy_ports_policy_file.empty() &&
            access(unhealthy_ports_dump_file.c_str(), F_OK) != -1 &&
            access(healthy_ports_policy_file.c_str(), F_OK) != -1) {
        errors = this->num_errors;
        warnings = this->num_warnings;

        rc = this->MarkOutUnhealthyPorts(unhealthy_ports);
        if (rc) {
            ERR_PRINT("Failed to mark unhealthy ports out the fabric\n");
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }

        if (this->num_errors > errors)
            ERR_PRINT("Marking of unhealthy ports finished with errors.\n");
        else if (this->num_warnings > warnings)
            WARN_PRINT("Marking of unhealthy ports finished with warnings.\n");
    }

    size_t fabric_switches_total = 0;
    size_t fabric_switches_fully = 0;
    size_t fabric_switches_partial = 0;
    size_t fabric_switches_ports = 0;

    size_t fabric_cas_total = 0;
    size_t fabric_cas_fully = 0;
    size_t fabric_cas_partial = 0;
    size_t fabric_cas_ports = 0;

    size_t fabric_routers_total = 0;
    size_t fabric_routers_fully = 0;
    size_t fabric_routers_partial = 0;
    size_t fabric_routers_ports = 0;

    IBFabric * p_fabric = this->p_ibdiag->GetDiscoverFabricPtr();

    for (map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end(); ++nI) {
        IBNode *p_node = nI->second;
        if (!p_node)
            continue;

        if (!p_node->getInSubFabric())
            continue;

        bool all_in_scope     = true;
        size_t  port_count    = 0;

        for (phys_port_t i = 1; i <= p_node->numPorts; ++i) {
            IBPort *p_port = p_node->getPort(i);
            if (!p_port)
                continue;

            if (!p_port->getInSubFabric()) {
                all_in_scope = false;
                continue;
            }

            ++port_count;
        }

        if (p_node->type == IB_SW_NODE)
            fabric_switches_total++;
        else if (p_node->type == IB_CA_NODE)
            fabric_cas_total++;
        else if (p_node->type == IB_RTR_NODE)
            fabric_routers_total++;

        if (port_count) {
            if (p_node->type == IB_SW_NODE) {

                fabric_switches_ports += port_count;
                if (all_in_scope)
                    ++fabric_switches_fully;
                else
                    ++fabric_switches_partial;

            } else if (p_node->type == IB_CA_NODE) {
                fabric_cas_ports += port_count;
                if (all_in_scope)
                    ++fabric_cas_fully;
                else
                    ++fabric_cas_partial;

            } else if (p_node->type == IB_RTR_NODE) {
                fabric_routers_ports += port_count;
                if (all_in_scope)
                    ++fabric_routers_fully;
                else
                    ++fabric_routers_partial;
            }
        }
    }

    char buff[1024];
    if (!scope_file.empty() || !exclude_scope_file.empty() ||
            (!unhealthy_ports_dump_file.empty() && !healthy_ports_policy_file.empty()))
        snprintf(buff, sizeof(buff), "In scope [total/fully/partial/ports] switches [%lu/%lu/%lu/%lu], "
                    "CAs [%lu/%lu/%lu/%lu], Routers [%lu/%lu/%lu/%lu];"
                    " scope lines %d, exclude scope lines %d, unhealthy ports: %d\n",
                    fabric_switches_total, fabric_switches_fully, fabric_switches_partial, fabric_switches_ports,
                    fabric_cas_total, fabric_cas_fully, fabric_cas_partial, fabric_cas_ports,
                    fabric_routers_total, fabric_routers_fully, fabric_routers_partial, fabric_routers_ports,
                    scope_lines, exclude_scope_lines, unhealthy_ports);
    else
        snprintf(buff, sizeof(buff), "No scope/unhealthy ports files. Total switches/ports [%lu/%lu], "
                    "CAs/ports [%lu/%lu], Routers/ports [%lu/%lu] \n",
                    fabric_switches_total, fabric_switches_ports,
                    fabric_cas_total, fabric_cas_ports,
                    fabric_routers_total, fabric_routers_ports);

    PRINT("\n-I- %s", buff);

    if (SysLogger::instance().get_enabled())
        SysLogger::instance().syslog(buff);

    IBDIAGNET_RETURN(0);
}

/////
//DiscoveryFabric public methods:
/////
DiscoveryFabricStage::DiscoveryFabricStage(IBDiag *p_ibdiag) :
        DiscoveryStage(p_ibdiag), switch_action(false), ca_action(false)

{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int DiscoveryFabricStage::DiscoverFabric()
{

    IBDIAGNET_ENTER;

    int rc = 0;

    if (load_from_file) {
        rc = p_ibdiag->DiscoverFabricFromFile(load_from_file_path);
    } else {
        rc = p_ibdiag->DiscoverFabric(this->max_hops);
    }
    IBDIAGNET_RETURN(rc);
}

int DiscoveryFabricStage::Prepare()
{
    IBDIAGNET_ENTER;
    DiscoveryStage::Prepare();
    if (!this->to_check_dup_guids) {        //skipped duplicated guids detection
        WARN_PRINT("Advanced %s skipped (Only Basic detection will be performed)\n",
                CHECK_NAME_DUPLICATED_GUIDS);
        ++this->num_warnings;
        this->p_ibdiag->GetRefToCheckDupGuids() = false;
    }
    IBDIAGNET_RETURN(0);
}

/******************************************************/
/////
//Discovery private methods:
/////
void DiscoveryStage::PrintDuplicatedGuidsDetectErrors()
{
    IBDIAGNET_ENTER;
    if (!this->p_ibdiag->GetRefToDupGuidsDetectionErrors().size())
        IBDIAGNET_RETURN_VOID;
    LOG_PRINT("\n");
    LOG_WARN_PRINT("Duplicated GUIDs detection finished with some warnings:\n");
    for (list_string::iterator eI = this->p_ibdiag->GetRefToDupGuidsDetectionErrors().begin();
            eI != this->p_ibdiag->GetRefToDupGuidsDetectionErrors().end(); ++eI) {
        //will be printed to both screen + log
        LOG_WARN_PRINT("%s", (*eI).c_str());
        ++this->num_warnings;
    }
    LOG_PRINT("\n");
    IBDIAGNET_RETURN_VOID;
}

void DiscoveryStage::CheckAPortSymmetry(APort* p_aport)
{
    // to_string doesn't have an overload for uint8_t, so we create a lambda
    // that converts uint8_t to string with the correct signature for CheckAttributeSymmetry's template.
    string (*uint8_to_str)(const uint8_t) = [](uint8_t i) {
        return std::to_string(i);
    };

    // similar lambda that returns a hex (ptr) string from uint64_t
    // for getAttributeArrayStr.
    string (*uint64_to_ptrstr)(const uint64_t) = [](uint64_t i) {
        stringstream ss;
        ss << PTR(i);
        return ss.str();
    };

    if (!p_aport)
        return;

    int missing = p_aport->findMissingIBPort();
    vec_p_fabric_err& aport_errors = this->p_ibdiag->GetErrorsByAPort(p_aport);
    if (missing != -1) {
        aport_errors.push_back(new APortMissingPlanes(p_aport));
        p_aport->asymmetry_mask.set(APORT_MISSING_PLANES);
        return;
    }

    if (p_aport->getAggregatedLabel().empty()) {
        aport_errors.push_back(new APortNoAggregatedLabel(p_aport));
        p_aport->asymmetry_mask.set(APORT_NO_AGG_LABEL);
        return;
    }

    if (!p_aport->is_data_worthy(DATA_WORTHY_SMP_MAD))
        return;

    if (CheckAttributeSymmetry(p_aport, "Port State",
                &IBPort::get_internal_state,
                IB_UNKNOWN_PORT_STATE,
                port_state_to_str))
        return;

    if (CheckAttributeSymmetry(p_aport, "Width",
                &IBPort::get_internal_width,
                IB_UNKNOWN_LINK_WIDTH,
                width_to_str))
        return;

    if (CheckAttributeSymmetry(p_aport, "Speed",
                &IBPort::get_internal_speed,
                IB_UNKNOWN_LINK_SPEED,
                speed_to_str))
        return;

    ibport_accessor_t<uint8_t> mtu_cap_accessor = [this] (const IBPort* p_port) {
        return this->p_ibdiag->GetIBDMExtendedInfoPtr()->getPortMTUCap(p_port);
    };

    if (CheckAttributeSymmetry(p_aport, "MTU Cap",
                mtu_cap_accessor,
                uint8_t(0),
                uint8_to_str))
        return;

    ibport_accessor_t<uint8_t> vl_cap_accessor = [this] (const IBPort* p_port) {
        return this->p_ibdiag->GetIBDMExtendedInfoPtr()->getPortVLCap(p_port);
    };
    if (CheckAttributeSymmetry(p_aport, "VL Cap",
                vl_cap_accessor,
                uint8_t(0),
                uint8_to_str))
        return;

    // If all planes are on the same node, all port GUIDs must be the same.
    // otherwise, all GUIDs must be unique.
    bool should_guid_equal = (p_aport->get_common_node() != nullptr);
    if ((should_guid_equal && !p_aport->isEqualAttribute(&IBPort::guid_get)) ||
            (!should_guid_equal && !p_aport->isUniqueAttribute(&IBPort::guid_get)) ) {
        string port_guids = p_aport->getAttributeArrayStr(&IBPort::guid_get, uint64_to_ptrstr);
        aport_errors.push_back(new APortInvalidPortGuids(p_aport, port_guids));
        p_aport->asymmetry_mask.set(APORT_INVALID_PORT_GUIDS);
        return;
    }
}

void DiscoveryStage::CheckRemotePlanes(APort* p_aport) {
    APort* remote_aport = nullptr;
    bool same_aport = true;

    if (!p_aport || p_aport->ports.empty() || p_aport->visited)
        return;

    vec_p_fabric_err& aport_errors = this->p_ibdiag->GetErrorsByAPort(p_aport);

    // check that all planes are connected to the same APort,
    // add all remote aports to visited set.
    for (size_t i = 1; i < p_aport->ports.size(); ++i) {
        IBPort* ibport = p_aport->ports[i];
        // remote doesn't exist or isn't in an aport
        if (!ibport || !ibport->p_remotePort ||
                !ibport->p_remotePort->p_aport) {
            same_aport = false;
            continue;
        }

        ibport->p_remotePort->p_aport->visited = true;

        // current remote aport not equal to former remote aports 
        if (remote_aport && 
                ibport->p_remotePort->p_aport != remote_aport) {
            same_aport = false;
            continue;
        }

        remote_aport = ibport->p_remotePort->p_aport;
    }

    if (!same_aport) {
        // APort is connected to multiple remote APorts or has null remote ports/aports
        aport_errors.push_back(new APortInvalidConnection(p_aport));
        p_aport->asymmetry_mask.set(APORT_INVALID_CONNECTION);
        return;
    }

    if (remote_aport->ports.size() != p_aport->ports.size()) {
        aport_errors.push_back(new APortInvalidNumOfPlanes(
                p_aport, 
                static_cast<int>(p_aport->ports.size() - 1),
                static_cast<int>(remote_aport->ports.size() - 1)));
        p_aport->asymmetry_mask.set(APORT_INVALID_NUM_PLANES);
        return;
    }

    for (size_t i = 1; i < p_aport->ports.size(); ++i) {
        IBPort* ibport = p_aport->ports[i];
        if (!ibport->p_port_hierarchy_info ||
            !ibport->p_remotePort->p_port_hierarchy_info)
            continue;
        if (ibport->p_port_hierarchy_info->m_plane 
                != ibport->p_remotePort->p_port_hierarchy_info->m_plane) {
            aport_errors.push_back(new APortInvalidRemotePlane(
                    p_aport, 
                    ibport->p_port_hierarchy_info->m_plane,
                    ibport->p_remotePort->p_port_hierarchy_info->m_plane
            ));
            p_aport->asymmetry_mask.set(APORT_INVALID_REMOTE_PLANE);
            continue;
        }
    }
}

void DiscoveryStage::FNMPortsChecks(list_p_fabric_general_err& errs)
{
    IBDIAGNET_ENTER;
    set< IBSystem *> fnm_systems;

    // find all systems with APorts - these systems should have FNM ports
    for (const auto& sys_guid_aports_pair : this->p_ibdiag->GetDiscoverFabricPtr()->APortsBySysGuid) {
        // go over all the plane ports of all the APorts, add each system we encounter to the set.
        for (const APort *p_curr_aport : sys_guid_aports_pair.second) {
            if (!p_curr_aport)
                continue;

            for (size_t i = 1; i < p_curr_aport->ports.size(); ++i) {
                if (p_curr_aport->ports[i] && p_curr_aport->ports[i]->p_node
                        && p_curr_aport->ports[i]->p_node->p_system)
                    fnm_systems.insert(p_curr_aport->ports[i]->p_node->p_system);
            }
        }
    }

    for (IBSystem *p_system : fnm_systems) {
        if (!CheckValidFNMConnections(errs, p_system))
            continue;

        if (!CheckFNMSpeeds(errs, p_system))
            continue;
    }

    IBDIAGNET_RETURN_VOID;
}

bool DiscoveryStage::CheckValidFNMConnections(list_p_fabric_general_err& errs,
                                              const IBSystem* p_system)
{
    set<IBNode *> visited;
    IBNode *first_node = nullptr;
    IBNode *curr_node = nullptr;
    // the port from which we found curr_node
    IBPort *p_entry_port = nullptr;
    IBPort *exit_port = nullptr;
    // save FNM exit ports path for printing
    vector<IBPort *> path;
    bool ret = true;

    // find a switch node with FNM ports for start
    for (auto& name_node : p_system->NodeByName) {
        IBNode *p_node = name_node.second;
        if (!p_node || p_node->type != IB_SW_NODE)
            continue;

        if(p_node->hasFNMPort()) {
            first_node = p_node;
            break;
        }
    }
    // no FNM ports on switches
    if (!first_node)
        return true;
    curr_node = first_node;

    // stop when we get to a non FNM node, or when we revisit a node.
    while (curr_node && !visited.count(curr_node)) {
        visited.insert(curr_node);

        vec_pport node_fnms;
        curr_node->getAllFNMPorts(node_fnms);

        // find a valid FNM to exit from
        exit_port = nullptr;
        for (IBPort *p_fnm : node_fnms) {
            if (!p_fnm || !p_fnm->is_data_worthy(DATA_WORTHY_SMP_MAD) || !p_fnm->p_remotePort)
                continue;

            //In Crocodile FNM Port is connected as 2x to one port (No Ring)
            // exit_port then must be also entry_port.
            // So in case of crocodile switch don't skip port if its the same as entry port
            if (p_fnm == p_entry_port && !curr_node->isPrismaSwitch())
                continue;

            exit_port = p_fnm;
            break;
        }

        if (!exit_port) {
            errs.push_back(new NoValidExitFNM(curr_node, node_fnms, path));
            return false;
        }

        if (!exit_port->p_remotePort->isFNMPort()) {
            errs.push_back(new NonFNMConnection(curr_node, exit_port, path));
            return false;
        }

        path.push_back(exit_port);
        // advance to the next node
        p_entry_port = exit_port->p_remotePort;
        curr_node = p_entry_port->p_node;
    }

    // reached a null node
    if (!curr_node)
        return false;

    // FNM ports lead to the same node twice
    // without passing through the first node
    if (curr_node != first_node) {
        errs.push_back(new FNMLoopInsideRing(p_system, path));
        return false;
    }

    for (auto& name_node : p_system->NodeByName) {
        IBNode *p_node = name_node.second;
        if (!p_node || p_node->type != IB_SW_NODE)
            continue;
        if (!visited.count(p_node)) {
            ret = false;
            errs.push_back(new FNMSwitchNotVisited(p_system, p_node, path));
            continue;
        }
    }

    return ret;
}

bool DiscoveryStage::CheckFNMSpeeds(list_p_fabric_general_err& errs,
                                    const IBSystem* p_system)
{
    bool uniform_speed = true;
    IBLinkSpeed first_speed = IB_UNKNOWN_LINK_SPEED;
    for (auto& name_node : p_system->NodeByName) {
        IBNode *p_node = name_node.second;
        if (!p_node || p_node->type != IB_SW_NODE)
            continue;
        vec_pport fnm_ports;
        p_node->getAllFNMPorts(fnm_ports);
        for (IBPort *p_fnm : fnm_ports) {
            if (p_fnm->get_common_speed() == IB_UNKNOWN_LINK_SPEED)
                continue;
            if (first_speed == IB_UNKNOWN_LINK_SPEED)
                first_speed = p_fnm->get_common_speed();

            if (p_fnm->get_common_speed() != first_speed){
                uniform_speed = false;
                break;
            }
        }
    }

    // create speeds array and report error
    if (!uniform_speed) {
        stringstream ss;
        ss << "[";
        bool is_first = true;
        for (auto& name_node : p_system->NodeByName) {
            IBNode *p_node = name_node.second;
            if (!p_node || p_node->type != IB_SW_NODE)
                continue;
            vec_pport fnm_ports;
            p_node->getAllFNMPorts(fnm_ports);
            for (IBPort *p_fnm : fnm_ports) {
                    if (is_first)
                        is_first = false;
                    else
                        ss << ", ";
                ss << p_fnm->getLabel().c_str() << ":"
                   << speed2char(p_fnm->get_internal_speed());
            }
        }
        ss << "]" << endl;

        errs.push_back(new FabricErrInvalidFNMSpeeds(p_system, ss.str()));
    }

    return uniform_speed;
}

int DiscoveryStage::CageManagerChecks(list_p_fabric_general_err& cage_manager_errs)
{
    IBDIAGNET_ENTER;
    int rc = 0;

    std::map<std::tuple<uint64_t, int, int>, bool> aports_map;
    for (const auto& entry : this->p_ibdiag->GetDiscoverFabricPtr()->APortsBySysGuid) {
        if (entry.second.empty())
            continue;

        for (size_t i = 0; i < entry.second.size(); i++) {
            APort* p_aport = entry.second[i];
            if (!p_aport)
                continue;

            bool is_cage_manager_exists = false;
            bool is_aport_relevant = false;
            for (IBPort* p_port : p_aport->ports) {
                if (!p_port || !p_port->p_port_hierarchy_info || p_port->p_port_hierarchy_info->m_template_guid != 4)
                    continue;
                is_aport_relevant = true;
                bool is_cage_manager = p_port->p_port_hierarchy_info->m_is_cage_manager == 1;
                int cage = p_port->p_port_hierarchy_info->m_cage;
                int plane = p_port->p_port_hierarchy_info->m_cage;
                const auto& it = aports_map.find(std::make_tuple(p_port->guid_get(), cage, plane));

                if (it == aports_map.end()) {
                    aports_map.insert(std::make_pair(std::make_tuple(p_port->guid_get(), cage, plane), is_cage_manager));
                }
                else if (it->second != is_cage_manager) { 
                    cage_manager_errs.push_back(new APortInvalidCageManagerSymmetryInCage(p_aport, cage, plane));
                    
                }

                is_cage_manager_exists |= is_cage_manager;
            }

            if (!is_cage_manager_exists && is_aport_relevant) {
                cage_manager_errs.push_back(new APortInvalidCageManager(p_aport));
            }
        }
    }

    IBDIAGNET_RETURN(rc);
}

int DiscoveryStage::APortChecks()
{
    IBDIAGNET_ENTER;
    // a set of APorts that passed a remote check
    int rc = 0;

    // run symmetry check on all aports
    for (const auto& entry : this->p_ibdiag->GetDiscoverFabricPtr()->APortsBySysGuid) {

        if (entry.second.empty())
            continue;

        for (size_t i = 0; i < entry.second.size(); i++) {
            CheckAPortSymmetry(entry.second[i]);
        }
    }

    // run remote checks on all aports
    this->p_ibdiag->GetDiscoverFabricPtr()->unvisitAllAPorts();
    for (const auto& entry : this->p_ibdiag->GetDiscoverFabricPtr()->APortsBySysGuid) {
        if (entry.second.empty())
            continue;

        for (size_t i = 0; i < entry.second.size(); i++) {
            APort *p_aport = entry.second[i];
            if (!p_aport || !p_aport->allPlanesConnected() || !p_aport->isSymmetric())
                continue;
            CheckRemotePlanes(p_aport);
        }
    }

    IBDIAGNET_RETURN(rc);
}

int DiscoveryStage::DumpAPortsReport(bool invalid_aports_only)
{
    IBDIAGNET_ENTER;

    int rc = 0;
    ofstream sout;

    if (!this->p_ibdiag->GetDiscoverFabricPtr()->IsHaveAPorts())
        IBDIAGNET_RETURN(rc);

    if (this->p_ibdiag->OpenFile("APorts", string(APORT_FILE), sout, false, "#"))
        WARN_PRINT("The file: %s with aport validation details"
                   " will not be provided.\n", APORT_FILE);

    for (const auto& entry : this->p_ibdiag->GetDiscoverFabricPtr()->APortsBySysGuid) {
        if (entry.second.empty())
            continue;

        // check if we have aports with errors in this SysGuid
        auto it = std::find_if(entry.second.begin(), entry.second.end(), 
            [this](APort *p){
                    if (!p)
                        return false;
                    return !this->p_ibdiag->GetErrorsByAPort(p).empty();
                }
        );

        if (invalid_aports_only && it == entry.second.end())
            continue;

        sout << "System GUID: " << PTR(entry.first) << endl;
        for (size_t i = 0; i < entry.second.size(); i++) {
            APort* p_aport = entry.second[i];
            if (!p_aport ||
                (invalid_aports_only &&
                    this->p_ibdiag->GetErrorsByAPort(p_aport).empty()))
                continue;
            DumpAPort(sout, p_aport);
            sout << endl;
        }
    }

    //stream does not crash on write attempt if it has no associated file
    this->p_ibdiag->CloseFile(sout);

    IBDIAGNET_RETURN(rc);
}

void DiscoveryStage::DumpAPort(ofstream& sout, APort* p_aport)
{
    vec_p_fabric_err& aport_errors = this->p_ibdiag->GetErrorsByAPort(p_aport);
    sout << "\tAPort " << DEC(p_aport->aport_index) << " Aggregated label: ";
    if (p_aport->getAggregatedLabel().empty())
        sout << "N/A";
    else 
        sout << p_aport->getAggregatedLabel();
    sout << endl;

    for (size_t i = 1; i < p_aport->ports.size(); ++i) {
        IBPort* curr_ibport =  p_aport->ports[i];
        if (!curr_ibport) {
            sout << "\t\t\tN/A" << endl;
            continue;
        }
        sout << "\t\t\t" << curr_ibport->p_port_hierarchy_info->extendedLabel() << ": "
            << port_state_to_str(curr_ibport->get_internal_state())
            << " PortGuid: " << PTR(curr_ibport->guid_get())
            << " NodeGuid: " << PTR(curr_ibport->p_node->guid_get())
            << " Port: " << DEC(curr_ibport->num);

        sout << ", ";
        if (curr_ibport->get_internal_width() == IB_UNKNOWN_LINK_WIDTH
            && curr_ibport->get_internal_speed() == IB_UNKNOWN_LINK_SPEED)
                sout << "N/A";
        else
            sout << ((curr_ibport->get_internal_width() == IB_UNKNOWN_LINK_WIDTH) ?\
                        "N/Ax" : width_to_str(curr_ibport->get_internal_width()))
                << ((curr_ibport->get_internal_speed() == IB_UNKNOWN_LINK_SPEED) ?\
                        "N/A" : speed_to_str(curr_ibport->get_internal_speed()));

        sout << ", MTU "
             << DEC(p_ibdiag->GetIBDMExtendedInfoPtr()->getPortMTUCap(curr_ibport))
             << ", VLCAP " 
             << DEC(p_ibdiag->GetIBDMExtendedInfoPtr()->getPortVLCap(curr_ibport))
             << endl;
    }

    if (aport_errors.empty())
        return;

    sout << "\t\tErrors:" << endl;
    for (auto err_it : aport_errors) {
        sout << "\t\t\t" << err_it->getDescription() << endl;    
    }

}

void DiscoveryStage::CopyAPortErrorsToList(list_p_fabric_general_err& errs)
{
    // copy errors from all the aports to the same list
    for (const auto& entry : this->p_ibdiag->GetDiscoverFabricPtr()->APortsBySysGuid) {
        if (entry.second.empty())
            continue;

        for (size_t i = 0; i < entry.second.size(); i++) {
            APort *p_aport = entry.second[i];
            vec_p_fabric_err& aport_errors = this->p_ibdiag->GetErrorsByAPort(p_aport);
            if (!p_aport)
                continue;
            for (auto err : aport_errors) {
                // copy the error only if it's an aport or ibport error
                if (FabricErrAPort* aport_err_p = dynamic_cast<FabricErrAPort *>(err)) {
                    errs.push_back(new FabricErrAPort(*aport_err_p));
                } else if (FabricErrPort* port_err_p = dynamic_cast<FabricErrPort *>(err)) {
                    errs.push_back(new FabricErrPort(*port_err_p));
                }
            }
        }
    }
}

/////
//Discovery public methods:
/////
DiscoveryStage::DiscoveryStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_DISCOVERY, p_ibdiag),
        device_name(""), port_num(0), port_guid(0),
        max_hops(IBDIAG_MAX_HOPS), to_check_dup_guids(true), to_check_dup_node_desc(true),
        initiliazed_with_device(false), initiliazed_with_guid(false),
        output_capability_file(""), output_full_capability_file(""),
        input_capability_file(""), show_ports_data_db2(true), dump_lst_with_lmc(false),
        node_name_map_file(""), p_mkey_manager(NULL), load_from_file(false), num_of_switches(0),
        num_of_channel_adapters(0), num_of_channel_adapters_sdm(0), num_of_routers(0),
        num_of_aggregation_nodes(0), num_of_GPUs(0), check_switch_duplicated_guids(false)
{
    IBDIAGNET_ENTER;
    //check if we have environment variable with path
    //to capability mask configuration file
    char *capability_file = getenv(CAPABILITY_FILE_ENV);

    if (capability_file)
        input_capability_file = capability_file;

    char *map_file = getenv(MAP_FILE_ENV);

    //check if we have environment variable with path
    //to node name map file
    if (map_file)
        node_name_map_file = map_file;

    IBDIAGNET_RETURN_VOID;
}

int DiscoveryStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    //init - set port
    if (this->initiliazed_with_device && initiliazed_with_guid) {
        WARN_PRINT("Local GUID were given together with Device name / Port number, initializing only by GUID\n");
        ++this->num_warnings;
    }

    this->p_ibdiag->SetCheckSwitchDuplicatedGuids(this->check_switch_duplicated_guids);

    if (!load_from_file) { // in load_from_file we set the device and port later
        if (initiliazed_with_guid) {    //init by guid
            if (this->p_ibdiag->SetPort(this->port_guid)) {
                ERR_PRINT("Failed to initialize - %s\n", this->p_ibdiag->GetLastError());
                ++this->num_errors;
                IBDIAGNET_RETURN(1);
            }
        } else {    //init by device - this is also default
            if (this->p_ibdiag->SetPort(this->device_name.c_str(), this->port_num)) {
                ERR_PRINT("Failed to initialize - %s\n", this->p_ibdiag->GetLastError());
                ++this->num_errors;
                IBDIAGNET_RETURN(1);
            }
        }

        INFO_PRINT("Using local IB device %s:%d\n",
                   this->p_ibdiag->GetIbisPtr()->GetDevName().c_str(),
                   this->p_ibdiag->GetIbisPtr()->GetPortNum());
        SysLogger::instance().syslog("Using local IB device %s:%d\n",
                                     this->p_ibdiag->GetIbisPtr()->GetDevName().c_str(),
                                     this->p_ibdiag->GetIbisPtr()->GetPortNum());
    }

    if (!this->to_check_dup_node_desc) {    //skipped duplicated node description detection
        WARN_PRINT("%s skipped\n", CHECK_NAME_DUPLICATED_NODE_DESC);
        ++this->num_warnings;
    }

    //try to read Node Name Map file
    if (!this->node_name_map_file.empty() &&
        access(this->node_name_map_file.c_str(), F_OK) != -1) {
        string parser_output = string("");
        IBDiag::PrintFileTimestamp(node_name_map_file, "Node Map");

        int rc = p_ibdiag->ParseNodeNameMapFile(node_name_map_file.c_str(),
                                                parser_output);
        LOG_PRINT("%s\n", parser_output.c_str());

        if (rc) {
            ERR_PRINT("Failed to parse Node Name Map file (%s).\n",
                      node_name_map_file.c_str());
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }

        u_int32_t errors = this->num_errors;
        u_int32_t warnings = this->num_warnings;
        //parse number of errors, warnings
        this->ParseIBDMOutput(parser_output);
        if (this->num_errors > errors) {
            ERR_PRINT("Parsing the Node Name Map file finished " \
                      "with errors.\n");
        } else if (this->num_warnings > warnings) {
            WARN_PRINT("Parsing the Node Name Map file " \
                       "finished with warnings.\n");
        }
    }


    IBDIAGNET_RETURN(0);
}

int DiscoveryStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;

    INFO_PRINT("Start Fabric Discover\n");
    int rc2 = DiscoverFabric();
    SCREEN_PRINT("\n");

    if (!this->load_from_file && !p_ibdiag->IsNodeDescriptionSent()){
        list_p_fabric_general_err& pi_errors = p_ibdiag->GetErrors();
        INFO_PRINT("Fill NodeDesc data\n");
        int local_rc = p_ibdiag->FillInNodeDescription(pi_errors);
        SCREEN_PRINT("\n");
        if (local_rc){
            ERR_PRINT("Failed to fill NodeDesc data: %s\n", p_ibdiag->GetLastError());
            IBDIAGNET_RETURN(1);
        }
        else {
            INFO_PRINT("NodeDesc finished successfully \n");
        }
    }

    int rc = this->AnalyzeCheckResults(p_ibdiag->GetErrors(),
                CHECK_FABRIC_DISCOVER,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                false);

    if (rc) {
        /*
        PRINT("\n");
        ERR_PRINT("Failed to discover fabric, err=%s\n",
                this->p_ibdiag->GetLastError());
        ++this->num_errors;
        */
        if (this->to_check_dup_guids)   //maybe we failed because duplicated GUIDs detection had errors
            this->PrintDuplicatedGuidsDetectErrors();

        IBDIAGNET_RETURN(1);
    }

    if (!this->load_from_file){
        SCREEN_PRINT("\n");

        list_p_fabric_general_err pi_errors;
        INFO_PRINT("Fill PortInfo data\n");
        rc = p_ibdiag->BuildPortInfo(pi_errors);
        SCREEN_PRINT("\n");
        if (rc){
            this->AnalyzeCheckResults(pi_errors,
                    CHECK_PORT_INFO,
                    rc,
                    IBDIAG_ERR_CODE_FABRIC_ERROR,
                    this->num_errors,
                    this->num_warnings,
                    false);

            ERR_PRINT("Failed to fill PortInfo data\n");
        }
        else {
            INFO_PRINT("PortInfo finished successfully\n");
        }
    }

    if (this->p_mkey_manager) {
        if (p_ibdiag->GetIbisPtr() && p_ibdiag->GetIbisPtr()->GetMKeyManager())
        {
            if (!p_mkey_manager->getConstMKey()) {
                CreateLid2GuidDB();
                p_mkey_manager->setLidToNodeGuidMap(lid2guid_map);
            }
        }
        else
            ERR_PRINT("can't get MKeyManager Ptr\n");
    }

    this->p_ibdiag->GetKeyUpdaterPtr()->UpdateKeysPerPort({IBIS_VS_KEY, IBIS_CC_KEY,
                                                           IBIS_CLASS_C_KEY, IBIS_PM_KEY});
    //this is essential part of discovery, we can't continue without it
    //(example, scope failed)
    if (PostDiscover())
        IBDIAGNET_RETURN(1);

    string parser_output;

    if(!psl_dump_file.empty()) {
        parser_output = "";
        INFO_PRINT("Load PSL file (%s).\n", psl_dump_file.c_str());

        rc = p_ibdiag->ParsePSLFile(psl_dump_file, parser_output);
        LOG_PRINT("%s\n", parser_output.c_str());

        if (rc) {
            ERR_PRINT("Failed to parse PSL file (%s).\n", psl_dump_file.c_str());
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }
    }

    //try to read capability mask configuration file
    if (!this->input_capability_file.empty() &&
        access(this->input_capability_file.c_str(), F_OK) != -1) {
        parser_output = "";

        IBDiag::PrintFileTimestamp(input_capability_file, "Cpability Mask Configuration");

        rc = p_ibdiag->ParseCapabilityMaskFile(
            this->input_capability_file.c_str(), parser_output);
        LOG_PRINT("%s\n", parser_output.c_str());

        if (rc) {
            ERR_PRINT("Failed to parse Capability mask configuration file (%s).\n",
                      this->input_capability_file.c_str());
            ++this->num_errors;
        }
    }

    //check if we need to write the capability masks file
    if (!this->output_capability_file.empty()) {

    	parser_output = "";

    	OutputControl::Identity identity(this->output_capability_file,
    	                                 OutputControl::OutputControl_Flag_UserFile);

        if (p_ibdiag->DumpCapabilityMaskFile(identity, parser_output)) {
            WARN_PRINT("Writing Capability Masks file to %s failed\n", this->output_capability_file.c_str());
            ++this->num_errors;
        }
        LOG_PRINT("%s\n", parser_output.c_str());
    }

    parser_output = "";
    //initialize capability mask caches
    rc2 = p_ibdiag->BuildCapabilityCache(parser_output);
    if (rc2) {
        LOG_PRINT("%s\n", parser_output.c_str());
        ERR_PRINT("Failed to init capability mask inner cache.\n");
        ++this->num_errors;
    }
    SCREEN_PRINT("\n");

    //SMP capability
    if (!load_from_file) {
        list_p_fabric_general_err vs_cap_smp_errors;
        INFO_PRINT("Build VS Capability SMP\n");
        rc2 = p_ibdiag->BuildVsCapSmp(vs_cap_smp_errors);
        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(vs_cap_smp_errors,
                CHECK_VS_CAP_SMP,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
    }

    // Build VS Extended Node Info
    if (!load_from_file) {

        list_p_fabric_general_err ext_nodes_errors;
        INFO_PRINT("Build VS Extended Node Info\n");
        rc2  = this->p_ibdiag->BuildExtendedNodeInfo(ext_nodes_errors);

        if (rc2 == IBDIAG_SUCCESS_CODE)
            this->p_ibdiag->retrieved_ext_node_info_ok = true;

        SCREEN_PRINT("\n");
        this->p_ibdiag->ext_node_info_check_results =
            this->AnalyzeCheckResults(ext_nodes_errors,
                                      CHECK_VS_EXT_NODE_INFO_RETRIEVE,
                                      rc2,
                                      IBDIAG_ERR_CODE_FABRIC_ERROR,
                                      this->num_errors,
                                      this->num_warnings,
                                      true);
    } else {

        // check if ExtendedNodeInfo exist in file for all switches
        if (!this->p_ibdiag->ValidateExtendedNodeInfoForSwitches()) {
            this->p_ibdiag->retrieved_ext_node_info_ok  = true;
            this->p_ibdiag->ext_node_info_check_results = true;
        }
    }

    // Build VS Extended Port Info
    if (!load_from_file) {
        //now we have capability -> can run ExtendedPortInfo MADs
        //to determine extended mellanox speeds speed and LLR
        //note: even if we haven't ran the SMP and GMP sub stages,
        //we might have a capability from initial configuration
        list_p_fabric_general_err vs_ext_pi_errors;
        INFO_PRINT("Build VS Extended Port Info\n");
        rc2 = p_ibdiag->BuildExtendedPortInfo(vs_ext_pi_errors);
        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(vs_ext_pi_errors,
                CHECK_VS_EXTENDED_PORT_INFO,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
    }

    //GMP capability
    if (load_from_file) {
        g_vs_cap_gmp_run = true;
    } else {
        bool can_send_mads_by_lid = false;
        if (check_if_can_send_mads_by_lid(this->p_ibdiag, can_send_mads_by_lid) ||
            !can_send_mads_by_lid) {
            INFO_PRINT("%s skipped\n", CHECK_VS_CAP_GMP);
            PRINT("\n");
        } else {
            list_p_fabric_general_err vs_cap_gmp_errors;
            INFO_PRINT("Build VS Capability GMP\n");
            int rc2 = p_ibdiag->BuildVsCapGmp(vs_cap_gmp_errors);
            SCREEN_PRINT("\n");
            this->AnalyzeCheckResults(vs_cap_gmp_errors,
                    CHECK_VS_CAP_GMP,
                    rc2,
                    IBDIAG_ERR_CODE_FABRIC_ERROR,
                    this->num_errors,
                    this->num_warnings,
                    true);

            g_vs_cap_gmp_run = true;
        }
    }

    if (!load_from_file) {
        //collect port info extended info for FEC data
        list_p_fabric_general_err pi_ext_errors;
        INFO_PRINT("Build VS Port Info Extended\n");
        rc2 = p_ibdiag->BuildPortInfoExtended(pi_ext_errors);
        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(pi_ext_errors,
                CHECK_PORT_INFO_EXTENDED,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
    }

    if (!load_from_file) {
        //bring switch info
        list_p_fabric_general_err switch_errors;
        INFO_PRINT("Build Switch Info\n");
        rc2 = this->p_ibdiag->BuildSwitchInfo(switch_errors);
        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(switch_errors,
                CHECK_NAME_SW_INFO_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings);

        //bring extended switch info
        INFO_PRINT("Build Extended Switch Info\n");
        rc2 = this->p_ibdiag->BuildExtendedSwitchInfo(switch_errors);
        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(switch_errors,
                CHECK_NAME_EXT_SW_INFO_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
    }

    if (!load_from_file) {
        //bring chassis info
        list_p_fabric_general_err chassis_errors;
        INFO_PRINT("Build Chassis Info\n");
        rc2 = this->p_ibdiag->BuildChassisInfo(chassis_errors);

        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(chassis_errors,
                                  CHECK_NAME_CHASSIS_INFO_RETRIEVE,
                                  rc2,
                                  IBDIAG_ERR_CODE_FABRIC_ERROR,
                                  this->num_errors,
                                  this->num_warnings);
    }

    if (!load_from_file) {
        //bring hierarchy info
        list_p_fabric_general_err hierarchy_errors;
        INFO_PRINT("Build Hierarchy Info\n");
        rc2 = this->p_ibdiag->BuildHierarchyInfo(hierarchy_errors);
        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(hierarchy_errors,
                CHECK_NAME_HIERARCHY_INFO_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings);
    }

    if (!load_from_file) {
        //bring AR info
        list_p_fabric_general_err ar_errors;
        INFO_PRINT("Build AR Info\n");
        rc2 = this->p_ibdiag->BuildARInfo(ar_errors);

        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(ar_errors,
                CHECK_NAME_AR_INFO_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings);
    }

    if (!load_from_file) {
        list_p_fabric_general_err aports_errors;
        INFO_PRINT("Build APorts Info\n");
        rc2 = p_ibdiag->CollectAPortsData(aports_errors);

        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(aports_errors,
                CHECK_APORTS_COLLECT_DATA,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings);
    }

    //here we have correct speeds after VS ExtendePortInfo MADs were sent
    //dump db file
    if (this->p_ibdiag->DumpInternalToCSV(*this->p_csv_out,
                                               this->show_ports_data_db2)) {
        ERR_PRINT("Writing discovery info to DB failed, err=%s\n",
                this->p_ibdiag->GetLastError());
        ++this->num_errors;
        IBDIAGNET_RETURN(1);    /* exiting - there are tools
                                    that need to parse this file */
    }

    // port hierarchy info file
    if (this->p_ibdiag->WritePortHierarchyInfoFile(PORT_HI_INFO_FILE)) {
        ERR_PRINT("Write port hierarchy info file failed\n");
        ++this->num_errors;
    }


    //dump lst file
    if (this->p_ibdiag->WriteLSTFile(LST_FILE, this->dump_lst_with_lmc)) {
        WARN_PRINT("Writing lst file failed\n");
        ++this->num_errors;
    }

    //network dump file
    if (this->p_ibdiag->WriteNetDumpFile(NET_DUMP_FILE)) {
        WARN_PRINT("Writing network dump file failed\n");
        ++this->num_errors;
    }

    //network aggregated dump file
    if (this->p_ibdiag->WriteNetDumpAggregatedFile(NET_DUMP_AGG_FILE)) {
        WARN_PRINT("Writing network aggregated dump file failed\n");
        ++this->num_errors;
    }

    p_ibdiag->DumpGeneralInfoSMPToCSV(*this->p_csv_out);

    if (this->p_ibdiag->DumpExtendedSwitchInfoToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing Extended Switch info to DB failed, err=%s\n",
                this->p_ibdiag->GetLastError());
        ++this->num_errors;
    }

    p_ibdiag->DumpChassisInfoToCSV(*this->p_csv_out);

    //dump the capability mask configurations after we are done with both stages
    //using it as a hidden option for debug
    if (!this->output_full_capability_file.empty()) {
        string parser_output;

        OutputControl::Identity identity(this->output_full_capability_file,
                                         OutputControl::OutputControl_Flag_UserFile);


        if (p_ibdiag->DumpFullCapabilityMaskFile(identity, parser_output))
        {
            WARN_PRINT("Writing Guid to Capability Masks file to %s failed\n",
                       this->output_full_capability_file.c_str());
            ++this->num_errors;
        }

        LOG_PRINT("%s\n", parser_output.c_str());
    }

    IBDIAGNET_RETURN(0);
}

int DiscoveryStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc;
    list_p_fabric_general_err nodes_errors;

    //print duplicated GUIDs detection error if not skipped
    if (this->to_check_dup_guids)
        this->PrintDuplicatedGuidsDetectErrors();

    //duplicated GUIDs detection - we can find duplicated GUID even if
    //skipped in some cases (like duplicated port GUID)
    list_p_fabric_general_err guids_errors;
    int rc2 = this->p_ibdiag->CheckDuplicatedGUIDs(guids_errors);
    rc = this->AnalyzeCheckResults(guids_errors,
            CHECK_NAME_DUPLICATED_GUIDS,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    if (rc2 == IBDIAG_ERR_CODE_CHECK_FAILED)    //abort and exit
        IBDIAGNET_RETURN(1);

    //if needed than check for duplicated node description
    if (this->to_check_dup_node_desc) {
        rc2 = this->p_ibdiag->CheckDuplicatedNodeDescription(nodes_errors);
        if (rc2 == IBDIAG_ERR_CODE_CHECK_FAILED)
            WARN_PRINT("Note: If you have unmanaged systems then duplication can occur\n");
        rc |= this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_DUPLICATED_NODE_DESC,
                rc2,
                IBDIAG_ERR_CODE_CHECK_FAILED,
                this->num_errors,
                this->num_warnings,
                true);
    }

    //check port hierarchy info
    list_p_fabric_general_err hierarchy_errors;
    rc2 = this->p_ibdiag->CheckPortHierarchyInfo(hierarchy_errors);
    rc |= this->AnalyzeCheckResults(hierarchy_errors,
            CHECK_NAME_PORT_HIERARCHY_INFO,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    
    list_p_fabric_general_err cage_manager_errs;
    CageManagerChecks(cage_manager_errs);
    rc |= this->AnalyzeCheckResults(cage_manager_errs,
            CHECK_CAGE_MANAGER_APORT,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);


    //if the port is down than exit
    bool local_link_is_down;
    if (check_if_local_link_is_down(this->p_ibdiag, local_link_is_down))
        IBDIAGNET_RETURN(1);
    if (local_link_is_down) {
        ERR_PRINT("The Local link is in DOWN state\n");
        IBDIAGNET_RETURN(1);
    }

    this->p_ibdiag->GetIBDiagSMDB().Apply(*(this->p_ibdiag->GetDiscoverFabricPtr()));


    list_p_fabric_general_err aport_errors;
    APortChecks();

    DumpAPortsReport(false);
    CopyAPortErrorsToList(aport_errors);
    rc |= this->AnalyzeCheckResults(aport_errors,
            CHECK_NAME_APORT_SYMMETRY,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);

    list_p_fabric_general_err fnm_errors;
    FNMPortsChecks(fnm_errors);
    rc |= this->AnalyzeCheckResults(fnm_errors,
            CHECK_NAME_FNM_PORT,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);

    IBDIAGNET_RETURN(rc);
}

int DiscoveryStage::CleanResources()
{
    IBDIAGNET_ENTER;
    p_ibdiag->GetDiscoverFabricPtr()->CleanUpInternalDB();
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / IBDIAG_ERR_CODE_IBDM_ERR
int DiscoveryStage::CountNodeTypes()
{
    IBDIAG_ENTER;

    unordered_set<IBSystem *> counted_systems;
    this->num_of_switches = 0;
    this->num_of_channel_adapters = 0;
    this->num_of_channel_adapters_sdm = 0;
    this->num_of_routers = 0;
    this->num_of_aggregation_nodes = 0;
    this->num_of_GPUs = 0;

    // Get over all nodes
    for (map_str_pnode::iterator nI = this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.begin();
            nI != this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.end(); ++nI) {

        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->p_ibdiag->SetLastError("DB error - found null node in NodeByName map"
                "for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (p_curr_node->type == IB_SW_NODE) {
            // some nodes should only by counted as a system
            IBPort *p_port = p_curr_node->getPortWithAsicName();
            if (p_port && PortHierarchyInfo::IsMultiNodeSwitch(p_port->p_port_hierarchy_info)) {
                if (counted_systems.count(p_curr_node->p_system))
                    continue;
                else
                    counted_systems.insert(p_curr_node->p_system);
            }
            this->num_of_switches++;
            continue;
        }

        if (p_curr_node->type == IB_RTR_NODE) {
            this->num_of_routers++;
            continue;
        }

        if (p_curr_node->type == IB_CA_NODE) {

            IBPort *p_port = p_curr_node->getPort(1);

            if (p_port &&
                    p_port->p_remotePort &&
                    p_port->p_remotePort->getSpecialPortType() == IB_SPECIAL_PORT_AN) {
                this->num_of_aggregation_nodes++;
                continue;
            }

            if (p_curr_node->isSDM()) {
                ++this->num_of_channel_adapters_sdm;
                continue;
            }

            if (p_curr_node->ext_type == IB_GPU_NODE) {
                ++this->num_of_GPUs;
                continue;
            }

            this->num_of_channel_adapters++;
        }
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

bool DiscoveryStage::GetFabricSummarySection(stringstream &summary_str)
{
    IBDIAGNET_ENTER;

    if (this->CountNodeTypes())
        IBDIAGNET_RETURN(1);

    if (num_of_channel_adapters_sdm)
        this->max_spaces_fabric_summary = MAX_SPACES_FABRIC_SUMMARY_LONG;

    char buffer[1024] = {0};
    string nodes_str = "Total Nodes";
    snprintf(buffer, sizeof(buffer), "%*.s: %lu", this->max_spaces_fabric_summary - int(nodes_str.length()), " ",
             this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.size());
    summary_str << nodes_str << buffer << "\n";

    CLEAR_STRUCT(buffer);
    nodes_str = "IB Switches";
    snprintf(buffer, sizeof(buffer), "%*.s: " U32D_FMT, this->max_spaces_fabric_summary - int(nodes_str.length()), " ",
            this->num_of_switches);
    summary_str << nodes_str << buffer << "\n";

    CLEAR_STRUCT(buffer);
    nodes_str = "IB Channel Adapters";
    snprintf(buffer, sizeof(buffer), "%*.s: " U32D_FMT, this->max_spaces_fabric_summary - int(nodes_str.length()), " ",
             this->num_of_channel_adapters);
    summary_str << nodes_str << buffer << "\n";

    if (num_of_channel_adapters_sdm) {
        CLEAR_STRUCT(buffer);
        nodes_str = "IB Channel Adapters Socket-Direct";
        snprintf(buffer, sizeof(buffer), "%*.s: " U32D_FMT, this->max_spaces_fabric_summary - int(nodes_str.length()), " ",
                 this->num_of_channel_adapters_sdm);
        summary_str << nodes_str << buffer << "\n";
    }

    CLEAR_STRUCT(buffer);
    nodes_str = "IB Aggregation Nodes";
    snprintf(buffer, sizeof(buffer), "%*.s: " U32D_FMT, this->max_spaces_fabric_summary - int(nodes_str.length()), " ",
             this->num_of_aggregation_nodes);
    summary_str << nodes_str << buffer << "\n";

    CLEAR_STRUCT(buffer);
    nodes_str = "IB Routers";
    snprintf(buffer, sizeof(buffer), "%*.s: " U32D_FMT, this->max_spaces_fabric_summary - int(nodes_str.length()), " ",
             this->num_of_routers);
    summary_str << nodes_str << buffer << "\n";

    if (this->num_of_GPUs) {
        CLEAR_STRUCT(buffer);
        nodes_str = "IB GPUs";
        snprintf(buffer, sizeof(buffer), "%*.s: " U32D_FMT, this->max_spaces_fabric_summary - int(nodes_str.length()), " ",
                 this->num_of_GPUs);
        summary_str << nodes_str << buffer << "\n";
    }

    u_int64_t num_ar_switches = 0;
    u_int64_t num_hbf_en_switches = 0;
    u_int64_t num_hbf_sup_switches = 0;

    this->p_ibdiag->GetAREnabledNum(num_ar_switches, num_hbf_en_switches, num_hbf_sup_switches);
    summary_str << "\nAdaptive Routing is enabled on " << num_ar_switches << " switches.\n";

    if (num_hbf_sup_switches)
        summary_str << "Hashed Based Forwarding is enabled on " << num_hbf_en_switches << " switches.\n";

    IBDIAGNET_RETURN(true);
}

void DiscoveryStage::CreateLid2GuidDB()
{
    IBDIAGNET_ENTER;

    if (!p_ibdiag->GetDiscoverFabricPtr())
        IBDIAGNET_RETURN_VOID;

    for (map_str_pnode::const_iterator itr = this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.begin();
            itr != this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.end(); itr++) {
        const IBNode *p_node = itr->second;
        if (!p_node)
            continue;

        phys_port_t start = (p_node->type == IB_SW_NODE ? 0 : 1);
        phys_port_t end = (p_node->type == IB_SW_NODE ? 0 : p_node->numPorts);
        for (phys_port_t i = start; i <= end; ++i) {
            const IBPort *p_port = p_node->getPort(i);
            if (!p_port)
                continue;

            this->lid2guid_map[p_port->base_lid] = p_port->guid_get();
        }
    }

    IBDIAGNET_RETURN_VOID;
}

/******************************************************/
LidsStage::LidsStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_LIDS, p_ibdiag)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int LidsStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int LidsStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    //dummy one
    IBDIAGNET_RETURN(0);
}

int LidsStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc;

    list_p_fabric_general_err lids_errors;
    int rc2 = this->p_ibdiag->CheckLids(lids_errors);
    if (rc2 == IBDIAG_ERR_CODE_CHECK_FAILED)
        found_lids_errors = true;
    rc = this->AnalyzeCheckResults(lids_errors,
            CHECK_NAME_LIDS,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
/////
//ToplogtMatchStage public methods:
/////
TopologyMatchStage::TopologyMatchStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_TOPOLOGY_MATCH, p_ibdiag),
        spec_fabric(),
        output_topo_file(""), input_topo_file(""), topo_ibnl_dir(""),
        use_smart_names(false)
{
    IBDIAGNET_ENTER;
    this->MarkStageAsSkipped(); //this stage will be skipped by default
	//this->compare_nodes_names = 0;
    IBDIAGNET_RETURN_VOID;
}

int TopologyMatchStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int TopologyMatchStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    if (topo_ibnl_dir.empty()) {
        topo_ibnl_dir = generate_file_name(*this->p_base_path, IBNL_OUTPUT_DIR);
        INFO_PRINT("Topo ibnl dir is not given, use default: %s\n", topo_ibnl_dir.c_str());
    }

    //create a topology file if needed
    if (!this->output_topo_file.empty()) {

        //construct Systems - creates IBSystem ports and check IBNL definitions
        ibdmClearInternalLog();

        list< string > dirs;
        dirs.push_back(topo_ibnl_dir);
        theSysDefsCollection()->parseSysDefsFromDirs(dirs);

        this->p_ibdiag->GetDiscoverFabricPtr()->constructSystems();
        char *buffer= ibdmGetAndClearInternalLog();
        if (!buffer) {
            ERR_PRINT("Failed to allocate buffer for ibdm output\n");
            IBDIAGNET_RETURN(1);
        } else {
            u_int32_t errors = num_errors;
            u_int32_t warnings = num_warnings;
            string construct_res(buffer);
            this->ParseIBDMOutput(construct_res);

            if (num_errors > errors) {
                ERR_PRINT("Preparing topology for discovered fabric"
                          " finished with errors.\n");
            } else if (num_warnings > warnings) {
                WARN_PRINT("Preparing topology for discovered fabric"
                           " finished with warnings.\n");
            }
            LOG_PRINT("%s", buffer);
            free(buffer);
        }

        //try to create topo file and ibnl files
        INFO_PRINT("Creating topology file for discovered fabric - %s\n",
                this->output_topo_file.c_str());

        ibdmClearInternalLog();

        OutputControl::Identity identity(this->output_topo_file,
        		                         OutputControl::OutputControl_Flag_UserFile);

        OutputControl::Properties properties(identity);

        if (!this->p_ibdiag->GetDiscoverFabricPtr()->dumpTopology(identity))
            this->p_ibdiag->AddGeneratedFile("Topology",  properties.path());
        else
        	this->p_ibdiag->AddGeneratedFile("Topology",  "Write file failure.");

        buffer = ibdmGetAndClearInternalLog();
        if (!buffer) {
            ERR_PRINT("Failed to allocate buffer for ibdm output\n");
            IBDIAGNET_RETURN(1);
        }
        PRINT("%s", buffer);
        free(buffer);
    }

    if (!this->input_topo_file.empty()) {

         //parsing the topology file
        ibdmClearInternalLog();

        //reparse topo_ibnl_dir in case aditional ibnls were added
        list< string > dirs;
        dirs.push_back(topo_ibnl_dir);
        theSysDefsCollection()->parseSysDefsFromDirs(dirs);

        IBDiag::PrintFileTimestamp(this->input_topo_file, "Topology File");
        int rc = GetSpecFabricPtr()->parseTopology(input_topo_file);
        char *buffer = ibdmGetAndClearInternalLog();
        if (!buffer) {
            ERR_PRINT("Failed to allocate buffer for ibdm output\n");
            IBDIAGNET_RETURN(1);
        }
        PRINT("%s", buffer);
        free(buffer);
        if (rc)
            ERR_PRINT("Parse topology file: %s finished with errors\n",
                      input_topo_file.c_str());
    }

    IBDIAGNET_RETURN(0);
}

int TopologyMatchStage::RunCheck()
{
    IBDIAGNET_ENTER;
    if (this->input_topo_file.empty())
        IBDIAGNET_RETURN(0);
    if (GetSpecFabricPtr()->NodeByName.empty()) {
        ERR_PRINT("Topology Matching skipped due to parsing errors\n");
        IBDIAGNET_RETURN(0);
    }
    INFO_PRINT("Topology Matching\n");
    //get root port
    IBPort *p_root_port = this->p_ibdiag->GetRootPort();
    if (!p_root_port) {
        ERR_PRINT("Topology Matching failed, err=%s\n", this->p_ibdiag->GetLastError());
        IBDIAGNET_RETURN(1);
    }

    //match discovered fabric with spec fabric
    char *messages;
    TopoMatchOptions opts;

    /*
    if (TopoMatchFabrics(&this->spec_fabric,        // The specification fabric
            this->p_ibdiag->GetDiscoverFabricPtr(), // The discovered fabric
            p_root_port->p_node->name.c_str(),
                                                    // The name of the node to be the anchor point - in the spec fabric,
            p_root_port->num,                       // The port number of the anchor port - in the spec fabric
            p_root_port->guid_get(),                // Guid of the anchor port - in the discovered fabric
            &messages,
			this->compare_nodes_names)) {
           */
    //opts.mapFileName = mapFileName;
    opts.matchCAsByName = true; // match non-SW by name
    opts.matchSWsByName = true; // match SW by name
    opts.smartSWsName = this->use_smart_names; // use search switches by description if by name failed
    //opts.csvFileName = csvFileName;
    opts.outMapFileName = generate_file_name(*this->p_base_path, "topodiff.names");

    if (TopoMatchFabricsFromEdge(&this->spec_fabric,        // The specification fabric
            this->p_ibdiag->GetDiscoverFabricPtr(), // The discovered fabric
            opts, &messages)) {
        WARN_PRINT("Found mismatches between the topology defined in %s and the discovered fabric.\n",
                this->input_topo_file.c_str());
        if (messages) {
            string topo_result = messages;
            this->ParseIBDMOutput(topo_result);
            LOG_PRINT("%s\n", topo_result.c_str());
        }
    } else {
        INFO_PRINT("The topology defined in %s perfectly matches the discovered fabric\n",
                this->input_topo_file.c_str());
    }

    PRINT("\n");
    free(messages);
    IBDIAGNET_RETURN(0);
}


/******************************************************/
LinksStage::LinksStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_LINKS, p_ibdiag)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

void LinksStage::printLinksSummary(stringstream &summary_str,
                                   string name,
                                   uint32_t num_of_links,
                                   vec_vec_uint32 &link_width_speed_matrix)
{
    IBLinkWidth width;
    IBLinkSpeed speed;
    string width_str;
    string speed_str;
    char buffer[1024] = {0};

    string msg = string("Total number of ") + name + string("links");
    snprintf(buffer,
             sizeof(buffer),
             "%*.s: " U32D_FMT,
             this->max_spaces_fabric_summary - int(msg.length()),
             " ",
             num_of_links);

    summary_str << msg << buffer << "\n";

    msg = name + string("Links at ");

    for (int width_index = 1; width_index < IB_LINK_WIDTH_END_INDEX; width_index++) {

         width = width_index2width((IBLinkWidthIndex)width_index);
         width_str = width2char(width);

         for (int speed_index = 1; speed_index < IB_LINK_SPEED_END_INDEX; speed_index++) {

              if (!link_width_speed_matrix[width_index][speed_index])
                  continue;

              speed = speed_index2speed((IBLinkSpeedIndex)speed_index);
              speed_str = speed2char(speed);

              CLEAR_STRUCT(buffer);
              string width_speed_str = width_str + speed_str;

              snprintf(buffer, sizeof(buffer), "%s%*.s: " U32D_FMT, width_speed_str.c_str(),
                       this->max_spaces_fabric_summary -
                       int(msg.length() + width_speed_str.length()), " ",
                       link_width_speed_matrix[width_index][speed_index]);

              summary_str << msg << buffer << "\n";
         }
    }
}

bool LinksStage::GetFabricSummarySection(stringstream &summary_str)
{
    IBDIAGNET_ENTER;

    char buffer[1024] = {0};
    vec_vec_uint32 &ib_link_width_speed_matrix = this->ib_links_info.GetIBLinkWidthSpeedMatrix();

    // Legacy
    printLinksSummary(summary_str,
                      "",
                      this->ib_links_info.GetNumOfIBLinks(),
                      ib_link_width_speed_matrix);

    // FNM
    if (this->ib_links_info.GetNumOfFNMLinks()) {
        summary_str << endl;
        printLinksSummary(summary_str,
                          "FNM ",
                          this->ib_links_info.GetNumOfFNMLinks(),
                          this->ib_links_info.GetFNMLinkWidthSpeedMatrix());
    }

    // FNM1
    if (this->ib_links_info.GetNumOfFNM1Links()) {
        summary_str << endl;
        printLinksSummary(summary_str,
                          "FNM1 ",
                          this->ib_links_info.GetNumOfFNM1Links(),
                          this->ib_links_info.GetFNM1LinkWidthSpeedMatrix());
    }

    string asymmetrical_links_message = "Asymmetrical Links";
    if (ib_link_width_speed_matrix[IB_UNKNOWN_LINK_WIDTH_INDEX][IB_UNKNOWN_LINK_SPEED_INDEX]) {

        CLEAR_STRUCT(buffer);
        snprintf(buffer, sizeof(buffer), "%s%*.s: " U32D_FMT, asymmetrical_links_message.c_str(),
                this->max_spaces_fabric_summary -
                int(asymmetrical_links_message.length()), " ",
                ib_link_width_speed_matrix[IB_UNKNOWN_LINK_WIDTH_INDEX][IB_UNKNOWN_LINK_SPEED_INDEX]);

        summary_str << endl << buffer << endl;
    }

    IBDIAGNET_RETURN(true);
}

int LinksStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int LinksStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    //dummy one
    IBDIAGNET_RETURN(0);
}

int LinksStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc;

    list_p_fabric_general_err links_errors;
    int rc2 = this->p_ibdiag->CheckLinks(links_errors, &this->ib_links_info);
    if (rc2 == IBDIAG_ERR_CODE_CHECK_FAILED)
        found_links_errors = true;
    rc = this->AnalyzeCheckResults(links_errors,
            CHECK_NAME_LINKS,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
SMStage::SMStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_SM, p_ibdiag)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

bool SMStage::GetFabricSummarySection(stringstream &summary_str)
{
    IBDIAGNET_ENTER;

    string p_sm_masters_list_str = "";
    string p_sm_standby_list_str = "";
    set < APort* > aports_set;

    list_p_sm_info_obj &sm_info_list = this->p_ibdiag->GetIBDMExtendedInfoPtr()->
                                             getSMPSMInfoListRef();

    char buffer[1024] = {0};

    //collect info
    for (auto const& p_sm_info : sm_info_list) {

        IBPort* p_port = p_sm_info->p_port;
        SMP_SMInfo sm_info = p_sm_info->smp_sm_info;

        if (!(sm_info.SmState == IBIS_IB_SM_STATE_MASTER) &&
                !(sm_info.SmState == IBIS_IB_SM_STATE_STANDBY)) {
            continue;
        }

        if (p_port->p_aport) {
            if (!aports_set.count(p_port->p_aport))
                aports_set.insert(p_port->p_aport);
            else
                continue;
        }

        CLEAR_STRUCT(buffer);
        snprintf(buffer, sizeof(buffer), "Port=" U32D_FMT " LID=%u" " GUID=" U64H_FMT \
                 " devid=" U32D_FMT " Priority:" U32D_FMT " Node_Type=%s Node_Description=%s",
                 p_port->num,
                 p_port->base_lid,
                 p_port->guid_get(),
                 p_port->p_node->devId,
                 sm_info.Priority,
                 nodetype2char(p_port->p_node->type),
                 p_port->p_node->description.c_str());

        if (sm_info.SmState == IBIS_IB_SM_STATE_MASTER) {
            p_sm_masters_list_str += "Master SM: " + string(buffer) + "\n";
        }

        if (sm_info.SmState == IBIS_IB_SM_STATE_STANDBY) {
            p_sm_standby_list_str += "Standby SM: " + string(buffer) + "\n";
        }
    }

    if (p_sm_masters_list_str == "")
        p_sm_masters_list_str = "Master SM  : No Master SM\n";

    if (p_sm_standby_list_str == "")
        p_sm_standby_list_str = "Standby SM : No Standby SM\n";

    summary_str << p_sm_masters_list_str << p_sm_standby_list_str;

    IBDIAGNET_RETURN(true);
}

int SMStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int SMStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc;

    list_p_fabric_general_err sm_errors;
    int rc2 = this->p_ibdiag->BuildSMInfo(sm_errors);
    rc = this->AnalyzeCheckResults(sm_errors,
            CHECK_NAME_SM_INFO_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (!rc) {
        if (this->p_ibdiag->DumpSMInfoToCSV(*this->p_csv_out)) {
            ERR_PRINT("Writing SM info to DB failed\n");
            ++this->num_errors;
        }
        if (this->p_ibdiag->WriteSMFile(SM_FILE)) {
            ERR_PRINT("Writing Subnet Manager file failed.\n");
            ++this->num_errors;
        }
    }
    IBDIAGNET_RETURN(rc);
}

int SMStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc;

    list_p_fabric_general_err sm_errors;
    int rc2 = this->p_ibdiag->CheckSMInfo(sm_errors);
    rc = this->AnalyzeCheckResults(sm_errors,
            CHECK_NAME_SM_INFO,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    IBDIAGNET_RETURN(rc);
}

int SMStage::CleanResources()
{
    IBDIAGNET_ENTER;
    release_container_data(this->p_ibdiag->GetIBDMExtendedInfoPtr()->getSMPSMInfoListRef());
    IBDIAGNET_RETURN(0);
}


/******************************************************/
/////
//pm private methods:
/////
static bool pm_caught_signal = false;

#ifdef __WIN__
    int signals_array[] = {SIGINT};
#else
    int signals_array[] = {SIGINT, SIGQUIT, SIGABRT, SIGPIPE, SIGTERM};
#endif


void pm_stage_sig_handler(int signum)
{
    IBDIAGNET_ENTER;
    // no need to set handler again because we want the default handler
    // from here

    //mark we caught the signal
    pm_caught_signal = true;

    SCREEN_PRINT("\n");
    if (signum == SIGINT) {
        INFO_PRINT("Received User ^C (%s). Stopping\n", strsignal(signum));
    } else {
        INFO_PRINT("Received signal %d (%s). Stopping\n", signum, strsignal(signum));
    }
    IBDIAGNET_RETURN_VOID;
}

void PMStage::Sleep(u_int32_t seconds)
{
    IBDIAGNET_ENTER;
    u_int32_t seconds_left = seconds;
    INFO_PRINT("Going to sleep for %u seconds until next counters sample\n", seconds);
    while (!pm_caught_signal && seconds_left) {
        SCREEN_INFO_PRINT("Time left to sleep ... %u seconds.\r", seconds_left);
        fflush(stdout);
        sleep(1);
        --seconds_left;
    }
    SCREEN_PRINT("\n");
    IBDIAGNET_RETURN_VOID;
}

/////
//pm public methods:
/////
PMStage::PMStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_PM, p_ibdiag),
        to_reset_counters(false),
        to_reset_histogram_data(false),
        to_calc_ber(false),
        m_check_counters_bitset(0x00),
        en_per_lane_cnts(false),
        en_per_slvl_cntrs(false),
        en_per_slvl_sup_warn(false),
        en_port_samples_result(false),
        ber_threshold(DEFAULT_BER_THRESHOLD),
        pause_time_between_samples(DEFAULT_SECONDS_BETWEEN_COUNTERS_SAMPLES),
        diff_time_between_samples(0),
        can_send_mads_by_lid(false),
        total_pm_symbol_errors(0)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int PMStage::CleanResources()
{
    IBDIAGNET_ENTER;

    this->total_pm_symbol_errors = 0;

    this->p_ibdiag->GetIBDMExtendedInfoPtr()->
           CleanPMInfoObjVector(this->first_sample_pm_info_obj_vector);
    this->p_ibdiag->GetIBDMExtendedInfoPtr()->
           CleanPMInfoObjVector(this->p_ibdiag->GetIBDMExtendedInfoPtr()->getPMInfoObjVectorRef());

    release_container_data(m_slvl_cntrs_vec);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

PMStage::~PMStage()
{
    IBDIAGNET_ENTER;

    this->CleanResources();

    IBDIAGNET_RETURN_VOID;
}

int PMStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag, this->can_send_mads_by_lid))
        IBDIAGNET_RETURN(1);
    IBDIAGNET_RETURN(0);
}

void PMStage::PreparePerSlVlCounters()
{

    IBDIAGNET_ENTER;
    //VS Class
    CountersPerSLVL * port_rcv_data_vl = new PortRcvDataVL();
    m_slvl_cntrs_vec.push_back(port_rcv_data_vl);

    CountersPerSLVL * port_xmit_data_vl = new PortXmitDataVL();
    m_slvl_cntrs_vec.push_back(port_xmit_data_vl);

    CountersPerSLVL * port_rcv_ext_data_vl = new PortRcvDataVLExt();
    m_slvl_cntrs_vec.push_back(port_rcv_ext_data_vl);

    CountersPerSLVL * port_xmit_ext_data_vl = new PortXmitDataVLExt();
    m_slvl_cntrs_vec.push_back(port_xmit_ext_data_vl);

    CountersPerSLVL * port_rcv_pkt_vl = new PortRcvPktVL();
    m_slvl_cntrs_vec.push_back(port_rcv_pkt_vl);

    CountersPerSLVL * port_xmit_pkt_vl = new PortXmitPktVL();
    m_slvl_cntrs_vec.push_back(port_xmit_pkt_vl);

    CountersPerSLVL * port_rcv_ext_pkt_vl = new PortRcvPktVLExt();
    m_slvl_cntrs_vec.push_back(port_rcv_ext_pkt_vl);

    CountersPerSLVL * port_xmit_ext_pkt_vl = new PortXmitPktVLExt();
    m_slvl_cntrs_vec.push_back(port_xmit_ext_pkt_vl);

    CountersPerSLVL * port_xmit_ext_wait_vl = new PortXmitWaitVLExt();
    m_slvl_cntrs_vec.push_back(port_xmit_ext_wait_vl);

    //PM Class
    CountersPerSLVL * port_xmit_data_sl = new PortXmitDataSL();
    m_slvl_cntrs_vec.push_back(port_xmit_data_sl);

    CountersPerSLVL * port_rcv_data_sl = new PortRcvDataSL();
    m_slvl_cntrs_vec.push_back(port_rcv_data_sl);

    CountersPerSLVL * port_xmit_data_sl_ext = new PortXmitDataSLExt();
    m_slvl_cntrs_vec.push_back(port_xmit_data_sl_ext);

    CountersPerSLVL * port_rcv_data_sl_ext = new PortRcvDataSLExt();
    m_slvl_cntrs_vec.push_back(port_rcv_data_sl_ext);

    CountersPerSLVL* port_vl_xmit_flow_ctl_update_errors = new PortVLXmitFlowCtlUpdateErrors();
    m_slvl_cntrs_vec.push_back(port_vl_xmit_flow_ctl_update_errors);

    /*

    //
    // Counters was replaced by PortXmitWaitVLExt from VS specs
    //
    CountersPerSLVL* port_vl_xmit_wait_vl = new PortXmitWaitVL();
    m_slvl_cntrs_vec.push_back(port_vl_xmit_wait_vl);

    */

    IBDIAGNET_RETURN_VOID;
}


void PMStage::CountTotalPMSymbolError(u_int64_t pm_threshold)
{
    IBDIAG_ENTER;

    IBDMExtendedInfo* fabric_extended_info = this->p_ibdiag->GetIBDMExtendedInfoPtr();
    for (u_int32_t i = 0; i < fabric_extended_info->getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = fabric_extended_info->getPortPtr(i);
        if (!p_curr_port || !p_curr_port->getInSubFabric())
            continue;

        struct PM_PortCountersExtended *p_curr_port_counters_extended =
                                                 fabric_extended_info->getPMPortCountersExtended(i);
        if (!p_curr_port_counters_extended)      //info not exist
            continue;

        if (p_curr_port_counters_extended->SymbolErrorCounter > pm_threshold)
            this->total_pm_symbol_errors++;
    }

    IBDIAG_RETURN_VOID;
}


bool PMStage::GetFabricSummarySection(stringstream &summary_str)
{
    IBDIAGNET_ENTER;

    // check if threshold was given to symbol_error_counter_extended counter
    map_str_uint64::iterator it = this->counters_to_threshold_map.
            find(string("symbol_error_counter_extended"));

    if (it == this->counters_to_threshold_map.end())
        IBDIAGNET_RETURN(false);

    this->CountTotalPMSymbolError(it->second);

    if (this->total_pm_symbol_errors == 0)
        IBDIAGNET_RETURN(false);

    summary_str << "High symbol errors reported by "
                << this->total_pm_symbol_errors
                << " ports"
                << endl;

    IBDIAGNET_RETURN(true);
}


int PMStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    // SMP MADs are sent by direct route; they do not depend on LIDs
    static bool get_policy_config = true;
    if (get_policy_config) {
        list_p_fabric_general_err errors;
        INFO_PRINT("Build Ports Recovery Policy Config Data\n");
        rc = p_ibdiag->BuildPortRecoveryPolicyConfig(errors);

        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(errors,
                CHECK_PORTS_RECOVERY_POLICY_CONFIG,
                rc,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings);

        if (p_ibdiag->DumpPortRecoveryPolicyConfigToCSV(*this->p_csv_out)) {
            ERR_PRINT("Writing Port Recovery Policy Config Data to DB failed\n");
            ++this->num_errors;
        }

        get_policy_config = false;
    }

    // The rest of MADs depend on LID and should be skipped if LIDs are not available
    if (!this->can_send_mads_by_lid) {
        if (this->to_reset_counters)
            INFO_PRINT("%s skipped\n", CHECK_NAME_PM_RESET);
        INFO_PRINT("%s skipped\n", CHECK_NAME_PM_RETRIEVE);
        PRINT("\n");
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
    }

    rc  = IBDIAG_SUCCESS_CODE;
    list_p_fabric_general_err ports_errors;
    //if need to reset
    if (this->to_reset_counters) {
        int rc2  = this->p_ibdiag->ResetPortCounters(ports_errors,
                                                     m_check_counters_bitset);
        SCREEN_PRINT("\n\n");
        rc = this->AnalyzeCheckResults(ports_errors,
                CHECK_NAME_PM_RESET,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        if (rc)
            IBDIAGNET_RETURN(rc);

        // Reset RN counters
        rc2 = this->p_ibdiag->ClearRNCounters(ports_errors);

        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(ports_errors,
                CHECK_NAME_ROUTE_RN_COUNTERS_CLEAR,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        if (rc)
            IBDIAGNET_RETURN(rc);

        // Reset HBF counters
        rc2 = this->p_ibdiag->ClearHBFCounters(ports_errors);

        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(ports_errors,
                CHECK_NAME_HBF_COUNTERS_CLEAR,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        if (rc)
            IBDIAGNET_RETURN(rc);

        // Reset Credit Watchdog Timeout Counters
        rc2 = this->p_ibdiag->ClearCreditWatchdogTimeoutCounters(ports_errors);

        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(ports_errors,
                CHECK_CREDIT_WATCHDOG_TIMEOUT_COUNTERS_RESET,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        if (rc)
            IBDIAGNET_RETURN(rc);

        // Reset Fast Recovery Counters
        rc2 = this->p_ibdiag->ClearFastRecoveryCounters(ports_errors);

        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(ports_errors,
                                       CHECK_FAST_RECOVERY_COUNTERS_RESET,
                                       rc2,
                                       IBDIAG_ERR_CODE_FABRIC_ERROR,
                                       this->num_errors,
                                       this->num_warnings,
                                       true);

        if (rc)
            IBDIAGNET_RETURN(rc);

        //Reset Port General Counters
        rc2 = this->p_ibdiag->BuildVSPortGeneralCounters(ports_errors, true);

        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(ports_errors,
                                       CHECK_PORT_GENERAL_COUNTERS_RESET,
                                       rc2,
                                       IBDIAG_ERR_CODE_FABRIC_ERROR,
                                       this->num_errors,
                                       this->num_warnings,
                                       true);

        if (rc)
            IBDIAGNET_RETURN(rc);

        //Reset Port Recovery Policy Counters
        rc2 = this->p_ibdiag->BuildVSPortPolicyRecoveryCounters(ports_errors, true);

        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(ports_errors,
                                       CHECK_PORT_POLICY_RECOVERY_COUNTERS_RESET,
                                       rc2,
                                       IBDIAG_ERR_CODE_FABRIC_ERROR,
                                       this->num_errors,
                                       this->num_warnings,
                                       true);

        if (rc)
            IBDIAGNET_RETURN(rc);
    }

    //retrieve all counters - first sample
    struct timeval start_measure;
    if ((rc = gettimeofday(&start_measure, NULL)) != 0) {
        IBDIAG_LOG(TT_LOG_LEVEL_INFO,
                   "Failed to get gettimeofday before first sample.");
        ERR_PRINT("Failed to retrieve first sample time\n");
        ++this->num_errors;
        IBDIAGNET_RETURN(1);
    }

    int rc2 = this->p_ibdiag->BuildPortCounters(ports_errors,
                                                  m_check_counters_bitset, this->en_port_samples_result);
    PRINT("\n");
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_NAME_PM_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (rc)
        IBDIAGNET_RETURN(rc);

    // Build RN counters
    rc2 = this->p_ibdiag->BuildRNCounters(ports_errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_NAME_ROUTE_RN_COUNTERS_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (rc)
        IBDIAGNET_RETURN(rc);

    // Build HBF counters
    rc2 = this->p_ibdiag->BuildHBFCounters(ports_errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_NAME_HBF_COUNTERS_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (rc)
        IBDIAGNET_RETURN(rc);

    // Build Credit Watchdog Timeout Counters
    rc2 = this->p_ibdiag->BuildCreditWatchdogTimeoutCounters(ports_errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_CREDIT_WATCHDOG_TIMEOUT_COUNTERS_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (rc)
        IBDIAGNET_RETURN(rc);

    // Build Fast Recovery Counters
    rc2 = this->p_ibdiag->BuildFastRecoveryCounters(ports_errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_FAST_RECOVERY_COUNTERS_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (rc)
        IBDIAGNET_RETURN(rc);

    // Build Port General Counters
    rc2 = this->p_ibdiag->BuildVSPortGeneralCounters(ports_errors, false);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_PORT_GENERAL_COUNTERS_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (rc)
        IBDIAGNET_RETURN(rc);

    //Reset Port Recovery Policy Counters
    rc2 = this->p_ibdiag->BuildVSPortPolicyRecoveryCounters(ports_errors, false);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(ports_errors,
                                       CHECK_PORT_POLICY_RECOVERY_COUNTERS_RETRIEVE,
                                       rc2,
                                       IBDIAG_ERR_CODE_FABRIC_ERROR,
                                       this->num_errors,
                                       this->num_warnings,
                                       true);

    if (rc)
        IBDIAGNET_RETURN(rc);


    // Build Performance Histogram
    rc2 = this->p_ibdiag->BuildPerformanceHistogramInfo(ports_errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_PERFORMANCE_HISTOGRAM_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (rc)
        IBDIAGNET_RETURN(rc);

    //Build Performance Histogram Buffer Control
    rc2 = this->p_ibdiag->BuildPerformanceHistogramBufferControl(ports_errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_PERFORMANCE_HISTOGRAM_BUFFER_CONTROL,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    //Build Performance Histogram Buffer Data
    rc2 = this->p_ibdiag->BuildPerformanceHistogramBufferData(ports_errors, this->to_reset_histogram_data);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_PERFORMANCE_HISTOGRAM_BUFFER_DATA,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    rc2 = this->p_ibdiag->BuildPerformanceHistogramPortsControl(ports_errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_PERFORMANCE_HISTOGRAM_PORTS_CONTROL,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    rc2 = this->p_ibdiag->BuildPerformanceHistogramPortsData(ports_errors, this->to_reset_histogram_data);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_PERFORMANCE_HISTOGRAM_PORTS_DATA,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (this->pause_time_between_samples != 0) {   //need to take another sample
        //copy counters vector of first sample
        this->p_ibdiag->CopyPMInfoObjVector(this->first_sample_pm_info_obj_vector);

        //set a signal handler for kill signals
        for (unsigned int i = 0; i < (sizeof(signals_array) / sizeof(signals_array[0])); ++i) {
            signal(signals_array[i], pm_stage_sig_handler);
        }
        //pausing
        this->Sleep(this->pause_time_between_samples);
        //set default handlers for kill signals
        for (unsigned int i = 0; i < (sizeof(signals_array) / sizeof(signals_array[0])); ++i) {
            signal(signals_array[i], SIG_DFL);
        }

        //retrieve all counters - second sample
        struct timeval end_measure;
        if ((rc = gettimeofday(&end_measure, NULL)) != 0) {
            IBDIAG_LOG(TT_LOG_LEVEL_INFO,
                       "Failed to get gettimeofday before second sample.");
            ERR_PRINT("Failed to retrieve second sample time\n");
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }

        int rc2 = this->p_ibdiag->BuildPortCounters(ports_errors,
                                                      m_check_counters_bitset, this->en_port_samples_result);
        SCREEN_PRINT("\n\n");

        long long elapsed_mic_sec = (end_measure.tv_sec - start_measure.tv_sec)*1000000LL +
                                    (end_measure.tv_usec - start_measure.tv_usec);

        this->diff_time_between_samples = (double)elapsed_mic_sec / (double)1000000;

        rc = this->AnalyzeCheckResults(ports_errors,
                CHECK_NAME_PM_RETRIEVE2,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (rc)
            IBDIAGNET_RETURN(rc);
    }

    if (this->en_per_slvl_cntrs) {
        this->PreparePerSlVlCounters();
        if (this->to_reset_counters) {
            for (vec_slvl_cntrs::iterator it = m_slvl_cntrs_vec.begin();
                    it != m_slvl_cntrs_vec.end(); it++) {
                CountersPerSLVL * cntrs_per_slvl = *it;
                int rc2 = this->p_ibdiag->BuildDBOrResetSLVLCntrs(
                        ports_errors,
                        true,//clear slvl counters
                        this->en_per_slvl_sup_warn,
                        cntrs_per_slvl);

                SCREEN_PRINT("\n");
                rc = this->AnalyzeCheckResults(ports_errors,
                        cntrs_per_slvl->GetCntrHeader(),
                        rc2,
                        IBDIAG_ERR_CODE_FABRIC_ERROR,
                        this->num_errors,
                        this->num_warnings);
                ports_errors.clear();
            }
        }

        for (vec_slvl_cntrs::iterator it = m_slvl_cntrs_vec.begin();
             it != m_slvl_cntrs_vec.end(); it++) {
            CountersPerSLVL * cntrs_per_slvl = *it;
            int rc2 = this->p_ibdiag->BuildDBOrResetSLVLCntrs(
                    ports_errors,
                    false,//build slvl counters DB
                    this->en_per_slvl_sup_warn,
                    cntrs_per_slvl);

            SCREEN_PRINT("\n");
            rc = this->AnalyzeCheckResults(ports_errors,
                    cntrs_per_slvl->GetCntrHeader(),
                    rc2,
                    IBDIAG_ERR_CODE_FABRIC_ERROR,
                    this->num_errors,
                    this->num_warnings);
            ports_errors.clear();
        }
    }

    IBDIAGNET_RETURN(rc);
}

int PMStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("Ports counter Checks skipped\n");
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }
    list_p_fabric_general_err ports_errors;

    //calculate counters calculation
    int rc2 = this->p_ibdiag->CalcCounters(this->first_sample_pm_info_obj_vector,
                                           this->diff_time_between_samples,
                                           ports_errors);

    list_p_fabric_general_err pm_threshold_overflow_warnings;
    // Errors accumulated in the CalcCounters will be analyzed with CheckAllPMValues later
    //counters value check
    rc2 |= this->p_ibdiag->CheckAllPMValues(ports_errors,
                                            pm_threshold_overflow_warnings,
                                            this->counters_to_threshold_map,
                                            m_check_counters_bitset);

    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_NAME_PM_VALUE,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    if (rc)
        IBDIAGNET_RETURN(rc);

    rc = this->AnalyzeCheckResults(pm_threshold_overflow_warnings,
            CHECK_NAME_OVERFLOW_PM_VALUE,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    if (this->p_ibdiag->DumpPortSamplesControlToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing Port Samples Control to DB CSV failed\n");
        ++this->num_errors;
    }

    if (this->en_port_samples_result) {
        if (this->p_ibdiag->DumpPortSamplesResultToCSVTable(*this->p_csv_out)) {
            ERR_PRINT("Writing Port Samples Result to DB CSV failed\n");
            ++this->num_errors;
        }
    }

    // now when we have the calculated counters, we can print to the files
    if (this->p_ibdiag->DumpPortCountersToCSV(*this->p_csv_out, m_check_counters_bitset)) {
        ERR_PRINT("Writing ports counters to DB failed\n");
        ++this->num_errors;
    }

    if (this->en_per_slvl_cntrs) {
        if (this->p_ibdiag->DumpPerSLVLPortCountersToCSV(*this->p_csv_out,
                                                         this->m_slvl_cntrs_vec)) {
            ERR_PRINT("Write SLVL Counters PM Data to DB failed\n");
            ++this->num_errors;
        }
    }

    if (this->p_ibdiag->DumpHBFCountersToCSV(*this->p_csv_out)) {
        ERR_PRINT("Write HBF counters to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpCreditWatchdogTimeoutToCSV(*this->p_csv_out)) {
        ERR_PRINT("Write Credit Watchdog Timeout counters to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpFastRecoveryCountersToCSV(*this->p_csv_out)) {
        ERR_PRINT("Write Fast Recovery Counters to DB failed\n");
        ++this->num_errors;
    }
    
    if (this->p_ibdiag->DumpPortGeneralCountersToCSV(*this->p_csv_out)) {
        ERR_PRINT("Write Port General Counters to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpPortPolicyRecoveryCountersToCSV(*this->p_csv_out)) {
        ERR_PRINT("Write Port General Counters to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpPerformanceHistogramInfoToCSV(*this->p_csv_out)) {
        ERR_PRINT("Write Performance Histogram Info to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpPerformanceHistogramBufferControlToCSV(*this->p_csv_out)) {
        ERR_PRINT("Write Performance Histogram Buffer Control Info to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpPerformanceHistogramBufferDataToCSV(*this->p_csv_out)) {
        ERR_PRINT("Write Performance Histogram Buffer Data Info to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpPerformanceHistogramPortsControlToCSV(*this->p_csv_out)) {
        ERR_PRINT("Write Performance Histogram Ports Control to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpPerformanceHistogramPortsDataToCSV(*this->p_csv_out)) {
        ERR_PRINT("Write Performance Histogram Buffer Data Info to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->WritePMFile(PM_FILE,
                                    m_check_counters_bitset,
                                    en_per_lane_cnts)) {
        ERR_PRINT("Writing ports counters file failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->WritePMAggregatedFile(PM_AGG_FILE,
                                             m_check_counters_bitset,
                                             en_per_lane_cnts)) {
        ERR_PRINT("Writing aggregated ports counters file failed\n");
        ++this->num_errors;
    }

    list_p_fabric_general_err pfrn_errors;

    if (this->p_ibdiag->DumpRNCountersToCSV(*this->p_csv_out, pfrn_errors)) {
        ERR_PRINT("Writing RN counters to DB failed\n");
        ++this->num_errors;
    }

    // check for bad values in pfrn_received_error counter
    this->AnalyzeCheckResults(pfrn_errors,
            CHECK_NAME_PFRN_RECEIVED_ERROR,
            0,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings,
            true);

    if (this->p_ibdiag->WriteRNCounters_2_File(RNC_FILE_2)) {
        ERR_PRINT("Writing RN counters 2 file failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->WriteRNCountersFile(RNC_FILE)) {
        ERR_PRINT("Writing RN counters file failed\n");
        ++this->num_errors;
    }

    //abort all other checks - we don't have another sample
    if (this->pause_time_between_samples == 0) {
        INFO_PRINT("Ports counters Difference Check will be skipped - pause time is zero\n");
        IBDIAGNET_RETURN(rc);
    }

    //no increase check
    rc2 = this->p_ibdiag->CheckCountersDiff(this->first_sample_pm_info_obj_vector,
            ports_errors);

    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_NAME_PM_DIFF,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);

    if (rc)
        IBDIAGNET_RETURN(rc);

    rc = this->p_ibdiag->DumpPortCountersDeltaToCSV(*this->p_csv_out,
                                    this->first_sample_pm_info_obj_vector,
                                    m_check_counters_bitset,
                                    ports_errors);
    if (!ports_errors.empty())
        INFO_PRINT("Negative values were found while delta for port counters was calculated\n");

    this->AnalyzeCheckResults(ports_errors,
            CHECK_NAME_PM_DELTA,
            rc,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);

    //BER check - only if needed
    if (this->to_calc_ber) {
        rc2 = this->p_ibdiag->CalcBERErrors(this->first_sample_pm_info_obj_vector,
                (u_int64_t)(this->ber_threshold),
                this->diff_time_between_samples,
                ports_errors,
                *this->p_csv_out);
        rc = this->AnalyzeCheckResults(ports_errors,
                CHECK_NAME_BER,
                rc2,
                IBDIAG_ERR_CODE_CHECK_FAILED,
                this->num_errors,
                this->num_warnings);

        rc2 = this->p_ibdiag->CalcPhyTest(this->first_sample_pm_info_obj_vector,
                                          this->diff_time_between_samples,
                                          *this->p_csv_out);
    }
    IBDIAGNET_RETURN(rc);
}

/******************************************************/

/////
//NVLTelemetryStage public methods:
/////

NVLinkStage::NVLinkStage(IBDiag *p_ibdiag)
    : Stage(STAGE_NAME_NVLINK, p_ibdiag), m_reset_counters(false)
{
    IBDIAGNET_ENTER;
    this->MarkStageAsSkipped(); //this stage will be skipped by default
    IBDIAGNET_RETURN_VOID;
}

int NVLinkStage::CleanResources()
{
	IBDIAGNET_ENTER;
	this->p_ibdiag->GetIBDMExtendedInfoPtr()->CleanNVLDB();
	IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int NVLinkStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int NVLinkStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;

    int                         rc, rc2;
    list_p_fabric_general_err   errors;

    rc2 = this->p_ibdiag->BuildNVLAnycastLIDInfo(errors);
    rc = this->AnalyzeCheckResults(errors,
            CHECK_NAME_NVL_ANYCAST_LID_INFO_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (this->p_ibdiag->DumpAnycastLIDInfoToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing NVL Anycast LID Info data to DB failed\n");
        ++this->num_errors;
    }

    rc2 = this->p_ibdiag->BuildNVLHBFConfig(errors);
    rc = this->AnalyzeCheckResults(errors,
            CHECK_NAME_NVL_HBF_CONFIG_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (this->p_ibdiag->DumpNVLHBFConfigToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing NVL Anycast LID Info data to DB failed\n");
        ++this->num_errors;
    }

    rc2 = this->p_ibdiag->BuildNVLContainAndDrainInfo(errors);
    rc = this->AnalyzeCheckResults(errors,
            CHECK_NAME_NVL_CONTAIN_AND_DRAIN_INFO_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (this->p_ibdiag->DumpNVLContainAndDrainInfoToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing NVL Contain and Drain Info data to DB failed\n");
        ++this->num_errors;
    }

    rc2 = this->p_ibdiag->BuildNVLContainAndDrainPortState(errors);
    rc = this->AnalyzeCheckResults(errors,
            CHECK_NAME_NVL_CONTAIN_AND_DRAIN_PORT_STATE_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (this->p_ibdiag->DumpNVLContainAndDrainPortStateToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing NVL Contain and Drain Port State data to DB failed\n");
        ++this->num_errors;
    }

    rc2 = this->p_ibdiag->BuildNVLClassPortInfo(errors);
    rc = this->AnalyzeCheckResults(errors,
            CHECK_NAME_NVL_CLASS_PORT_INFO_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (this->p_ibdiag->DumpNVLClassPortInfoToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing NVL Class Port Info data to DB failed\n");
        ++this->num_errors;
    }


    rc2 = this->p_ibdiag->BuildNVLReductionInfo(errors);
    rc = this->AnalyzeCheckResults(errors,
            CHECK_NAME_NVL_REDUCTION_INFO_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (this->p_ibdiag->DumpNVLReductionInfoToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing NVL Reduction Info data to DB failed\n");
        ++this->num_errors;
    }

    rc2 = this->p_ibdiag->BuildNVLReductionPortInfo(errors);
    rc = this->AnalyzeCheckResults(errors,
            CHECK_NAME_NVL_REDUCTION_PORT_INFO_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (this->p_ibdiag->DumpNVLReductionPortInfoToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing NVL Reduction Port Info data to DB failed\n");
        ++this->num_errors;
    }

    rc2 = this->p_ibdiag->BuildNVLReductionForwardingTable(errors);
    rc = this->AnalyzeCheckResults(errors,
            CHECK_NAME_NVL_REDUCTION_FORWARDING_TABLE_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (this->p_ibdiag->DumpNVLReductionForwardingTableToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing NVL Forwarding_Table Info data to DB failed\n");
        ++this->num_errors;
    }

    rc2 = this->p_ibdiag->BuildNVLPenaltyBoxConfig(errors);
    rc = this->AnalyzeCheckResults(errors,
            CHECK_NAME_NVL_PENALTY_BOX_CONFIG_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (this->p_ibdiag->DumpNVLPenaltyBoxConfigToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing NVL Penalty Box Config data to DB failed\n");
        ++this->num_errors;
    }

    rc2 = this->p_ibdiag->BuildNVLReductionConfigureMLIDMonitors(errors);
    rc = this->AnalyzeCheckResults(errors,
            CHECK_NAME_NVL_REDUCTION_CONFIGURE_MLID_MONITORS_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (this->p_ibdiag->DumpNVLReductionConfigureMLIDMonitorsToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing NVL Reduction Configure MLID Monitors info to DB failed\n");
        ++this->num_errors;
    }

    rc2 = this->p_ibdiag->BuildNVLReductionCounters(errors, this->m_reset_counters);
    rc = this->AnalyzeCheckResults(errors,
            CHECK_NAME_NVL_REDUCTION_COUNTERS_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (this->p_ibdiag->DumpNVLReductionCountersToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing NVL Reduction Counters info to DB failed\n");
        ++this->num_errors;
    }

    rc2 = this->p_ibdiag->BuildNVLReductionRoundingMode(errors);
    rc = this->AnalyzeCheckResults(errors,
            CHECK_NAME_NVL_REDUCTION_ROUNDING_MODE_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (this->p_ibdiag->DumpNVLReductionRoundingModeToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing NVL Reduction Rounding Mode data to DB failed\n");
        ++this->num_errors;
    }

    IBDIAGNET_RETURN(rc);
}

/******************************************************/

/////
//RoutingStage public methods:
/////
RoutingStage::RoutingStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_ROUTING, p_ibdiag),
        retrieved_switch_info_ok(false), retrieved_uc_fdbs_ok(false),
        retrieved_mc_fdbs_ok(false),
        retrieved_plft_data_ok(false), retrieved_ar_data_ok(false),
        retrieved_rn_data_ok(false), ar_enabled(false), is_fat_tree(false), use_vlr(false),
        use_vs_route(true), check_mcast_credloop(false), check_sl(false),
        dump_only(false), dump_only_skip_routing_tables(false),
        static_ca2ca(false), skip_slvl(false), default_sl(0), num_ar_switches(0),
        num_hbf_switches(0), is_entry_plane_filter_sup(false), is_rail_filter_sup(false),
        is_end_port_plane_filter_sup(false)
{
    IBDIAGNET_ENTER;
    this->MarkStageAsSkipped(); //this stage will be skipped by default
    IBDIAGNET_RETURN_VOID;
}

int RoutingStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

void RoutingStage::DumpHBFTables()
{
    IBDIAGNET_ENTER;

    if (this->p_ibdiag->DumpHBFConfigToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing of HBF table failed with: %s\n", p_ibdiag->GetLastError());
        ++this->num_errors;
    }

    IBDIAGNET_RETURN_VOID;
}

int RoutingStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    int rc2 = IBDIAG_SUCCESS_CODE;
    int rc3 = IBDIAG_SUCCESS_CODE;
    list_p_fabric_general_err nodes_errors;

    string parser_output;
    if(!sa_dump_file.empty()) {
        INFO_PRINT("Load SA Dump file (%s).\n", sa_dump_file.c_str());

        //To check vport and aguids in the sadump file,
        //need to call VirtualizationStage & AGuidStage before this function
        rc = p_ibdiag->ParseSADumpFile(sa_dump_file, parser_output);
        LOG_PRINT("%s\n", parser_output.c_str());

        if (rc) {
            ERR_PRINT("Failed to parse SA Dump file (%s).\n", sa_dump_file.c_str());
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }
    }

    if (use_vs_route) {

        if (!this->p_ibdiag->ext_node_info_check_results)
            if (this->p_ibdiag->WriteVL2VLFile(VL2VL_FILE)) {
                ERR_PRINT("Writing VL2VL file failed\n");
                ++this->num_errors;
            }

        //bring PLFT Data
        unsigned int supportedDev;
        rc2  = this->p_ibdiag->BuildPLFTData(nodes_errors, supportedDev);
        if (rc2 == IBDIAG_SUCCESS_CODE){
            this->retrieved_plft_data_ok = true;
            INFO_PRINT("PLFT is enabled on %d switches.\n",supportedDev);
        }
        rc = this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_PLFT_DATA_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        if (!rc)
            if (this->p_ibdiag->WritePLFTFile(PLFT_FILE)) {
                ERR_PRINT("Writing PLFT file failed\n");
                ++this->num_errors;
            }

        //bring AR Data
        AdditionalRoutingDataMap routing_data_map;
        AdditionalRoutingDataMap* p_routing_data_map = NULL;
        p_routing_data_map = &routing_data_map;

        rc2  = this->p_ibdiag->BuildARData(nodes_errors,
                                              supportedDev,
                                              p_routing_data_map,
                                              this->dump_only_skip_routing_tables);
        //includes switches with either HBF or AR enabled
        this->num_ar_switches = (u_int32_t)supportedDev;
        INFO_PRINT("Adaptive Routing is enabled on %d switches.\n", supportedDev);
        if (rc2 == IBDIAG_SUCCESS_CODE){
            this->retrieved_ar_data_ok = true;
            this->ar_enabled = (supportedDev > 0);
        }
        int ar_result_rc = this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_AR_DATA_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);


        //we have to retrive HBF data before the FAR file is written
        supportedDev = 0;
        rc2  = this->p_ibdiag->BuildHBFData(nodes_errors,
                                              supportedDev);
        INFO_PRINT("Hashed Based Forwarding is enabled on %d switches.\n", supportedDev);
        this->num_hbf_switches = (u_int32_t)supportedDev;
        rc = this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_HBF_DATA_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        this->DumpHBFTables();

        if (!this->dump_only_skip_routing_tables) {
            if (!ar_result_rc) {
                if (this->p_ibdiag->WriteARFile(AR_FILE)) {
                    ERR_PRINT("Writing AR file failed\n");
                    ++this->num_errors;
                }
            }

            if (this->p_ibdiag->WriteFARFile(&routing_data_map, FAR_FILE, false)) {
                ERR_PRINT("Writing FAR file failed\n");
                ++this->num_errors;
            }

            if (this->p_ibdiag->WriteFARFile(&routing_data_map, FAR_FLID_FILE, true)) {
                ERR_PRINT("Writing FAR FLID file failed\n");
                ++this->num_errors;
            }
        }

        INFO_PRINT("Routing Notification is supported on %d switches.\n",
                AdditionalRoutingData::sw_supports_rn_count);

        if (AdditionalRoutingData::sw_supports_rn_count == 0)
            PRINT("\n");

        if (AdditionalRoutingData::sw_supports_rn_count != 0) {

            rc2 = this->p_ibdiag->BuildRNData(nodes_errors, &routing_data_map);

            if (rc2 == IBDIAG_SUCCESS_CODE)
                this->retrieved_rn_data_ok = true;

            this->AnalyzeCheckResults(nodes_errors,
                    CHECK_NAME_ROUTE_RN_DATA_RETRIEVE,
                    rc2,
                    IBDIAG_ERR_CODE_FABRIC_ERROR,
                    this->num_errors,
                    this->num_warnings,
                    true);

            rc2 = this->p_ibdiag->WriteRNFile(nodes_errors, &routing_data_map, RN_FILE);

            if (rc2)
                ERR_PRINT("Writing RN file failed\n");

            this->AnalyzeCheckResults(nodes_errors,
                    CHECK_NAME_ROUTE_DUMP_RN_DATA,
                    rc2,
                    IBDIAG_ERR_CODE_FABRIC_ERROR,
                    this->num_errors,
                    this->num_warnings,
                    true); // only_warning?
        }

        // Build pFRN Data
        INFO_PRINT("Build pFRN Data\n");
        rc2 = this->p_ibdiag->BuildpFRNData(nodes_errors);
        this->p_ibdiag->Dump_pFRNConfigToCSV(*this->p_csv_out);

        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_PFRN_DATA_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        INFO_PRINT("Build N2N Class Port Info\n");
        rc2 = this->p_ibdiag->BuildN2NClassPortInfo(nodes_errors);
        this->p_ibdiag->Dump_N2NClassPortInfoToCSV(*this->p_csv_out);

        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_N2N_CPI_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        INFO_PRINT("Build N2N Neighbors Info\n");
        rc2 = this->p_ibdiag->BuildNeighborsInfo(nodes_errors);
        this->p_ibdiag->Dump_NeighborsInfoToCSV(*this->p_csv_out);

        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_NEIGHBORS_INFO_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        INFO_PRINT("Build N2N Key Info\n");
        rc2 = this->p_ibdiag->BuildN2NKeyInfo(nodes_errors);
        this->p_ibdiag->Dump_N2NKeyInfoToCSV(*this->p_csv_out);

        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_N2N_KEY_INFO_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        // Entry Plane Filter
        INFO_PRINT("Build Entry Plane Filter\n");
        rc2 = this->p_ibdiag->BuildEntryPlaneFilter(nodes_errors, this->is_entry_plane_filter_sup);

        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ENTRY_PLANE_FILTER_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        if (this->is_entry_plane_filter_sup)
            if (this->p_ibdiag->WriteEPFFile(EPF_FILE)) {
                ERR_PRINT("Writing Entry Plane Filter file failed\n");
                ++this->num_errors;
            }

        // Rail Filter
        INFO_PRINT("Build Rail Filter\n");
        rc2 = this->p_ibdiag->BuildRailFilter(nodes_errors, this->is_rail_filter_sup);

        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_RAIL_FILTER_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        if (this->is_rail_filter_sup)
            if (this->p_ibdiag->WriteRailFilterFile(RAIL_FILTER_FILE)) {
                ERR_PRINT("Writing Rail Filter file failed\n");
                ++this->num_errors;
            }

        // Linear Forwarding Table Split
        INFO_PRINT("Build Linear Forwarding Table Split\n");
        rc2 = this->p_ibdiag->BuildLFTSplit(nodes_errors);
        this->p_ibdiag->DumpLFTSplitToCSV(*this->p_csv_out);

        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_LFT_SPLIT,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        // End Port Plane Filter
        INFO_PRINT("Build End Port Plane Filter\n");
        rc2 = this->p_ibdiag->BuildEndPortPlaneFilter(nodes_errors, this->is_end_port_plane_filter_sup);

        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_END_PORT_PLANE_FILTER_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        if (this->is_end_port_plane_filter_sup)
            if (this->p_ibdiag->WriteEndPortPlaneFilterFile(END_PORT_PF_FILE)) {
                ERR_PRINT("Writing End Port Plane Filter file failed\n");
                ++this->num_errors;
            }
    }

    if (this->dump_only_skip_routing_tables) {
        LOG_INFO_PRINT("Requested to skip routing tables (LFTs) retrieving.\n");
    } else {

        //bring unicast FDBs info
        INFO_PRINT("Build Unicast FDBS Info\n");
        rc2  = this->p_ibdiag->BuildUCFDBSInfo(nodes_errors);
        if (rc2 == IBDIAG_SUCCESS_CODE)
            this->retrieved_uc_fdbs_ok = true;
        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_UCFDBS_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (!rc)
            if (this->p_ibdiag->WriteUCFDBSFile(UC_FDBS_FILE)) {
                ERR_PRINT("Writing unicast FDBS file failed\n");
                ++this->num_errors;
            }

        if (!this->retrieved_uc_fdbs_ok)    //abort
            IBDIAGNET_RETURN(rc);

        //bring multicast FDBs info
        INFO_PRINT("Build Multicast FDBS Info\n");
        rc2  = this->p_ibdiag->BuildMCFDBSInfo(nodes_errors);
        if (rc2 == IBDIAG_SUCCESS_CODE)
            this->retrieved_mc_fdbs_ok = true;
        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_MCASTDBS_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (!rc)
            if (this->p_ibdiag->WriteMCFDBSFile(MC_FDBS_FILE)) {
                ERR_PRINT("Writing multicast FDBS file failed\n");
                ++this->num_errors;
            }

        //Write SLVL File
        if (skip_slvl) {
            INFO_PRINT("Skipping SLVL Table Building\n");
        } else if(use_vlr || use_vs_route || check_mcast_credloop || check_sl) {
            INFO_PRINT("Build SLVL Table\n");
            rc2  = this->p_ibdiag->WriteSLVLFile(SLVL_FILE,
                    nodes_errors);
            SCREEN_PRINT("\n");
            rc = this->AnalyzeCheckResults(nodes_errors,
                    CHECK_NAME_ROUTE_DUMP_SLVL_TABLE,
                    rc2,
                    IBDIAG_ERR_CODE_FABRIC_ERROR,
                    this->num_errors,
                    this->num_warnings,
                    true);

        }

        // bring VL Arbitration Table info
        INFO_PRINT("Build VL Arbitration Table\n");
        rc2 = this->p_ibdiag->BuildVLArbitrationTable(nodes_errors);
        this->p_ibdiag->DumpVLArbitrationToCSV(*this->p_csv_out);

        SCREEN_PRINT("\n");
        rc3 = this->AnalyzeCheckResults(nodes_errors,
                CHECK_VL_ARBITRATION_TABLE_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
    }

    if (rc)
        IBDIAGNET_RETURN(rc);
    if (rc3)
        IBDIAGNET_RETURN(rc3);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int RoutingStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    if (this->dump_only || this->dump_only_skip_routing_tables) {
        LOG_INFO_PRINT("Requested to skip Routing checks\n");
        IBDIAGNET_RETURN(rc);
    }

    if ((!this->retrieved_uc_fdbs_ok) || (!this->retrieved_mc_fdbs_ok)) {
        ERR_PRINT("Skipping Routing checks because not all information retrieved successfully\n");
        ++this->num_errors;
        IBDIAGNET_RETURN(rc);
    }

    this->p_ibdiag->SetDefaultSL(default_sl);

    if (use_vs_route) {

        if ((!this->p_ibdiag->retrieved_ext_node_info_ok) ||
            (!this->retrieved_plft_data_ok) ||
            (!this->retrieved_ar_data_ok) ) {
            ERR_PRINT("Skipping Adaptive Routing checks because" \
                      " not all information retrieved successfully\n");
            ++this->num_errors;
            this->ar_enabled = false;
        }
    }

    LOG_PRINT("\n\n");
    string routing_output;

    this->p_ibdiag->CountSkipRoutingChecksNodes(routing_output);
    this->ParseIBDMOutput(routing_output);
    LOG_PRINT("%s\n", routing_output.c_str());

    if (check_sl) {
        LOG_PRINT("###################################\n");
        LOG_INFO_PRINT("SL2VL table check:\n");
        LOG_PRINT("###################################\n");
        this->p_ibdiag->CheckSL2VLTables(routing_output);
        this->ParseIBDMOutput(routing_output);
        LOG_PRINT("%s\n", routing_output.c_str());
    }

    LOG_PRINT("###################################\n");
    LOG_INFO_PRINT("Fabric Qualities Report:\n");
    LOG_PRINT("###################################\n");
    this->p_ibdiag->ReportFabricQualities(routing_output, this->p_base_path->c_str(),
                                          ar_enabled, static_ca2ca);
    this->ParseIBDMOutput(routing_output);
    LOG_PRINT("%s\n", routing_output.c_str());

    if (use_vs_route && this->ar_enabled) {

        LOG_PRINT("###################################\n");
        LOG_INFO_PRINT("Fabric AR Empty Groups Report:\n");
        LOG_PRINT("###################################\n");
        this->p_ibdiag->ReportFabricAREmptyGroups(routing_output);
        this->ParseIBDMOutput(routing_output);
        LOG_PRINT("%s\n", routing_output.c_str());

        LOG_PRINT("###################################\n");
        LOG_INFO_PRINT("Fabric AR Connectivity Report:\n");
        LOG_PRINT("###################################\n");
        this->p_ibdiag->ReportFabricARConnectivity(routing_output);
        this->ParseIBDMOutput(routing_output);
        LOG_PRINT("%s\n", routing_output.c_str());

        // AR validation - validate that AR routing tables are configured correctly.
        LOG_PRINT("###################################\n");
        LOG_INFO_PRINT("AR Validation Report:\n");
        LOG_PRINT("###################################\n");
        this->p_ibdiag->ReportFabricARValidation(routing_output);
        this->ParseIBDMOutput(routing_output);
        LOG_PRINT("%s\n", routing_output.c_str());
    }

    LOG_PRINT("###################################\n");
    LOG_INFO_PRINT("Credit Loops Report:\n");
    LOG_PRINT("###################################\n");
    if(check_mcast_credloop)
        CredLoopMode(0,1);

    this->p_ibdiag->ReportCreditLoops(routing_output, this->is_fat_tree,
                                      this->ar_enabled);
    this->ParseIBDMOutput(routing_output);
    LOG_PRINT("%s\n", routing_output.c_str());
    LOG_PRINT("\n\n");

    // pFRN validations
    list_p_fabric_general_err pfrn_errors;
    int rc2 = 0;

    rc2 = this->p_ibdiag->pFRNSupportAndTrapsValidation(pfrn_errors);
    this->AnalyzeCheckResults(pfrn_errors,
            CHECK_NAME_PFRN_SUPPORT_AND_TRAPS,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    rc2 = this->p_ibdiag->pFRNNeighborsValidation(pfrn_errors);
    this->AnalyzeCheckResults(pfrn_errors,
            CHECK_NAME_PFRN_NEIGHBORS,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    rc2 = this->p_ibdiag->ARGroupsUniformValidation(pfrn_errors);
    this->AnalyzeCheckResults(pfrn_errors,
            CHECK_NAME_AR_GROUPS_UNIFORM,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    list_p_fabric_general_err errors_list;

    // EntryPlaneFilter validations
    rc2 = this->p_ibdiag->EntryPlaneFilterValidation(errors_list);
    this->AnalyzeCheckResults(errors_list,
                              CHECK_NAME_ENTRY_PLANE_FILTER_VALIDATION,
                              rc2,
                              IBDIAG_ERR_CODE_FABRIC_ERROR,
                              this->num_errors,
                              this->num_warnings,
                              true);

    rc2 = this->p_ibdiag->EndPortPlaneFilterValidation(errors_list);
    SCREEN_PRINT("\n");
    this->AnalyzeCheckResults(errors_list,
                              CHECK_NAME_END_PORT_PLANE_FILTER_VALIDATION,
                              rc2,
                              IBDIAG_ERR_CODE_FABRIC_ERROR,
                              this->num_errors,
                              this->num_warnings,
                              true);

    // Static Routing Symmetric Link Validation
    rc2 = this->p_ibdiag->StaticRoutingSymmetricLinkValidation(errors_list);
    this->AnalyzeCheckResults(errors_list,
                              CHECK_NAME_STATIC_ROUTING_ASYMMETRIC_VALIDATION,
                              rc2,
                              IBDIAG_ERR_CODE_FABRIC_ERROR,
                              this->num_errors,
                              this->num_warnings,
                              true);

    if (use_vs_route && this->ar_enabled) {

        // Adaptive Routing Symmetric Link Validation
        rc2 = this->p_ibdiag->AdaptiveRoutingSymmetricLinkValidation(errors_list);
        this->AnalyzeCheckResults(errors_list,
                                  CHECK_NAME_ADAPTIVE_ROUTING_ASYMMETRIC_VALIDATION,
                                  rc2,
                                  IBDIAG_ERR_CODE_FABRIC_ERROR,
                                  this->num_errors,
                                  this->num_warnings,
                                  true);

    }

    IBDIAGNET_RETURN(rc);
}


/******************************************************/
NodesInfoStage::NodesInfoStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_NODES_INFO, p_ibdiag),
        can_send_mads_by_lid(false)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int NodesInfoStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag, this->can_send_mads_by_lid))
            IBDIAGNET_RETURN(1);
    IBDIAGNET_RETURN(0);
}

int NodesInfoStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_NODES_INFO_RETRIEVE);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    if (!g_vs_cap_gmp_run) {
        INFO_PRINT("%s wasn't ran. Skipping %s\n",
                   CHECK_VS_CAP_GMP, CHECK_NAME_NODES_INFO_RETRIEVE);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err nodes_errors;

    if (this->p_ibdiag->DumpVSNodeInfoToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing Nodes Info to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->WriteNodesInfoFile(NODES_INFO_FILE)) {
        ERR_PRINT("Writing Nodes Info file failed\n");
        ++this->num_errors;
    }

    IBDIAGNET_RETURN(rc);
}

int NodesInfoStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_NODES_INFO_FW);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }
    if (!g_vs_cap_gmp_run) {
        INFO_PRINT("%s wasn't ran. Skipping %s\n",
                   CHECK_VS_CAP_GMP, CHECK_NAME_NODES_INFO_FW);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err fw_errors;
    int rc2 = this->p_ibdiag->CheckFWVersion(fw_errors);
    rc = this->AnalyzeCheckResults(fw_errors,
            CHECK_NAME_NODES_INFO_FW,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
Speed_WidthStage::Speed_WidthStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_SPEED_WIDTH, p_ibdiag),
        expected_link_speed(""), expected_link_width("")
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int Speed_WidthStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int Speed_WidthStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    list_p_fabric_general_err links_errors;

    //check link speed
    if (this->expected_link_speed != "") {
        INFO_PRINT("Link Speed Check (Expected value given = %s)\n", this->expected_link_speed.c_str());
    } else {
        INFO_PRINT("Link Speed Check (Compare to supported link speed)\n");
    }
    int rc2 = this->p_ibdiag->CheckLinkSpeed(links_errors, this->expected_link_speed);
    rc = this->AnalyzeCheckResults(links_errors,
            CHECK_NAME_LINK_SPEED,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);

    //check link width
    if (this->expected_link_width != "") {
        INFO_PRINT("Link Width Check (Expected value given = %s)\n", this->expected_link_width.c_str());
    } else {
        INFO_PRINT("Link Width Check (Compare to supported link width)\n");
    }
    rc2 = this->p_ibdiag->CheckLinkWidth(links_errors, this->expected_link_width);
    rc = this->AnalyzeCheckResults(links_errors,
            CHECK_NAME_LINK_WIDTH,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
PKeyStage::PKeyStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_PKEY, p_ibdiag),
        can_send_mads_by_lid(false), skip_aport_membership(false)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

PKeyStage::~PKeyStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int PKeyStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int PKeyStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    list_p_fabric_general_err pkey_errors;
    int rc2 = this->p_ibdiag->BuildPartitionKeys(pkey_errors);
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(pkey_errors,
            CHECK_NAME_PKEY_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (!rc) {
        this->p_ibdiag->BuildPKeyMap(pkey_to_p_port_mem);
    }

    IBDIAGNET_RETURN(rc);
}

int PKeyStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    list_p_fabric_general_err pkey_errors;
    int rc2 = this->p_ibdiag->CheckPartitionKeys(pkey_errors);
    if (!rc2)
        this->p_ibdiag->CheckAPortsPKeys(pkey_errors, this->skip_aport_membership);
    if (this->p_ibdiag->DumpPartitionKeysToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing partition key info to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->WritePKeyFile(PKEY_FILE, pkey_to_p_port_mem)) {
        ERR_PRINT("Writing partition keys info file failed\n");
        ++this->num_errors;
    }

    rc = this->AnalyzeCheckResults(pkey_errors,
        STAGE_NAME_PKEY,
        rc2,
        IBDIAG_ERR_CODE_CHECK_FAILED,
        this->num_errors,
        this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
AGUIDStage::AGUIDStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_AGUID, p_ibdiag),
        can_send_mads_by_lid(false)
{
    IBDIAGNET_ENTER;
    this->MarkStageAsSkipped(); //this stage will be skipped by default
    IBDIAGNET_RETURN_VOID;
}

AGUIDStage::~AGUIDStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int AGUIDStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag, this->can_send_mads_by_lid))
            IBDIAGNET_RETURN(1);
    IBDIAGNET_RETURN(0);
}

int AGUIDStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_AGUID_RETRIEVE);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err aguid_errors;
    int rc2 = this->p_ibdiag->BuildAliasGuids(aguid_errors);
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(aguid_errors,
            CHECK_NAME_AGUID_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (!rc) {
        if (this->p_ibdiag->DumpAliasGUIDToCSV(*this->p_csv_out)) {
            ERR_PRINT("Writing alias gudis info to DB failed\n");
            ++this->num_errors;
        }

        if (this->p_ibdiag->WriteAliasGUIDFile(AGUID_FILE)) {
            ERR_PRINT("Writing alias gudis info file failed\n");
            ++this->num_errors;
        }
    }
    IBDIAGNET_RETURN(rc);
}

int AGUIDStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_AGUID);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err aguid_errors;
    int rc2 = this->p_ibdiag->CheckDuplicatedAliasGuids(aguid_errors);
    rc = this->AnalyzeCheckResults(aguid_errors,
            STAGE_NAME_AGUID,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
MlnxCntrsStage::MlnxCntrsStage(IBDiag *p_ibdiag) :
    Stage(STAGE_NAME_MLNX_CNTRS, p_ibdiag),
    can_send_mads_by_lid(false), to_reset_counters(false)
{
    IBDIAGNET_ENTER;
    this->MarkStageAsSkipped(); //this stage will be skipped by default
    IBDIAGNET_RETURN_VOID;
}

MlnxCntrsStage::~MlnxCntrsStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int MlnxCntrsStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag,
        this->can_send_mads_by_lid))
            IBDIAGNET_RETURN(1);
    IBDIAGNET_RETURN(0);
}

int MlnxCntrsStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_MLNX_CNTRS_RETRIEVE);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err mlnx_cntrs_errors;
    //if need to reset
    if (this->to_reset_counters) {
        int rc2  = this->p_ibdiag->ResetDiagnosticCounters(mlnx_cntrs_errors);
        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(mlnx_cntrs_errors,
                CHECK_NAME_MLNX_CNTRS_RESET,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
    }
    if (rc)
        IBDIAGNET_RETURN(rc);

    int rc2 = this->p_ibdiag->BuildVsDiagnosticCounters(
            mlnx_cntrs_errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(mlnx_cntrs_errors,
            CHECK_NAME_MLNX_CNTRS_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    IBDIAGNET_RETURN(rc);
}

int MlnxCntrsStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    // now when we have the calculated counters, we can print to the files
    if (this->p_ibdiag->DumpDiagnosticCountersToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing ports counters to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->WriteMlnxCntrsFile(MLNX_CNTRS_FILE)) {
        ERR_PRINT("Writing Mellanox diagnostic counters file failed\n");
        ++this->num_errors;
    }

    IBDIAGNET_RETURN(rc);
}

//**********************************************
VirtualizationStage::VirtualizationStage(IBDiag *p_ibdiag) :
    Stage(STAGE_NAME_VIRTUALIZATION, p_ibdiag),
    can_send_mads_by_lid(false)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

VirtualizationStage::~VirtualizationStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int VirtualizationStage::CleanResources()
{
    IBDIAGNET_ENTER;

    for (IBNode *p_ca : this->p_ibdiag->GetDiscoverFabricPtr()->HCAs) {
        if (!p_ca)
            continue;

        for (phys_port_t i = 1; i <= p_ca->numPorts; ++i) {
            IBPort *p_port = p_ca->getPort(i);
            if (!p_port)
                continue;

            p_port->CleanVPorts();
        }
    }
    this->p_ibdiag->GetIBDMExtendedInfoPtr()->CleanVPortDB();
    this->p_ibdiag->GetDiscoverFabricPtr()->ResetVPortIndex();
    this->p_ibdiag->GetDiscoverFabricPtr()->CleanVNodes();
    this->p_ibdiag->GetIBDMExtendedInfoPtr()->CleanVNodeDB();
    this->p_ibdiag->GetDiscoverFabricPtr()->ResetVNodeIndex();

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int VirtualizationStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag,
        this->can_send_mads_by_lid))
            IBDIAGNET_RETURN(1);
    IBDIAGNET_RETURN(0);
}

int VirtualizationStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_VIRTUALIZATION_RETRIEVE);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err vport_errors;
    int rc2 = p_ibdiag->BuildVirtualization(vport_errors);
    PRINT("\n");
    rc = this->AnalyzeCheckResults(vport_errors,
                                   CHECK_NAME_VIRTUALIZATION_RETRIEVE,
                                   rc2,
                                   IBDIAG_ERR_CODE_FABRIC_ERROR,
                                   this->num_errors,
                                   this->num_warnings,
                                   true);


    IBDIAGNET_RETURN(rc);
}

int VirtualizationStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    list_p_fabric_general_err vport_errors;
    this->p_ibdiag->CheckAndSetVPortLid(vport_errors);
    rc = this->AnalyzeCheckResults(vport_errors,
            CHECK_NAME_VPORT_LIDS,
            rc,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);

    this->p_ibdiag->CheckVPortDuplicatedGuids(vport_errors);
    rc = this->AnalyzeCheckResults(vport_errors,
            CHECK_NAME_VPORT_DUPLICATED_GUID,
            rc,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);

    if (this->p_ibdiag->DumpVNodeInfoToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing VNodes to DB failed\n");
        ++this->num_errors;
    }
    if (this->p_ibdiag->DumpVPortInfoToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing VPorts to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpVPortsGUIDInfoToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing VPorts GUID Info to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->WriteVPortsFile(VPORTS_FILE)) {
        ERR_PRINT("Writing vports info file failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->WriteVPortsPKeyTblFile(VPORTS_PKEY_FILE)) {
        ERR_PRINT("Writing vports pkey table file failed\n");
        ++this->num_errors;
    }

    IBDIAGNET_RETURN(rc);
}

TemperatureSensing::TemperatureSensing (IBDiag *p_ibdiag) :
    Stage(STAGE_NAME_TEMPSENSING, p_ibdiag)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

TemperatureSensing::~TemperatureSensing()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int TemperatureSensing::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int TemperatureSensing::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    list_p_fabric_general_err temp_sensing_errors;

    int rc2 = this->p_ibdiag->BuildSMPTempSensing(
            temp_sensing_errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(temp_sensing_errors,
            STAGE_NAME_TEMPSENSING,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    IBDIAGNET_RETURN(rc);
}

int TemperatureSensing::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    if (this->p_ibdiag->DumpTempSensingToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing ports counters to DB failed\n");
        ++this->num_errors;
    }

    IBDIAGNET_RETURN(rc);
}

/*************************************************************************************************/
QoSStage::QoSStage (IBDiag *p_ibdiag) :
    Stage(STAGE_NAME_QOS, p_ibdiag)
{
    IBDIAGNET_ENTER;
    //mark this stage as skipped by default
    MarkStageAsSkipped();
    IBDIAGNET_RETURN_VOID;
}

QoSStage::~QoSStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int QoSStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int QoSStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;

    list_p_fabric_general_err errors;

    int rc = this->p_ibdiag->BuildSMPQoSConfigSL(errors, false);

    SCREEN_PRINT("\n");
    int rc2 = this->AnalyzeCheckResults(errors,
            CHECK_NAME_QOS_CONFIG_SL_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    rc = this->p_ibdiag->BuildSMPQoSConfigSL(errors, true);

    SCREEN_PRINT("\n");
    int rc3 = this->AnalyzeCheckResults(errors,
            CHECK_NAME_VPORTS_QOS_CONFIG_SL_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (rc2)
        IBDIAGNET_RETURN(rc2);
    if (rc3)
        IBDIAGNET_RETURN(rc3);

    rc2 = this->p_ibdiag->BuildSMPQoSConfigVL(errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(errors,
            CHECK_NAME_QOS_CONFIG_VL_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int QoSStage::RunCheck()
{
    IBDIAGNET_ENTER;

    list_p_fabric_general_err aport_qos_errors;
    int rc = IBDIAG_SUCCESS_CODE;

    if (this->p_ibdiag->DumpQoSConfigSLToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing QoS config SL to csv file failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpVPortQoSConfigSLToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing VPorts QoS config SL to csv file failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpQoSConfigVLToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing QoS config VL to csv file failed\n");
        ++this->num_errors;
    }

    int rc1 = p_ibdiag->CheckAPortsQosSymmetry(aport_qos_errors);
    rc = this->AnalyzeCheckResults(aport_qos_errors,
            CHECK_NAME_APORT_QOS_SYMMETRY,
            rc1,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings,
            true);

    IBDIAGNET_RETURN(rc);
}


RoutersStage::RoutersStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_ROUTERS, p_ibdiag), p_FLIDsManager(NULL)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

RoutersStage::~RoutersStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

const char* RoutersStage::GetLastError()
{
    IBDIAGNET_ENTER;

    if (this->p_FLIDsManager)
        IBDIAGNET_RETURN(this->p_FLIDsManager->GetLastError());

    IBDIAGNET_RETURN(Stage::GetLastError());
}

int RoutersStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int RoutersStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    list_p_fabric_general_err errors;

    int rc2 = this->p_ibdiag->BuildRouterInfo(errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(errors,
                                   CHECK_ROUTERS_INFO_DB,
                                   rc2,
                                   IBDIAG_ERR_CODE_FABRIC_ERROR,
                                   this->num_errors,
                                   this->num_warnings);
    errors.clear();
    rc2 = this->p_ibdiag->BuildAdjSiteLocalSubnetsTable(errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(errors,
                                   CHECK_ROUTERS_TABLE,
                                   rc2,
                                   IBDIAG_ERR_CODE_FABRIC_ERROR,
                                   this->num_errors,
                                   this->num_warnings);

    SCREEN_PRINT("\n");
    rc2 = this->p_ibdiag->BuildAdjSubnetsRouterLIDInfoTable(errors);
    rc = this->AnalyzeCheckResults(errors,
                                   CHECK_RETRIEVE_ADJ_SUBNETS_FLID,
                                   rc2,
                                   IBDIAG_ERR_CODE_FABRIC_ERROR,
                                   this->num_errors,
                                   this->num_warnings);

    SCREEN_PRINT("\n");
    rc2 = this->p_ibdiag->BuildRouterLIDTable(errors);
    rc = this->AnalyzeCheckResults(errors,
                                   CHECK_RETRIEVE_ROUTRS_FLID,
                                   rc2,
                                   IBDIAG_ERR_CODE_FABRIC_ERROR,
                                   this->num_errors,
                                   this->num_warnings);


    SCREEN_PRINT("\n");
    rc2 = this->p_ibdiag->BuildARGroupToRouterLIDTable(errors);
    rc = this->AnalyzeCheckResults(errors,
                                   CHECK_RETRIEVE_AR_GROUP_TO_ROUTER_FLID,
                                   rc2,
                                   IBDIAG_ERR_CODE_FABRIC_ERROR,
                                   this->num_errors,
                                   this->num_warnings);

    IBDIAGNET_RETURN(rc);
}

int RoutersStage::RunCheck()
{
    IBDIAGNET_ENTER;

    if (this->p_ibdiag->DumpRouterInfoToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing routers info to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpRouterAdjSiteLocalSubnetTableToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing routers adjacent site local subnets table to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpRouterNextHopToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing routers next hop table to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpAdjSubnetsRouterLIDInfoTableToCSV(*this->p_csv_out)) {
        ERR_PRINT("Writing routers adjacent subnets FLID data to DB failed\n");
        ++this->num_errors;
    }

    this->CheckFLIDs();

    IBDIAGNET_RETURN(0);
}

void RoutersStage::CheckFLIDs()
{
    IBDIAGNET_ENTER;

    p_FLIDsManager = new FLIDsManager(this->p_ibdiag);

    list_p_fabric_general_err errorsList;
    int rc = p_FLIDsManager->CheckLocalSubnet(errorsList);
    this->AnalyzeCheckResults(errorsList,
            CHECK_LOCAL_SUBNET_FLID,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            false);

    if (!p_FLIDsManager->IsApplicable()) {
        INFO_PRINT("Skipping FLID verification\n");

        delete this->p_FLIDsManager;
        //last error will be taken from ibdiag
        this->p_FLIDsManager = NULL;
        IBDIAGNET_RETURN_VOID;
    }

    if (this->p_ibdiag->WriteARGroupToRouterFLIDData(AR_GROUP_TO_FLID_DUMP_FILE)) {
        ERR_PRINT("Writing routers adaptive routing groups to FLIDs data failed\n");
        ++this->num_errors;
    }

    //if network has any router we should have information about FLIDs
    rc = p_FLIDsManager->CheckAdjacentSubnets(errorsList);
    this->AnalyzeCheckResults(errorsList,
            CHECK_ADJ_SUBNETS_FLID,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            false);

    rc = p_FLIDsManager->CheckHCAsAndSwitches(errorsList);
    this->AnalyzeCheckResults(errorsList,
            CHECK_HCA_SWITCHES_FLID,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            false);

    //should be called after CheckAdjacentSubnets
    rc = p_FLIDsManager->CheckPFRNSettings(errorsList);
    this->AnalyzeCheckResults(errorsList,
            CHECK_PFRN_SETTINGS,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            false);


    rc = p_FLIDsManager->CheckRemoteEnabledFLIDs(errorsList);
    this->AnalyzeCheckResults(errorsList,
            CHECK_REMOTE_ENABLED_FLIDS,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            false);

    if (p_FLIDsManager->Dump(FLID_DUMP_FILE)) {
        ERR_PRINT("Failed to dump FLID data with: %s\n", p_FLIDsManager->GetLastError());
        this->num_errors++;
    }

    INFO_PRINT("Additional data could be found in the output(ibdiagnet2.fat_tree) "
            "generated by the Fat-Tree validation stage\n");

    delete this->p_FLIDsManager;
    //last error will be taken from ibdiag
    this->p_FLIDsManager = NULL;

    IBDIAGNET_RETURN_VOID;
}

SharpStage::SharpStage (IBDiag *p_ibdiag):
    Stage(STAGE_NAME_SHARP, p_ibdiag), can_send_mads_by_lid(false),
    dump_counters(false), dump_hba_port_counters(false), clear_counters(false),
    check_trees(false), version(0),
    p_sharp_mngr(NULL), sharp_versions_counter(SHARP_CONTROL_VERSION_MAX, 0), num_of_llts(0),
    num_of_sats(0), num_of_active_jobs(0)
{
    IBDIAGNET_ENTER;
    this->MarkStageAsSkipped(); //this stage will be skipped by default

    IBDIAGNET_RETURN_VOID;
}

SharpStage::~SharpStage()
{
    IBDIAGNET_ENTER;
    delete p_sharp_mngr;
    IBDIAGNET_RETURN_VOID;
}

void SharpStage::CountSharpVersionsAndNumOfActiveJobs()
{
    IBDIAGNET_ENTER;

    u_int32_t accum[AN_ACTIVE_JOBS_DWORDS_NUM] = {0};

    const list_sharp_an &sharp_ans = this->p_sharp_mngr->GetSharpANList();
    for (list_sharp_an::const_iterator an_iter = sharp_ans.begin(); an_iter != sharp_ans.end();
         ++an_iter) {
         SharpAggNode *p_sharp_aggnode = *an_iter;
         if (!p_sharp_aggnode)
             continue;

         const AM_ANActiveJobs &an_active_jobs = p_sharp_aggnode->GetANActiveJobs();

         for (uint i = 0; i < AN_ACTIVE_JOBS_DWORDS_NUM; ++i)
             accum[i] |= an_active_jobs.active_jobs[i];

         AM_ANInfo cur_an = p_sharp_aggnode->GetANInfo();
         u_int16_t bit = 0x1;
         if (!(cur_an.active_sharp_version_bit_mask & bit))
             this->sharp_versions_counter[0]++;
         for (u_int16_t version_idx = 1; version_idx < SHARP_CONTROL_VERSION_MAX; version_idx++) {
              if (cur_an.active_sharp_version_bit_mask & (bit << version_idx))
                  this->sharp_versions_counter[version_idx]++;
         }
    }

    this->num_of_active_jobs = 0;
    for (uint i = 0; i < AN_ACTIVE_JOBS_DWORDS_NUM; ++i) {
        u_int32_t x = accum[i];
        for (uint j = 0; j < sizeof(accum[0]) * 8; ++j, x >>= 1)
            this->num_of_active_jobs += (x & 0x01);
    }

    IBDIAGNET_RETURN_VOID
}

void SharpStage::CountSharpTreeTypes()
{
    IBDIAGNET_ENTER;

    const list_sharp_an& sharp_ans = this->p_sharp_mngr->GetSharpANList();

    for (list_sharp_an::const_iterator an_it = sharp_ans.begin(); an_it != sharp_ans.end(); ++an_it) {

        SharpAggNode *p_sharp_an = *an_it;
        if (!p_sharp_an)
          continue;

        const vector_sharp_root_node& root_nodes = p_sharp_an->GetSharpRootNodeVec();

        for (u_int64_t tree_idx = 0; tree_idx < root_nodes.size(); ++tree_idx) {

            SharpTree* root_node = root_nodes[tree_idx];
            if (!root_node || !root_node->GetRoot())
              continue;

            if (root_node->GetRoot()->GetTreeConfig().tree_mode == TREE_MODE_SAT)
              this->num_of_sats++;
            else
              this->num_of_llts++;

        } // roots
    } // ANs

    IBDIAGNET_RETURN_VOID
}

bool SharpStage::GetFabricSummarySection(stringstream &summary_str)
{
    IBDIAGNET_ENTER;

    for (u_int16_t version_idx = 0; version_idx < SHARP_CONTROL_VERSION_MAX; version_idx++) {
         summary_str << "Number of aggregation nodes version " << version_idx + 1 << ": "
                     << this->sharp_versions_counter[version_idx] << endl;
    }
    if (check_trees) {
        summary_str << "Number of LLT aggregation trees: " << num_of_llts << endl;
        summary_str << "Number of streaming aggregation trees: " << num_of_sats << endl;
    }
    summary_str << "Number of active jobs: " << num_of_active_jobs << endl;

    IBDIAGNET_RETURN(true);
}

int SharpStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag,
        this->can_send_mads_by_lid))
            IBDIAGNET_RETURN(1);

    p_sharp_mngr = new SharpMngr(p_ibdiag, version);
    IBDIAGNET_RETURN(0);
}

int SharpStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    if (!p_ibdiag->GetDiscoverFabricPtr() ||
        !p_ibdiag->GetIbisPtr())
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    KeyUpdater* key_updater = this->p_ibdiag->GetKeyUpdaterPtr();
    if (key_updater->IsKeyAvailable(IBIS_AM_KEY))
        key_updater->UpdateKeysPerPort({IBIS_AM_KEY});

    list_p_fabric_general_err sharp_discovery_errors;

    int rc2 = p_sharp_mngr->BuildSharpConfiguration(sharp_discovery_errors, check_trees);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(sharp_discovery_errors,
            STAGE_NAME_SHARP,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    if (dump_counters) {
        INFO_PRINT("Build AM PerfCounters DB\n");
        rc2 = p_sharp_mngr->BuildPerformanceCounters(sharp_discovery_errors, false);
        SCREEN_PRINT("\n");
        if (rc) {
            ERR_PRINT("Failed to build AM PerfCounters DB.\n");
            IBDIAG_RETURN(rc);
        }

        rc = this->AnalyzeCheckResults(sharp_discovery_errors,
                CHECK_NAME_SHARP_PM_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (rc)
            IBDIAGNET_RETURN(rc);
    }

    if (dump_hba_port_counters) {
        INFO_PRINT("Build AM HBA Ports PerfCounters DB\n");
        rc2 = p_sharp_mngr->BuildPerformanceCounters(sharp_discovery_errors, true);
        SCREEN_PRINT("\n");
        if (rc) {
            ERR_PRINT("Failed to build AM HBA Ports PerfCounters DB.\n");
            IBDIAG_RETURN(rc);
        }

        rc = this->AnalyzeCheckResults(sharp_discovery_errors,
                CHECK_NAME_HBA_PORT_COUNTERS_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (rc)
            IBDIAGNET_RETURN(rc);
    }

    if (clear_counters) {
        INFO_PRINT("Reset AM PerfCounters DB\n");
        rc2 = p_sharp_mngr->ResetPerformanceCounters(sharp_discovery_errors);
        SCREEN_PRINT("\n");
        if (rc) {
            ERR_PRINT("Failed to reset AM PerfCounters DB.\n");
            IBDIAG_RETURN(rc);
        }

        rc = this->AnalyzeCheckResults(sharp_discovery_errors,
                CHECK_NAME_SHARP_PM_RESET,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (rc)
            IBDIAGNET_RETURN(rc);
    }

    IBDIAGNET_RETURN(rc);
}

int SharpStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    list_p_fabric_general_err sharp_discovery_errors;

    int rc2 = p_sharp_mngr->VerifyTrapsLids(sharp_discovery_errors);
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(sharp_discovery_errors,
            CHECK_NAME_SHARP_VERIFY_TRAPS_LIDS,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    rc2 = p_sharp_mngr->VerifyVersions(sharp_discovery_errors);
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(sharp_discovery_errors,
            CHECK_NAME_SHARP_VERIFY_VERSIONS,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    if (check_trees) {
        rc2 = p_sharp_mngr->ConnectTreeEdges(sharp_discovery_errors);
        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(sharp_discovery_errors,
                CHECK_NAME_SHARP_CONNECT_TREE_EDGES,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings);
        if (rc)
            IBDIAGNET_RETURN(rc);

        rc2 = p_sharp_mngr->CheckSharpTrees(sharp_discovery_errors);
        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(sharp_discovery_errors,
                CHECK_NAME_SHARP_CHECK_TREES,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (rc)
            IBDIAGNET_RETURN(rc);

        if (p_sharp_mngr->WriteSharpFile(SHARP_FILE)) {
            ERR_PRINT("Write sharp data file failed\n");
            ++this->num_errors;
        }
    }

    if (p_sharp_mngr->DumpSharpANInfoToCSV(*p_csv_out)) {
        ERR_PRINT("Dump sharp ANs Info DB to CSV file failed \n");
        ++this->num_errors;
    }

    if (p_sharp_mngr->WriteSharpANInfoFile(SHARP_AN_FILE)) {
        ERR_PRINT("Write sharp ANs file failed\n");
        ++this->num_errors;
    }

    if (dump_counters)
        p_sharp_mngr->DumpSharpPMCountersToCSV(*p_csv_out);

    if (dump_hba_port_counters)
        p_sharp_mngr->DumpSharpPMHBAPortCountersToCSV(*p_csv_out);

    this->CountSharpVersionsAndNumOfActiveJobs();
    if (check_trees) {
        this->CountSharpTreeTypes();
    }

    IBDIAGNET_RETURN(rc);
}


/**********************************
 ***  Congestion Control Stage  ***
 **********************************/
CongestionControlStage::CongestionControlStage(IBDiag *p_ibdiag) :
    Stage(STAGE_NAME_CC, p_ibdiag), to_get_congestion_counters(false),
    to_clear_congestion_counters(false), enabled_SWs(0), enabled_CAs(0),
    enabled_algo_CAs(0), enabled_algo_CA_ports(0), disabled_algo_CA_ports(0)

{
    IBDIAGNET_ENTER;
    //mark this stage as skipped by default
    MarkStageAsSkipped();
    IBDIAGNET_RETURN_VOID;
}

CongestionControlStage::~CongestionControlStage()
{
    IBDIAGNET_ENTER;

    this->CleanResources();

    IBDIAGNET_RETURN_VOID;
}

void CongestionControlStage::PreparePerSlVlCounters()
{
    IBDIAGNET_ENTER;

    this->m_slvl_cntrs_vec.push_back(new PortXmitConCtrl());
    this->m_slvl_cntrs_vec.push_back(new PortVLXmitTimeCong());
    this->m_slvl_cntrs_vec.push_back(new PortVLXmitTimeCongExt());

    IBDIAGNET_RETURN_VOID;
}

int CongestionControlStage::Prepare()
{
    IBDIAGNET_ENTER;

    HDR_PRINT("%s\n", this->stage_name.c_str());

    if (!this->ppcc_algo_path.empty()) {
        if (this->ppcc_algo_database.ParseDir(this->ppcc_algo_path))
            IBDIAGNET_RETURN(1);
    }

    this->PreparePerSlVlCounters();

    IBDIAGNET_RETURN(0);
}

int CongestionControlStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;

    list_p_fabric_general_err congestion_control_errors;

    int rc = this->p_ibdiag->BuildCCEnhancedCongestionInfo(congestion_control_errors);
    SCREEN_PRINT("\n");
    int rc2 = this->AnalyzeCheckResults(congestion_control_errors,
            CHECK_NAME_ENHANCED_CC_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    rc = this->p_ibdiag->BuildCCSwithGeneralSettings(congestion_control_errors);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(congestion_control_errors,
            CHECK_NAME_ENHANCED_CC_SWITCH_GENERAL_SETTINGS_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    rc = this->p_ibdiag->BuildCCSwithConfig(congestion_control_errors, this->enabled_SWs);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(congestion_control_errors,
            CHECK_NAME_ENHANCED_CC_SWITCH_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    rc = this->p_ibdiag->BuildCCHCAGeneralSettings(congestion_control_errors);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(congestion_control_errors,
            CHECK_NAME_ENHANCED_CC_HCA_GENERAL_SETTINGS_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    rc = this->p_ibdiag->BuildCCHCAConfig(congestion_control_errors, this->enabled_CAs);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(congestion_control_errors,
            CHECK_NAME_ENHANCED_CC_HCA_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    if (this->to_get_congestion_counters) {
        rc = this->p_ibdiag->BuildCCHCAStatisticsQuery(congestion_control_errors,
                                                       this->to_clear_congestion_counters);
        SCREEN_PRINT("\n");
        rc2 = this->AnalyzeCheckResults(congestion_control_errors,
                CHECK_NAME_ENHANCED_CC_HCA_COUNTERS_RETRIEVE,
                rc,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        if (rc2)
            IBDIAGNET_RETURN(rc2);
    }

    rc = this->p_ibdiag->BuildCCHCAAlgoConfigSup(congestion_control_errors);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(congestion_control_errors,
            CHECK_NAME_CC_HCA_ALGO_CONFIG_SUP_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    rc = this->p_ibdiag->BuildCCHCAAlgoConfig(congestion_control_errors);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(congestion_control_errors,
            CHECK_NAME_CC_HCA_ALGO_CONFIG_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    rc = this->p_ibdiag->BuildCCHCAAlgoConfigParams(congestion_control_errors);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(congestion_control_errors,
            CHECK_NAME_CC_HCA_ALGO_PARAMS_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    if (this->to_get_congestion_counters) {
        // CC HCA Algo Counters
        rc = this->p_ibdiag->BuildCCHCAAlgoCounters(congestion_control_errors,
                                                       this->to_clear_congestion_counters);
        SCREEN_PRINT("\n");

        rc2 = this->AnalyzeCheckResults(congestion_control_errors,
                CHECK_NAME_CC_HCA_ALGO_COUNTERS_RETRIEVE,
                rc,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (rc2)
            IBDIAGNET_RETURN(rc2);

        // CC SLVL Counters
        for (CountersPerSLVL* p_slvl_cntrs : this->m_slvl_cntrs_vec) {

            rc = this->p_ibdiag->BuildCCSLVLCounters(congestion_control_errors,
                                                    this->to_clear_congestion_counters,
                                                    p_slvl_cntrs);

            SCREEN_PRINT("\n");
            rc2 = this->AnalyzeCheckResults(congestion_control_errors,
                                            p_slvl_cntrs->GetCntrHeader(),
                                            rc,
                                            IBDIAG_ERR_CODE_FABRIC_ERROR,
                                            this->num_errors,
                                            this->num_warnings,
                                            true);
            if (rc2)
                IBDIAGNET_RETURN(rc2);
        }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int CongestionControlStage::RunCheck()
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;
    int rc2 = IBDIAG_SUCCESS_CODE;
    list_p_fabric_general_err cc_errors;

    this->p_ibdiag->DumpCCEnhancedInfoToCSV(*this->p_csv_out);
    this->p_ibdiag->DumpCCSwitchGeneralSettingsToCSV(*this->p_csv_out);
    this->p_ibdiag->DumpCCPortProfileSettingsToCSV(*this->p_csv_out);
    this->p_ibdiag->DumpCCSLMappingSettingsToCSV(*this->p_csv_out);
    this->p_ibdiag->DumpCCHCAGeneralSettingsToCSV(*this->p_csv_out);
    this->p_ibdiag->DumpCCHCARPParametersToCSV(*this->p_csv_out);
    this->p_ibdiag->DumpCCHCANPParametersToCSV(*this->p_csv_out);

    // CC Algo
    rc = this->p_ibdiag->DumpCCHCAAlgoConfigSupToCSV(*this->p_csv_out, cc_errors);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(cc_errors,
            CHECK_NAME_CC_HCA_ALGO_CONFIG_SUP_VALUES,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    this->p_ibdiag->DumpCCHCAAlgoConfigToCSV(*this->p_csv_out,
                                                  this->enabled_algo_CAs,
                                                  this->enabled_algo_CA_ports,
                                                  this->disabled_algo_CA_ports);

    rc = this->p_ibdiag->DumpCCHCAAlgoConfigParamsToCSV(*this->p_csv_out, cc_errors);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(cc_errors,
            CHECK_NAME_CC_HCA_ALGO_PARAMS_VALUES,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    if (this->to_get_congestion_counters) {

        this->p_ibdiag->DumpCCHCAStatisticsQueryToCSV(*this->p_csv_out);

        rc = this->p_ibdiag->DumpCCHCAAlgoCountersToCSV(*this->p_csv_out, cc_errors);
        SCREEN_PRINT("\n");
        rc2 = this->AnalyzeCheckResults(cc_errors,
                CHECK_NAME_CC_HCA_ALGO_COUNTERS_VALUES,
                rc,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (rc2)
            IBDIAGNET_RETURN(rc2);

        if (this->p_ibdiag->DumpPerSLVLPortCountersToCSV(*this->p_csv_out,
                                                         this->m_slvl_cntrs_vec)) {
            ERR_PRINT("Write CC SLVL Counters Data to DB failed\n");
            ++this->num_errors;
        }
    }

    // PPCC file
    if (!this->ppcc_algo_path.empty()) {
        if (this->p_ibdiag->WriteCCHCAAlgoDataToFile(PPCC_FILE, this->ppcc_algo_database)) {
            ERR_PRINT("Writing CC HCA Algo Data fo file failed\n");
            ++this->num_errors;
        }
    }

    // CC Algo Validations
    rc = this->p_ibdiag->CC_HCA_AlgoValidations(cc_errors, this->ppcc_algo_database);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(cc_errors,
            CHECK_NAME_CC_HCA_ALGO_VALIDATIONS,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    IBDIAGNET_RETURN(rc);
}

bool CongestionControlStage::GetFabricSummarySection(stringstream &summary_str)
{
    IBDIAGNET_ENTER;

    summary_str << "Congestion Control is enabled on "
                << this->enabled_SWs
                << " switches and "
                << this->enabled_CAs
                << " Channel Adapters."
                << endl;

    summary_str << "Congestion Control Algo is enabled on "
                << this->enabled_algo_CAs
                << " Channel Adapters (enabled ports: "
                << this->enabled_algo_CA_ports;

    if (disabled_algo_CA_ports)
        summary_str << ", disabled ports: "
                    << disabled_algo_CA_ports;

    summary_str << ")."
                << endl;

    IBDIAGNET_RETURN(true);
}

int CongestionControlStage::CleanResources()
{
    IBDIAGNET_ENTER;

    release_container_data(m_slvl_cntrs_vec);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}
/*****************************
 ***  Fast Recovery Stage  ***
 *****************************/
FastRecoveryStage::FastRecoveryStage(IBDiag *p_ibdiag) :
    Stage(STAGE_NAME_FAST_RECOVERY, p_ibdiag)
{
    IBDIAGNET_ENTER;
    //mark this stage as skipped by default
    MarkStageAsSkipped();
    IBDIAGNET_RETURN_VOID;
}

FastRecoveryStage::~FastRecoveryStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int FastRecoveryStage::Prepare()
{
    IBDIAGNET_ENTER;

    HDR_PRINT("%s\n", this->stage_name.c_str());

    IBDIAGNET_RETURN(0);
}

int FastRecoveryStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;

    list_p_fabric_general_err fast_recovery_errors;

    // Build Profiles Config
    int rc = this->p_ibdiag->BuildProfilesConfig(fast_recovery_errors);
    SCREEN_PRINT("\n");
    int rc2 = this->AnalyzeCheckResults(fast_recovery_errors,
            CHECK_PROFILES_CONFIG_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    // Build Credit Watchdog Config
    rc = this->p_ibdiag->BuildCreditWatchdogConfig(fast_recovery_errors);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(fast_recovery_errors,
            CHECK_CREDIT_WATCHDOG_CONFIG_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    // Build BER Config
    rc = this->p_ibdiag->BuildBERConfig(fast_recovery_errors);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(fast_recovery_errors,
            CHECK_BER_CONFIG_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FastRecoveryStage::RunCheck()
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    this->p_ibdiag->DumpProfilesConfigToCSV(*this->p_csv_out);
    this->p_ibdiag->DumpCreditWatchdogConfigToCSV(*this->p_csv_out);
    this->p_ibdiag->DumpBERConfigToCSV(*this->p_csv_out);

    IBDIAGNET_RETURN(rc);
}

CheckRailOptimizedTopologyStage::CheckRailOptimizedTopologyStage(IBDiag *p_ibdiag):
            Stage(STAGE_NAME_RAIL_OPT_TOPO_CHECK, p_ibdiag),
            source(PCI_BDF_SOURCE_AUTO), factor(90), railFlow(0), p_railRegex(NULL)
{
    IBDIAGNET_ENTER;
    this->MarkStage(false);
    IBDIAGNET_RETURN_VOID;
}

CheckRailOptimizedTopologyStage::~CheckRailOptimizedTopologyStage()
{
    IBDIAGNET_ENTER;

    delete p_railRegex;

    IBDIAGNET_RETURN_VOID;
}

int CheckRailOptimizedTopologyStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int CheckRailOptimizedTopologyStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int CheckRailOptimizedTopologyStage::RunCheck()
{
    IBDIAGNET_ENTER;

    ExcludedData excludedData;
    PCI_LeafSwitchesInfoVec pciLeafSwitchInfoVec;
    list_p_fabric_general_err errorsList;
    int rc = this->RetrieveRailsData(errorsList, excludedData, pciLeafSwitchInfoVec);
    this->AnalyzeCheckResults(errorsList,
            CHECK_RAIL_OPTIMIZED_TOPOLOGY_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            false);

    if (rc)
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);

    this->PrintValidationReport(pciLeafSwitchInfoVec, excludedData);
    SCREEN_PRINT("\n");

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int CheckRailOptimizedTopologyStage::RetrieveRailsData(list_p_fabric_general_err &errorsList,
                                                       ExcludedData &excludedData,
                                                       PCI_LeafSwitchesInfoVec &leafSwitchInfoVec)
{
    IBDIAGNET_ENTER;

    if (!this->p_ibdiag->IsRailDataCollected()) {
        this->lastError = "Error: The PCIe data needed for performing the "
                 "test has not been properly collected. "
                 "Verify that the Phy plugin has been loaded";
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    for (const auto p_switch : this->p_ibdiag->GetDiscoverFabricPtr()->Switches) {

        if (!p_switch) {
            this->lastError = "DB error - found null node in Switches set";
            IBDIAG_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
        }

        if (!p_switch->getInSubFabric())
            continue;

        PCI_AddressToNodesMap pciAddresses;
        PCI_BDF_SOURCE source;

        int rc = this->GetPCIRelatedData(*p_switch, pciAddresses, excludedData, errorsList, source);
        if (rc)
            IBDIAG_RETURN(rc);

        size_t size = pciAddresses.size();
        if (size > 0 && size != this->railFlow) {
            leafSwitchInfoVec.push_back(PCI_LeafSwitchInfo());

            PCI_LeafSwitchInfo& back = leafSwitchInfoVec.back();
            back.p_switch = p_switch;
            back.source = source;
            back.pciAddressMap.swap(pciAddresses);
        }
    }//end nodes loop

    if (excludedData.socketDirect)
        errorsList.insert(errorsList.begin(), new RailsSDMCardsError(excludedData.socketDirect));

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

bool CheckRailOptimizedTopologyStage::ShouldFilterNode(const string &description) const
{
    IBDIAG_ENTER;

    if (!this->p_railRegex)
        IBDIAG_RETURN(false);

    rexMatch *p_rexRes = p_railRegex->apply(description.c_str());
    if (p_rexRes) {
        delete p_rexRes;
        IBDIAG_RETURN(false);
    }

    IBDIAG_RETURN(true);
}

bool CheckRailOptimizedTopologyStage::SetRailRegex(const string &val)
{
    IBDIAG_ENTER;

    //p_railRegex will be deleted at stage's destructor if created
    p_railRegex = new regExp(val.c_str());
    IBDIAG_RETURN(p_railRegex->valid());
}

PCI_BDF_SOURCE CheckRailOptimizedTopologyStage::ResolveBDFSource(const IBNode &curr_node) const
{
    IBDIAG_ENTER;

    if (this->source == PCI_BDF_SOURCE_HI || this->source == PCI_BDF_SOURCE_PHY)
        IBDIAG_RETURN(this->source);

    size_t no_bdf = 0;
    size_t hcaCount = 0;
    for (phys_port_t i = 1; i <= curr_node.numPorts; ++i) {
        const IBPort *p_curr_port = curr_node.getPort(i);
        //Skip ports that are down or have undefined state
        //or are not in the scope of the fabric.
        if (!p_curr_port || !p_curr_port->is_data_worthy(DATA_WORTHY_SMP_MAD))
            continue;

        const IBPort *p_remote_port = p_curr_port->p_remotePort;
        if (!p_remote_port)
           continue;

        const IBNode *p_remote_node = p_remote_port->p_node;
        if (!p_remote_node || p_remote_node->type != IB_CA_NODE || p_remote_node->isSpecialNode())
            continue;

        hcaCount++;

        if (!p_remote_port->p_port_hierarchy_info ||
                p_remote_port->p_port_hierarchy_info->bdf() == -1) {
            no_bdf++;
            continue;
        }
    }

    //all CAs support HI
    if (!no_bdf)
        IBDIAG_RETURN(PCI_BDF_SOURCE_HI);

    //if less then (100 - this->factor) percent of CAs does not support HI
    //we apply HI otherwise PHY
    IBDIAG_RETURN(
            100*no_bdf/hcaCount <= static_cast<size_t>(100 - this->factor)
            ? PCI_BDF_SOURCE_HI : PCI_BDF_SOURCE_PHY);
}

int CheckRailOptimizedTopologyStage::GetPCIRelatedData(const IBNode &switch_node,
                                                       PCI_AddressToNodesMap &pciAddresses,
                                                       ExcludedData &excludedCards,
                                                       list_p_fabric_general_err &errorsList,
                                                       PCI_BDF_SOURCE &inSource)
{
    IBDIAG_ENTER;

    inSource = this->ResolveBDFSource(switch_node);

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
            "collecting PCI addresses for the switch: %s GUID=" U64H_FMT ", apply_hi_bdf: %d\n",
            switch_node.name.c_str(), switch_node.guid_get(), (int)inSource);

    for (phys_port_t i = 1; i <= switch_node.numPorts; ++i) {
        IBPort *p_curr_port = switch_node.getPort(i);
        //Skip ports that are down or have undefined state
        //or are not in the scope of the fabric.
        if (!p_curr_port || !p_curr_port->is_data_worthy(DATA_WORTHY_SMP_MAD))
            continue;

        IBPort *p_remote_port = p_curr_port->p_remotePort;
        if (!p_remote_port)
           continue;

        IBNode *p_remote_node = p_remote_port->p_node;
        if (!p_remote_node) {
            stringstream sstr;
            sstr << "DB error - the port with GUID="
                 << PTR(p_remote_port->guid_get())
                 << endl;
            this->lastError = sstr.str();
            IBDIAG_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
        }

        if (p_remote_node->type != IB_CA_NODE || p_remote_node->isSpecialNode())
            continue;

        if (this->ShouldFilterNode(p_remote_node->description))
            continue;

        PCI_Address address;
        if (inSource == PCI_BDF_SOURCE_PHY) {
            if (p_remote_node->isSDM()) {
                excludedCards.socketDirect++;

                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                       "adding port: %s to exclude list\n", p_remote_port->getExtendedName().c_str());
                continue;
            }

            address = p_remote_node->getPCIaddress();
        }
        else
            address = p_remote_port->getPCIaddress();

        //port is marked as exluded because we failed to get its PCI data in mode
        if (!address.IsValid()) {
            excludedCards.noAddress++;
            errorsList.push_back(new RailsInvalidPCIAddress(p_remote_port, inSource));

            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                "exclude port %s GUID=" U64H_FMT " from verification \n",
                p_remote_port->getName().c_str(), p_remote_port->guid_get());
            continue;
        }

        pciAddresses[address].push_back(p_remote_node);

        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                "adding node %s GUID=" U64H_FMT " PCI address:  %x.%x.%x\n",
                p_remote_node->name.c_str(), p_remote_node->guid_get(),
                address.bus, address.device, address.function);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int CheckRailOptimizedTopologyStage::PrintValidationReport(const PCI_LeafSwitchesInfoVec &pciLeafSwitchInfoVec,
                                                 const ExcludedData &excludedCards)
{
    IBDIAGNET_ENTER;

    INFO_PRINT("Starting %s\n", CHECK_RAIL_OPTIMIZED_TOPOLOGY_VALIDATION);

    if (pciLeafSwitchInfoVec.empty() && !excludedCards.noAddress &&
            !excludedCards.socketDirect) {
        INFO_PRINT("%s finished successfully\n", CHECK_RAIL_OPTIMIZED_TOPOLOGY_VALIDATION);
        IBDIAGNET_RETURN(0);
    }

    //we have dump file in case if any mismatch is found
    ofstream  sout;
    if (this->p_ibdiag->OpenFile("Rails Dump", string(RAILS_DUMP_FILE), sout, false, "#"))
        WARN_PRINT("The file: %s with rails validation details"
                   " will not be provided.\n", RAILS_DUMP_FILE);

    IBFabric::GetSwitchLabelPortNumExplanation(sout);

    //stream does not crash on write attempt if it has no associated file
    int rc = this->ReportOnExcludedData(excludedCards, sout);

    for (PCI_LeafSwitchesInfoVec::const_iterator it = pciLeafSwitchInfoVec.begin();
            !rc && it != pciLeafSwitchInfoVec.end(); it++)
        rc = this->ReportOnLeafSwitch(*it, sout);

    this->p_ibdiag->CloseFile(sout);

    this->num_warnings += static_cast<u_int32_t>(pciLeafSwitchInfoVec.size());
    WARN_PRINT("%s ended with %d warnings\n", CHECK_RAIL_OPTIMIZED_TOPOLOGY_VALIDATION,
            this->num_warnings);
    INFO_PRINT("Rail Optimized Topology validation is usually applicable to the compute nodes.\n");
    INFO_PRINT("If detected mis-cabled nodes are not compute ones, please apply Rail Optimized " \
               "Topology check for specific set of nodes by invoking: " \
               "--rail_validation_opt regex='reg expression'\n");

    IBDIAGNET_RETURN(rc);
}

int CheckRailOptimizedTopologyStage::ReportOnExcludedData(const ExcludedData &excludedData,
                                                           ostream &ofs)
{
    IBDIAGNET_ENTER;

    if (excludedData.socketDirect)
        ofs << excludedData.socketDirect
            << " socket-direct cards excluded from rail-optimized validation"
            << endl;

    if (excludedData.noAddress)
        ofs << "Failed to obtain PCI address for "
            << excludedData.noAddress << " ports in applied mode. "
            << "They are excluded from rail-optimized validation"
            << endl;

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

#define IDX_TO_STR_RANGE    ('Z' - 'A' + 1)
static char* idx_to_str(int index, char * buffer)
{
    if (index >= IDX_TO_STR_RANGE)
        buffer = idx_to_str(index / IDX_TO_STR_RANGE - 1, buffer);

    *buffer++ = static_cast<char>('A' + (index % IDX_TO_STR_RANGE));
    *(buffer) = 0;
    return buffer;
}

#define PCIe_fmt "PCIe 0000:%02x:%02x.%x"
int CheckRailOptimizedTopologyStage::ReportOnLeafSwitch(const PCI_LeafSwitchInfo &leafSwitchInfo, ostream &ofs)
{
    IBDIAGNET_ENTER;

    const IBNode *p_switch = leafSwitchInfo.p_switch;
    if (!p_switch) {
        ERR_PRINT("\tError: Null pointer to swicth node\n");
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }

    this->PrintSwitchData(leafSwitchInfo, ofs);

    if (leafSwitchInfo.pciAddressMap.empty()) {
        WARN_PRINT("\t\t has no PCI address on all connected HCAs\n");

        ofs << "\t\t has no PCI address on all connected HCAs" << endl;
    } else {
        int index = 0;
        for (PCI_AddressToNodesMap::const_iterator it = leafSwitchInfo.pciAddressMap.begin();
            it != leafSwitchInfo.pciAddressMap.end(); it++) {

            const PCI_Address& address = it->first;
            const vector_p_node &nodes = it->second;
            if (nodes.empty()) {
                ERR_PRINT("\tError: Empty nodes list was provided "
                          "for the " PCIe_fmt "\n",
                          address.bus, address.device, address.function);
                IBDIAGNET_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
            }

            int rc = this->PrintRailData(address, nodes, index, ofs);
            if (rc)
               IBDIAGNET_RETURN(rc);

            index++;
        }
    }

    IBDIAGNET_RETURN(0);
}

int CheckRailOptimizedTopologyStage::PrintRailData(const PCI_Address& address,
                                                   const vector_p_node &nodes,
                                                   int index,
                                                   ostream &ofs)
{
    IBDIAGNET_ENTER;

    char buffer[32] = {'\0'};
    idx_to_str(index, buffer);
    const size_t size = nodes.size();

    for (size_t i = 0; i < size; i++)
    {
        IBNode *p_node = const_cast<IBNode*>(nodes[i]);
        if (!p_node) {
            ERR_PRINT("\tError: Null pointer found in HCAs "
                      "for the " PCIe_fmt "\n",
                      address.bus, address.device, address.function);
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }

        IBPort *p_switch_port = NULL;
        IBPort *p_node_port = NULL;
        if (this->GetSwitchAndNodePorts(p_node, p_switch_port, p_node_port)) {
            ERR_PRINT("\tError: "
                      " HCA GUID=" U64H_FMT " is not connected to any switch\n",
                      p_node->guid_get());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }

        if (i == 0) {
            WARN_PRINT("\t\t rail %s (" PCIe_fmt "): "
                    "%lu ports ==> "
                    "%s <--> %s%s\n",
                    buffer,
                    address.bus, address.device, address.function,
                    size,
                    p_node_port->getExtendedName().c_str(),
                    p_switch_port->getExtendedName().c_str(), (size == 1) ? "" : ", ...");

            ofs << "    rail " << buffer
                << "(0000:" << HEX(address.bus, 2)
                << '.' << HEX(address.device, 2)
                <<  '.' << HEX(address.function) << "): "
                << size << " ports:" << endl
                << "                            " << p_node_port->getExtendedName()
                << " <--> " << p_switch_port->getExtendedName().c_str() << endl;
        } else
            ofs << "                            "
                << p_node_port->getExtendedName()
                << " <--> " << p_switch_port->getExtendedName().c_str() << endl;
    }//end for

    IBDIAGNET_RETURN(0);
}

void CheckRailOptimizedTopologyStage::PrintSwitchData(const PCI_LeafSwitchInfo &leafSwitchInfo,
                                                      ostream &ofs)
{
    IBDIAGNET_ENTER;

    const char *source_str = "Unknown";

    if(leafSwitchInfo.source == PCI_BDF_SOURCE_HI)
        source_str = "HI";
    else if(leafSwitchInfo.source == PCI_BDF_SOURCE_PHY)
        source_str = "PHY";

    WARN_PRINT("\tNode rail connectivity mismatch by %s source on the switch: \"%s\" GUID=" U64H_FMT "\n",
            source_str,
            leafSwitchInfo.p_switch->description.c_str(), leafSwitchInfo.p_switch->guid_get());

    ofs << endl << "Node rail connectivity mismatch by " << source_str
        << " source on the switch: " << leafSwitchInfo.p_switch->description
        << " GUID=" << PTR(leafSwitchInfo.p_switch->guid_get()) << endl;

    IBDIAGNET_RETURN_VOID;
}

int CheckRailOptimizedTopologyStage::GetSwitchAndNodePorts(IBNode *p_node,
                                IBPort *&p_switch_port, IBPort *&p_node_port)
{
    IBDIAGNET_ENTER;

    int rc = 0;
    p_switch_port = NULL;
    p_node_port = NULL;

    for (phys_port_t i = 0; i <= p_node->numPorts; ++i) {
        IBPort *p_curr_port = p_node->getPort(i);
        //Skip ports that are down or have undefined state
        //or are not in the scope of the fabric.
        if (!p_curr_port || !p_curr_port->is_data_worthy(DATA_WORTHY_SMP_MAD))
            continue;

        IBPort *p_remote_port = p_curr_port->p_remotePort;
        if (p_remote_port) {
           p_switch_port = p_remote_port;
           p_node_port = p_curr_port;
           break;
        }
    }

    if (!p_switch_port || !p_node_port)
        rc = 1;

    IBDIAGNET_RETURN(rc);
}

PostReportsStage::PostReportsStage(IBDiag *p_ibdiag) : Stage(STAGE_NAME_POST_REPORTS, p_ibdiag),
                                                       to_check_sm(true), num_of_diff_val(0)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

PostReportsStage::~PostReportsStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int PostReportsStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int PostReportsStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int PostReportsStage::RunCheck()
{
    IBDIAGNET_ENTER;

    warnings_list warnings;
    if (p_ibdiag->CreateIBNetDiscoverFile(IBNET_DISCOVER_FILE, warnings)) {
        this->num_errors++;

        ERR_PRINT("Error: %s\n", p_ibdiag->GetLastError());
        ERR_PRINT("Writing of IBNetdDscover file failed\n");
    } else if (warnings.empty())
        INFO_PRINT("Writing of IBNetdDscover file finished successfully\n");
    else {
        this->num_warnings += static_cast<u_int32_t>(warnings.size());

        //we do not print any warning on screen and the log file.
        WARN_PRINT("Writing of IBNetdDscover file ended with warnings\n");
    }

    SCREEN_PRINT("\n");

    if (this->p_ibdiag->CreateIBLinkInfoFile(IBLINKINFO_FILE)) {
        ERR_PRINT("Creating IB Link Info file failed\n");
        ++this->num_errors;
    }
#ifdef DEBUG

    SCREEN_PRINT("\n");
    if (!this->generated_class_name.empty()) {
       if (this->p_ibdiag->CreateIBCPPSimulatorInfo(this->generated_class_name, this->simulator_guids)) {
           ERR_PRINT("Creating Info for C++ Simulator file failed\n");
           ++this->num_errors;
        }
    }

#endif

    if (this->to_check_sm) {
        // Validation of SM configuration
        list_p_fabric_general_err errors;

        int rc = this->p_ibdiag->PostReportsSMValidations(errors, this->num_of_diff_val);
        SCREEN_PRINT("\n");
        int rc2 = this->AnalyzeCheckResults(errors,
                                            CHECK_POST_REPORTS_SM_VALIDATIONS,
                                            rc,
                                            IBDIAG_ERR_CODE_FABRIC_ERROR,
                                            this->num_errors,
                                            this->num_warnings,
                                            true);
        if (rc2)
            IBDIAGNET_RETURN(rc2);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void PostReportsStage::ParseNodeGUIDs(const std::string &guids)
{
    IBDIAGNET_ENTER;

    istringstream tokenStream(guids);
    string token;
    while (std::getline(tokenStream, token, ',')) {
        uint64_t result = std::stoull(token, NULL, 16);
        this->simulator_guids.insert(result);
    }

    IBDIAGNET_RETURN_VOID;
}

void PostReportsStage::PrintWarnings(const warnings_list& warnings) const
{
    IBDIAGNET_ENTER;

    for (warnings_list::const_iterator itr = warnings.begin();
         itr != warnings.end(); itr++) {
        const string& ref = *itr;
        WARN_PRINT("%s\n", ref.c_str());
    }

    IBDIAGNET_RETURN_VOID;
}

CheckDFPTopologyStage::CheckDFPTopologyStage(IBDiag *p_ibdiag):Stage(STAGE_NAME_DFP_TOPO_CHECK, p_ibdiag),
                                        maxCAsOnSpine(-1), is_roots_from_file(false)
{
    IBDIAGNET_ENTER;
    this->MarkStage(false);
    IBDIAGNET_RETURN_VOID;

}

int CheckDFPTopologyStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int CheckDFPTopologyStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int CheckDFPTopologyStage::RunCheck()
{
    IBDIAGNET_ENTER;

    if (this->BuildAndValidate())
        ERR_PRINT("DFP validation has failed\n");
    else if (!this->num_warnings && !this->num_errors)
             INFO_PRINT("DFP validation finished successfully\n");
    else if (this->num_errors)
        WARN_PRINT("DFP validation finished with errors\n");
    else if (this->num_warnings && !this->num_errors)
        WARN_PRINT("DFP validation finished with warnings\n");

    SCREEN_PRINT("\n");

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int CheckDFPTopologyStage::DiscoverDFPRootsBySMDB()
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    for (set_pnode::iterator it = this->p_ibdiag->GetDiscoverFabricPtr()->Switches.begin();
          it != this->p_ibdiag->GetDiscoverFabricPtr()->Switches.end();
          it++) {
         IBNode *p_switch = *it;
         if (p_switch->rank == RANK_OF_SPINE)
             this->dfp_roots[p_switch->guid_get()] = p_switch;
    }

    if (this->dfp_roots.empty()) {
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
        ERR_PRINT("No DFP Root switches are founded in fabric\n");
    }

    IBDIAGNET_RETURN(rc);
}

int CheckDFPTopologyStage::DiscoverDFPRootsByMaxCAsOnSpine()
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    for (set_pnode::iterator it = this->p_ibdiag->GetDiscoverFabricPtr()->Switches.begin();
          it != this->p_ibdiag->GetDiscoverFabricPtr()->Switches.end();
          it++) {
         int num_of_cas = 0;
         IBNode *p_switch = *it;

         //Take care of planarized topologies
         for (u_int32_t i = 1 ; i <=  p_switch->numPorts ; i++) {
              IBPort *p_curr_port = p_switch->getPort((phys_port_t)i);
              if (!p_curr_port)                       //port not exist
                  continue;
              //Skip ports that are down or have undefined state
              //or are not in the scope of the fabric.
              if (!p_curr_port->is_data_worthy(DATA_WORTHY_SMP_MAD))
                  continue;
              if (!p_curr_port->p_remotePort || !p_curr_port->p_remotePort->p_node)
                  continue;
              // Skip on special ports
              if (p_curr_port->isSpecialPort())
                  continue;

              IBNode *p_remote_node = p_curr_port->p_remotePort->p_node;
              if (p_remote_node->type == IB_CA_NODE)
                  num_of_cas++;
         }

         if (num_of_cas <= this->maxCAsOnSpine)
             this->dfp_roots[p_switch->guid_get()] = p_switch;
    }

    if (this->dfp_roots.empty()) {
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
        ERR_PRINT("No DFP Root switches are founded in fabric\n");
    }

    IBDIAGNET_RETURN(rc);
}

int CheckDFPTopologyStage::BuildAndValidate()
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    if (this->IsRootFromFile())
        rc = this->DiscoverDFPRootsBySMDB();
    else
        rc = this->DiscoverDFPRootsByMaxCAsOnSpine();

    if (rc) {
        this->num_errors++;
        ERR_PRINT("Failed to obtain DFP Roots\n");
        IBDIAGNET_RETURN(rc);
    }

    DFPTopology topology(*this->p_ibdiag->GetDiscoverFabricPtr());
    rc = topology.Build(this->dfp_roots);
    if (rc) {
        this->num_errors++;
        ERR_PRINT("Failed to build DFP topology\n");
        IBDIAGNET_RETURN(rc);
    }

    rc = topology.Validate(this->num_warnings, this->num_errors);
    if (rc) {
       this->num_errors++;
       IBDIAGNET_RETURN(rc);
    }

    ofstream  sout;

    rc = p_ibdiag->OpenFile("DFPToplogy", string(DFP_DUMP_FILE), sout, false, "#");

    if (rc) {
        this->num_errors++;
        ERR_PRINT("Failed to create DFP topolgy dump file with: %s\n", p_ibdiag->GetLastError());
        IBDIAGNET_RETURN(rc);
    }

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    if (this->IsRootFromFile())
        sout << "List of DFP Root switches is loaded from the SMDB file" << endl;
    else
        sout << "DFP Root switches are discovered by parameter "
             << "\"maximum CAs\"" << ": "
             << this->maxCAsOnSpine
             << endl;

    if (topology.DumpToStream(sout)) {
        this->num_errors++;
        ERR_PRINT("Failed to dump DFP topology\n");
    }

    this->p_ibdiag->CloseFile(sout);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}


CheckFatTreeTopologyStage::CheckFatTreeTopologyStage(IBDiag *p_ibdiag):
                                Stage(STAGE_NAME_FAT_TREE_TOPO_CHECK, p_ibdiag),
                                retries(7), equalResults(2), p_rootsRegExp(NULL),
                                rootsFromFile(false)
{
    IBDIAGNET_ENTER;
    this->MarkStage(false);
    IBDIAGNET_RETURN_VOID;
}

int CheckFatTreeTopologyStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int CheckFatTreeTopologyStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int CheckFatTreeTopologyStage::RunCheck()
{
    IBDIAGNET_ENTER;
    //todo think of backup algorithm

    this->p_ibdiag->SetLastError("");
    ofstream  outStream;
    if (p_ibdiag->OpenFile("FTToplogy", string(FAT_TREE_DUMP_FILE), outStream, false, "#")) {
        this->num_errors++;
        ERR_PRINT("Failed to create Fat-Tree topology dump file with: %s\n",
                p_ibdiag->GetLastError());
    }

    FTTopology ftTopology(*this->p_ibdiag->GetDiscoverFabricPtr(), outStream);

    list_p_fabric_general_err errorsList;
    int rc = IBDIAG_SUCCESS_CODE;

    if (this->rootsFromFile)
        rc =  ftTopology.Build(errorsList, this->lastError);
    else if (this->p_rootsRegExp)
        rc =  ftTopology.Build(errorsList, this->lastError, *this->p_rootsRegExp);
    else
        rc =  ftTopology.Build(errorsList, this->lastError, this->retries, this->equalResults);

    this->AnalyzeCheckResults(errorsList,
            "Fat-Tree topology detection",
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            false);

    if (rc) {
        this->p_ibdiag->CloseFile(outStream);
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
    }

    INFO_PRINT("%s\n", ftTopology.LevelsReport().c_str());

    rc =  ftTopology.Validate(errorsList, this->lastError);
    SCREEN_PRINT("\n");
    this->AnalyzeCheckResults(errorsList,
            "Fat-Tree topology validation",
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            false);

    if (ftTopology.GetInternalErrors())
        ERR_PRINT("For more errors see the dump file: ibdiagnet2.fat_tree\n\n");

    INFO_PRINT("Calculated Fat-Tree bisectional bandwidth: %.2f [Gb/s]\n",
             ftTopology.GetMinimalRatio());

    if (ftTopology.Dump()) {
        this->num_errors++;
        ERR_PRINT("Failed to dump Fat-Tree topology\n");
    }

    this->p_ibdiag->CloseFile(outStream);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

bool CheckFatTreeTopologyStage::SetRootsRegex(const string &val)
{
    IBDIAG_ENTER;

    //p_rootsRegExp will be deleted at IBDiag's destructor if created
    p_rootsRegExp = new regExp(val.c_str());
    if (!p_rootsRegExp->valid())
        IBDIAG_RETURN(false);

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "FT roots regular expressioon: %s\n", val.c_str());

    IBDIAG_RETURN(true);
}

ReportCableInfoStage::ReportCableInfoStage(IBDiag *p_ibdiag):
        Stage(STAGE_NAME_REPORT_CABLE_INFO, p_ibdiag),
        parse_prtl_mod_dp_latency(false), prtl_latency_from_table(false)
{
    IBDIAGNET_ENTER;

    this->MarkStage(false);
    this->InitCDRVendorTable();
    this->InitPNVendorTable();
    this->InitLocalPhyLatancyTable();

    IBDIAGNET_RETURN_VOID;
}

int ReportCableInfoStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());

    //load file only once in case of multiple runnings
    if (this->parse_prtl_mod_dp_latency) {
        this->parse_prtl_mod_dp_latency = false;

        if (!this->pn_vendor_latency_file.empty()) {
            if (this->ParsePNVendorLatencyFile())
                ERR_PRINT("Parsing of PN Vendor to PRTL Latency file was finished with errors\n\n");
            else
                INFO_PRINT("Parsing of PN Vendor to PRTL file was finished successfully\n\n");
        }

        if (!this->cdr_vendor_latency_file.empty()) {
            if (this->ParseCDRVendorLatencyFile())
                ERR_PRINT("Parsing of CDR Vendor to PRTL Latency file was finished with errors\n\n");
            else
                INFO_PRINT("Parsing of CDR Vendor to PRTL Latency file was finished successfully\n\n");
        }

        if (!this->local_phy_latency_file.empty()) {
            if (this->ParsePRTLLocalPHYLatencyFile())
                ERR_PRINT("Parsing of PRTL Local PHY Latency file was finished with errors\n\n");
            else
                INFO_PRINT("Parsing of PRTL Local PHY Latency file was finished successfully\n\n");
        }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ReportCableInfoStage::ParsePRTLLocalPHYLatencyFile()
{
    IBDIAGNET_ENTER;
        int rc = IBDIAG_SUCCESS_CODE;

    IBDiag::PrintFileTimestamp(this->local_phy_latency_file, "Local PHY latency table");

    ifstream fs(this->local_phy_latency_file);
    if (fs.fail()) {
        ERR_PRINT("Failed to open configuration file: %s\n",
                  this->local_phy_latency_file.c_str());
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_PARSE_FILE_FAILED);
    }

    int lineNum = 0;

    while (fs.good()) {
        char line[1024] = {'\0'};
        fs.getline(line, 1024);
        lineNum++;

        // skip empty lines and comments
        if (!line[0] || line[0] == '#')
            continue;

        char techn_str[128]= {0};
        char type_str[128] = {0};
        char width_str[128] = {0};
        char fec_op_str[128] = {0};
        u_int16_t latency = 0;
        char tail_str[128] = {0};

        if (sscanf(line, "%[^:] : %[^:] : %[^:] : %[^=]"
                " =%hd %s",
                techn_str, type_str, width_str, fec_op_str, &latency, tail_str) != 5) {
            ERR_PRINT("Failed to parse line %d, invalid format: %s\n", lineNum, line);
            rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
            continue;
        }

        SerDesVersion technology = char2serdesversion(techn_str);
        if (technology == SerDes_NA) {
            ERR_PRINT("Failed to parse line %d, invalid tecnology: %s\n", lineNum, line);
            rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
            continue;
        }

        IBNodeType type = char2nodetype(type_str);
        if (type == IB_UNKNOWN_NODE_TYPE) {
            ERR_PRINT("Failed to parse line %d, invalid node type: %s\n", lineNum, line);
            rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
            continue;
        }

        IBLinkWidth width = char2width(width_str);
        if (width == IB_UNKNOWN_LINK_WIDTH) {
            ERR_PRINT("Failed to parse line %d, invalid link width: %s\n", lineNum, line);
            rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
            continue;
        }

        IBFECMode fec_op = char2fec(fec_op_str);
        if (fec_op == IB_FEC_NA) {
            ERR_PRINT("Failed to parse line %d, invalid FEC option: %s\n", lineNum, line);
            rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
            continue;
        }

        this->local_phy_latency_table[LocalPhyLatencyKey(technology, type, width, fec_op)] =
                                    latency;
    }

    IBDIAGNET_RETURN(rc);
}

int ReportCableInfoStage::ParseCDRVendorLatencyFile()
{
    IBDIAGNET_ENTER;

    IBDiag::PrintFileTimestamp(this->cdr_vendor_latency_file, "CDR Vendor to PRTL Latency Table");

    ifstream fs(this->cdr_vendor_latency_file);
    if (fs.fail()) {
        ERR_PRINT("Failed to open configuration file: %s\n", this->cdr_vendor_latency_file.c_str());
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_PARSE_FILE_FAILED);
    }

    int rc = IBDIAG_SUCCESS_CODE;
    int lineNum = 0;

    while (fs.good()) {
        char line[1024] = {0};
        fs.getline(line, 1024);
        lineNum++;

        // skip empty lines and comments
        if (!line[0] || line[0] == '#')
            continue;

        char vendor_str[128] = {0};
        u_int16_t latency = {0};
        char tail_str[128] = {0};
        int ret = sscanf(line, "%[^=]=%hd %s", vendor_str, &latency, tail_str);
        if (ret != 2) {
            ERR_PRINT("Failed to parse line %d, invalid format: %s\n", lineNum, line);
            rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
            continue;
        }

        u_int16_t cdr_vendor = 0;
        ret = sscanf (vendor_str, "%hx %s", &cdr_vendor, tail_str);
        if (ret !=1 ) {
            ERR_PRINT("Failed to parse line %d, invalid format: %s\n", lineNum, line);
            rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
            continue;
        }

        this->cdr_vendor_prtl_latency_table[cdr_vendor] = latency;
    }

    IBDIAGNET_RETURN(rc);
 }

int ReportCableInfoStage::ParsePNVendorLatencyFile()
{
    IBDIAGNET_ENTER;

    IBDiag::PrintFileTimestamp(this->pn_vendor_latency_file, "PN Vendor to PRTL Latency Table");

    ifstream fs(this->pn_vendor_latency_file);
    if (fs.fail()) {
        ERR_PRINT("Failed to open configuration file: %s\n", this->pn_vendor_latency_file.c_str());
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_PARSE_FILE_FAILED);
    }

    int rc = IBDIAG_SUCCESS_CODE;
    int lineNum = 0;

    while (fs.good()) {
        char line[1024] = {0};
        fs.getline(line, 1024);
        lineNum++;

        // skip empty lines and comments
        if (!line[0] || line[0] == '#')
            continue;

        char pn_str[128] = {0};
        u_int16_t latency= {0};
        char tail_str[128] = {0};

        int ret = sscanf(line, "%[^=]=%hd %s", pn_str, &latency, tail_str);
        if (ret != 2) {
            ERR_PRINT("Failed to parse line %d, invalid format: %s\n", lineNum, line);
            rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
            continue;
        }

        string trimed = trim(pn_str);
        if (trimed.find(' ') != string::npos || trimed.empty()) {
            ERR_PRINT("Failed to parse line %d, invalid format: %s\n", lineNum, line);
            rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
            continue;
        }

        this->pn_vendor_prtl_latency_table[trimed] = latency;
    }

    IBDIAGNET_RETURN(rc);
}

int ReportCableInfoStage::DumpPRTLLocalPHYLatencyTable(const string &dump_file) const
{
    IBDIAGNET_ENTER;
    ofstream fs(dump_file);
    if (fs.fail()) {
        ERR_PRINT("Failed to create/open file %s "
                "to dump PRTL \"Local PHY Latency\" table\n",
                dump_file.c_str());
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    INFO_PRINT("Dumping PRTL \"Local PHY Latency\" table into the file: %s\n\n",
            dump_file.c_str());

    fs << "# " << IBDIAGNET_VERSION << endl;
    fs << "#" << endl;
    fs << "#" << endl;

    fs << "# " << "<Technology>:<Type>:<Width>:<FEC Mode> = <Latency>" << endl;
    fs << "#" << endl;
    fs << endl;

    fs << "# Technology Mapping:" << endl;
    fs << "#" << endl;
    fs << "# " << serdesversion2char(SerDes_7nm) << " = "
       << serdesversion2char(SerDes_7nm) << "nm" << endl;
    fs << "# " << serdesversion2char(SerDes_5nm) << " = "
       << serdesversion2char(SerDes_5nm) << "nm" << endl;
    fs << "#" << endl;

    fs << "# Type Values:" << endl;
    fs << "#" << endl;
    fs << "# " << nodetype2char(IB_CA_NODE) << ","
       << nodetype2char(IB_SW_NODE)         << endl;
    fs << "#" << endl;

    fs << "# Width Values:" << endl;
    fs << "#" << endl << "# ";
    for (int index = IB_LINK_WIDTH_1X_INDEX; index < IB_LINK_WIDTH_END_INDEX; index++) {
        fs << width2char(width_index2width(static_cast<IBLinkWidthIndex>(index)));

        if (index + 1 < IB_LINK_WIDTH_END_INDEX)
            fs << ",";
    }
    fs << endl << "#" << endl;

    fs << "# FEC mode Mapping:" << endl;
    fs << "#" << endl;
    for (int index = IB_FEC_NO_FEC; index <= IB_FEC_MLNX_RS_271_257_PLR; index++){
        if (index > IB_FEC_RS_FEC_544_514 && index < IB_FEC_MLNX_STRONG_RS_FEC)
            continue;

        fs << "# " << fec2char(static_cast<IBFECMode>(index)) << " = "
           << DEC(index) << endl;
    }
    fs << "#" << endl;

    for (const auto &entry : this->local_phy_latency_table)
        fs << serdesversion2char(entry.first.version) <<  ':'
           << nodetype2char(entry.first.type)         <<  ':'
           << width2char(entry.first.width)           <<  ':'
           << DEC(entry.first.fec_mode)               <<  " = "
           << entry.second << endl;

    fs.close();

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ReportCableInfoStage::DumpCDRVendorPRTLLatencyTable(const string &dump_file) const
{
    IBDIAGNET_ENTER;

    IBDIAGNET_ENTER;

    ofstream fs(dump_file);
    if (fs.fail()) {
        ERR_PRINT("Failed to create/open file %s "
                "to dump CDR Vendor to PRTL \"MOD DP\" latency table\n",
                dump_file.c_str());
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    INFO_PRINT("Dumping CDR Vendor to PRTL \"MOD DP\" latency table into the file: %s\n\n",
            dump_file.c_str());


    fs << "# " << IBDIAGNET_VERSION << endl;
    fs << "#" << endl;
    fs << "#" << endl;

    fs << "# " << "<CDR Vendor> = <latency_value>" << endl;
    fs << "#" << endl;
    fs << endl;

    for (const auto &entry : this->cdr_vendor_prtl_latency_table)
        fs << PTR(entry.first) << " = " << entry.second << endl;

    fs.close();


    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ReportCableInfoStage::DumpPNVendorPRTLLatencyTable(const string &dump_file) const
{
    IBDIAGNET_ENTER;

    ofstream fs(dump_file);
    if (fs.fail()) {
        ERR_PRINT("Failed to create/open file %s "
                "to dump PN Vendor to PRTL \"MOD DP\" latency table\n",
                dump_file.c_str());
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    INFO_PRINT("Dumping PN Vendor to PRTL \"MOD DP\" latency table into the file: %s\n\n",
            dump_file.c_str());


    fs << "# " << IBDIAGNET_VERSION << endl;
    fs << "#" << endl;
    fs << "#" << endl;

    fs << "# " << "<PN Vendor> = <latency_value>" << endl;
    fs << "#" << endl;
    fs << endl;

    for (const auto &entry : this->pn_vendor_prtl_latency_table)
        fs << entry.first << " = " << entry.second << endl;

    fs.close();

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ReportCableInfoStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ReportCableInfoStage::RunCheck()
{
    IBDIAGNET_ENTER;

    list_p_fabric_general_err errorsList;
    int rc =  this->GeneralReport(errorsList);
    this->AnalyzeCheckResults(errorsList,
            "Cable Report",
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            false);
 
    rc = this->FWVersionReport(errorsList);
    this->AnalyzeCheckResults(errorsList,
            "Checking cables FW version",
            rc,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings,
            false);

    rc =  this->TemperatureErrorsReport(errorsList);
    this->AnalyzeCheckResults(errorsList,
            "Cable Temperature Report",
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            false);

    this->ReleaseCableInfo();

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ReportCableInfoStage::TemperatureErrorsReport(list_p_fabric_general_err &errorsList)
{
    IBDIAGNET_ENTER;

    for (map_str_pnode::iterator nI = this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.begin();
            nI != this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->lastError =
                    "DB error - found null node in NodeByName map for key = " + (*nI).first;
            this->num_errors++;
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        for (phys_port_t port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {
            IBPort *p_curr_port = p_curr_node->getPort(port_num);
            if (!p_curr_port || !p_curr_port->p_combined_cable)
                continue;

            bool isDetected = false;

            int rc = this->CheckTemperature(p_curr_port->p_combined_cable->GetTemperatureAlarms(),
                                            "Temperature Alarm reported", *p_curr_port, errorsList, isDetected);
            if (rc)
                IBDIAGNET_RETURN(rc);

            if (isDetected)
                continue;

            rc = this->CheckTemperature(p_curr_port->p_combined_cable->GetTemperatureErrorsByTreshold(),
                                        "Temperature detected", *p_curr_port, errorsList, isDetected);
            if (rc)
                IBDIAGNET_RETURN(rc);
        }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ReportCableInfoStage::CheckTemperature(int tempCode, const string &message, const IBPort &port,
                                            list_p_fabric_general_err &errorsList, bool &isDetected)
{
    IBDIAGNET_ENTER;

    isDetected = false;

    if (tempCode < 0) {
        this->lastError = "DB error - found CombinedCableInfo object with wrong data. Port: "
                        + port.getExtendedName();
        this->num_errors++;
                IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    if (tempCode & CABLE_HIGH_TEMP_MASK) {
        isDetected = true;

        stringstream ss;
        ss << "High " << message;
        errorsList.push_back(new CableTemperatureErr(&port, ss.str(),
                                            port.p_combined_cable->GetTemperatureStr(),
                                            port.p_combined_cable->GetHighTemperatureThresholdStr()));
    }

    if (tempCode & CABLE_LOW_TEMP_MASK) {
        isDetected = true;

        stringstream ss;
        ss << "Low " << message;
        errorsList.push_back(new CableTemperatureErr(&port, ss.str(),
                                            port.p_combined_cable->GetTemperatureStr(),
                                            port.p_combined_cable->GetLowTemperatureThresholdStr()));
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ReportCableInfoStage::FWVersionReport(list_p_fabric_general_err &errorsList)
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;
    cable_id_to_fw_map maxFWVersionsMap;
    rc = this->CollectMaxFWVersions(maxFWVersionsMap);
    if (rc)
        IBDIAGNET_RETURN(rc);

    rc = this->CheckFWVersions(maxFWVersionsMap, errorsList);
    IBDIAGNET_RETURN(rc);
}

int ReportCableInfoStage::CollectMaxFWVersions(cable_id_to_fw_map &maxFWVersionsMap)
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    for (map_str_pnode::const_iterator nI = this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.begin();
            nI != this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.end(); ++nI) {
        const IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->lastError =
                    "DB error - found null node in NodeByName map for key = " + (*nI).first;
            this->num_errors++;
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        for (phys_port_t port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {
            const IBPort *p_curr_port = p_curr_node->getPort(port_num);
            if (!p_curr_port || !p_curr_port->getInSubFabric() || !p_curr_port->p_combined_cable)
                continue;

            cable_id_t cableId = p_curr_port->p_combined_cable->GetIdObj();
            if (!cableId.IsValid())
                continue;

            cable_fw_t cableFW = p_curr_port->p_combined_cable->GetFWObj();
            if (!cableFW.IsValid())
                continue;

            std::pair<cable_id_to_fw_map::iterator, bool> entry =
                    maxFWVersionsMap.insert(std::pair<cable_id_t, cable_fw_t>(cableId, cableFW));
            if (!entry.second && entry.first->second < cableFW)
                entry.first->second = cableFW;
        }
    }

    IBDIAGNET_RETURN(rc);
}

std::ostream& operator<<(std::ostream& os, const cable_fw_t& fw)
{
    os << +fw.hi << '.' << +fw.mid << '.' << +fw.lo;
    return os;
}

std::ostream& operator<<(std::ostream& os, const cable_id_t& id)
{
    os  << "Vendor: " << (id.vendor.empty() ? "NA" : id.vendor)
        << ", PN: "   << (id.pn.empty() ? "NA" : id.pn)
        << ", Identifier: " << (id.identifier.empty() ? "NA" : id.identifier)
        << ", Revision: " << (id.revision.empty() ? "NA" : id.revision);

    return os;
}

int ReportCableInfoStage::CheckFWVersions(const cable_id_to_fw_map &maxFWVersionsMap,
                                          list_p_fabric_general_err &errorsList)
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;
    PairsContainer<const IBPort*> checkedPairs;

    for (map_str_pnode::const_iterator nI = this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.begin();
            nI != this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.end(); ++nI) {
        const IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->lastError =
                    "DB error - found null node in NodeByName map for key = " + (*nI).first;
            this->num_errors++;
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        for (phys_port_t port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {
            const IBPort *p_curr_port = p_curr_node->getPort(port_num);
            if (!p_curr_port || !p_curr_port->getInSubFabric() || !p_curr_port->p_combined_cable)
                continue;

            const cable_id_t cableId = p_curr_port->p_combined_cable->GetIdObj();
            if (!cableId.IsValid())
                continue;

            const cable_fw_t cableFW = p_curr_port->p_combined_cable->GetFWObj();
            if (!cableFW.IsValid())
                continue;

            cable_id_to_fw_map::const_iterator itr = maxFWVersionsMap.find(cableId);
            if (itr != maxFWVersionsMap.end() && cableFW < itr->second) {
                stringstream ss;
                ss << "The cable {"  << cableId << '}'
                   << " has FW version " << cableFW
                   << " while the latest FW version for the same cable on this fabric is "
                   << itr->second
                   << endl;
                errorsList.push_back(new CableFWVersionError(p_curr_port, ss.str()));
            }

            const IBPort *p_remote_port =  p_curr_port->p_remotePort;
            if (!p_remote_port || !p_remote_port->getInSubFabric() ||
                    !p_remote_port->p_combined_cable)
                continue;

            if (checkedPairs.Contains(p_curr_port, p_remote_port))
                continue;

            checkedPairs.Add(p_curr_port, p_remote_port);

            const cable_fw_t remoteCableFW = p_remote_port->p_combined_cable->GetFWObj();
            if (!remoteCableFW.IsValid())
                continue;

            const cable_id_t remoteCableId = p_remote_port->p_combined_cable->GetIdObj();

            //task 3521765
            //we temporarely disable this verification
            // if (remoteCableFW != cableFW) {
            //     stringstream ss;
            //     ss << "The cable " << p_curr_port->getExtendedName()
            //        << "<-->" << p_remote_port->getExtendedName()
            //        << " has different FW versions on its transeivers"
            //        << " {" << cableId << ", FW:" << cableFW << '}'
            //        << " and {" << remoteCableId << ", FW:" << remoteCableFW << '}'
            //        << endl;
            //     errorsList.push_back(new CableFWVersionMismatchError(ss.str()));
            // }

        }
    }

    IBDIAGNET_RETURN(rc);
}

void ReportCableInfoStage::SetPRTLLength(const IBPort *p_curr_port, const IBPort *p_remote_port,
                                        PairsContainer<const IBPort*> &checkedPairs,
                                        list_p_fabric_general_err &errorsList)
{
    IBDIAGNET_ENTER;

    if (checkedPairs.Contains(p_curr_port, p_remote_port))
        IBDIAG_RETURN_VOID;

    checkedPairs.Add(p_curr_port, p_remote_port);

    if (p_curr_port->p_combined_cable->IsOpticCable() &&
            p_remote_port->p_combined_cable->IsOpticCable()) {
        //set length by either two cable's PRTL registers or a one that only exists
        if (p_curr_port->p_prtl && p_remote_port->p_prtl) {
            this->SetPRTLLength(p_curr_port, p_remote_port, errorsList);
        } else if (p_curr_port->p_prtl) {
            this->SetPRTLLength(p_curr_port->p_prtl, p_curr_port, p_remote_port, errorsList);
        } else if (p_remote_port->p_prtl) {
            this->SetPRTLLength(p_remote_port->p_prtl, p_remote_port, p_curr_port, errorsList);
        }
    } else {
        if (p_curr_port->p_combined_cable->IsOpticCable() || p_remote_port->p_combined_cable->IsOpticCable())
            errorsList.push_back(new CableTypeMismatchError(p_curr_port));
    }

    IBDIAG_RETURN_VOID;
}

void ReportCableInfoStage::SetPRTLLength(const PrtlRecord *p_prtl,
                                         const IBPort *p_port, const IBPort *p_other_port,
                                         list_p_fabric_general_err &errorsList)
{
    IBDIAGNET_ENTER;

    //p_other_port does not have any PRTL pointer at all
    if (!p_port->p_prtl->IsSupported())
        IBDIAG_RETURN_VOID;

    string message;
    PrtlRecord prtl(*p_port->p_prtl);
    if (!this->UpdatePrtlLatencyData(prtl, *p_port, message)) {
        errorsList.push_back(new PrtlRegisterInvalidError(p_port, message));
        IBDIAG_RETURN_VOID;
    }

    PrtlRecord other_prtl;
    //other_prtl is a dummy object;
    //it is used only because some of the data can be picked up from hashtables
    if (!this->UpdatePrtlLatencyData(other_prtl, *p_other_port, message)) {
        errorsList.push_back(new PrtlRegisterInvalidError(p_other_port, message));
        IBDIAG_RETURN_VOID;
    }

    float length = prtl.CalculateLength(other_prtl, message);
    if (length <= 0) {
        errorsList.push_back(new PrtlRegisterInvalidError(p_port, message));
        IBDIAG_RETURN_VOID;
    }

    p_port->p_combined_cable->SetPrtlLength(length);

    IBDIAG_RETURN_VOID;
}

void ReportCableInfoStage::SetPRTLLength(const IBPort *p_curr_port, const IBPort *p_remote_port,
                                         list_p_fabric_general_err &errorsList)
{
    IBDIAGNET_ENTER;

    if (p_curr_port->p_prtl->IsSupported() && p_remote_port->p_prtl->IsSupported()) {
        string message;
        string remoteMessage;

        PrtlRecord localPrtl(*p_curr_port->p_prtl);
        bool localUpdate = this->UpdatePrtlLatencyData(localPrtl, *p_curr_port, message);
        if (!localUpdate)
            errorsList.push_back(new PrtlRegisterInvalidError(p_curr_port, message));

        PrtlRecord remotePrtl(*p_remote_port->p_prtl);
        bool remoteUpdate = this->UpdatePrtlLatencyData(remotePrtl, *p_remote_port,
                                                remoteMessage);
        if (!remoteUpdate)
            errorsList.push_back(new PrtlRegisterInvalidError(p_remote_port, remoteMessage));

        if (!localUpdate || !remoteUpdate)
            IBDIAG_RETURN_VOID;

        float localLength = localPrtl.CalculateLength(remotePrtl, message);
        float remoteLength = remotePrtl.CalculateLength(localPrtl, remoteMessage);

        if (localLength <= 0)
            errorsList.push_back(new PrtlRegisterInvalidError(p_curr_port, message));

        if  (remoteLength <= 0)
            errorsList.push_back(new PrtlRegisterInvalidError(p_remote_port, remoteMessage));

        if (localLength > 0 && remoteLength > 0) {
            p_curr_port->p_combined_cable->SetPrtlLength(std::min(localLength, remoteLength));
            p_remote_port->p_combined_cable->SetPrtlLength(std::min(localLength,
                                                                            remoteLength));
        }
    } else {
        if (p_curr_port->p_prtl->IsSupported() || p_remote_port->p_prtl->IsSupported())
            errorsList.push_back(new PrtlRegisterMismatchError(p_curr_port));
    }


    IBDIAG_RETURN_VOID;
}

int ReportCableInfoStage::GeneralReport(list_p_fabric_general_err &errorsList)
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;
    ofstream sout;
    PairsContainer<const IBPort*> checkedPairs;

    //todo verify the stage is activated
    bool csv_disabled = this->p_csv_out->DumpStart(SECTION_COMBINED_CABLE_INFO);

    if (!csv_disabled)
        this->p_csv_out->WriteBuf(CombinedCableInfo::GetCSVHeader().c_str());

    this->InitFileStream(sout);

    if (!sout.is_open())
        IBFabric::GetSwitchLabelPortNumExplanation(sout);

    for (map_str_pnode::iterator nI = this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.begin();
            nI != this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->lastError =
                    "DB error - found null node in NodeByName map for key = " + (*nI).first;
            this->num_errors++;
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        for (phys_port_t port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {
            IBPort *p_curr_port = p_curr_node->getPort(port_num);
            if (!p_curr_port || !p_curr_port->p_combined_cable)
                continue;

            IBPort *p_remote_port = p_curr_port->p_remotePort;
            if (p_remote_port && p_remote_port->p_combined_cable
                    && (p_curr_port->p_prtl || p_remote_port->p_prtl))
                this->SetPRTLLength(p_curr_port, p_remote_port, checkedPairs, errorsList);

            this->ReportToCSV(*p_curr_port, *p_curr_port->p_combined_cable);

            if (sout.is_open())
                this->ReportToStream(sout, *p_curr_port, *p_curr_port->p_combined_cable);
        }
    }

    if (!csv_disabled)
        this->p_csv_out->DumpEnd(SECTION_COMBINED_CABLE_INFO);

    if (sout.is_open())
        this->p_ibdiag->CloseFile(sout);

    IBDIAGNET_RETURN(rc);
}

void ReportCableInfoStage::InitLocalPhyLatancyTable()
{
    IBDIAG_ENTER;

    //CA
    //QM2 NDR 4x RS544
    this->local_phy_latency_table[LocalPhyLatencyKey(SerDes_7nm, IB_CA_NODE,
                IB_LINK_WIDTH_4X, IB_FEC_RS_FEC_544_514)] = 142;
    this->local_phy_latency_table[LocalPhyLatencyKey(SerDes_7nm, IB_CA_NODE,
                IB_LINK_WIDTH_4X, IB_FEC_MLNX_RS_544_514_PLR)] = 142;

    //QM2 NDR 4x RS271
    this->local_phy_latency_table[LocalPhyLatencyKey(SerDes_7nm, IB_CA_NODE,
                IB_LINK_WIDTH_4X, IB_FEC_LL_RS_FEC)] = 111;
    this->local_phy_latency_table[LocalPhyLatencyKey(SerDes_7nm, IB_CA_NODE,
                IB_LINK_WIDTH_4X, IB_FEC_MLNX_RS_271_257_PLR)] = 111;

    //QM2 NDR 2x RS544
    this->local_phy_latency_table[LocalPhyLatencyKey(SerDes_7nm, IB_CA_NODE,
                IB_LINK_WIDTH_2X, IB_FEC_RS_FEC_544_514)] = 170;
    this->local_phy_latency_table[LocalPhyLatencyKey(SerDes_7nm, IB_CA_NODE,
                IB_LINK_WIDTH_2X, IB_FEC_MLNX_RS_544_514_PLR)] = 170;

    //QM2 NDR 2x RS271
    this->local_phy_latency_table[LocalPhyLatencyKey(SerDes_7nm, IB_CA_NODE,
                IB_LINK_WIDTH_2X, IB_FEC_LL_RS_FEC)] = 130;
    this->local_phy_latency_table[LocalPhyLatencyKey(SerDes_7nm, IB_CA_NODE,
                IB_LINK_WIDTH_2X, IB_FEC_MLNX_RS_271_257_PLR)] = 130;

    //SWITCH
    //CX7/BF3 NDR 4x RS544
    this->local_phy_latency_table[LocalPhyLatencyKey(SerDes_7nm, IB_SW_NODE,
                IB_LINK_WIDTH_4X, IB_FEC_RS_FEC_544_514)] = 138;
    this->local_phy_latency_table[LocalPhyLatencyKey(SerDes_7nm, IB_SW_NODE,
                IB_LINK_WIDTH_4X, IB_FEC_MLNX_RS_544_514_PLR)] = 138;

    //CX7/BF3 NDR 4x RS271
    this->local_phy_latency_table[LocalPhyLatencyKey(SerDes_7nm, IB_SW_NODE,
                IB_LINK_WIDTH_4X, IB_FEC_LL_RS_FEC)] = 107;
    this->local_phy_latency_table[LocalPhyLatencyKey(SerDes_7nm, IB_SW_NODE,
                IB_LINK_WIDTH_4X, IB_FEC_MLNX_RS_271_257_PLR)] = 107;

    //CX7/BF3 NDR 2x RS544
    this->local_phy_latency_table[LocalPhyLatencyKey(SerDes_7nm, IB_SW_NODE,
                IB_LINK_WIDTH_2X, IB_FEC_RS_FEC_544_514)] = 166;
    this->local_phy_latency_table[LocalPhyLatencyKey(SerDes_7nm, IB_SW_NODE,
                IB_LINK_WIDTH_2X, IB_FEC_MLNX_RS_544_514_PLR)] = 166;

    //CX7/BF3 NDR 2x RS271
    this->local_phy_latency_table[LocalPhyLatencyKey(SerDes_7nm, IB_SW_NODE,
                IB_LINK_WIDTH_2X, IB_FEC_LL_RS_FEC)] = 126;
    this->local_phy_latency_table[LocalPhyLatencyKey(SerDes_7nm, IB_SW_NODE,
                IB_LINK_WIDTH_2X, IB_FEC_MLNX_RS_271_257_PLR)] = 126;

    IBDIAGNET_RETURN_VOID;
}

void ReportCableInfoStage::InitCDRVendorTable()
{
    IBDIAGNET_ENTER;

    //According to the spec 0 stands for Unknown_or_no_CDR
    this->cdr_vendor_prtl_latency_table[0x00] = 2;
    this->cdr_vendor_prtl_latency_table[0x03] = 4;
    //this->cdr_vendor_prtl_latency_table[0x04] =  TBD

    IBDIAGNET_RETURN_VOID;
}

void ReportCableInfoStage::InitPNVendorTable()
{
    IBDIAGNET_ENTER;

    //Bagheera800
    this->pn_vendor_prtl_latency_table["MMS4X00-NM"] = 80;
    this->pn_vendor_prtl_latency_table["MMS4X00-NM-FLT"] = 80;
    this->pn_vendor_prtl_latency_table["MMS4X00-NS"] = 80;
    this->pn_vendor_prtl_latency_table["MMS4X00-NS-FLT"] = 80;
    this->pn_vendor_prtl_latency_table["P45669-001"] = 80;

    //Bagheera400
    this->pn_vendor_prtl_latency_table["MMS4X00-NS400"] = 80;

    //Bagheera DiD 2ns
    //not provided

    //Louie800
    this->pn_vendor_prtl_latency_table["01LL927"] = 80;
    this->pn_vendor_prtl_latency_table["MMA4Z00-NS"] = 80;
    this->pn_vendor_prtl_latency_table["MMA4Z00-NS-FLT"] = 80;
    this->pn_vendor_prtl_latency_table["MMA4Z00-NV4"] = 80;
    this->pn_vendor_prtl_latency_table["MMA4Z00-NV4-FLT"] = 80;
    this->pn_vendor_prtl_latency_table["P49384-001"] = 80;
    this->pn_vendor_prtl_latency_table["STC7B05673"] = 80;
    this->pn_vendor_prtl_latency_table["STC7B23528"] = 80;


    //Louie
    this->pn_vendor_prtl_latency_table["01LL928"] = 80;
    this->pn_vendor_prtl_latency_table["MMA4Z00-NS400"] = 80;
    this->pn_vendor_prtl_latency_table["P49386-001"] = 80;
    this->pn_vendor_prtl_latency_table["STC7B05888"] = 80;

    //Qlouie400
    this->pn_vendor_prtl_latency_table["01LL929"] = 80;
    this->pn_vendor_prtl_latency_table["MMA1Z00-NS400"] = 80;
    this->pn_vendor_prtl_latency_table["P65325-001"] = 80;
    this->pn_vendor_prtl_latency_table["STC7B05898"] = 80;

    //Louie800 DiD 2ns
    //not provided

    //Qlouie400 DiD
    this->pn_vendor_prtl_latency_table["MMA4Z00-NS400-LP"] = 2;

    //Hydra (2xFR4)
    this->pn_vendor_prtl_latency_table["MMS4X50-NM"] = 80;
    this->pn_vendor_prtl_latency_table["MMS4X50-NS"] = 80;

    IBDIAGNET_RETURN_VOID;
}

bool ReportCableInfoStage::UpdatePrtlLatencyData(PrtlRecord &prtl, const IBPort &port,
                                                 string &message)
{
    IBDIAGNET_ENTER;

    const CombinedCableInfo &combined_cable = *port.p_combined_cable;
    //Order is changed. Firstly we look with PN_Vendor.
    //We are sure in its value
    if (this->prtl_latency_from_table || !prtl.local_mod_dp_latency) {
        map_pn_vendor_latency_t::const_iterator itr_pn =
                this->pn_vendor_prtl_latency_table.find(combined_cable.GetPNVendor());
        if (itr_pn != this->pn_vendor_prtl_latency_table.end())
            prtl.local_mod_dp_latency = itr_pn->second;
        else {
            map_cdr_vendor_latency_t::const_iterator  itr_cdr =
                    this->cdr_vendor_prtl_latency_table.find(combined_cable.GetCDRVendor());
            if (itr_cdr != this->cdr_vendor_prtl_latency_table.end())
                prtl.local_mod_dp_latency = itr_cdr->second;
            else  {
                message = "No \"MOD DP\" latency is provided";
                IBDIAG_RETURN(false);
            }
        }
    }

    if (this->prtl_latency_from_table || !prtl.local_phy_latency)
        prtl.local_phy_latency = this->GetPhyLatency(port);

    if (!prtl.local_phy_latency) {
        message = "No PHY latency is provided";
        IBDIAG_RETURN(false);
    }

    IBDIAGNET_RETURN(true);
}

u_int16_t ReportCableInfoStage::GetPhyLatency(const IBPort &port) const
{
    IBDIAGNET_ENTER;

    if (!port.p_node)
        IBDIAGNET_RETURN(0);

    map_local_phy_latency_t::const_iterator itr =
           this->local_phy_latency_table.find(LocalPhyLatencyKey(port.p_node->serdes_version,
                                                                port.p_node->type,
                                                                port.get_internal_width(),
                                                                port.get_fec_mode()));
    if (itr == this->local_phy_latency_table.end())
        IBDIAGNET_RETURN(0);


    IBDIAGNET_RETURN(itr->second);
}

void ReportCableInfoStage::InitFileStream(ofstream &fileStream)
{
    IBDIAGNET_ENTER;

    if (this->p_ibdiag->OpenFile("Cables Information",
                                string(COMBINED_CABLE_DUMP_FILE), fileStream, false, "#")) {
        ERR_PRINT("Failed to open Cables Information file. Reason: %s\n", p_ibdiag->GetLastError());
        this->num_errors++;
    }

    IBDIAGNET_RETURN_VOID;
}

void ReportCableInfoStage::ReportToCSV(const IBPort &port, const CombinedCableInfo &combined)
{
    IBDIAGNET_ENTER;

    stringstream sstream;

    sstream << PTR(port.p_node->guid_get())      << ','
           << PTR(port.guid_get())               << ','
           << +port.num                          << ',';

    combined.ToCSVStream(sstream);
    this->p_csv_out->WriteBuf(sstream.str());

    IBDIAGNET_RETURN_VOID;
}

void ReportCableInfoStage::ReportToStream(ostream &stream,
                                        const IBPort &port,
                                        const CombinedCableInfo &combined) const
{
    IBDIAGNET_ENTER;

    stream << "-------------------------------------------------------" << endl
           <<  "Port="      << (int)port.num                            << ' '
           <<  "Lid="       << PTR(port.base_lid, 4)                    << ' '
           <<  "GUID="      << PTR(port.guid_get())                     << ' '
           <<  "Port Name=" << port.getName()                           << endl
           << "-------------------------------------------------------" << endl;

    combined.ToFileStream(stream);
    stream << endl << endl;

    IBDIAGNET_RETURN_VOID;
}

void ReportCableInfoStage::ReleaseCableInfo()
{
    IBDIAGNET_ENTER;

    for (map_str_pnode::iterator nI = this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.begin();
            nI != this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            ERR_PRINT("DB error - found null node in NodeByName map for key = %s\n",
                    (*nI).first.c_str());
            continue;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        for (phys_port_t port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {
            IBPort *p_curr_port = p_curr_node->getPort(port_num);
            if (!p_curr_port)
                continue;

            delete p_curr_port->p_combined_cable;
            p_curr_port->p_combined_cable = NULL;

            delete p_curr_port->p_prtl;
            p_curr_port->p_prtl = NULL;
        }
    }

    p_ibdiag->SetCableExported(false);

    IBDIAGNET_RETURN_VOID;
}

bool ReportCableInfoStage::IsSkipped()
{
    IBDIAGNET_ENTER;

    bool is_skipped = Stage::IsSkipped();
    if (!is_skipped) {
       IBDIAGNET_RETURN(false);
    }

    if (this->p_ibdiag->IsCableExported()) {
        this->MarkStageAsActive();
        IBDIAGNET_RETURN(false);
    }

    IBDIAGNET_RETURN(is_skipped);
}

ExportDataStage::ExportDataStage(IBDiag *p_ibdiag, list_p_symbols_plugin *p_loaded_plugins_list)
    : Stage(STAGE_NAME_EXPORT_DATA, p_ibdiag), p_loaded_plugins_list(p_loaded_plugins_list)
{
    IBDIAGNET_ENTER;

    this->MarkStageAsSkipped();

    IBDIAGNET_RETURN_VOID;
}

ExportDataStage::~ExportDataStage()
{
}

int ExportDataStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ExportDataStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ExportDataStage::RunCheck()
{
    IBDIAGNET_ENTER;

    int rc, rc2 = 1;
    list_p_fabric_general_err export_data_errors;
    export_session_handle_t session_handle;

    CLEAR_STRUCT(session_handle);

    session_handle = this->p_ibdiag->ExportOpenSession(this->p_ibdiag->curr_iteration);
    if (!session_handle) {
        export_data_errors.push_back(new ExportDataErr(NULL, NULL,
                                     "Got an empty export session handle"));
        goto exit;
    }

    rc2 = this->p_ibdiag->ExportData(session_handle, export_data_errors);

    for (list_p_symbols_plugin::iterator it = this->p_loaded_plugins_list->begin();
            it != this->p_loaded_plugins_list->end(); ++it) {
        Plugin *p_plugin = (*it)->p_plugin_obj;
        if (!p_plugin->IsFinished())
            continue;

        rc2 |= p_plugin->ExportData(session_handle, export_data_errors);
    }

    this->p_ibdiag->ExportCloseSession(session_handle, rc2);

exit:
    rc = this->AnalyzeCheckResults(export_data_errors,
            CHECK_EXPORT_DATA,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings);

    IBDIAGNET_RETURN(rc);
}

ScopeBuilderStage::ScopeBuilderStage(IBDiag *p_ibdiag)
    : Stage(STAGE_NAME_SCOPE_BUILDER, p_ibdiag), p_filter(nullptr)
{
    IBDIAGNET_ENTER;

    this->MarkStageAsSkipped();

    IBDIAGNET_RETURN_VOID;
}

int ScopeBuilderStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ScopeBuilderStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ScopeBuilderStage::RunCheck()
{
    IBDIAGNET_ENTER;

    set_pnode dest_nodes;
    int rc = this->ReadHostFile(dest_nodes);
    if (rc == IBDIAG_SUCCESS_CODE) {
        INFO_PRINT("Parsing the Host file finished successfully\n\n");
    } else if (rc == IBDIAG_ERR_CODE_PARSE_FILE_FAILED) {
        this->num_errors++;
        ERR_PRINT("Parsing the Host file was finished with errors\n\n");
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
    } else if (rc == IBDIAG_ERR_CODE_NOT_READY) {
        this->num_warnings++;
        WARN_PRINT("Parsing the Host file was finished with warnings\n\n");
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
    } else {
        this->num_warnings++;
        WARN_PRINT("Parsing the Host file was finished with warnings\n\n");
    }

    set_pnode scope_nodes;
    set_p_port scope_ports;
    list_p_fabric_general_err errors;
    rc = this->p_ibdiag->ApplyPathScope(dest_nodes, scope_nodes, scope_ports, errors);
    this->AnalyzeCheckResults(errors,
            CHECK_PATH_SCOPE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings);

    if (!scope_nodes.empty()) {
        if (this->p_ibdiag->CreateScopeFile(scope_nodes, "ibdiagnet2.guid")) {
            this->num_warnings++;
            WARN_PRINT("-W- Failed to open file to save Scope info.\n");
        }
    } else {
        this->num_warnings++;
        WARN_PRINT("-W- No data was collected for Scope file.\n");
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ScopeBuilderStage::ParseHostsString(const char *p_buffer, set_pnode &dest_nodes) const
{
    IBDIAGNET_ENTER;

    if (!p_buffer)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);

    hostlist* p_hl = NULL;
    char* p_host = NULL;

    p_hl = hostlist_create(p_buffer);
    if (!p_hl)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);

    hostlist_uniq(p_hl);

    while (NULL != (p_host = hostlist_shift(p_hl))) {
        for (auto entry : p_ibdiag->GetDiscoverFabricPtr()->NodeByName) {
            IBNode *p_node = entry.second;
            if (!p_node || p_node->type == IB_SW_NODE)
                continue;

            //here we try to distinguish between r-ufm7 and r-ufm77
            //if r-ufm7 is given we try to avoid taking r-ufm77
            //name looks like r-ufm123/mlnx_5
            string host = string(p_host) + '/';
            std::string::size_type pos = p_node->name.find(host);
            if (pos == string::npos || pos)
                continue;

            if (this->p_filter) {//regular expression is applied to node's description
                rexMatch *p_match = this->p_filter->apply(p_node->description.c_str());
                if (p_match) {
                    dest_nodes.insert(p_node);
                    delete p_match;
                }
            } else
                dest_nodes.insert(p_node);
        }

        free(p_host);
    }

    hostlist_destroy(p_hl);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ScopeBuilderStage::ReadHostFile(set_pnode &dest_nodes) const
{
    IBDIAGNET_ENTER;
    int ret = IBDIAG_SUCCESS_CODE;

    IBDiag::PrintFileTimestamp(this->host_file, "Host file");

    ifstream fs(this->host_file);
    if (fs.fail()) {
        ERR_PRINT("Failed to open the Host file: %s\n", this->host_file.c_str());
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_PARSE_FILE_FAILED);
    }

    int line_num = 1;
    while (fs.good()) {
        char line[1024] = {'\0'};
        fs.getline(line, ARRAY_SIZE(line));

        if (!line[0] || line[0] == '#')
            continue;

        if (this->ParseHostsString(line, dest_nodes)) {
            WARN_PRINT("Failed to parse line: %d\n", line_num);
            ret = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
        }

        line_num++;
    }

    if (dest_nodes.empty()) {
        WARN_PRINT("Destination nodes list is empty\n");
        ret = IBDIAG_ERR_CODE_NOT_READY;
    }

    IBDIAGNET_RETURN(ret);
}

bool PostReportsStage::GetFabricSummarySection(stringstream &summary_str)
{
    IBDIAGNET_ENTER;

    if (this->num_of_diff_val) {
        summary_str << "Post Reports SM Configuration Validations: "
                    << this->num_of_diff_val
                    << " fields have different value across the fabric."
                    << endl;

        IBDIAGNET_RETURN(true);
    }

    IBDIAGNET_RETURN(false);
}

ParseKeyFilesStage::ParseKeyFilesStage(IBDiag *p_ibdiag)
    : Stage(STAGE_NAME_PARSE_KEY_FILES, p_ibdiag),
    is_long_iteration(false)
{
    IBDIAGNET_ENTER;

    this->MarkStageAsSkipped();

    IBDIAGNET_RETURN_VOID;
}

int ParseKeyFilesStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ParseKeyFilesStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ParseKeyFilesStage::RunCheck()
{
    if (!this->is_long_iteration)
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);

    IBDIAGNET_RETURN(this->p_ibdiag->GetKeyUpdaterPtr()->UpdateSetKeysIfNeeded());
}
